﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Storage;
using Windows.Storage.AccessCache;

namespace pdftron.PDF.Tools.Collections
{
    public sealed class PinnedRootFolders
    {
        private const string NUMBER_OF_PINNED_FOLDERS_SETTINGS_NAME = "CompleteReader_NumberOfPinnedFolders";
        private const string PINNED_FOLDER_SETTINGS_NAME = "CompleteReader_PinnedFolderNumber_";

        private static ApplicationDataContainer _localSettings;
        private static ApplicationDataContainer LocalSettings => _localSettings ?? (_localSettings = ApplicationData.Current.LocalSettings);

        private int NumberOfPinnedFolders
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(NUMBER_OF_PINNED_FOLDERS_SETTINGS_NAME))
                        return (int)LocalSettings.Values[NUMBER_OF_PINNED_FOLDERS_SETTINGS_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return 0;
            }
            set => LocalSettings.Values[NUMBER_OF_PINNED_FOLDERS_SETTINGS_NAME] = value;
        }

        private PinnedFolderInfo GetPinnedFolder(int index)
        {
            try
            {
                string settingsName = PINNED_FOLDER_SETTINGS_NAME + index;
                if (LocalSettings.Values.ContainsKey(settingsName))
                    return new PinnedFolderInfo((string)LocalSettings.Values[settingsName]);
            }
            catch
            {
                //Swallow exception
            }

            return null;
        }

        private void SetPinnedFolder(PinnedFolderInfo info, int index)
        {
            string settingsName = PINNED_FOLDER_SETTINGS_NAME + index;
            LocalSettings.Values[settingsName] = info.GetStorageString();
        }

        /// <summary>
        /// Shifts all the pinned folders down by 1, starting at startingIndex
        /// </summary>
        /// <param name="startingIndex">The index at which to start shifting folders</param>
        private void DeletePinnedFolder(int startingIndex)
        {
            NumberOfPinnedFolders--;
            for (int index = startingIndex; index < NumberOfPinnedFolders; ++index)
            {
                SetPinnedFolder(GetPinnedFolder(index + 1), index);
            }
        }

        private static PinnedRootFolders _instance;
        private static PinnedRootFolders Instance => _instance ?? (_instance = new PinnedRootFolders());

        private readonly SemaphoreSlim _loadingSemaphore = new SemaphoreSlim(1);

        private List<PinnedItem> _pinnedItems;
        public List<PinnedItem> PinnedItems => _pinnedItems;

        private PinnedRootFolders()
        {
        }

        /// <summary>
        /// This functions will always wait until the RecentItems list is populated, and will then return an instance of RecentItemsData.
        /// This should always be the way to get an instance of this class.
        /// </summary>
        /// <param name="parent">The owner of the recent items list. Used by this class to invoke the dispatcher</param>
        /// <returns>An instance of RecentItem guaranteed to have the list of recent items populated.</returns>
        public static async Task<PinnedRootFolders> GetItemSourceAsync()
        {
            await Instance._loadingSemaphore.WaitAsync().ConfigureAwait(false);

            try
            {
                if (Instance._pinnedItems == null)
                    Instance._pinnedItems = await Instance.GetPinnedItemsAsync();
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error: " + e.ToString());
            }
            finally
            {
                Instance._loadingSemaphore.Release();
            }

            return _instance;
        }

        private IAsyncOperation<List<PinnedItem>> GetPinnedItemsAsync()
        {
            Task<List<PinnedItem>> t = new Task<List<PinnedItem>>(() => GetPinnedItems());
            t.Start();

            return t.AsAsyncOperation();
        }

        private List<PinnedItem> GetPinnedItems()
        {
            List<PinnedItem> pinnedItemsList = new List<PinnedItem>();
            try
            {
                for (int index = 0; index < NumberOfPinnedFolders; index++)
                {
                    PinnedFolderInfo folderInfo = GetPinnedFolder(index);
                    PinnedItem item = new PinnedItem(Path.GetFileName(folderInfo.FolderPath), folderInfo.Token, folderInfo.FolderPath);
                    item.SetToFolder();
                    pinnedItemsList.Add(item);
                }
            }
            catch
            {
                //Swallow exception
            }

            return pinnedItemsList;
        }

        /// <summary>
        /// Adds the folder as a pinned item. Will return the PinnedItem created if the folder was added,
        /// and null if the folder was already pinned.
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        public PinnedItem AddPinnedFolder(StorageFolder item)
        {
            int index = NumberOfPinnedFolders;
            try
            {
                string token = StorageApplicationPermissions.FutureAccessList.Add(item);
                PinnedItem pinnedItem = new PinnedItem(item.Name, token, item.Path);
                pinnedItem.SetToFolder();

                if (!PinnedItems.Any(x => x.Token == pinnedItem.Token))
                {
                    PinnedItems.Add(pinnedItem);
                    SetPinnedFolder(new PinnedFolderInfo(token, pinnedItem.DocumentPath), index);
                    NumberOfPinnedFolders++;
                    return pinnedItem;
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error adding folder: " + e.ToString());
            }

            return null;
        }

        public void DeletePinnedFolder(PinnedItem item)
        {
            int index = PinnedItems.IndexOf(item);
            PinnedItems.Remove(item);
            DeletePinnedFolder(index);
        }

        public void ClearAllPinnedFolders()
        {
            const int desiredNumberOfItems = 0;
            while (PinnedItems.Count > desiredNumberOfItems)
            {
                PinnedItems.RemoveAt(0);
            }

            NumberOfPinnedFolders = 0;
        }

        public async Task<StorageFolder> GetFolderFromPinnedItemAsync(PinnedItem item)
        {
            StorageFolder folder = null;
            try
            {
                if (!string.IsNullOrWhiteSpace(item.Token))
                    folder = await StorageApplicationPermissions.FutureAccessList.GetFolderAsync(item.Token);
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error getting storage folder: " + e.ToString());
            }

            return folder;
        }

        /// <summary>
        /// Creates a recent item from a storage item without adding it to the future access list
        /// </summary>
        /// <param name="file"></param>
        /// <returns>The recent item created</returns>
        public PinnedItem CreateNewPinnedItem(IStorageItem file)
        {
            PinnedItem item = new PinnedItem(file);
            if (file is StorageFolder)
            {
                item.SetToFolder();
            }

            return item;
        }

        public bool ContainsToken(string token)
        {
            foreach (PinnedItem item in PinnedItems)
            {
                if (item.Token?.Equals(token) == true)
                    return true;
            }

            return false;
        }
    }
}
