﻿using pdftron.PDF.Tools.Controls.ControlBase;
using pdftron.PDF.Tools.Controls.ViewModels.Multimedia;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Storage;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;


namespace pdftron.PDF.Tools.Controls.Multimedia
{
    public sealed partial class AudioControl : UserControl, ControlBase.ICloseableControl, IDisposable
    {
        private OnCloseRequestedHandler _closeRequested;
        public delegate void OnCloseRequestedHandler();

        /// <summary>
        /// Raised when Audio Control is requesting to be closed
        /// </summary>
        public event OnCloseRequestedHandler CloseRequested
        {
            add
            {
                if (_closeRequested?.GetInvocationList().Contains(value) != true)
                    _closeRequested += value;
            }
            remove => _closeRequested -= value;
        }

        public event ControlBase.ControlClosedDelegate ControlClosed = delegate { };

        private OnRecordingCompletedHandler _recordingCompleted;
        public delegate void OnRecordingCompletedHandler(StorageFile file, pdftron.PDF.Rect target, int pageNumber);

        /// <summary>
        /// Raised when the recording is completed and audio file saved to temp folder
        /// </summary>
        public event OnRecordingCompletedHandler RecordingCompleted
        {
            add
            {
                if (_recordingCompleted?.GetInvocationList().Contains(value) != true)
                    _recordingCompleted += value;
            }
            remove => _recordingCompleted -= value;
        }

        public AudioControl(PDFViewCtrl mPDFView, Windows.Foundation.Point target)
        {
            this.InitializeComponent();

            if (mPDFView == null)
                throw new Exception("PDFViewCtrl not initialized for AudioControl");

            this.DataContext = ViewModel = new AudioControlViewModel(mPDFView, target);
            ViewModel.CloseRequested += ViewModel_CloseRequested;
            ViewModel.RecordingCompleted += ViewModel_RecordingCompleted;
        }

        #region Public Properties

        public AudioControlViewModel ViewModel { get; private set; }

        /// <summary>
        /// Audio Control is initialized and ready to be used
        /// </summary>
        public bool IsReady { get; set; }

        #endregion

        #region Public Methods

        /// <summary>
        /// Initialize all resources needed for Audio/Sound recording and playback
        /// </summary>
        public async Task<bool> InitializeAsync()
        {
            if (ViewModel != null)
                return await ViewModel.InitAsync();

            return false;
        }

        /// <summary>
        /// The result file recorded once <see cref="RecordingCompleted"/> is raised
        /// </summary>
        /// <returns></returns>
        public StorageFile RecordedFile() 
        {
            if (ViewModel == null)
                return null;

            return ViewModel.RecordedFile;
        }

        public void CloseControl()
        {

        }

        public void Dispose()
        {
            if (ViewModel != null)
            {
                ViewModel.CloseRequested -= ViewModel_CloseRequested;
                ViewModel.RecordingCompleted -= ViewModel_RecordingCompleted;
                
                ViewModel = null;
            }
        }
        #endregion

        #region Events

        private void ViewModel_CloseRequested()
        {
            _closeRequested?.Invoke();
        }

        private void ViewModel_RecordingCompleted(StorageFile file, pdftron.PDF.Rect target, int pageNumber)
        {
            _recordingCompleted?.Invoke(file, target, pageNumber);
        }

        #endregion
    }
}
