﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls.Resources
{
    /// <summary>
    /// This control gives the user control over the buttons for previous and next page.
    /// It is added because using the Magnifying View, it was possible to make the system deadlock when using the default buttons
    /// on the PDFViewCtrl.The deadlock appears to be entirely within code for UWP (for example an undetected layout loop)
    /// The buttons here ensure that we call functions on the PDFViewCtrl, instead of triggering a selection change on the internal FlipView.
    /// </summary>
    public sealed partial class PDFViewCtrlNavButtonHost : UserControl
    {
        #region Dependency Properties

        public static readonly DependencyProperty HostedPDFViewCtrlProperty = DependencyProperty.Register(
              "HostedPDFViewCtrl",
              typeof(PDFViewCtrl),
              typeof(PDFViewCtrlNavButtonHost),
              new PropertyMetadata(null, new PropertyChangedCallback(OnHostedPDFViewCtrlChanged))
            );

        public PDFViewCtrl HostedPDFViewCtrl
        {
            get { return (PDFViewCtrl)GetValue(HostedPDFViewCtrlProperty); }
            set { SetValue(HostedPDFViewCtrlProperty, value); }
        }

        private PDFViewCtrl _oldView = null;

        private static void OnHostedPDFViewCtrlChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            PDFViewCtrlNavButtonHost host = d as PDFViewCtrlNavButtonHost;
            if (host != null)
            {
                if (host._oldView != null)
                {
                    host.PDFViewCtrlBorder.Child = null;
                    host.UnsubscribePDFViewCtrl(host._oldView);
                }

                PDFViewCtrl newView = e.NewValue as PDFViewCtrl;
                if (newView != null)
                {
                    newView.PreviousAndNextButtonVisibility = Visibility.Collapsed;
                    host.PDFViewCtrlBorder.Child = newView;
                    host.ResolveVisiblity();
                    host.SubscribePDFViewCtrl(newView);
                }
                host._oldView = newView;
            }
        }

        #endregion Dependency Properties

        public PDFViewCtrlNavButtonHost()
        {
            this.InitializeComponent();

            PrevButton.Click += PrevButton_Click;
            NextButton.Click += NextButton_Click;
        }

        private void PrevButton_Click(object sender, RoutedEventArgs e)
        {
            PDFViewCtrl ctrl = HostedPDFViewCtrl;
            if (ctrl != null)
            {
                ctrl.GotoPreviousPage();
            }
        }

        private void NextButton_Click(object sender, RoutedEventArgs e)
        {
            PDFViewCtrl ctrl = HostedPDFViewCtrl;
            if (ctrl != null)
            {
                ctrl.GotoNextPage();
            }
        }

        private void SubscribePDFViewCtrl(PDFViewCtrl ctrl)
        {
            PDFViewCtrl Ctrl = ctrl;

            ctrl.PointerPressed += Ctrl_PointerPressed;
            ctrl.PointerMoved += Ctrl_PointerMoved;
            ctrl.OnLayoutChanged += Ctrl_OnLayoutChanged;
            ctrl.OnPageNumberChanged += Ctrl_OnPageNumberChanged;
        }

        private void UnsubscribePDFViewCtrl(PDFViewCtrl ctrl)
        {
            ctrl.PointerPressed -= Ctrl_PointerPressed;
            ctrl.PointerMoved -= Ctrl_PointerMoved;
            ctrl.OnPageNumberChanged -= Ctrl_OnPageNumberChanged;
        }

        #region Button Visibility

        private bool _WasLastInteractionTouch = true;
        private bool WasLastInteractionTouch
        {
            get { return _WasLastInteractionTouch; }
            set
            {
                if (_WasLastInteractionTouch != value)
                {
                    _WasLastInteractionTouch = value;
                    ResolveVisiblity();
                }
            }
        }

        private bool ShowPrevButton
        {
            get { return PrevButtonEntranceControl.IsOpen; }
            set { PrevButtonEntranceControl.IsOpen = value; }
        }

        private bool ShowNextButton
        {
            get { return NextButtonEntranceControl.IsOpen; }
            set { NextButtonEntranceControl.IsOpen = value; }
        }

        private void ResolveVisiblity()
        { 
            PDFViewCtrl ctrl = HostedPDFViewCtrl;
            if (UtilityFunctions.IsContinuous(ctrl.GetPagePresentationMode()) || _WasLastInteractionTouch)
            {
                ShowNextButton = false;
                ShowPrevButton = false;
            }
            else
            {
                int currentPage = ctrl.GetCurrentPage();
                int maxPage = ctrl.GetPageCount();
                PDFViewCtrlPagePresentationMode mode = ctrl.GetPagePresentationMode();
                if (mode == PDFViewCtrlPagePresentationMode.e_single_page)
                {
                    ShowPrevButton = currentPage > 1;
                    ShowNextButton = currentPage < maxPage;
                }
                else if (mode == PDFViewCtrlPagePresentationMode.e_facing)
                {
                    ShowPrevButton = currentPage > 2;
                    ShowNextButton = (currentPage + 1 < maxPage || (currentPage < maxPage && currentPage % 2 == 0));
                }
                else if (mode == PDFViewCtrlPagePresentationMode.e_facing_cover)
                {
                    ShowPrevButton = currentPage > 1;
                    ShowNextButton = (currentPage + 1 < maxPage || (currentPage < maxPage && currentPage % 2 == 1));
                }
            }
        }


        private void Ctrl_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            WasLastInteractionTouch = e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch;
        }

        private void Ctrl_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            WasLastInteractionTouch = e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch;
        }

        private void Ctrl_OnLayoutChanged()
        {
            ResolveVisiblity();
        }

        private void Ctrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            ResolveVisiblity();
        }

        #endregion Button Visibility
    }
}
