﻿using pdftron.Common;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using PDFRect = pdftron.PDF.Rect;   

namespace pdftron.PDF.Tools.Controls
{
    /// <summary>
    /// Helper class that own a FreeHandCreateTool as and a PDFViewCtrl, and replicates the strokes drawn by the 
    /// tool on the PDFViewCtrl
    /// </summary>
    public class StrokeReplicator : Canvas
    {
        private class StrokeProperties
        {
            public FreehandCreate.PenStroke StrokeInPDFSpace;
            public FreehandCreate.StrokeAppearance StrokeAppearance;
            public Path Path;
            public PathFigure PathFigure;
            public UIPoint LastPoint;
            public UIPoint CurrentPoint;
            public PDFRect BoundingRect
            {
                get
                {
                    return GetBoundingRect();
                }
            }

            public StrokeProperties(FreehandCreate.PenStroke stroke, FreehandCreate.StrokeAppearance appearance)
            {
                StrokeInPDFSpace = stroke;
                StrokeAppearance = appearance;
                Path = new Path();
                Path.StrokeLineJoin = PenLineJoin.Round;
                Path.StrokeStartLineCap = PenLineCap.Round;
                Path.StrokeEndLineCap = PenLineCap.Round;

                PathGeometry geom = new PathGeometry();
                Path.Data = geom;

                PathFigure = new PathFigure();
                geom.Figures.Add(PathFigure);
            }

            public void ApplyAppearance(double zoomFactor)
            {
                Path.StrokeThickness = zoomFactor * StrokeAppearance.Thickness;
                Path.Stroke = StrokeAppearance.Brush;
                Path.Opacity = StrokeAppearance.Opacity;
            }

            private PDFRect GetBoundingRect()
            {
                if (StrokeInPDFSpace.StrokePoints.Count == 0)
                {
                    return null;
                }
                PDFRect rect = new PDFRect(StrokeInPDFSpace.Bounds.x1, StrokeInPDFSpace.Bounds.y1, StrokeInPDFSpace.Bounds.x2, StrokeInPDFSpace.Bounds.y2);
                rect.Inflate(10);
                return rect;
            }
        }


        private PDFViewCtrl _Ctrl;
        private FreehandCreate _Tool;

        private Dictionary<FreehandCreate.PenStroke, StrokeProperties> _VisiblePaths;
        private FreehandCreate.PenStroke _CurrentPenStroke;
        private int _CurrentStrokePoints = 0;
        private CubicCurveBuilder _InkcurveBuilder;
        private Canvas _AnnotationCanvas;

        public StrokeReplicator(PDFViewCtrl ctrl, FreehandCreate tool)
        {
            _Ctrl = ctrl;
            _Tool = tool;

            _Tool.StrokeMadeVisible += Tool_StrokeMadeVisible;
            _Tool.StrokeMadeInvisible += Tool_StrokeMadeInvisible;
            _Tool.CurrentStrokeModifed += Tool_CurrentStrokeModifed;
            _Tool.AllStrokesMadeInvisible += Tool_AllStrokesMadeInvisible;

            _VisiblePaths = new Dictionary<FreehandCreate.PenStroke, StrokeProperties>();
            _InkcurveBuilder = Annots.Ink.GetCurveBuilder();

            IsCommited = false;
        }

        public void CleanUp()
        {
            _Tool.StrokeMadeVisible -= Tool_StrokeMadeVisible;
            _Tool.StrokeMadeInvisible -= Tool_StrokeMadeInvisible;
            _Tool.CurrentStrokeModifed -= Tool_CurrentStrokeModifed;
            _Tool.AllStrokesMadeInvisible -= Tool_AllStrokesMadeInvisible;
        }

        public void Commit()
        {
            CommitInk();
        }

        public bool HasStroke
        {
            get { return _VisiblePaths.Count > 0; }
        }

        public bool IsCommited { get; private set; }

        #region Responding To Drawing

        private void Tool_StrokeMadeVisible(FreehandCreate.StrokeAppearance appearance, FreehandCreate.PenStroke stroke)
        {
            _CurrentPenStroke = stroke;
            _VisiblePaths[_CurrentPenStroke] = new StrokeProperties(_CurrentPenStroke, appearance);
            _VisiblePaths[_CurrentPenStroke].ApplyAppearance(_Ctrl.GetZoom());
            Canvas annotationCanvas = _Ctrl.GetAnnotationCanvas();
            if (annotationCanvas != _AnnotationCanvas && _AnnotationCanvas != null)
            {
                //remove them all!
                if (_AnnotationCanvas.Children.Contains(this))
                {
                    _AnnotationCanvas.Children.Remove(this);
                }
            }
            _AnnotationCanvas = annotationCanvas;
            if (!_AnnotationCanvas.Children.Contains(this))
            {
                _AnnotationCanvas.Children.Add(this);
            }
            this.Children.Add(_VisiblePaths[_CurrentPenStroke].Path);
            _CurrentStrokePoints = 0;
            _InkcurveBuilder = Annots.Ink.GetCurveBuilder();
            Tool_CurrentStrokeModifed(stroke);
        }

        private void Tool_StrokeMadeInvisible(FreehandCreate.PenStroke stroke)
        {
            if (_VisiblePaths.ContainsKey(stroke) && this.Children.Contains(_VisiblePaths[stroke].Path))
            {
                this.Children.Remove(_VisiblePaths[stroke].Path);
            }
        }

        private void Tool_CurrentStrokeModifed(FreehandCreate.PenStroke stroke)
        {
            DrawPathForStroke(stroke, _CurrentStrokePoints);
            _CurrentStrokePoints = stroke.StrokePoints.Count;
        }

        private void DrawPathForStroke(FreehandCreate.PenStroke stroke, int startingPosition)
        {
            int pointCount = stroke.StrokePoints.Count;
            if (pointCount <= startingPosition)
            {
                // TODO erase
                return;
            }
            StrokeProperties path = _VisiblePaths[stroke];

            if (startingPosition == 0)
            {
                DoubleRef x = new DoubleRef(stroke.StrokePoints[0].x);
                DoubleRef y = new DoubleRef(stroke.StrokePoints[0].y);
                _Ctrl.ConvPagePtToScreenPt(x, y, path.StrokeAppearance.PageNumber);
                _Ctrl.ConvScreenPtToAnnotationCanvasPt(x, y, path.StrokeAppearance.PageNumber);
                _InkcurveBuilder.AddSourcePoint(x.Value, y.Value);
                path.LastPoint = new Windows.Foundation.Point(x.Value, y.Value);
                path.PathFigure.StartPoint = path.LastPoint;
                if (pointCount == 1)
                {
                    path.PathFigure.Segments.Add(new LineSegment() { Point = new UIPoint(path.LastPoint.X + 0.5, path.LastPoint.Y + 0.5) });
                }
                ++startingPosition;
            }

            uint cubicPointsAtStart = _InkcurveBuilder.NumCubicPoints();

            for (int i = startingPosition; i < pointCount; ++i)
            {
                DoubleRef x = new DoubleRef(stroke.StrokePoints[i].x);
                DoubleRef y = new DoubleRef(stroke.StrokePoints[i].y);
                _Ctrl.ConvPagePtToScreenPt(x, y, path.StrokeAppearance.PageNumber);
                _Ctrl.ConvScreenPtToAnnotationCanvasPt(x, y, path.StrokeAppearance.PageNumber);
                _InkcurveBuilder.AddSourcePoint(x.Value, y.Value);
                if (!path.StrokeAppearance.UseSmoothing)
                {
                    path.CurrentPoint = new UIPoint(x.Value, y.Value);
                    path.PathFigure.Segments.Add(new LineSegment() { Point = path.CurrentPoint });
                }
            }

            if (path.StrokeAppearance.UseSmoothing)
            {
                if (path.PathFigure.Segments.Count >= 1)
                {
                    path.PathFigure.Segments.RemoveAt(path.PathFigure.Segments.Count - 1);
                }

                if (cubicPointsAtStart >= 4)
                {
                    cubicPointsAtStart -= 3;
                }

                uint totalCubicPoint = _InkcurveBuilder.NumCubicPoints();
                for (uint i = cubicPointsAtStart; i < totalCubicPoint; i += 3)
                {
                    BezierSegment qbc = new BezierSegment();

                    qbc.Point1 = new UIPoint(_InkcurveBuilder.GetCubicXCoord(i), _InkcurveBuilder.GetCubicYCoord(i));
                    qbc.Point2 = new UIPoint(_InkcurveBuilder.GetCubicXCoord(i + 1), _InkcurveBuilder.GetCubicYCoord(i + 1));
                    qbc.Point3 = new UIPoint(_InkcurveBuilder.GetCubicXCoord(i + 2), _InkcurveBuilder.GetCubicYCoord(i + 2));

                    path.PathFigure.Segments.Add(qbc);
                }
            }
        }

        private void Tool_AllStrokesMadeInvisible()
        {
            foreach (StrokeProperties pathProps in _VisiblePaths.Values)
            {
                if (this.Children.Contains(pathProps.Path))
                {
                    this.Children.Remove(pathProps.Path);
                }
            }
            _VisiblePaths.Clear();
        }

        #endregion Responding To Drawing

        #region Finishing

        private void CommitInk()
        {
            PDFRect updateRect = null;
            foreach (StrokeProperties strokes in _VisiblePaths.Values)
            {
                PDFRect rect = strokes.BoundingRect;
                rect = UtilityFunctions.ConvertFromPageRectToScreenRect(rect, strokes.StrokeAppearance.PageNumber, _Ctrl);
                rect.Normalize();
                if (updateRect == null)
                {
                    updateRect = rect;
                }
                else
                {
                    updateRect.x1 = Math.Min(updateRect.x1, rect.x1);
                    updateRect.y1 = Math.Min(updateRect.y1, rect.y1);
                    updateRect.x2 = Math.Max(updateRect.x2, rect.x2);
                    updateRect.y2 = Math.Max(updateRect.y2, rect.y2);
                }
            }
            if (updateRect != null)
            {
                IsCommited = true;
                _Ctrl.Update(updateRect);
            }
        }

        #endregion Finishing
    }
}
