﻿using pdftron.Common;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.Graphics.Imaging;
using Windows.Media.Capture;
using Windows.Storage;
using Windows.Storage.Pickers;
using Windows.Storage.Streams;
using Windows.System.UserProfile;
using Windows.UI.Xaml.Media.Imaging;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class CreateDocControlViewModel : ViewModelBase
    {
        #region Enums
        public enum CreateDocFrom
        {
            Blank,
            Image
        }

        public enum ImageCreationType
        {
            File,
            Camera,
            None
        }
        #endregion

        #region Constants
        protected const double A0_Width = 2383.2;
        protected const double A0_Height = 3369.6;
        protected const double A1_Width = 1684.8;
        protected const double A1_Height = 2383.2;
        protected const double A2_Width = 1188.0;
        protected const double A2_Height = 1684.8;
        protected const double A3_Width = 842.4;
        protected const double A3_Height = 1188.0;
        protected const double A4_Width = 595.44;
        protected const double A4_Height = 842.4;
        protected const double A5_Width = 419.76;
        protected const double A5_Height = 595.44;

        protected const double B0_Width = 2835.8;
        protected const double B0_Height = 4010.4;
        protected const double B1_Width = 2005.2;
        protected const double B1_Height = 2835.8;
        protected const double B2_Width = 1417.9;
        protected const double B2_Height = 2005.2;
        protected const double B3_Width = 1002.6;
        protected const double B3_Height = 1417.9;
        protected const double B4_Width = 708.9;
        protected const double B4_Height = 1002.6;
        protected const double B5_Width = 499.0;
        protected const double B5_Height = 708.9;

        protected const double Letter_Width = 612.0;
        protected const double Letter_Height = 792.0;
        protected const double Legal_Width = 612.0;
        protected const double Legal_Height = 1008;

        protected const double _inConversionRate = 1 / 72.0;
        protected const double _mmConversionRate = 0.3528;

        private const int MAX_PAGE_COUNT = 50;
        #endregion

        #region Private Properties
        private string _DocumentTitle;

        private int _TotalPages;

        private CreateDocFrom _DocumentType;
        private ImageCreationType _ImageType;

        private bool _IsPortrait = true;

        private PaperSizeHolder _SelectedPaperSize;
        private PaperTypeHolder _SelectedPaperType;

        private Dictionary<PaperSizes, Tuple<double, double>> _PaperSizeMap;

        private BitmapImage _previewImage;

        private readonly List<string> letterSizeCultures = new List<string> { "US", "CA", "MX", "CU", "DO", "GT", "CR", "SV", "HN", "CO", "VE", "PH", "CL" };
        private readonly IList<string> Valid_Image_Types = new List<string> { ".jpg", ".bmp", ".jpeg", ".tif", ".tiff", ".png", ".gif", ".ico", ".jpeg-xr" };
        private readonly IDictionary<ImageCreationType, StorageFile> _filesPicked = new Dictionary<ImageCreationType, StorageFile>();
        #endregion

        #region Constructors
        /// <summary>
        /// Creates a new CreateDocControlViewModel
        /// </summary>
        public CreateDocControlViewModel()
        {
            InitCollections();

            CMDPickImageFromFile = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PickImageFromFileAsync);
                command.Execute(parameter);
            });

            CMDPickImageFromCamera = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PickImageFromCameraAsync);
                command.Execute(parameter);
            });

            CMDCreateDocument = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(CreateDocumentAsync);
                command.Execute(parameter);
            });
        }
        #endregion

        #region Public Properties
        public ObservableCollection<PaperTypeHolder> AvailablePaperTypes { get; private set; }
        public ObservableCollection<PaperSizeHolder> AvailablePaperSizes { get; private set; }

        public string DocumentTitle
        {
            get { return _DocumentTitle; }
            set
            {
                if (Set(ref _DocumentTitle, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public CreateDocFrom DocumentType
        {
            get { return _DocumentType; }
            set
            {
                if (Set(ref _DocumentType, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public bool IsPortrait
        {
            get { return _IsPortrait; }
            set
            {
                if (Set(ref _IsPortrait, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public PaperSizeHolder SelectedPaperSize
        {
            get { return _SelectedPaperSize; }
            set
            {
                if (Set(ref _SelectedPaperSize, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public PaperTypeHolder SelectedPaperType
        {
            get { return _SelectedPaperType; }
            set
            {
                if (Set(ref _SelectedPaperType, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public int TotalPages
        {
            get { return _TotalPages; }
            set
            {
                if (Set(ref _TotalPages, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public string DefaultDocumentName
        {
            get
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                return loader.GetString("DocumentCreation_AllDocuments_DefaultTitle");
            }
        }

        public BitmapImage PreviewImage
        {
            get => _previewImage;
            set
            {
                _previewImage = value;
                RaisePropertyChanged();
                RaisePropertyChanged(nameof(IsImageSelected));
            }
        }

        public bool IsImageSelected => _previewImage != null;

        public ImageCreationType ImageType
        {
            get { return _ImageType; }
            set
            {
                if (Set(ref _ImageType, value))
                {
                    RaisePropertyChanged();
                }
            }
        }
        #endregion

        #region Commands
        public RelayCommand CMDPickImageFromFile { get; }

        public RelayCommand CMDPickImageFromCamera { get; }

        public RelayCommand CMDCreateDocument { get; set; }
        #endregion

        #region Private Methods
        private string GetFileName()
        {
            if (string.IsNullOrEmpty(_DocumentTitle))
                return DefaultDocumentName;

            return _DocumentTitle;
        }

        private Task<PDFDoc> GetBlankPDFDocAsync()
        {
            AppCenterHandlerBase.CURRENT.SendEvent("PDF Creation",
                new Dictionary<string, string>()
                {
                    ["Type"] = "Blank Document"
                });

            AppCenterHandlerBase.CURRENT.SendEvent("PDF Creation",
                new Dictionary<string, string>()
                {
                    ["Paper Type"] = Enum.GetName(typeof(PaperType), SelectedPaperType.PaperType)
                });

            return Task.Run(() =>
            {
                GetPaperDimensionsFromSize(SelectedPaperSize.PaperSize, out double width, out double height);
                if (!IsPortrait)    //landscape
                {
                    double temp = width;
                    width = height;
                    height = temp;
                }

                return PDFDocCreator.Create(SelectedPaperType.PaperType, width, height, _TotalPages);
            });
        }

        private async Task<PDFDoc> GetImagePDFDOC()
        {

            if(_filesPicked.Count == 0)
            {
                return null;
            }

            PDFDoc doc = new PDFDoc();

            if (_filesPicked.Count > 0)
            {
                var file = _filesPicked.Values.FirstOrDefault();
                if (file.FileType == ".xps" || file.FileType == ".oxps")
                {
                    await GetDocFromConverter(doc, _filesPicked[0]).ConfigureAwait(false);
                    return doc;
                }
            }

            SDFDoc sdoc = doc.GetSDFDoc();

            ElementBuilder builder = new ElementBuilder();		// ElementBuilder is used to build new Element objects
            ElementWriter writer = new ElementWriter();	// ElementWriter is used to write Elements to the page

            foreach (StorageFile file in _filesPicked.Values)
            {
                IRandomAccessStream ras = await file.OpenReadAsync();
                Page page = doc.PageCreate();
                writer.Begin(page);	// begin writing to this page

                Image img;
                try
                {
                    img = await Image.CreateAsync(sdoc, ras);
                }
                catch (Exception)
                {
                    img = null;
                }

                try
                {
                    if (img == null)
                    {
                        BitmapDecoder decoder = await BitmapDecoder.CreateAsync(ras);

                        uint height = decoder.PixelHeight;
                        uint width = decoder.PixelWidth;

                        PixelDataProvider pdp2 = await decoder.GetPixelDataAsync(
                            BitmapPixelFormat.Rgba8,
                            BitmapAlphaMode.Straight,
                            new BitmapTransform(),
                            ExifOrientationMode.RespectExifOrientation,
                            ColorManagementMode.DoNotColorManage);

                        byte[] pixels = pdp2.DetachPixelData();
                        byte[] noAlphaPixels = new byte[(pixels.Length * 3) / 4];

                        int j = -1;
                        for (int i = 0; i < pixels.Length; i += 4)
                        {
                            noAlphaPixels[++j] = pixels[i];
                            noAlphaPixels[++j] = pixels[i + 1];
                            noAlphaPixels[++j] = pixels[i + 2];
                        }

                        try
                        {
                            img = Image.Create(sdoc, noAlphaPixels, (int)width, (int)height, 8, ColorSpace.CreateDeviceRGB(), ImageInputFilter.e_none);
                        }
                        catch (Exception)
                        {
                            img = null;
                        }
                    }
                }
                catch
                {
                    //Swallow exception
                }

                if (img == null)
                    return null;

                page.SetCropBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));
                page.SetMediaBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));

                Element element = builder.CreateImage(img, new Matrix2D(img.GetImageWidth(), 0, 0, img.GetImageHeight(), 0, 0));

                writer.WritePlacedElement(element);

                writer.End();  // save changes to the current page
                doc.PagePushBack(page);
            }

            return doc;
        }

        private async Task GetDocFromConverter(PDFDoc doc, StorageFile file)
        {
            StorageFile newfile = await file.CopyAsync(ApplicationData.Current.TemporaryFolder, file.Name, NameCollisionOption.ReplaceExisting).AsTask().ConfigureAwait(false);

            pdftron.PDF.Convert.FromXps(doc, newfile.Path);
        }

        private string GetFileNameWithExtension()
        {
            string fileName = GetFileName();
            if (fileName.Length < 4)
                fileName += ".pdf";

            if (!fileName.Substring(fileName.Length - 4).Equals(".pdf", StringComparison.OrdinalIgnoreCase))
                fileName += ".pdf";

            return fileName;
        }

        private async Task<IReadOnlyList<StorageFile>> GetOpenFileFromPicker(IList<string> fileTypes)
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            FileOpenPicker fileOpenPicker = new FileOpenPicker
            {
                ViewMode = PickerViewMode.Thumbnail,
                CommitButtonText = loader.GetString("DocumentCreation_ImageDocument_PickFileConform")
            };

            foreach (string filetype in fileTypes)
            {
                fileOpenPicker.FileTypeFilter.Add(filetype);
            }

            StorageFile theFile = await fileOpenPicker.PickSingleFileAsync();
            return new List<StorageFile>
            {
                theFile
            };
        }

        private async Task PickImageFromFileSystemAsync()
        {
            try
            {

                IReadOnlyList<StorageFile> pickedFiles = await GetOpenFileFromPicker(Valid_Image_Types);
                if (pickedFiles[0] == null)
                    return;

                _filesPicked.Clear();
                _filesPicked.Add(ImageCreationType.File, pickedFiles[0]);

                await AddFileToBeShownAsync(_filesPicked[0]);
                RaisePropertyChanged("IsInputValid");
            }
            finally
            {
            }
        }

        private async Task AddFileToBeShownAsync(StorageFile file)
        {
            try
            {
                if (file.FileType == ".xps" || file.FileType == ".oxps")
                {
                    PreviewImage = new BitmapImage
                    {
                        UriSource = new Uri("ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf.png", UriKind.Absolute)
                    };
                }
                else
                {
                    using (IRandomAccessStream fileStream = await file.OpenReadAsync())
                    {
                        BitmapImage bitmapImage = new BitmapImage();
                        await bitmapImage.SetSourceAsync(fileStream);
                        PreviewImage = bitmapImage;
                    }
                }
            }
            catch
            {
                //Swallow exception
            }
        }

        private async Task GetImageFromCameraAsync()
        {
            try
            {
                CameraCaptureUI cameraUI = new CameraCaptureUI();
                cameraUI.PhotoSettings.AllowCropping = false;
                cameraUI.PhotoSettings.MaxResolution = CameraCaptureUIMaxPhotoResolution.HighestAvailable;
                cameraUI.PhotoSettings.AllowCropping = true;

                StorageFile capturedMedia = await cameraUI.CaptureFileAsync(CameraCaptureUIMode.Photo);
                if (capturedMedia != null)
                {
                    _filesPicked.Clear();
                    _filesPicked.Add(ImageCreationType.Camera, capturedMedia);

                    await AddFileToBeShownAsync(capturedMedia);
                    RaisePropertyChanged("IsInputValid");
                }
            }
            finally
            {
            }
        }

        private async Task PickImageFromFileAsync(object parameter) => await PickImageFromFileSystemAsync();

        private async Task PickImageFromCameraAsync(object sender) => await GetImageFromCameraAsync();

        #endregion

        #region Public Methods
        /// <summary>
        /// Create a PDF Document and invokes 
        /// NewDocumentCreated which passes the StorageFile document as the parameter
        /// </summary>
        private async Task CreateDocumentAsync(object sender)
        {
            bool success = false;
            StorageFile savedFile = null;
            PDFDoc doc = null;
            try
            {
                if(_DocumentType == CreateDocFrom.Blank)
                    doc = await GetBlankPDFDocAsync();

                else if(_DocumentType == CreateDocFrom.Image)
                {
                    doc = await GetImagePDFDOC();
                }

                if(doc != null)
                {
                    savedFile = await DocumentManager.Instance.SaveToTemporaryFileAsync(doc, GetFileNameWithExtension());
                    doc.GetDocInfo().SetTitle(GetFileName());

                    if (doc != null && savedFile != null)
                        success = true;
                }
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }

            if (savedFile != null)
            {
                if (success)
                {
                    _NewDocumentCreated?.Invoke(savedFile);
                }
                else
                {
                    _documentCreationFailed?.Invoke();
                }
            }
            else
            {
                _documentCreationFailed?.Invoke();
            }
        }
        #endregion

        #region Init
        private void InitCollections()
        {
            AvailablePaperSizes = new ObservableCollection<PaperSizeHolder>();
            AvailablePaperTypes = new ObservableCollection<PaperTypeHolder>();

            InitPapers();
            DocumentTitle = GetFileName();
        }

        private void InitPapers()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            PaperSizes regionSize = GetRegionDefaultPageSize();
            double convRate = _inConversionRate;
            string units = loader.GetString("DocumentCreation_AllDocuments_UnitsInch");
            if (regionSize == PaperSizes.A4)
            {
                convRate = _mmConversionRate;
                units = loader.GetString("DocumentCreation_AllDocuments_UnitsMilliMeter");
            }

            foreach (PaperSizes size in (PaperSizes[])Enum.GetValues(typeof(PaperSizes)))
            {
                GetPaperDimensionsFromSize(size, out double height, out double width);
                PaperSizeHolder psh = new PaperSizeHolder(size, width * convRate, height * convRate, units);
                AvailablePaperSizes.Add(psh);
                if (size == regionSize)
                    SelectedPaperSize = psh;
            }

            foreach (PaperType type in (PaperType[])Enum.GetValues(typeof(PaperType)))
            {
                AvailablePaperTypes.Add(new PaperTypeHolder(type));
            }

            SelectedPaperType = AvailablePaperTypes[0]; // Default is Blank Page
        }

        private void GetPaperDimensionsFromSize(PaperSizes size, out double width, out double height)
        {
            width = PaperSizeMap[size].Item1;
            height = PaperSizeMap[size].Item2;
        }

        private Dictionary<PaperSizes, Tuple<double, double>> PaperSizeMap
        {
            get
            {
                return _PaperSizeMap ?? (_PaperSizeMap = new Dictionary<PaperSizes, Tuple<double, double>>
                {
                    [PaperSizes.Legal] = new Tuple<double, double>(Legal_Width, Legal_Height),
                    [PaperSizes.Letter] = new Tuple<double, double>(Letter_Width, Letter_Height),
                    [PaperSizes.A0] = new Tuple<double, double>(A0_Width, A0_Height),
                    [PaperSizes.A1] = new Tuple<double, double>(A1_Width, A1_Height),
                    [PaperSizes.A2] = new Tuple<double, double>(A2_Width, A2_Height),
                    [PaperSizes.A3] = new Tuple<double, double>(A3_Width, A3_Height),
                    [PaperSizes.A4] = new Tuple<double, double>(A4_Width, A4_Height),
                    [PaperSizes.A5] = new Tuple<double, double>(A5_Width, A5_Height),
                    [PaperSizes.B0] = new Tuple<double, double>(B0_Width, B0_Height),
                    [PaperSizes.B1] = new Tuple<double, double>(B1_Width, B1_Height),
                    [PaperSizes.B2] = new Tuple<double, double>(B2_Width, B2_Height),
                    [PaperSizes.B3] = new Tuple<double, double>(B3_Width, B3_Height),
                    [PaperSizes.B4] = new Tuple<double, double>(B4_Width, B4_Height),
                    [PaperSizes.B5] = new Tuple<double, double>(B5_Width, B5_Height)
                });
            }
        }

        private PaperSizes GetRegionDefaultPageSize()
        {
            PaperSizes paperSize = PaperSizes.A4;

            string language = string.Empty;
            if (GlobalizationPreferences.Languages.Count > 0)
                language = GlobalizationPreferences.Languages[0];

            if (!string.IsNullOrWhiteSpace(language) && language.Length > 2 && letterSizeCultures.Contains(language.Substring(language.Length - 2).ToUpper()))
                paperSize = PaperSizes.Letter;

            return paperSize;
        }
        #endregion

        #region Events
        public delegate void DocumentCreationFailedDelegate();
        private DocumentCreationFailedDelegate _documentCreationFailed;

        /// <summary>
        /// Raised when a document creation attempt fails
        /// </summary>
        public event DocumentCreationFailedDelegate DocumentCreationFailed
        {
            add
            {
                if (_documentCreationFailed?.GetInvocationList().Contains(value) != true)
                    _documentCreationFailed += value;
            }
            remove => _documentCreationFailed -= value;
        }

        private NewDocumentCreatedDelegate _NewDocumentCreated;
        public delegate void NewDocumentCreatedDelegate(StorageFile file);

        /// <summary>
        /// Raised when a document is successfully created
        /// </summary>
        public event NewDocumentCreatedDelegate NewDocumentCreated
        {
            add
            {
                if (_NewDocumentCreated?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _NewDocumentCreated += value;
            }
            remove => _NewDocumentCreated -= value;
        }
        #endregion
    }
}