﻿using pdftron.Common;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.UI.Xaml.Media.Imaging;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class RasterizerFactory
    {
        private BitmapAvailableDelegate _bitmapAvailable;
        public delegate void BitmapAvailableDelegate(int pageNumber);
        public event BitmapAvailableDelegate BitmapAvailable
        {
            add
            {
                if (_bitmapAvailable?.GetInvocationList().Contains(value) != true)
                    _bitmapAvailable += value;
            }
            remove => _bitmapAvailable -= value;
        }

        private PDFDoc _Doc;

        private Dictionary<int, ImageProperties> _Images;
        private int _TotalPages;
        private int _CurrentPage;

        private int _CurrentJobPage;
        private PDFRasterizer _Rasterizer;
        private bool _IsWorking = false;

        private const int MAX_RENDERED_IMAGES_PER_DIRECTION = 4;

        private bool _Destroyed = false;

        public RasterizerFactory(PDFDoc doc, int totalPages)
        {
            _Doc = doc;
            _TotalPages = totalPages;
            _CurrentPage = 1;
            _CurrentJobPage = 0;

            _Images = new Dictionary<int, ImageProperties>();
            _Rasterizer = new PDFRasterizer();
            _Rasterizer.SetDrawAnnotations(true);
        }

        public void ShutDown()
        {
            _Destroyed = true;
            _Rasterizer.Cancel = true;
            _Images.Clear();
        }

        public void SetColorPostProcessing(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color? background, Windows.UI.Color? foreground)
        {
            if (mode == PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map && background != null && foreground != null)
            {
                _Rasterizer.SetColorPostProcessColors(background.Value, foreground.Value);
                _Rasterizer.SetColorPostProcessColors(foreground.Value, background.Value);
            }
            else
            {
                _Rasterizer.SetColorPostProcessMode(mode);
            }
            _Rasterizer.SetBackgroundColor(Windows.UI.Colors.White);
        }

        public bool HasImage(int pageNumber)
        {
            if (_Images.ContainsKey(pageNumber) && _Images[pageNumber].Bitmap != null)
            {
                return true;
            }
            return false;
        }

        public WriteableBitmap GetImage(int pageNumber)
        {
            if (HasImage(pageNumber))
            {
                return _Images[pageNumber].Bitmap;
            }
            return null;
        }

        public void SwitchToPage(int pageNumber)
        {
            if (!HasImage(pageNumber) && _CurrentJobPage != pageNumber)
            {
                _Rasterizer.Cancel = true;
            }
            _CurrentPage = pageNumber;
            StartWork();
        }

        private async void StartWork()
        {
            if (_IsWorking)
            {
                return;
            }
            _IsWorking = true;
            while (true)
            {
                if (_Destroyed)
                {
                    return;
                }
                try
                {
                    _CurrentJobPage = GetNextJobPage();
                    string pagesWithImages = "";
                    foreach (int key in _Images.Keys)
                    {
                        pagesWithImages += " " + key;
                    }

                    if (_CurrentJobPage >= 0 && _CurrentJobPage <= _TotalPages)
                    {
                        _Rasterizer.Cancel = false;
                        ImageProperties props = await GetBitMapFromPageAsync(_CurrentJobPage, _Doc);
                        if (props != null && !_Destroyed)
                        {
                            Windows.UI.Xaml.Media.Imaging.WriteableBitmap wb = new Windows.UI.Xaml.Media.Imaging.WriteableBitmap(props.Width, props.Height);
                            System.IO.Stream pixelStream = wb.PixelBuffer.AsStream();
                            await Task<WriteableBitmap>.Run(() =>
                            {
                                pixelStream.Seek(0, System.IO.SeekOrigin.Begin);
                                pixelStream.Write(props.ImageBuffer.Value, 0, props.ImageBuffer.Value.Length);
                            });
                            if (!_Destroyed)
                            {
                                props.Bitmap = wb;
                                _Images[_CurrentJobPage] = props;
                                RemoveFurthestBitmapIfNeeded(_CurrentPage);
                                _bitmapAvailable?.Invoke(_CurrentJobPage);
                            }
                        }
                    }
                    else
                    {
                        _IsWorking = false;
                        return;
                    }
                }
                catch (Exception e)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }
                finally
                {

                }
            }
        }

        private int GetNextJobPage()
        {
            int nextJobPage = _CurrentPage;
            while (Math.Abs(nextJobPage - _CurrentPage) <= MAX_RENDERED_IMAGES_PER_DIRECTION)
            {
                if (nextJobPage < 1 || nextJobPage > _TotalPages || _Images.ContainsKey(nextJobPage))
                {
                    nextJobPage = _CurrentPage + (_CurrentPage - nextJobPage);
                    if (nextJobPage >= _CurrentPage)
                    {
                        ++nextJobPage;
                    }
                }
                else
                {
                    return nextJobPage;
                }
            }
            return -1;
        }

        private void RemoveFurthestBitmapIfNeeded(int pageNumber)
        {
            if (_Images.Count > MAX_RENDERED_IMAGES_PER_DIRECTION * 2 + 1)
            {
                int furthestKey = pageNumber;
                foreach (int key in _Images.Keys)
                {
                    int dist = Math.Abs(key - pageNumber);
                    if (dist > MAX_RENDERED_IMAGES_PER_DIRECTION
                        && dist > Math.Abs(furthestKey - pageNumber))
                    {
                        furthestKey = key;
                    }
                }
                if (furthestKey != pageNumber)
                {
                    _Images.Remove(furthestKey);
                }
            }
        }

        private Windows.Foundation.IAsyncOperation<ImageProperties> GetBitMapFromPageAsync(int pageNumber, PDFDoc doc)
        {
            Task<ImageProperties> t = new Task<ImageProperties>(() =>
            {
                return GetBitMapFromPage(pageNumber, doc);
            });
            t.Start();
            return t.AsAsyncOperation<ImageProperties>();
        }

        private ImageProperties GetBitMapFromPage(int pageNum, PDFDoc pdfDoc)
        {
            try
            {
                ImageProperties props = new ImageProperties();

                pdfDoc.LockRead();
                Page page = pdfDoc.GetPage(pageNum);

                double scaleFactor = 1;
                Rect bbox = page.GetBox(PageBox.e_crop);
                bbox.Normalize();

                double pageWidth = page.GetPageWidth();
                double pageHeight = page.GetPageHeight();

                double pageScaleFactorX = 1;
                double pageScaleFactorY = 1;

                scaleFactor *= Math.Max(pageScaleFactorX, pageScaleFactorY);

                if (scaleFactor <= 0)
                {
                    return null;
                }

                Matrix2D mtx = new Matrix2D(scaleFactor, 0, 0, scaleFactor, 0, 0);
                mtx = Matrix2D.Mult(mtx, page.GetDefaultMatrix(true, PageBox.e_crop, PageRotate.e_0));

                int width = (int)(pageWidth * scaleFactor);
                int height = (int)(pageHeight * scaleFactor);

                int comps = 4; // BGRA
                int stride = ((width * comps + 3) / 4) * 4;

                ByteArrayRef img = new ByteArrayRef(stride * height);

                _Rasterizer.Rasterize(page, img, width, height, stride, comps, false, mtx, bbox);
                if (_Rasterizer.Cancel)
                {
                    return null;
                }
                props.Width = width;
                props.Height = height;
                props.ImageBuffer = img;

                return props;
            }
            catch (Exception e)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
            }

            finally
            {
                pdfDoc.UnlockRead();
            }

            return null;
        }
    }
}
