﻿using pdftron.Common;
using pdftron.Filters;
using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;
using System;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.Foundation;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Popups;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Controls.ViewModels.FileOpening
{
    public class DocumentOpener : ViewModelBase
    {
        private bool _openingNonePDF;
        private ConvertCanceller _currentConvertCanceller;
        private bool _isModal;
        private bool _isInputDisabled;
        private bool _isFileOpeningProgressBarVisible;
        private bool _isPasswordDialogOpen;
        private bool _isPackageDialogOpen;
        private bool _isConvertingDocument;
        private FileOpeningCanceler _currentFileOpeningCanceler;
        private NewDocumentProperties _selectedFileProperties;
        private PasswordViewModel _passwordViewModel;
        private string _currentPassword = string.Empty;
        private bool _isIncorrectPasswordNotificationVisible;
        private PdfPackageViewModel _packageViewModel;

        /// <summary>
        /// Raised when a document is ready to be opened.
        /// </summary>
        private DocumentReadyDelegate _documentReady;
        public delegate void DocumentReadyDelegate(NewDocumentProperties selectedFileProperties);
        public event DocumentReadyDelegate DocumentReady
        {
            add
            {
                if (_documentReady?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _documentReady += value;
            }
            remove
            {
                if (_documentReady != null)
                {
                    DocumentReadyDelegate eventMethod = (DocumentReadyDelegate)Array.Find(
                        _documentReady.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _documentReady -= eventMethod;
                }
            }
        }

        /// <summary>
        /// Raised when a document couldn't be opened.
        /// </summary>
        private EventHandler<RoutedEventArgs> _documentLoadingFailed;
        public event EventHandler<RoutedEventArgs> DocumentLoadingFailed
        {
            add
            {
                if (_documentLoadingFailed?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _documentLoadingFailed += value;
            }
            remove
            {
                if (_documentLoadingFailed != null)
                {
                    EventHandler<RoutedEventArgs> eventMethod = (EventHandler<RoutedEventArgs>)Array.Find(
                        _documentLoadingFailed.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _documentLoadingFailed -= eventMethod;
                }
            }
        }

        /// <summary>
        /// Raised when the modality is changed.
        /// </summary>
        private ModalityChangedDelegate _modalityChanged;
        public delegate void ModalityChangedDelegate(bool isModal);
        public event ModalityChangedDelegate ModalityChanged
        {
            add
            {
                if (_modalityChanged?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _modalityChanged += value;
            }
            remove
            {
                if (_modalityChanged != null)
                {
                    ModalityChangedDelegate eventMethod = (ModalityChangedDelegate)Array.Find(
                        _modalityChanged.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _modalityChanged -= eventMethod;
                }
            }
        }

        public DocumentOpener()
        {
            InitCommands();
            CancelConversionCommandImpl(null);
        }

        /// <summary>
        /// Cancel any current document current in phase of opening
        /// </summary>
        public void CancelDocumentOpening()
        {
            CancelCurrentDocLoad();
            CancelConversionCommandImpl(null);
        }

        public async Task OpenAsync(StorageFile file, NewDocumentProperties documentProperties = null)
        {
            await OpenInternalAsync(file, documentProperties);
        }

        // gets or sets whether the current dialog is modal or not.
        public bool IsModal
        {
            get => _isModal;
            set
            {
                if (Set(ref _isModal, value))
                    _modalityChanged?.Invoke(_isModal);
            }
        }

        public bool IsInputDisabled
        {
            get => _isInputDisabled;
            set => Set(ref _isInputDisabled, value);
        }

        public bool IsFileOpeningProgressBarVisible
        {
            get => _isFileOpeningProgressBarVisible;
            set => Set(ref _isFileOpeningProgressBarVisible, value);
        }

        public bool IsPasswordDialogOpen
        {
            get => _isPasswordDialogOpen;
            set
            {
                Set(ref _isPasswordDialogOpen, value);
                IsModal = value;
            }
        }

        public bool IsPackageDialogOpen
        {
            get => _isPackageDialogOpen;
            set
            {
                Set(ref _isPackageDialogOpen, value);
                IsModal = value;
            }
        }

        /// <summary>
        /// Gets or sets whether a document is currently being converted. Setting this to false when it was true will cancel conversion.
        /// </summary>
        public bool IsConvertingDocument
        {
            get => _isConvertingDocument;
            set
            {
                if (value != _isConvertingDocument)
                {
                    _isConvertingDocument = value;
                    RaisePropertyChanged();
                    if (!_isConvertingDocument)
                        CancelCurrentConversion();

                    IsModal = value;
                }
            }
        }

        private void InitCommands()
        {
            PackageDialogBackCommand = new RelayCommand(PackageDialogBackCommandImpl);
            CancelConversionCommand = new RelayCommand(CancelConversionCommandImpl);
        }

        public RelayCommand PackageDialogBackCommand { get; private set; }

        public RelayCommand CancelConversionCommand { get; private set; }

        private void PackageDialogBackCommandImpl(object sender)
        {
            PackageDialogGoBack(); 
        }

        private void CancelConversionCommandImpl(object sender)
        {
            IsConvertingDocument = false;
        }

        private async Task OpenInternalAsync(StorageFile file, NewDocumentProperties documentProperties = null)
        {
            if (file == null)
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog
                {
                    Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                    Content = loader.GetString("DocumentsPage_FileOpeningError_FailedToCreateTempFile"),
                    CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);

                RaiseDocumentLoadingFailed();
                return;
            }

            if (IsAlreadyOpeningfile(file))
                return;

            if (!file.FileType.Equals(".pdf", StringComparison.OrdinalIgnoreCase))
            {
                if (file.FileType.Equals(".oxps", StringComparison.OrdinalIgnoreCase)
                    || file.FileType.Equals(".xps", StringComparison.OrdinalIgnoreCase)
                    || file.FileType.Equals(".txt", StringComparison.OrdinalIgnoreCase)
                    || file.FileType.Equals(".xml", StringComparison.OrdinalIgnoreCase)
                    || file.FileType.Equals(".md", StringComparison.OrdinalIgnoreCase))
                {
                    await OpenNonPdfPreConvertAsync(file);
                }
                else
                {
                    await OpenNonPdfUniversalConversionAsync(file);
                }

                return;
            }

            _selectedFileProperties = documentProperties ?? new NewDocumentProperties
            {
                OpenedDocumentState = OpenedDocumentStates.Normal
            };

            if (_selectedFileProperties.OpenedDocumentState == OpenedDocumentStates.None)
                _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.Normal;

            CancelCurrentDocLoad();
            FileOpeningCanceler canceler = new FileOpeningCanceler
            {
                File = file
            };

            _currentFileOpeningCanceler = canceler;
            StartProgressBarAfterWait(canceler);

            if ((file.Attributes & FileAttributes.Temporary) == FileAttributes.Temporary
                || ((file.Attributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly && !_selectedFileProperties.OpenedThroughDrop))
            {
                _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.ReadOnly;
                StorageFile temporaryFile = await UtilityFunctions.GetTemporarySaveFileAsync(file.DisplayName);
                if (temporaryFile == null)
                {
                    ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                    ContentDialog dialog = new ContentDialog 
                    {
                        Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                        Content = loader.GetString("DocumentsPage_FileOpeningError_FailedToCreateTempFile"),
                        CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                    };

                    await DialogHelper.ShowContentDialogAsync(dialog);
                    RaiseDocumentLoadingFailed();
                    return;
                }

                await file.CopyAndReplaceAsync(temporaryFile);
                file = temporaryFile;
                await UtilityFunctions.RemoveReadOnlyAndTemporaryFlagsAsync(file);
            }
            else if (_selectedFileProperties.OpenedThroughDrop)
            {
                _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.ReadOnly;
            }

            if (file != null)
            {
                _selectedFileProperties.File = file;
                _selectedFileProperties.Password = null;
                _selectedFileProperties.MRUToken = null;

                if (!_selectedFileProperties.RecentPropertiesRestored &&
                    _selectedFileProperties.OpenedDocumentState == OpenedDocumentStates.Normal
                    && RecentItemsData.IsReady)
                {
                    RecentDocumentProperties recentProps = await RecentItemsData.Instance.GetPropertiesForFileIfInList(file);
                    if (recentProps != null)
                        _selectedFileProperties.ApplyRecentProperties(recentProps);
                }

                if (!DocumentManager.IsReady)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        "DocumentOpener invoked without DocumentManager ready");
                    await DocumentManager.Instance.InitializeInstanceAsync();
                }

                StorageFile tempFile = await DocumentManager.Instance.OpenTemporaryCopyAsync(file);
                if (tempFile == null)
                {
                    _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.Uneditable;
                    tempFile = file;
                }

                _selectedFileProperties.TemporaryFile = tempFile;

                Windows.Storage.FileProperties.BasicProperties basicProps = await file.GetBasicPropertiesAsync();
                _selectedFileProperties.LastModifiedDate = basicProps.DateModified;
                _selectedFileProperties.Filesize = basicProps.Size;

                bool success = true;
                bool isUsedByOtherProgram = false;
                bool pdfDocCreatedSuccessfully = false;

                PDFDoc doc = null;
                try
                {
                    string path = file.Path;

                    bool isBox = UtilityFunctions.IsBox(path);
                    bool isDropBox = UtilityFunctions.IsDropBox(path);
                    bool isOneDrive = UtilityFunctions.IsOneDrive(path);
                    bool isOneDriveTemp = UtilityFunctions.IsOneDriveTempFile(file.Path);

                    if (_selectedFileProperties.OpenedDocumentState != OpenedDocumentStates.Uneditable)
                    {
                        if (isDropBox || isOneDriveTemp)
                            _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.ReadOnly;

                        if (isBox)
                            _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.ReadOnly;
                    }

                    if (canceler.Cancel)
                        return;

                    try
                    {
                        isUsedByOtherProgram = true;
                        try
                        {
                            using (IRandomAccessStream iras = await file.OpenReadAsync())
                            {
                                isUsedByOtherProgram = false;
                            }
                        }
                        catch (Exception ex)
                        {
                            AppCenterHandlerBase.CURRENT.TrackError(ex);
                        }

                        try
                        {
                            doc = await GetPDFDocAsync(tempFile);
                        }
                        catch (Exception e)
                        {
                            PDFNetException pdfNetEx = new PDFNetException(e.HResult);
                            if (pdfNetEx.IsPDFNetException)
                                Debug.WriteLine("Error creating PDFDoc: " + pdfNetEx);

                            AppCenterHandlerBase.CURRENT.TrackError(e);
                        }

                        if (doc != null)
                        {
                            pdfDocCreatedSuccessfully = true;

                            if (doc.IsModified() && _selectedFileProperties.OpenedDocumentState != OpenedDocumentStates.Uneditable)
                                _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.Corrupted;
                        }
                    }
                    catch (Exception e)
                    {
                        success = false;
                        Debug.WriteLine("Failed to create PDFDoc:\n" + e.ToString());
                    }

                    if (pdfDocCreatedSuccessfully)
                    {
                        _selectedFileProperties.Doc = doc;
                        if (canceler.Cancel)
                            return;

                        success = true;
                        _selectedFileProperties.File = file;
                    }
                }
                catch (Exception)
                {
                    success = false;
                }

                bool corrupt = false;
                if (pdfDocCreatedSuccessfully)
                {
                    try
                    {
                        if (!doc.InitStdSecurityHandler(string.Empty))
                        {
                            AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, "Document Opened Encrypted");

                            CancelCurrentDocLoad();
                            HandlePassword();
                            return;
                        }
                    }
                    catch (Exception)
                    {
                        corrupt = true;
                    }
                }
                else if (!pdfDocCreatedSuccessfully)
                {
                    CancelCurrentDocLoad();
                    IsModal = false;

                    if (isUsedByOtherProgram)
                    {
                        CancelCurrentDocLoad();
                        ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                        ContentDialog dialog = new ContentDialog
                        {
                            Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                            Content = loader.GetString("DocumentsPage_FileOpeningError_UsedByAnother_Info"),
                            CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);
                        RaiseDocumentLoadingFailed();
                        return;
                    }
                    else
                    {
                        CancelCurrentDocLoad();
                        ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                        ContentDialog dialog = new ContentDialog
                        {
                            Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                            Content = string.Format(loader.GetString("DocumentsPage_FileOpeningError_CorruptFile_Info"), ViewerSettings.DisplayName),
                            CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);
                        RaiseDocumentLoadingFailed();
                        return;
                    }
                }
                if (corrupt)
                {
                    IsModal = false;
                    CancelCurrentDocLoad();
                    ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                    ContentDialog dialog = new ContentDialog
                    {
                        Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                        Content = string.Format(loader.GetString("DocumentsPage_FileOpeningError_CorruptFile_Info"), ViewerSettings.DisplayName),
                        CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                    };

                    await DialogHelper.ShowContentDialogAsync(dialog);
                    RaiseDocumentLoadingFailed();
                    return;
                }

                if (success && !corrupt)
                {
                    PDFNet.EnableJavaScript(ViewerSettings.EnableJavaScript);
                    if (IsXFA())
                    {
                        await OpenXFAAsync();
                    }
                    else if (IsPackage())
                    {
                        await OpenPackageAsync();
                    }
                    else
                    {
                        await CheckIntegrityAsync();
                    }
                }
                else
                {
                    RaiseDocumentLoadingFailed();
                }
            }
        }

        private IAsyncOperation<PDFDoc> GetPDFDocAsync(IRandomAccessStream stream, string path)
        {
            Task<PDFDoc> t = new Task<PDFDoc>(() =>
                {
                    return new PDFDoc(stream, path);
                });

            t.Start();

            return t.AsAsyncOperation();
        }

        private IAsyncOperation<PDFDoc> GetPDFDocAsync(StorageFile file)
        {
            Task<PDFDoc> t = new Task<PDFDoc>(() =>
                {
                    return new PDFDoc(file);
                });

            t.Start();

            return t.AsAsyncOperation();
        }

        private async void TemporaryilyBlockInput()
        {
            try
            {
                IsInputDisabled = true;
                await Task.Delay(1000);
            }
            finally
            {
                IsInputDisabled = false;
            }
        }

        private void StartProgressBarAfterWait(FileOpeningCanceler canceler)
        {
            if (!canceler.Cancel)
                IsFileOpeningProgressBarVisible = true;
        }

        private void CloseProgressBar() => IsFileOpeningProgressBarVisible = false;

        private bool IsAlreadyOpeningfile(StorageFile file)
        {
            if (_currentFileOpeningCanceler == null)
                return false;

            if (_currentFileOpeningCanceler.File != null)
                return _currentFileOpeningCanceler.File.Equals(file);

            return false;
        }

        private void CancelCurrentDocLoad()
        {
            if (_currentFileOpeningCanceler != null)
            {
                _currentFileOpeningCanceler.Cancel = true;
                _currentFileOpeningCanceler.File = null;
            }

            CloseProgressBar();
        }

        private void RaiseDocumentReady()
        {
            if (_documentReady == null)
                return;

            _documentReady.Invoke(_selectedFileProperties);
        }

        private void RaiseDocumentLoadingFailed() => _documentLoadingFailed?.Invoke(this, new RoutedEventArgs());

        public PasswordViewModel PasswordViewModel
        {
            get => _passwordViewModel;
            set => Set(ref _passwordViewModel, value);
        }

        public string CurrentPassword
        {
            get => _currentPassword;
            set
            {
                if (Set(ref _currentPassword, value))
                {
                    RaisePropertyChanged(nameof(HasPasswordBoxGotContent));
                    IsIncorrectPasswordNotificationVisible = false;
                }
            }
        }

        public bool IsIncorrectPasswordNotificationVisible
        {
            get => _isIncorrectPasswordNotificationVisible;
            set
            {
                Set(ref _isIncorrectPasswordNotificationVisible, value);
                RaisePropertyChanged(nameof(HasPasswordBoxGotContent));
            }
        }

        public bool HasPasswordBoxGotContent => !string.IsNullOrEmpty(CurrentPassword) && CurrentPassword.Length > 0 && !IsIncorrectPasswordNotificationVisible;

        private void HandlePassword()
        {
            IsPasswordDialogOpen = true;
            PasswordViewModel = new PasswordViewModel(_selectedFileProperties.Doc);
            PasswordViewModel.PasswordHandled += PasswordViewModel_PasswordHandled;
        }

        private async void PasswordViewModel_PasswordHandled(bool success, string password)
        {
            IsModal = false;
            if (success)
            {
                // check for XFA or package
                if (IsXFA()) await OpenXFAAsync();
                if (IsPackage()) await OpenPackageAsync();

                _selectedFileProperties.Password = password;

                await CheckIntegrityAsync();
            }
            else
            {
                CancelCurrentDocLoad();
                _selectedFileProperties.Doc.Dispose();
                RaiseDocumentLoadingFailed();
            }

            // On Windows Phone 8.1, it's important that IsPasswordDialogOpen is set to false after CheckIntegrity.
            // IsPasswordDialogOpen = false will trigger modality to go away, causing the phone to reattach the AppBar
            // CheckIntegrity will trigger navigation, which for some reason breaks the app bar if it's just been attached.
            IsPasswordDialogOpen = false;
            PasswordViewModel = null;
        }

        public PdfPackageViewModel PackageViewModel
        {
            get => _packageViewModel;
            set => Set(ref _packageViewModel, value);
        }

        private bool IsPackage() => _selectedFileProperties.Doc.GetRoot().FindObj("Collection") != null;

        private async Task OpenPackageAsync()
        {
            Obj collectionObj = _selectedFileProperties.Doc.GetRoot().FindObj("Collection");
            if (collectionObj != null)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, "Document Opened Package");
                HandlePackage();
            }
            else
            {
                await CheckIntegrityAsync();
            }
        }

        private void HandlePackage()
        {
            CancelCurrentDocLoad();
            IsPackageDialogOpen = true;
            _selectedFileProperties.OpenedDocumentState = OpenedDocumentStates.ReadOnly;
            PdfPackageViewModel vm = new PdfPackageViewModel(_selectedFileProperties.Doc, _selectedFileProperties.File.Name);
            vm.FileSelcted += PackageDialog_FileSelcted;
            PackageViewModel = vm;
        }

        private async void PackageDialog_FileSelcted(PDFDoc doc, StorageFile file, bool err, string errorMessage)
        {
            if (doc != null)
            {
                _selectedFileProperties.Doc = doc;
                _selectedFileProperties.File = file;
                _selectedFileProperties.TemporaryFile = await DocumentManager.Instance.OpenTemporaryCopyAsync(file);
                if (!doc.InitStdSecurityHandler(""))
                {
                    HandlePassword();
                }
                else
                {
                    await CheckIntegrityAsync();
                }
            }
            else if (err)
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog
                {
                    Title = loader.GetString("PackageDialog_FileExtractionError_Title"),
                    Content = errorMessage,
                    CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                RaiseDocumentLoadingFailed();
            }

            // On Windows Phone 8.1, it's important that IsPackageDialogOpen is set to false after CheckIntegrity.
            // IsPackageDialogOpen = false will trigger modality to go away, causing the phone to reattach the AppBar
            // CheckIntegrity will trigger navigation, which for some reason breaks the app bar if it's just been attached.
            PackageViewModel = null;
            IsPackageDialogOpen = false;
        }

        private void PackageDialogGoBack()
        {
            IsPackageDialogOpen = false;
            RaiseDocumentLoadingFailed();
        }

        private bool IsXFA()
        {
            Obj needsRenderingObj = _selectedFileProperties.Doc.GetRoot().FindObj("NeedsRendering");
            return needsRenderingObj?.IsBool() == true && needsRenderingObj.GetBool();
        }

        private async Task OpenXFAAsync()
        {
            AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, "Document Opened XFA");
            CancelCurrentDocLoad();
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            ContentDialog dialog = new ContentDialog
            {
                Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                Content = loader.GetString("DocumentsPage_FileOpeningError_XFA_Info"),
                CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
            };

            await DialogHelper.ShowContentDialogAsync(dialog);

            // TODO Phone
            //IsModal = false;

            RaiseDocumentLoadingFailed();
        }

        private async Task CheckIntegrityAsync()
        {
            CancelCurrentDocLoad();
            if (_selectedFileProperties.Doc != null)
            {
                try
                {
                    _selectedFileProperties.Doc.LockRead();
                    int pageCount = _selectedFileProperties.Doc.GetPageCount();
                    if (pageCount <= 0)
                    {
                        ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                        ContentDialog dialog = new ContentDialog
                        {
                            Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                            Content = loader.GetString("DocumentsPage_FileOpeningError_NoPages_Info"),
                            CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);
                        RaiseDocumentLoadingFailed();
                        return;
                    }
                }
                finally
                {
                    _selectedFileProperties.Doc.UnlockRead();
                }

                RaiseDocumentReady();
            }
        }

        public async Task OpenNonPdfPreConvertAsync(StorageFile file)
        {
            if (_openingNonePDF)
                return;

            try
            {
                _openingNonePDF = true;
                IsConvertingDocument = true;
                ConvertCanceller canceller = new ConvertCanceller();
                _currentConvertCanceller = canceller;
                StorageFile originalFile = file;

                if (file.FileType == ".xps" || file.FileType == ".oxps")
                {
                    StorageFile tempFile = null;
                    if ((file.Attributes & FileAttributes.Temporary) == FileAttributes.Temporary)
                    {
                        Guid guid = Guid.NewGuid();
                        string tempName = guid.ToString();
                        tempFile = await file.CopyAsync(ApplicationData.Current.TemporaryFolder, tempName);
                        file = tempFile;
                    }

                    PDFDoc doc = null;
                    try
                    {
                        doc = await OpenFromXPSAsync(file);
                    }
                    catch (Exception e)
                    {
                        AnalyticsHandlerBase.CURRENT.SendException(e);
                    }

                    if (canceller.Cancel)
                    {
                        Debug.WriteLine("Cancelled XPS Conversion");
                        return;
                    }

                    if (tempFile != null)
                    {
                        try
                        {
                            await tempFile.DeleteAsync();
                        }
                        catch (Exception e)
                        {
                            AnalyticsHandlerBase.CURRENT.SendException(e);
                        }
                    }

                    if (doc != null)
                    {
                        doc.InitStdSecurityHandler(string.Empty); // we know there is no password

                        StorageFile tempPDFFile = null;
                        try
                        {
                            if (!DocumentManager.IsReady)
                            {
                                AnalyticsHandlerBase.CURRENT.SendEvent(
                                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                                    "DocumentOpener invoked without DocumentManager ready");
                                await DocumentManager.Instance.InitializeInstanceAsync().ConfigureAwait(false);
                            }

                            tempPDFFile = await DocumentManager.Instance.OpenTemporaryCopyAsync(file, false);
                        }
                        catch
                        {
                            //Swallow exception
                        }

                        if (tempPDFFile == null)
                            tempPDFFile = await ApplicationData.Current.TemporaryFolder.CreateFileAsync("FromXPS.pdf", CreationCollisionOption.GenerateUniqueName);

                        if (tempPDFFile != null)
                            await doc.SaveToNewLocationAsync(tempPDFFile, SDFDocSaveOptions.e_remove_unused);

                        if (canceller.Cancel)
                        {
                            Debug.WriteLine("Cancelled XPS Conversion");
                            return;
                        }

                        if (tempPDFFile != null)
                        {
                            _selectedFileProperties = new NewDocumentProperties
                            {
                                OpenedDocumentState = OpenedDocumentStates.NonePDF,
                                Doc = doc,
                                File = file,
                                TemporaryFile = tempPDFFile,
                                OriginalFileFromSystem = originalFile
                            };

                            await CheckIntegrityAsync();
                        }
                    }
                    else
                    {
                        RaiseDocumentLoadingFailed();
                        try
                        {
                            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                            ContentDialog dialog = new ContentDialog
                            {
                                Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                                Content = string.Format(loader.GetString("DocumentsPage_FileOpeningError_ConversionFailed_Content"), ViewerSettings.DisplayName),
                                CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                            };

                            await DialogHelper.ShowContentDialogAsync(dialog);
                        }
                        catch (Exception e)
                        {
                            AnalyticsHandlerBase.CURRENT.SendException(e);
                        }
                    }
                }
            }
            finally
            {
                _openingNonePDF = false;
                IsConvertingDocument = false;
            }
        }

        private async Task<PDFDoc> OpenFromXPSAsync(StorageFile file)
        {
            PDFDoc doc = new PDFDoc();
            using (IRandomAccessStream stream = await file.OpenReadAsync())
            {
                IFilter filter = new RandomAccessStreamFilter(stream);
                await pdftron.PDF.Convert.FromXpsAsync(doc, filter).AsTask().ConfigureAwait(false);
            }

            return doc;
        }

        private void CancelCurrentConversion()
        {
            if (_currentConvertCanceller != null)
            {
                _currentConvertCanceller.Cancel = true;
                _openingNonePDF = false;
                _currentConvertCanceller = null;
            }
        }

        private async Task OpenNonPdfUniversalConversionAsync(StorageFile file)
        {
            try
            {
                _selectedFileProperties = new NewDocumentProperties
                {
                    OpenedDocumentState = OpenedDocumentStates.Universal,
                    OriginalFileFromSystem = file
                };

                CancelCurrentDocLoad();
                FileOpeningCanceler canceler = new FileOpeningCanceler
                {
                    File = file
                };

                _currentFileOpeningCanceler = canceler;
                StartProgressBarAfterWait(canceler);

                bool mustCopy = (file.Attributes & FileAttributes.Temporary) == FileAttributes.Temporary || (file.Attributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly;
                mustCopy |= ViewerSettings.GetFileOpeningType(file) == FileOpeningType.UniversalConversionFromString;

                if (mustCopy)
                {
                    StorageFile temporaryFile = await UtilityFunctions.GetTemporarySaveFileAsync(file.Name);
                    if (canceler.Cancel)
                        return;

                    if (temporaryFile == null)
                    {
                        ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                        ContentDialog dialog = new ContentDialog
                        {
                            Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                            Content = loader.GetString("DocumentsPage_FileOpeningError_FailedToCreateTempFile"),
                            CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);

                        RaiseDocumentLoadingFailed();
                        return;
                    }

                    await file.CopyAndReplaceAsync(temporaryFile);
                    file = temporaryFile;
                    await UtilityFunctions.RemoveReadOnlyAndTemporaryFlagsAsync(file);
                    if (canceler.Cancel)
                        return;
                }

                if (ViewerSettings.GetFileOpeningType(file) != FileOpeningType.UniversalConversionFromString)
                    _selectedFileProperties.FileStream = await file.OpenReadAsync();

                _selectedFileProperties.File = file;
                RaiseDocumentReady();
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        public override bool GoBack()
        {
            if (IsPasswordDialogOpen)
            {
                PasswordViewModel_PasswordHandled(false, null); // mimic failing
                return true;
            }

            if (IsPackageDialogOpen && PackageViewModel?.GoBack() == true)
                return true;

            if (_currentFileOpeningCanceler != null)
            {
                _currentFileOpeningCanceler.Cancel = true;
                _currentFileOpeningCanceler = null;
                IsFileOpeningProgressBarVisible = false;
                IsModal = false;
            }

            return false;
        }
    }
}
