﻿using pdftron.Common;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using PDFRect = pdftron.PDF.Rect;
using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;
using UISize = Windows.Foundation.Size;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class MagnifyingControlViewModel : ViewModelBase
    {
        /// <summary>
        /// Represents the start position for when the view is first created
        /// </summary>
        public class StartPosition
        {
            /// <summary>
            /// The page to show.
            /// The default is -1, which means use the current page.
            /// </summary>
            public int StartPageNumber { get; set; }
            /// <summary>
            /// The Coordinate in page space that will be in the top left corner.
            /// The default is NaN, which will use the top left corner of the current view.
            /// </summary>
            public UIPoint StartPageTopLeftCorner { get; set; }
            /// <summary>
            /// The Zoom factor to start the view in.
            /// NaN will set it to 4x the Zoom of the MagnifyingPDFViewCtrl, and is the default
            /// </summary>
            public double StartZoomFactor { get; set; }

            public StartPosition()
            {
                StartPageNumber = -1;
                StartPageTopLeftCorner = new UIPoint(double.NaN, double.NaN);
                StartZoomFactor = double.NaN;
            }

            public StartPosition(int pageNumber, double xInPage, double yInPage, double zoom)
                : this(pageNumber, new UIPoint(xInPage, yInPage), zoom)
            {
            }

            public StartPosition(int pageNumber, UIPoint positionInPage, double zoom)
            {
                StartPageNumber = pageNumber;
                StartPageTopLeftCorner = positionInPage;
                StartZoomFactor = zoom;
            }
        }

        private StartPosition _StartPosition;

        public MagnifyingControlViewModel() : base()
        {
            _StartPosition = new StartPosition();
        }

        private void Init()
        {
            if (MagnifyingTool != null && MagnifyingPDFViewCtrl != null && MagnifiedPDFViewCtrl == null)
            {
                MagnifiedPDFViewCtrl = new PDFViewCtrl();
                MagnifiedToolManager = new ToolManager(MagnifiedPDFViewCtrl);
                MagnifiedToolManager.EnablePopupMenuOnLongPress = false;
                MagnifiedToolManager.PanToolTextSelectionMode = ToolManager.TextSelectionBehaviour.AlwaysPan;
                MagnifiedToolManager.CreateDefaultTool();
                MagnifiedToolManager.NewToolCreated += MagnifiedToolManager_NewToolCreated;
                MagnifiedToolManager.AnnotationAdded += MagnifiedToolManager_AnnotationAdded;
                MagnifiedToolManager.AnnotationEdited += MagnifiedToolManager_AnnotationEdited;
                MagnifiedToolManager.AnnotationPreRemoved += MagnifiedToolManager_AnnotationPreRemoved;
                MagnifiedToolManager.AnnotationRemoved += MagnifiedToolManager_AnnotationRemoved;
                MagnifiedToolManager.AnnotationGroupAdded += MagnifiedToolManager_AnnotationGroupAdded;
                MagnifiedToolManager.AnnotationGroupEdited += MagnifiedToolManager_AnnotationGroupEdited;
                MagnifiedToolManager.AnnotationGroupPreEdited += MagnifiedToolManager_AnnotationGroupPreEdited;
                MagnifiedToolManager.AnnotationGroupRemoved += MagnifiedToolManager_AnnotationGroupRemoved;
                MagnifyingTool.Opacity = 0.0;

                MagnifiedPDFViewCtrl.CanFlipPages = false;
                MagnifiedPDFViewCtrl.AnimateAllPageFlipping = false;
                MagnifiedPDFViewCtrl.PreviousAndNextButtonVisibility = Windows.UI.Xaml.Visibility.Collapsed;
                MagnifiedPDFViewCtrl.SetPagePresentationMode(MagnifyingPDFViewCtrl.GetPagePresentationMode());
                MagnifiedPDFViewCtrl.SetupThumbnails(false, true, true, 500, 100, 0.1);
                MagnifiedPDFViewCtrl.SetDoc(MagnifyingPDFViewCtrl.GetDoc());
            }
        }

        #region Methods

        public void SetPDFViewCtrl(PDFViewCtrl ctrl)
        {
            MagnifyingPDFViewCtrl = ctrl;
            Init();
        }

        public void SetMagnifyingTool(MagnifyingTool tool)
        {
            MagnifyingTool = tool;
            Init();
        }

        /// <summary>
        /// Sets the starting position for the view.
        /// Must be set before the PDFViewCtrl and MagnifyingTool are set.
        /// </summary>
        /// <param name="zoom"></param>
        public void SetStartPosition(StartPosition position)
        {
            _StartPosition = position;
        }

        public void CleanUp()
        {
            if (MagnifiedToolManager != null)
            {
                MagnifiedToolManager.NewToolCreated -= MagnifiedToolManager_NewToolCreated;
                MagnifiedToolManager.AnnotationAdded -= MagnifiedToolManager_AnnotationAdded;
                MagnifiedToolManager.AnnotationGroupAdded -= MagnifiedToolManager_AnnotationGroupAdded;
            }
            MagnifiedToolManager.Dispose();
            MagnifiedToolManager = null;
            MagnifyingTool = null;
            if (MagnifiedPDFViewCtrl != null)
            {
                PDFDoc doc = MagnifiedPDFViewCtrl.GetDoc();
                MagnifiedPDFViewCtrl = null;
            }
            MagnifyingPDFViewCtrl = null;
        }

        #endregion Methods


        #region Properties

        public PDFViewCtrl MagnifiedPDFViewCtrl
        {
            get { return _MagnifiedPDFViewCtrl; }
            private set
            {
                PDFViewCtrl oldView = _MagnifiedPDFViewCtrl;
                if (Set(ref _MagnifiedPDFViewCtrl, value))
                {
                    if (oldView != null)
                    {
                        oldView.CloseDoc();
                        oldView.OnSetDoc -= MagnifiedPDFViewCtrl_OnSetDoc;
                        oldView.SizeChanged -= MagnifiedPDFViewCtrl_SizeChanged;
                        oldView.OnScale -= MagnifiedPDFViewCtrl_OnScale;
                        oldView.OnViewChanged -= MagnifiedPDFViewCtrl_OnViewChanged;
                        oldView.OnPageNumberChanged -= MagnifiedPDFViewCtrl_OnPageNumberChanged;
                    }
                    if (_MagnifiedPDFViewCtrl != null)
                    {
                        MagnifiedPDFViewCtrl.OnSetDoc += MagnifiedPDFViewCtrl_OnSetDoc;
                        MagnifiedPDFViewCtrl.SizeChanged += MagnifiedPDFViewCtrl_SizeChanged;
                        MagnifiedPDFViewCtrl.OnScale += MagnifiedPDFViewCtrl_OnScale;
                        MagnifiedPDFViewCtrl.OnViewChanged += MagnifiedPDFViewCtrl_OnViewChanged;
                        MagnifiedPDFViewCtrl.OnPageNumberChanged += MagnifiedPDFViewCtrl_OnPageNumberChanged;
                    }
                }
            }
        }

        public ToolManager MagnifiedToolManager
        {
            get { return _MagnifiedToolManager; }
            set { _MagnifiedToolManager = value; }
        }

        private double _EffectiveHeight = double.NaN;
        public double EffectiveHeight
        {
            get { return _EffectiveHeight; }
            set
            {
                if (Set(ref _EffectiveHeight, value))
                {
                    if (MagnifiedPDFViewCtrl != null && MagnifyingPDFViewCtrl != null && !double.IsNaN(MagnifiedPDFViewCtrl.ActualHeight)
                        && !double.IsNaN(MagnifiedPDFViewCtrl.ActualWidth) && MagnifiedPDFViewCtrl.ActualWidth > 0)
                    {
                        UpdateMagnifyingToolSize(MagnifiedPDFViewCtrl.ActualWidth, MagnifiedPDFViewCtrl.ActualHeight, MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom());
                    }
                }
            }
        }

        #endregion Properties


        #region View Syncing

        private MagnifyingTool _MagnifyingTool;
        private PDFViewCtrl _MagnifyingPDFViewCtrl;

        private PDFViewCtrl _MagnifiedPDFViewCtrl;
        private ToolManager _MagnifiedToolManager;

        private bool IsInitialized
        {
            get { return _MagnifyingTool != null && _MagnifyingPDFViewCtrl != null && _MagnifiedPDFViewCtrl != null; }
        }

        private MagnifyingTool MagnifyingTool
        {
            get { return _MagnifyingTool; }
            set
            {
                MagnifyingTool oldTool = _MagnifyingTool;
                if (Set(ref _MagnifyingTool, value))
                {
                    if (oldTool != null)
                    {
                        oldTool.PositionChanged -= MagnifyingTool_PositionChanged;
                    }
                    if (_MagnifyingTool != null)
                    {
                        MagnifyingTool.PositionChanged += MagnifyingTool_PositionChanged;
                    }
                }
            }
        }

        private PDFViewCtrl MagnifyingPDFViewCtrl
        {
            get { return _MagnifyingPDFViewCtrl; }
            set
            {
                PDFViewCtrl oldView = _MagnifyingPDFViewCtrl;
                if (Set(ref _MagnifyingPDFViewCtrl, value))
                {
                    if (oldView != null)
                    {
                        oldView.SizeChanged -= MagnifyingPDFViewCtrl_SizeChanged;
                        oldView.OnScale -= MagnifyingPDFViewCtrl_OnScale;
                        oldView.OnViewChanged -= MagnifyingPDFViewCtrl_OnViewChanged;
                        oldView.OnPageNumberChanged -= MagnifyingPDFViewCtrl_OnPageNumberChanged;
                        oldView.OnPageFlipFinished -= MagnifyingPDFViewCtrl_OnPageFlipFinished;
                    }
                    if (_MagnifyingPDFViewCtrl != null)
                    {
                        MagnifyingPDFViewCtrl.SizeChanged += MagnifyingPDFViewCtrl_SizeChanged;
                        MagnifyingPDFViewCtrl.OnScale += MagnifyingPDFViewCtrl_OnScale;
                        MagnifyingPDFViewCtrl.OnViewChanged += MagnifyingPDFViewCtrl_OnViewChanged;
                        MagnifyingPDFViewCtrl.OnPageNumberChanged += MagnifyingPDFViewCtrl_OnPageNumberChanged;
                        MagnifyingPDFViewCtrl.OnPageFlipFinished += MagnifyingPDFViewCtrl_OnPageFlipFinished;
                    }
                }
            }
        }

        #endregion View Syncing


        #region Event Handling

        private bool _DocReady = false;
        private UIPoint _LastScrollPosition;
        private bool _IsContinuousInteraction = false;
        private bool _WaitingForUIToSync = true;
        public double _DesiredHorizontalScrollOnLayoutUpdate = -1;
        public double _DesiredVerticalScrollOnLayoutUpdate = -1;

        private bool IsUIAvailable
        {
            get { return _DocReady && !_WaitingForUIToSync; }
        }

        #region MagnifiedPDFViewCtrl
        private async void MagnifiedPDFViewCtrl_OnSetDoc()
        {
            try
            {
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.User;
                _WaitingForUIToSync = true;

                MagnifyingTool.Opacity = 1.0;

                int startPage = _StartPosition.StartPageNumber;
                if (startPage < 1)
                {
                    startPage = MagnifyingPDFViewCtrl.GetCurrentPage();
                }
                MagnifiedPDFViewCtrl.SetCurrentPage(startPage);

                double zoom = _StartPosition.StartZoomFactor;
                if (double.IsNaN(zoom))
                {
                    zoom = MagnifyingPDFViewCtrl.GetZoom() * 4;
                }
                MagnifiedPDFViewCtrl.SetZoom(zoom);

                DoubleRef x = new DoubleRef(_StartPosition.StartPageTopLeftCorner.X);
                DoubleRef y = new DoubleRef(_StartPosition.StartPageTopLeftCorner.Y);
                if (double.IsNaN(x.Value) || double.IsNaN(y.Value))
                {

                    try
                    {
                        MagnifyingPDFViewCtrl.DocLockRead();
                        Rect pageRect = MagnifyingPDFViewCtrl.GetDoc().GetPage(startPage).GetBox(MagnifyingPDFViewCtrl.GetPageBox());
                        Rect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, startPage, MagnifyingPDFViewCtrl);
                        screenRect.Normalize();
                        if (screenRect.x2 > 0 && screenRect.x1 < MagnifyingPDFViewCtrl.ActualWidth && screenRect.y2 > 0 && screenRect.y1 < MagnifyingPDFViewCtrl.ActualHeight)
                        {
                            x.Value = Math.Max(0, screenRect.x1);
                            y.Value = Math.Max(0, screenRect.y1);
                            MagnifyingPDFViewCtrl.ConvScreenPtToPagePt(x, y, startPage);
                        }
                        else
                        {
                            x.Value = screenRect.x1;
                            y.Value = screenRect.y1;
                            MagnifyingPDFViewCtrl.ConvScreenPtToPagePt(x, y, startPage);
                        }
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    finally
                    {
                        MagnifyingPDFViewCtrl.DocUnlockRead();
                    }
                    if (double.IsNaN(x.Value) || double.IsNaN(y.Value))
                    {
                        UIRect rect = GetVisibleContentBoundsInPDFViewCtrl(MagnifyingPDFViewCtrl);
                        x.Value = rect.X;
                        y.Value = rect.Y;
                        MagnifyingPDFViewCtrl.ConvScreenPtToPagePt(x, y, startPage);
                    }
                }
                else
                {

                }

                MagnifiedPDFViewCtrl.ConvPagePtToScreenPt(x, y, startPage);

                double hs = MagnifiedPDFViewCtrl.GetHScrollPos();
                double vs = MagnifiedPDFViewCtrl.GetVScrollPos();
                MagnifiedPDFViewCtrl.SetHScrollPos(hs + x.Value);
                MagnifiedPDFViewCtrl.SetVScrollPos(vs + y.Value);
                _LastScrollPosition.X = MagnifiedPDFViewCtrl.GetHScrollPos();
                _LastScrollPosition.Y = MagnifiedPDFViewCtrl.GetVScrollPos();

                double zoomRatio = MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom();
                UpdateMagnifyingToolSize(MagnifiedPDFViewCtrl.ActualWidth, MagnifiedPDFViewCtrl.ActualHeight, zoomRatio);
                SyncPositionFromMagnifyingControl(false);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
                _DocReady = true;
                await MagnifiedPDFViewCtrl.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
                {
                    _WaitingForUIToSync = false;
                });
            }
        }

        private void MagnifiedPDFViewCtrl_SizeChanged(object sender, Windows.UI.Xaml.SizeChangedEventArgs e)
        {
            if (_WaitingForUIToSync)
            {
                return;
            }
            MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.User;
            try
            {
                double zoomRatio = MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom();
                UpdateMagnifyingToolSize(e.NewSize.Width, e.NewSize.Height, zoomRatio);
                SyncPositionFromMagnifyingControl(false);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
            }


        }

        private void MagnifiedPDFViewCtrl_OnScale()
        {
            if (IsUIAvailable && MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.None)
            {
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.User;
                try
                {
                    double zoomRatio = MagnifyingPDFViewCtrl.GetZoom() / MagnifiedPDFViewCtrl.GetZoom();
                    double sizeRatio = MagnifyingTool.Size.Width / MagnifiedPDFViewCtrl.ActualWidth;
                    double difference = zoomRatio > sizeRatio ? zoomRatio / sizeRatio : sizeRatio / zoomRatio;
                    if (difference > 1.001)
                    {
                        UpdateMagnifyingToolSize(MagnifiedPDFViewCtrl.ActualWidth, MagnifiedPDFViewCtrl.ActualHeight, 1.0 / zoomRatio);
                    }
                    SyncPositionFromMagnifyingControl(false);
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
                }
            }
        }


        private DateTime _HoldMagnifiedViewUntil = DateTime.Now;
        private void MagnifiedPDFViewCtrl_OnViewChanged(object sender, Windows.UI.Xaml.Controls.ScrollViewerViewChangedEventArgs e)
        {
            if (_WaitingForUIToSync || MagnifyingPDFViewCtrl.IsCurrentlyFlipping)
            {
                return;
            }
            if (_DesiredHorizontalScrollOnLayoutUpdate >= 0)
            {
                MagnifiedPDFViewCtrl.SetHScrollPos(_DesiredHorizontalScrollOnLayoutUpdate);
            }
            if (_DesiredVerticalScrollOnLayoutUpdate >= 0)
            {
                MagnifiedPDFViewCtrl.SetVScrollPos(_DesiredVerticalScrollOnLayoutUpdate);
            }

            if (DateTime.Now < _HoldMagnifiedViewUntil)
            {
                return;
            }

            bool interactionAvailable = MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.None;
            bool alreadyHolding = _IsContinuousInteraction && MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.User;

            if (IsUIAvailable && !MagnifiedPDFViewCtrl.IsCurrentlyZooming)
            {
                if (interactionAvailable || alreadyHolding)
                {
                    MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.User;
                    try
                    {
                        SyncPositionFromMagnifyingControl(true);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    finally
                    {
                        _IsContinuousInteraction = e.IsIntermediate;
                        if (!_IsContinuousInteraction)
                        {
                            MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
                        }
                    }
                }
            }

            if (alreadyHolding && !e.IsIntermediate)
            {
                _IsContinuousInteraction = false;
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
            }
        }

        private void MagnifiedPDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            if (MagnifiedToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
            {
                (MagnifiedToolManager.CurrentTool as FreehandCreate).CommitAnnotation();
            }
        }

        #endregion MagnifiedPDFViewCtrl

        #region MagnifyingPDFViewCtrl

        private void MagnifyingPDFViewCtrl_SizeChanged(object sender, Windows.UI.Xaml.SizeChangedEventArgs e)
        {
            if (_WaitingForUIToSync)
            {
                return;
            }
            MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.Tool;
            try
            {
                SyncPositionFromMagnifyingTool(-1, -1);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
            }
        }

        private void MagnifyingPDFViewCtrl_OnScale()
        {
            if (IsUIAvailable && MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.None)
            {
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.Tool;
                try
                {
                    double zoomRatio = MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom();
                    double sizeRatio = MagnifiedPDFViewCtrl.ActualWidth / MagnifyingTool.Size.Width;
                    double difference = zoomRatio > sizeRatio ? zoomRatio / sizeRatio : sizeRatio / zoomRatio;
                    double scrollableWidth = -1;
                    double scrollableHeight = -1;
                    if (difference > 1.001)
                    {
                        int pageNumber = MagnifyingPDFViewCtrl.GetCurrentPage();

                        scrollableWidth = MagnifiedPDFViewCtrl.GetAnnotationCanvas().Width - MagnifiedPDFViewCtrl.ActualWidth;
                        scrollableHeight = MagnifiedPDFViewCtrl.GetAnnotationCanvas().Height - MagnifiedPDFViewCtrl.ActualHeight;
                        MagnifiedPDFViewCtrl.SetZoom(MagnifyingPDFViewCtrl.GetZoom() * sizeRatio);

                        double newMagnifiedZoom =  MagnifiedPDFViewCtrl.GetZoom();
                        double newZoomRatio = MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom();
                        double newSizeRatio = MagnifiedPDFViewCtrl.ActualWidth / MagnifyingTool.Size.Width;
                        difference = newZoomRatio > newSizeRatio ? newZoomRatio / newSizeRatio : newSizeRatio / newZoomRatio;
                        if (difference > 1.001)
                        {
                            UpdateMagnifyingToolSize(MagnifiedPDFViewCtrl.ActualWidth, MagnifiedPDFViewCtrl.ActualHeight, newZoomRatio);
                        }

                    }
                    SyncPositionFromMagnifyingTool(scrollableWidth, scrollableHeight);
                    SyncPositionFromMagnifyingControl(false);
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
                }
            }
        }

        private DateTime _HoldMagnifyingViewUntil = DateTime.Now;
        private void MagnifyingPDFViewCtrl_OnViewChanged(object sender, Windows.UI.Xaml.Controls.ScrollViewerViewChangedEventArgs e)
        {
            if (_WaitingForUIToSync || DateTime.Now < _HoldMagnifyingViewUntil)
            {
                return;
            }
            bool interactionAvailable = MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.None;
            bool alreadyHolding = _IsContinuousInteraction && MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.Tool;

            if (IsUIAvailable && !MagnifyingPDFViewCtrl.IsCurrentlyZooming)
            {
                if (interactionAvailable || alreadyHolding)
                {
                    MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.Tool;
                    try
                    {
                        SyncPositionFromMagnifyingTool(-1, -1);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    finally
                    {
                        _IsContinuousInteraction = e.IsIntermediate;
                        if (!_IsContinuousInteraction)
                        {
                            MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
                        }
                    }
                }
            }

            if (alreadyHolding && !e.IsIntermediate)
            {
                _IsContinuousInteraction = false;
                MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.None;
            }
        }
        private void MagnifyingPDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            if (MagnifyingPDFViewCtrl.IsCurrentlyFlipping || UtilityFunctions.IsContinuous(MagnifyingPDFViewCtrl.GetPagePresentationMode()))
            {
                return;
            }
            SyncAfterNonContinuousPageChange();
        }

        private void MagnifyingPDFViewCtrl_OnPageFlipFinished()
        {
            if (MagnifyingPDFViewCtrl.GetCurrentPage() != MagnifiedPDFViewCtrl.GetCurrentPage())
            {
                SyncAfterNonContinuousPageChange();
            }
        }

        private void SyncAfterNonContinuousPageChange()
        {
            if (MagnifiedToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
            {
                (MagnifiedToolManager.CurrentTool as FreehandCreate).CommitAnnotation();
            }
            MagnifyingTool.InteractionDrivers oldDriver = MagnifyingTool.InteractionDriver;
            MagnifyingTool.InteractionDriver = MagnifyingTool.InteractionDrivers.User;
            try
            {
                int beforePage = MagnifiedPDFViewCtrl.GetCurrentPage();
                double beforeZoom = MagnifiedPDFViewCtrl.GetZoom();
                MagnifiedPDFViewCtrl.SetCurrentPage(MagnifyingPDFViewCtrl.GetCurrentPage());

                int afterPage = MagnifiedPDFViewCtrl.GetCurrentPage();
                double afterZoom = MagnifiedPDFViewCtrl.GetZoom();

                PDFRect canvasRect = MagnifyingPDFViewCtrl.GetContentBoundingBox();
                double widthRatio = canvasRect.Width() / MagnifyingTool.Size.Width;
                double heightRatio = canvasRect.Height() / MagnifyingTool.Size.Height;
                double smallestRatio = Math.Min(widthRatio, heightRatio);

                if (smallestRatio < 1)
                {
                    MagnifyingTool.Size = new UISize(MagnifyingTool.Size.Width * smallestRatio, MagnifyingTool.Size.Height * smallestRatio);
                }
                MagnifyingTool.SnapWidget();

                double zoomRatio = MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom();
                double sizeRatio = MagnifiedPDFViewCtrl.ActualWidth / MagnifyingTool.Size.Width;
                double difference = zoomRatio > sizeRatio ? zoomRatio / sizeRatio : sizeRatio / zoomRatio;
                double scrollableWidth = -1;
                double scrollableHeight = -1;
                if (difference > 1.001)
                {
                    int pageNumber = MagnifyingPDFViewCtrl.GetCurrentPage();

                    scrollableWidth = MagnifiedPDFViewCtrl.GetAnnotationCanvas().Width - MagnifiedPDFViewCtrl.ActualWidth;
                    scrollableHeight = MagnifiedPDFViewCtrl.GetAnnotationCanvas().Height - MagnifiedPDFViewCtrl.ActualHeight;
                    MagnifiedPDFViewCtrl.SetZoom(0, 0, MagnifyingPDFViewCtrl.GetZoom() * sizeRatio);

                    double newMagnifiedZoom = MagnifiedPDFViewCtrl.GetZoom();
                    double newZoomRatio = MagnifiedPDFViewCtrl.GetZoom() / MagnifyingPDFViewCtrl.GetZoom();
                    double newSizeRatio = MagnifiedPDFViewCtrl.ActualWidth / MagnifyingTool.Size.Width;
                    difference = newZoomRatio > newSizeRatio ? newZoomRatio / newSizeRatio : newSizeRatio / newZoomRatio;
                    if (difference > 1.001)
                    {
                        UpdateMagnifyingToolSize(MagnifiedPDFViewCtrl.ActualWidth, MagnifiedPDFViewCtrl.ActualHeight, newZoomRatio);
                    }

                }
                SyncPositionFromMagnifyingTool(scrollableWidth, scrollableHeight);
                if (_DesiredHorizontalScrollOnLayoutUpdate < 0 && _DesiredVerticalScrollOnLayoutUpdate < 0)
                {
                    SyncPositionFromMagnifyingControl(false);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                MagnifyingTool.InteractionDriver = oldDriver;
            }
        }

        #endregion MagnifyingPDFViewCtrl

        #region MagnifyingTool

        private void MagnifyingTool_PositionChanged(object sender, Windows.UI.Xaml.RoutedEventArgs e)
        {
            if (MagnifyingTool.InteractionDriver == MagnifyingTool.InteractionDrivers.Tool)
            {
                double thisZoom = MagnifiedPDFViewCtrl.GetZoom();
                double newZoom = MagnifyingPDFViewCtrl.GetZoom() * (MagnifiedPDFViewCtrl.ActualWidth / MagnifyingTool.Size.Width);
                double difference = thisZoom > newZoom ? thisZoom / newZoom : newZoom / thisZoom;
                double scrollableWidth = -1;
                double scrollableHeight = -1;
                if (difference > 1.001)
                {
                    scrollableWidth = MagnifiedPDFViewCtrl.GetAnnotationCanvas().Width - MagnifiedPDFViewCtrl.ActualWidth;
                    scrollableHeight = MagnifiedPDFViewCtrl.GetAnnotationCanvas().Height - MagnifiedPDFViewCtrl.ActualHeight;
                    MagnifiedPDFViewCtrl.SetZoom(0, 0, newZoom);
                }
                SyncPositionFromMagnifyingTool(scrollableWidth, scrollableHeight);
            }
        }

        #endregion MagnifyingTool

        #region Magnified Ink

        private StrokeReplicator _StrokeReplicator;
        private void MagnifiedToolManager_NewToolCreated(ToolType toolType)
        {
            if (_StrokeReplicator != null)
            {
                if (_StrokeReplicator.IsCommited)
                {
                    MagnifyingTool.mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(MagnifyingPDFViewCtrl.GetAnnotationCanvas(),
                        _StrokeReplicator, _StrokeReplicator, MagnifiedPDFViewCtrl, MagnifyingTool.mToolManager));
                }
                else if (MagnifiedPDFViewCtrl.GetAnnotationCanvas().Children.Contains(_StrokeReplicator))
                {
                    MagnifiedPDFViewCtrl.GetAnnotationCanvas().Children.Remove(_StrokeReplicator);
                }
                _StrokeReplicator.CleanUp();
                _StrokeReplicator = null;
            }
            if (MagnifiedToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
            {
                _StrokeReplicator = new StrokeReplicator(MagnifyingPDFViewCtrl, MagnifiedToolManager.CurrentTool as FreehandCreate);
            }
        }

        private void MagnifiedToolManager_AnnotationAdded(IAnnot annotation, int pageNumber)
        {
            if (_StrokeReplicator != null)
            {
                _StrokeReplicator.Commit();
            }
            else
            {
                _MagnifyingPDFViewCtrl.UpdateWithAnnot(annotation, pageNumber);
            }
        }

        private void MagnifiedToolManager_AnnotationGroupAdded(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                MagnifiedToolManager_AnnotationAdded(annotationItem.Key, annotationItem.Value);
            }
        }

        private void MagnifiedToolManager_AnnotationEdited(IAnnot annotation, int pageNumber)
        {
            MagnifyingPDFViewCtrl.DocLockRead();
            try
            {
                if (annotation.GetAnnotType() == AnnotType.e_Widget)
                {
                    Annots.Widget annotWidget = new Annots.Widget(annotation.GetSDFObj());
                    Field annotField = annotWidget.GetField();
                    MagnifyingPDFViewCtrl.Update(annotField);
                }
                else
                {
                    UpdatePage(MagnifyingPDFViewCtrl, pageNumber);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                MagnifyingPDFViewCtrl.DocUnlockRead();
            }
        }

        private void MagnifiedToolManager_AnnotationGroupEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                MagnifiedToolManager_AnnotationEdited(annotationItem.Key, annotationItem.Value);
            }
        }

        private void MagnifiedToolManager_AnnotationGroupPreEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                MagnifiedToolManager_AnnotationEdited(annotationItem.Key, annotationItem.Value);
            }
        }

        private int _PageToUpdateOnDelete = 0;
        private void MagnifiedToolManager_AnnotationPreRemoved(IAnnot annotation, int pageNumber)
        {
            _PageToUpdateOnDelete = pageNumber;
        }

        private void MagnifiedToolManager_AnnotationRemoved(IAnnot annotation, int pageNumber)
        {
            if (_PageToUpdateOnDelete > 0)
            {
                MagnifyingPDFViewCtrl.DocLockRead();
                try
                {
                    UpdatePage(MagnifyingPDFViewCtrl, pageNumber);
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    MagnifyingPDFViewCtrl.DocUnlockRead();
                }
            }
        }

        private void MagnifiedToolManager_AnnotationGroupRemoved(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                MagnifiedToolManager_AnnotationRemoved(annotationItem.Key, annotationItem.Value);
            }
        }

        /// <summary>
        /// Makes ctrl update the page pageNumber
        /// Assume the document is already locked
        /// </summary>
        /// <param name="ctrl"></param>
        /// <param name="pageNumber"></param>
        /// <returns></returns>
        private void UpdatePage(PDFViewCtrl ctrl, int pageNumber)
        {
            PDFRect rect = ctrl.GetDoc().GetPage(pageNumber).GetBox(MagnifyingPDFViewCtrl.GetPageBox());
            rect = UtilityFunctions.ConvertFromPageRectToScreenRect(rect, pageNumber, MagnifyingPDFViewCtrl);
            ctrl.Update(rect);
        }

        #endregion Magnified Ink

        #endregion Event Handling


        #region Utility Functions
        /// <summary>
        /// Returns a rectangle describing what visible areas of the PDFViewCtrl contain actual content.
        /// Example: At very low zoom, it might be a small centered rectange, while a very high zoom it
        /// will be the entire PDFViewCtrl
        /// </summary>
        /// <returns></returns>
        public static UIRect GetVisibleContentBoundsInPDFViewCtrl(PDFViewCtrl ctrl)
        {
            PDFRect rect = ctrl.GetContentBoundingBox();

            UIRect canvasRect = UtilityFunctions.GetElementRect(ctrl.GetAnnotationCanvas(), ctrl);
            UIRect returnCanvasRect;
            returnCanvasRect.X = Math.Max(rect.x1, 0);
            returnCanvasRect.Y = Math.Max(rect.y1, 0);
            returnCanvasRect.Width = Math.Min(rect.x2 - returnCanvasRect.X, ctrl.ActualWidth - returnCanvasRect.X);
            returnCanvasRect.Height = Math.Min(rect.y2 - returnCanvasRect.Y, ctrl.ActualHeight - returnCanvasRect.Y);
            return returnCanvasRect;
        }

        /// <summary>
        /// Syncs the
        /// </summary>
        /// <param name="preferPan">If true, try to pan first and update widget position later if panning was not enough,
        /// otherwise, try to move the Widget position first, and then pan if this was not enough </param>
        private void SyncPositionFromMagnifyingControl(bool preferPan)
        {
            DoubleRef xPanH = new DoubleRef(0);
            DoubleRef yPanH = new DoubleRef(0);
            DoubleRef xPanV = new DoubleRef(0);
            DoubleRef yPanV = new DoubleRef(0);
            DoubleRef xPosH = new DoubleRef(0);
            DoubleRef yPosH = new DoubleRef(0);
            DoubleRef xPosV = new DoubleRef(0);
            DoubleRef yPosV = new DoubleRef(0);
            int pageNum = MagnifiedPDFViewCtrl.GetPageNumberFromScreenPoint(0, 0);
            int above = 0;
            int below = 0;
            double vertDistanceRatio = 0;
            int left = 0;
            int right = 0;
            double horizDistanceRatio = 0;

            double scrollWeWantToMatch = MagnifiedPDFViewCtrl.GetVScrollPos();

            GetPagesResponsibleForVerticalPosition(new UIPoint(0, 0), MagnifiedPDFViewCtrl, out above, out below, out vertDistanceRatio);
            if (above == below)
            {
                double yBefore = yPanV.Value;
                MagnifiedPDFViewCtrl.ConvScreenPtToPagePt(xPanV, yPanV, above);
                yPosV.Value = yPanV.Value;
                xPosV.Value = xPanV.Value;
                MagnifyingPDFViewCtrl.ConvPagePtToScreenPt(xPanV, yPanV, above);
            }
            else
            {
                yPanV.Value = GetVerticalScreenPointBetweenPage(above, below, vertDistanceRatio, MagnifyingPDFViewCtrl);
            }

            GetPagesResponsibleForHorizontalPosition(new UIPoint(0, 0), MagnifiedPDFViewCtrl, above, out left, out right, out horizDistanceRatio);
            if (left == right)
            {
                double xBefore = xPanH.Value;
                MagnifiedPDFViewCtrl.ConvScreenPtToPagePt(xPanH, yPanH, left);
                xPosH.Value = xPanH.Value;
                yPosH.Value = yPanH.Value;
                MagnifyingPDFViewCtrl.ConvPagePtToScreenPt(xPanH, yPanH, left);
            }
            else
            {
                xPanH.Value = GetHorizontalScreenPointBetweenPage(left, right, horizDistanceRatio, MagnifyingPDFViewCtrl);
            }

            if (preferPan)
            {
                MagnifyingPDFViewCtrl.SetHScrollPos(MagnifyingPDFViewCtrl.GetHScrollPos() + (xPanH.Value - MagnifyingTool.Position.X));
                MagnifyingPDFViewCtrl.SetVScrollPos(MagnifyingPDFViewCtrl.GetVScrollPos() + (yPanV.Value - MagnifyingTool.Position.Y));

                if (above == below)
                {
                    MagnifyingPDFViewCtrl.ConvPagePtToScreenPt(xPosV, yPosV, above);
                }
                else
                {
                    yPosV.Value = GetVerticalScreenPointBetweenPage(above, below, vertDistanceRatio, MagnifyingPDFViewCtrl);
                }
                if (left == right)
                {
                    double xBefore = xPosH.Value;
                    MagnifyingPDFViewCtrl.ConvPagePtToScreenPt(xPosH, yPosH, left);
                }
                else
                {
                    xPosH.Value = GetHorizontalScreenPointBetweenPage(left, right, horizDistanceRatio, MagnifyingPDFViewCtrl);
                }

                MagnifyingTool.Position = new UIPoint(xPosH.Value, yPosV.Value);
            }
            else
            {
                UIRect positionBounds = GetVisibleContentBoundsInPDFViewCtrl(MagnifyingPDFViewCtrl);
                if (xPanH.Value < positionBounds.Left)
                {
                    MagnifyingPDFViewCtrl.SetHScrollPos(MagnifyingPDFViewCtrl.GetHScrollPos() + xPanH.Value);
                }
                if (yPanV.Value < positionBounds.Top)
                {
                    MagnifyingPDFViewCtrl.SetVScrollPos(MagnifyingPDFViewCtrl.GetVScrollPos() + yPanV.Value);
                }
                if (xPanH.Value + MagnifyingTool.Size.Width > positionBounds.Right)
                {
                    MagnifyingPDFViewCtrl.SetHScrollPos(MagnifyingPDFViewCtrl.GetHScrollPos() + xPanH.Value + MagnifyingTool.Size.Width - positionBounds.Right);
                }
                if (yPanV.Value + MagnifyingTool.Size.Height > positionBounds.Bottom)
                {
                    MagnifyingPDFViewCtrl.SetVScrollPos(MagnifyingPDFViewCtrl.GetVScrollPos() + yPanV.Value + MagnifyingTool.Size.Height - positionBounds.Bottom);
                }

                if (above == below)
                {
                    MagnifyingPDFViewCtrl.ConvPagePtToScreenPt(xPosV, yPosV, above);
                }
                else
                {
                    yPosV.Value = GetVerticalScreenPointBetweenPage(above, below, vertDistanceRatio, MagnifyingPDFViewCtrl);
                }
                if (left == right)
                {
                    MagnifyingPDFViewCtrl.ConvPagePtToScreenPt(xPosH, yPosH, left);
                }
                else
                {
                    xPosH.Value = GetHorizontalScreenPointBetweenPage(left, right, horizDistanceRatio, MagnifyingPDFViewCtrl);
                }

                MagnifyingTool.Position = new UIPoint(xPosH.Value, yPosV.Value);
            }
            _HoldMagnifyingViewUntil = DateTime.Now + TimeSpan.FromSeconds(0.2);
        }

        private async void SyncPositionFromMagnifyingTool(double scrollableWidth, double scrollableHeight)
        {
            DoubleRef xPanH = new DoubleRef(MagnifyingTool.Position.X);
            DoubleRef yPanH = new DoubleRef(MagnifyingTool.Position.Y);
            DoubleRef xPanV = new DoubleRef(MagnifyingTool.Position.X);
            DoubleRef yPanV = new DoubleRef(MagnifyingTool.Position.Y);
            int pageNum = MagnifyingPDFViewCtrl.GetPageNumberFromScreenPoint(xPanH.Value, yPanV.Value);
            double newH = -1;
            double newV = -1;
            int above = 0;
            int below = 0;
            double vertDistanceRatio = 0;
            int left = 0;
            int right = 0;
            double horizDistanceRatio = 0;
            _HoldMagnifiedViewUntil = DateTime.Now + TimeSpan.FromSeconds(0.2);
            if (pageNum >= 1)
            {
                MagnifyingPDFViewCtrl.ConvScreenPtToPagePt(xPanH, yPanV, pageNum);
                MagnifiedPDFViewCtrl.ConvPagePtToScreenPt(xPanH, yPanV, pageNum);
                newH = MagnifiedPDFViewCtrl.GetHScrollPos() + xPanH.Value;
                newV = MagnifiedPDFViewCtrl.GetVScrollPos() + yPanV.Value;
            }
            else
            {
                DoubleRef dummy = new DoubleRef(0);
                GetPagesResponsibleForVerticalPosition(MagnifyingTool.Position, MagnifyingPDFViewCtrl, out above, out below, out vertDistanceRatio);
                if (above == below)
                {
                    MagnifyingPDFViewCtrl.ConvScreenPtToPagePt(xPanV, yPanV, above);
                    MagnifiedPDFViewCtrl.ConvPagePtToScreenPt(xPanV, yPanV, above);
                    newV = MagnifiedPDFViewCtrl.GetVScrollPos() + yPanV.Value;
                }
                else
                {
                    newV = MagnifiedPDFViewCtrl.GetVScrollPos() + GetVerticalScreenPointBetweenPage(above, below, vertDistanceRatio, MagnifiedPDFViewCtrl);
                }

                GetPagesResponsibleForHorizontalPosition(MagnifyingTool.Position, MagnifyingPDFViewCtrl, above, out left, out right, out horizDistanceRatio);
                if (left == right)
                {
                    MagnifyingPDFViewCtrl.ConvScreenPtToPagePt(xPanH, yPanH, left);
                    MagnifiedPDFViewCtrl.ConvPagePtToScreenPt(xPanH, yPanH, left);
                    newH = MagnifiedPDFViewCtrl.GetHScrollPos() + xPanH.Value;
                }
                else
                {
                    newH = MagnifiedPDFViewCtrl.GetHScrollPos() + GetHorizontalScreenPointBetweenPage(left, right, horizDistanceRatio, MagnifiedPDFViewCtrl);
                }
            }

            newH = Math.Max(newH, 0);
            newV = Math.Max(newV, 0);

            if ((scrollableWidth > 0 || scrollableHeight > 0) && (newH > scrollableWidth || newV > scrollableHeight))
            {
                // This means the UI will not be able to scroll to the desired position until layout is updated.
                _WaitingForUIToSync = true;
                if (newH > scrollableWidth)
                {
                    _DesiredHorizontalScrollOnLayoutUpdate = newH;
                    MagnifiedPDFViewCtrl.SetHScrollPos(newH);
                }
                else
                {
                    MagnifiedPDFViewCtrl.SetHScrollPos(newH);
                }
                if (newV > scrollableHeight)
                {
                    _DesiredVerticalScrollOnLayoutUpdate = newV;
                    MagnifiedPDFViewCtrl.SetVScrollPos(newV);
                }
                else
                {
                    MagnifiedPDFViewCtrl.SetVScrollPos(newV);
                }
                await MagnifiedPDFViewCtrl.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal,
                    () =>
                    {
                        if (_DesiredHorizontalScrollOnLayoutUpdate >= 0)
                        {
                            MagnifiedPDFViewCtrl.SetHScrollPos(_DesiredHorizontalScrollOnLayoutUpdate);
                        }
                        _DesiredHorizontalScrollOnLayoutUpdate = -1;
                        if (_DesiredVerticalScrollOnLayoutUpdate >= 0)
                        {
                            MagnifiedPDFViewCtrl.SetVScrollPos(_DesiredVerticalScrollOnLayoutUpdate);
                        }
                        _DesiredVerticalScrollOnLayoutUpdate = -1;
                        _WaitingForUIToSync = false;
                    });
            }
            else
            {
                MagnifiedPDFViewCtrl.SetHScrollPos(newH);
                MagnifiedPDFViewCtrl.SetVScrollPos(newV);
            }
            _HoldMagnifiedViewUntil = DateTime.Now + TimeSpan.FromSeconds(0.2);
        }

        private void UpdateMagnifyingToolSize(double newWidth, double newHeight, double zoomRatio)
        {
            double magnifiedHeight = EffectiveHeight;
            if (double.IsNaN(EffectiveHeight))
            {
                magnifiedHeight = newHeight;
            }
            Windows.Foundation.Size zoomedOutSize =
                new Windows.Foundation.Size(newWidth / zoomRatio, magnifiedHeight / zoomRatio);
            MagnifyingTool.Size = zoomedOutSize;
        }

        /// <summary>
        /// Finds the page above and below the screenPoint, as well as how close the point is to each
        /// </summary>
        /// <param name="screenPoint">The point whose position we want to find</param>
        /// <param name="above">The page above, or 0 if there is none</param>
        /// <param name="below">The page below, or 0 if there is none</param>
        /// <param name="distanceRatio">The closeness to the above page relative to the below page</param>
        /// <param name="ctrl">The PDFViewCtrl on which to do the calculations</param>
        private void GetPageAboveAndBelow(UIPoint screenPoint, out int above, out int below, out double distanceRatio, PDFViewCtrl ctrl)
        {
            above = 0;
            below = 0;
            distanceRatio = 0;
            double aboveDistance = 0;
            double belowDistance = 0;
            int currentPage = ctrl.GetCurrentPage();
            PDFDoc doc = ctrl.GetDoc();
            doc.LockRead();
            try
            {
                while (currentPage > 0 && currentPage < ctrl.GetPageCount())
                {
                    Page page = doc.GetPage(currentPage);
                    Rect pageRect = page.GetBox(ctrl.GetPageBox());
                    Rect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, currentPage, ctrl);
                    screenRect.Normalize();
                    if (screenRect.y1 + 1  > screenPoint.Y) // +1 is for rounding errors
                    {
                        below = currentPage;
                        aboveDistance = Math.Max(screenRect.y1 - screenPoint.Y, 0);
                        currentPage--;
                    }
                    else if (screenRect.y2 - 1 < screenPoint.Y)
                    {
                        above = currentPage;
                        belowDistance = Math.Max(screenPoint.Y - screenRect.y2, 0);
                        currentPage++;
                    }
                    else
                    {
                        return;
                    }

                    if (below > 0 && above > 0)
                    {
                        distanceRatio = aboveDistance / (belowDistance + aboveDistance);
                        return;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.UnlockRead();
            }
        }

        private int GetPageForCalculations(UIPoint screenPoint, PDFViewCtrl ctrl)
        {
            int pageNum = MagnifiedPDFViewCtrl.GetPageNumberFromScreenPoint(screenPoint.X, screenPoint.Y);
            if (pageNum > 0)
            {
                return pageNum;
            }

            PDFDoc doc = ctrl.GetDoc();
            doc.LockRead();
            try
            {
                int minPageOncanvas = 0;
                int maxPageOnCanvas = 0;
                GetPageRangeOnCanvas(out minPageOncanvas, out maxPageOnCanvas, ctrl);

                // See if a page is covering the same horizontal space.
                int matchingPage = ctrl.GetCurrentPage();
                int direction = 0;
                while (matchingPage >= minPageOncanvas && matchingPage <= maxPageOnCanvas)
                {
                    Page page = doc.GetPage(matchingPage);
                    Rect pageRect = page.GetBox(ctrl.GetPageBox());
                    Rect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, matchingPage, ctrl);
                    screenRect.Normalize();
                    if (screenRect.y1 + 1 > screenPoint.Y) // +1 is for rounding errors
                    {
                        if (direction > 0)
                        {
                            matchingPage = 0;
                            break;
                        }
                        direction = -1;
                        matchingPage--;
                    }
                    else if (screenRect.y2 - 1 < screenPoint.Y)
                    {
                        if (direction < 0)
                        {
                            matchingPage = 0;
                            break;
                        }
                        direction = 1;
                        matchingPage++;
                    }
                    else
                    {
                        break;
                    }
                }

                if (matchingPage > 0)
                {
                    PDFViewCtrlPagePresentationMode presentationMode = ctrl.GetPagePresentationMode();
                    if (presentationMode == PDFViewCtrlPagePresentationMode.e_single_page || presentationMode == PDFViewCtrlPagePresentationMode.e_single_continuous)
                    {
                        return matchingPage;
                    }
                    int otherPage = matchingPage;


                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.UnlockRead();
            }

            return 0;
        }

        private void GetPagesResponsibleForVerticalPosition(UIPoint screenPoint, PDFViewCtrl ctrl, out int above, out int below, out double distanceRatio)
        {
            above = 0;
            below = 0;
            distanceRatio = 0;
            double aboveDistance = -1;
            double belowDistance = -1;
            int currentPage = ctrl.GetCurrentPage();
            int minPageOncanvas = 0;
            int maxPageOnCanvas = 0;
            GetPageRangeOnCanvas(out minPageOncanvas, out maxPageOnCanvas, ctrl);
            PDFDoc doc = ctrl.GetDoc();

            PDFRect currentRectAbove = null;
            PDFRect currentRectBelow = null;
            int closestPageAbove = 0;

            // Get first page that is above the current point
            while (currentPage >= minPageOncanvas && currentPage <= maxPageOnCanvas)
            {
                Page page = doc.GetPage(currentPage);
                Rect pageRect = page.GetBox(ctrl.GetPageBox());
                Rect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, currentPage, ctrl);
                screenRect.Normalize();
                if (screenRect.y1 <= screenPoint.Y && screenRect.y2 >= screenPoint.Y)
                {
                    above = currentPage;
                    below = currentPage;
                    distanceRatio = 0;
                    return;
                }
                else if (screenRect.y2 - 1 < screenPoint.Y)
                {
                    closestPageAbove = currentPage;
                    currentRectAbove = screenRect;
                    aboveDistance = Math.Max(0, screenPoint.Y - screenRect.y2);
                    currentPage++;
                }
                else if (screenRect.y1 + 1 > screenPoint.Y)
                {
                    if (closestPageAbove > 0)
                    {
                        break;
                    }
                    else
                    {
                        currentPage--;
                    }
                }
            }
            above = closestPageAbove;
            if (above > 0)
            {
                // get tightest page above
                for (int pageToTest = closestPageAbove - 1; pageToTest >= minPageOncanvas; pageToTest--)
                {
                    Page page = doc.GetPage(pageToTest);
                    Rect pageRect = page.GetBox(ctrl.GetPageBox());
                    Rect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, pageToTest, ctrl);
                    screenRect.Normalize();

                    if (screenRect.y1 <= screenPoint.Y && screenRect.y2 >= screenPoint.Y)
                    {
                        above = pageToTest;
                        below = pageToTest;
                        distanceRatio = 0;
                        return;
                    }
                    else if (screenRect.y2 < currentRectAbove.y1)
                    {
                        break;
                    }
                    else
                    {
                        double distance = Math.Max(0, screenPoint.Y - screenRect.y2);
                        if (distance < aboveDistance)
                        {
                            above = pageToTest;
                            aboveDistance = distance;
                            currentRectAbove = screenRect;
                        }
                    }
                }
            }

            // get tightest page below
            for (int pageToTest = closestPageAbove + 1; pageToTest <= maxPageOnCanvas; pageToTest++)
            {
                Page page = doc.GetPage(pageToTest);
                Rect pageRect = page.GetBox(ctrl.GetPageBox());
                Rect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, pageToTest, ctrl);
                screenRect.Normalize();

                if (screenRect.y1 <= screenPoint.Y && screenRect.y2 >= screenPoint.Y)
                {
                    above = pageToTest;
                    below = pageToTest;
                    distanceRatio = 0;
                    return;
                }
                else if (currentRectBelow == null)
                {
                    currentRectBelow = screenRect;
                    below = pageToTest;
                    belowDistance = Math.Max(0, screenRect.y1 - screenPoint.Y);
                }
                else if (screenRect.y1 > currentRectBelow.y2)
                {
                    break;
                }
                else
                {
                    double distance = Math.Max(0, screenRect.y1 - screenPoint.Y);
                    if (distance < belowDistance)
                    {
                        below = pageToTest;
                        belowDistance = distance;
                        currentRectBelow = screenRect;
                    }
                }
            }

            if (above == 0)
            {
                above = below;
            }
            if (below == 0)
            {
                below = above;
            }
            if (above != below)
            {
                distanceRatio = belowDistance / (aboveDistance + belowDistance);
            }
        }

        private void GetPagesResponsibleForHorizontalPosition(UIPoint screenPoint, PDFViewCtrl ctrl, int startPage, out int leftPage, out int rightPage, out double distanceRatio)
        {
            leftPage = startPage;
            rightPage = startPage;
            distanceRatio = 0;

            PDFDoc doc = ctrl.GetDoc();
            Page page = doc.GetPage(startPage);
            PDFRect pageRect = page.GetBox(ctrl.GetPageBox());
            PDFRect screenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pageRect, startPage, ctrl);
            screenRect.Normalize();
            if (screenRect.x1 <= screenPoint.X && screenRect.x2 >= screenPoint.X)
            {
                return;
            }

            int pairedPageNumber = GetPairedPage(ctrl, startPage);
            Page pairedPage = doc.GetPage(pairedPageNumber);
            PDFRect pairedRect = pairedPage.GetBox(ctrl.GetPageBox());
            PDFRect pairedScreenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(pairedRect, pairedPageNumber, ctrl);
            pairedScreenRect.Normalize();

            if (pairedScreenRect.x1 <= screenPoint.X && pairedScreenRect.x2 >= screenPoint.X)
            {
                leftPage = pairedPageNumber;
                rightPage = pairedPageNumber;
                return;
            }

            if (screenRect.x1 + 1 > screenPoint.X && pairedScreenRect.x1 + 1 > screenPoint.X)
            {
                if (screenRect.x1 > pairedScreenRect.x1)
                {
                    leftPage = pairedPageNumber;
                    rightPage = pairedPageNumber;
                }
                return;
            }
            if (screenRect.x2 - 1 < screenPoint.X && pairedScreenRect.x2 - 1 < screenPoint.X)
            {
                if (screenRect.x2 < pairedScreenRect.x2)
                {
                    leftPage = pairedPageNumber;
                    rightPage = pairedPageNumber;
                }
                return;
            }

            PDFRect leftRect = screenRect;
            PDFRect rightRect = pairedScreenRect;
            leftPage = startPage;
            rightPage = pairedPageNumber;
            if (screenRect.x1 > pairedScreenRect.x1)
            {
                leftRect = pairedScreenRect;
                rightRect = screenRect;
                leftPage = pairedPageNumber;
                rightPage = startPage;
            }

            double leftDistance = Math.Max(screenPoint.X - leftRect.x2, 0);
            double rightDistance = Math.Max(rightRect.x1 - screenPoint.X, 0);
            distanceRatio = rightDistance / (leftDistance + rightDistance);
        }

        private int GetPairedPage(PDFViewCtrl ctrl, int pageNumber)
        {
            int minPage = 0;
            int maxPage = 0;
            GetPageRangeOnCanvas(out minPage, out maxPage, ctrl);
            PDFViewCtrlPagePresentationMode mode = ctrl.GetPagePresentationMode();
            if (mode == PDFViewCtrlPagePresentationMode.e_facing || mode == PDFViewCtrlPagePresentationMode.e_facing_continuous)
            {
                if (pageNumber % 2 == 0)
                {
                    return pageNumber - 1;
                }
                else
                {
                    if (pageNumber + 1 <= maxPage)
                    {
                        return pageNumber + 1;
                    }
                    else if (pageNumber - 1 >= minPage)
                    {
                        return pageNumber - 1;
                    }
                    return pageNumber;
                }
            }
            else if (mode == PDFViewCtrlPagePresentationMode.e_facing_cover || mode == PDFViewCtrlPagePresentationMode.e_facing_continuous_cover)
            {
                if (pageNumber % 2 == 1)
                {
                    if (pageNumber - 1 >= minPage)
                    {
                        return pageNumber - 1;
                    }
                    else if (pageNumber + 1 <= maxPage)
                    {
                        return pageNumber + 1;
                    }
                    return pageNumber;
                }
                else
                {
                    if (pageNumber + 1 <= maxPage)
                    {
                        return pageNumber + 1;
                    }
                    else if (pageNumber - 1 >= minPage)
                    {
                        return pageNumber - 1;
                    }
                    return pageNumber;
                }
            }
            return pageNumber;
        }

        private void GetPageRangeOnCanvas(out int minPage, out int maxPage, PDFViewCtrl ctrl)
        {
            minPage = 1;
            maxPage = ctrl.GetPageCount();
            PDFViewCtrlPagePresentationMode mode = ctrl.GetPagePresentationMode();
            if (!UtilityFunctions.IsContinuous(ctrl.GetPagePresentationMode()))
            {
                minPage = ctrl.GetCurrentPage();
                maxPage = ctrl.GetCurrentPage();
                if (mode == PDFViewCtrlPagePresentationMode.e_facing)
                {
                    if (minPage % 2 == 0)
                    {
                        minPage--;
                    }
                    else
                    {
                        maxPage = Math.Min(maxPage + 1, ctrl.GetPageCount());
                    }
                }
                else if (mode == PDFViewCtrlPagePresentationMode.e_facing_cover)
                {
                    if (minPage % 2 == 1)
                    {
                        minPage = Math.Max(1, minPage - 1);
                    }
                    else
                    {
                        maxPage = Math.Min(maxPage + 1, ctrl.GetPageCount());
                    }
                }
            }
        }

        /// <summary>
        /// Returns a vertical position on the PDFViewCtrl that is between the pages
        /// abovePage and belowPagem, weighted towards abovePage by distanceRatio
        /// </summary>
        /// <param name="abovePage">The page above</param>
        /// <param name="belowPage">The page below</param>
        /// <param name="distanceRatio">The weight of abovePage relative to belowPage (0 to 1)</param>
        /// <param name="ctrl">The PDFViewCtrl on which to do the calculations</param>
        /// <returns></returns>
        double GetVerticalScreenPointBetweenPage(int abovePage, int belowPage, double distanceRatio, PDFViewCtrl ctrl)
        {
            double position = 0;
            PDFDoc doc = ctrl.GetDoc();
            doc.LockRead();
            try
            {
                Rect abovePageRect = doc.GetPage(abovePage).GetBox(ctrl.GetPageBox());
                Rect belowPageRect = doc.GetPage(belowPage).GetBox(ctrl.GetPageBox());
                Rect aboveScreenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(abovePageRect, abovePage, ctrl);
                Rect belowScreenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(belowPageRect, belowPage, ctrl);
                aboveScreenRect.Normalize();
                belowScreenRect.Normalize();
                double y1 = aboveScreenRect.y2;
                double y2 = belowScreenRect.y1;
                position = (distanceRatio * y1) + ((1 - distanceRatio) * y2);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.UnlockRead();
            }


            return position;
        }

        double GetHorizontalScreenPointBetweenPage(int leftPage, int rightPage, double distanceRatio, PDFViewCtrl ctrl)
        {
            double position = 0;
            PDFDoc doc = ctrl.GetDoc();
            doc.LockRead();
            try
            {
                Rect leftPageRect = doc.GetPage(leftPage).GetBox(ctrl.GetPageBox());
                Rect rightPageRect = doc.GetPage(rightPage).GetBox(ctrl.GetPageBox());
                Rect leftScreenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(leftPageRect, leftPage, ctrl);
                Rect rightScreenRect = UtilityFunctions.ConvertFromPageRectToScreenRect(rightPageRect, rightPage, ctrl);
                leftScreenRect.Normalize();
                rightScreenRect.Normalize();
                double x1 = leftScreenRect.x2;
                double x2 = rightScreenRect.x1;
                position = (distanceRatio * x1) + ((1 - distanceRatio) * x2);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.UnlockRead();
            }

            return position;
        }

        #endregion Utility Functions
    }
}
