﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Media;


using UIRect = Windows.Foundation.Rect;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class MagnifyingToolbarViewModel : ViewModelBase
    {
        public MagnifyingToolbarViewModel()
        {
            InitCommands();
            ReadyInkPresets();
        }

        public MagnifyingToolbarViewModel(MagnifyingControlViewModel magnifyingControl)
            : this()
        {
            MagnifyingControlViewModel = magnifyingControl;
        }

        public MagnifyingToolbarViewModel(MagnifyingControlViewModel magnifyingControl, MagnifyingTool magnifyingTool)
            : this(magnifyingControl)
        {
            MagnifyingTool = magnifyingTool;
        }

        #region Public Functions

        public void CommitCurrentInk()
        {
            if (CurrentInkTool != null)
            {
                CurrentInkTool.CommitAnnotation();

            }
        }

        #endregion Public Functions


        #region Commands

        public RelayCommand InkButtonCommand { get; private set; }
        public RelayCommand ToolButtonCommand { get; private set; }
        public RelayCommand NavigationButtonCommand { get; private set; }
        public RelayCommand UndoRedoCommand { get; private set; }

        private void InitCommands()
        {
            InkButtonCommand = new RelayCommand(InkButtonCommandImpl);
            ToolButtonCommand = new RelayCommand(ToolButtonCommandImpl);
            NavigationButtonCommand = new RelayCommand(NavigationButtonCommandImpl);
            UndoRedoCommand = new RelayCommand(UndoRedoCommandImpl);
        }

        private void InkButtonCommandImpl(object parameter)
        {
            Button button = parameter as Button;
            if (button != null)
            {
                string tag = button.Tag as string;
                if (!string.IsNullOrEmpty(tag))
                {
                    int result = -1;
                    bool success = int.TryParse(tag, out result);
                    if (success)
                    {
                        if (result == CurrentPresetLabel)
                        {
                            ShowPopup(button);
                        }
                        else
                        {
                            if (CurrentInkTool == null && ToolManager != null)
                            {
                                ToolManager.CreateTool(ToolType.e_ink_create, ToolManager.CurrentTool, true);
                            }
                            CurrentPresetLabel = result;
                        }
                    }
                }
            }
        }

        private void ToolButtonCommandImpl(object parameter)
        {
            CurrentPresetLabel = -1;
            string toolName = parameter as string;
            if (!string.IsNullOrWhiteSpace(toolName))
            {
                ToolType toolType = AnnotationToolbarViewModel.GetToolTypeFromString(toolName);
                if (ToolManager != null)
                {
                    ToolManager.CreateTool(toolType, ToolManager.CurrentTool, false);
                }
            }
        }

        private void NavigationButtonCommandImpl(object parameter)
        {
            if (MagnifyingTool == null)
            {
                return;
            }
            string direction = parameter as string;
            if (CurrentInkTool != null)
            {
                CurrentInkTool.CommitAnnotation();
            }
            if (!string.IsNullOrWhiteSpace(direction))
            {
                if (direction.Equals("down", StringComparison.OrdinalIgnoreCase))
                {
                    MagnifyingTool.MoveWidget(MagnifyingTool.NavigationDirections.Down);
                }
                else if (direction.Equals("left", StringComparison.OrdinalIgnoreCase))
                {
                    MagnifyingTool.MoveWidget(MagnifyingTool.NavigationDirections.Left);
                }
                else if (direction.Equals("right", StringComparison.OrdinalIgnoreCase))
                {
                    MagnifyingTool.MoveWidget(MagnifyingTool.NavigationDirections.Right);
                }
            }
        }

        private void UndoRedoCommandImpl(object parameter)
        {
            string op = parameter as string;
            if (!string.IsNullOrEmpty(op) &&  CurrentInkTool != null)
            {
                if (op.Equals("undo", StringComparison.OrdinalIgnoreCase) && CurrentInkTool.CanUndoStroke())
                {
                    CurrentInkTool.UndoStroke();
                }
                else if (op.Equals("redo", StringComparison.OrdinalIgnoreCase) && CurrentInkTool.CanRedoStroke())
                {
                    CurrentInkTool.RedoStroke();
                }
                RaisePropertyChanged("CanUndo");
                RaisePropertyChanged("CanRedo");
            }
        }

        #endregion Commands


        #region Properties

        public ToolManager ToolManager
        {
            get
            {
                if (MagnifyingControlViewModel != null)
                {
                    return MagnifyingControlViewModel.MagnifiedToolManager;
                }
                return null;
            }
        }

        public PDFViewCtrl MagnifiedPDFViewCtrl
        {
            get
            {
                if (MagnifyingControlViewModel != null)
                {
                    return MagnifyingControlViewModel.MagnifiedPDFViewCtrl;
                }
                return null;
            }
        }

        private MagnifyingControlViewModel _MagnifyingControlViewModel;
        public MagnifyingControlViewModel MagnifyingControlViewModel
        {
            get { return _MagnifyingControlViewModel; }
            set
            {
                if (Set(ref _MagnifyingControlViewModel, value))
                {
                    if (_MagnifyingControlViewModel != null)
                    {
                        _MagnifyingControlViewModel.PropertyChanged += MagnifyingControlViewModel_PropertyChanged;
                    }
                    if (_MagnifyingControlViewModel.MagnifiedToolManager != null)
                    {
                        _CurrentSubscribedToolManager = _MagnifyingControlViewModel.MagnifiedToolManager;
                        _CurrentSubscribedToolManager.NewToolCreated += CurrentSubscribedToolManager_NewToolCreated;
                    }
                }
            }
        }

        private MagnifyingTool _MagnifyingTool;
        public MagnifyingTool MagnifyingTool
        {
            get { return _MagnifyingTool; }
            set
            {
                if (Set(ref _MagnifyingTool, value))
                {
                    RaisePropertyChanged("CanNavigate");
                }
            }
        }

        public bool CanNavigate { get { return _MagnifyingTool != null; } }

        public bool CanUndo { get { return CurrentInkTool != null && CurrentInkTool.CanUndoStroke(); } }

        public bool CanRedo { get { return CurrentInkTool != null && CurrentInkTool.CanRedoStroke(); } }

        private void MagnifyingControlViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("MagnifiedToolManager"))
            {
                if (_CurrentSubscribedToolManager != null)
                {
                    _CurrentSubscribedToolManager.NewToolCreated -= CurrentSubscribedToolManager_NewToolCreated;
                }
                if (_MagnifyingControlViewModel.MagnifiedToolManager != null)
                {
                    _CurrentSubscribedToolManager = _MagnifyingControlViewModel.MagnifiedToolManager;
                    _CurrentSubscribedToolManager.NewToolCreated += CurrentSubscribedToolManager_NewToolCreated;
                }
                RaisePropertyChanged("CurrentTool");
            }
        }

        private SolidColorBrush[] _PresetBrushes = new SolidColorBrush[3];
        public SolidColorBrush PresetBrush0
        {
            get
            {
                return _PresetBrushes[0];
            }
            private set
            {
                _PresetBrushes[0] = value;
                RaisePropertyChanged();
            }
        }
        public SolidColorBrush PresetBrush1
        {
            get { return _PresetBrushes[1]; }
            private set
            {
                _PresetBrushes[1] = value;
                RaisePropertyChanged();
            }
        }
        public SolidColorBrush PresetBrush2
        {
            get { return _PresetBrushes[2]; }
            private set
            {
                _PresetBrushes[2] = value;
                RaisePropertyChanged();
            }
        }

        private int _CurrentPresetLabel = -1;
        public int CurrentPresetLabel
        {
            get { return _CurrentPresetLabel; }
            set
            {
                if (value != _CurrentPresetLabel)
                {
                    _CurrentPresetLabel = value;
                    if (_CurrentPresetLabel < 100)
                    {
                        Settings.SetPropertylLabel(ToolType.e_ink_create, _CurrentPresetLabel);
                        SetInkPropertiesFromLabel();
                    }
                    else
                    {
                        SetEraser();
                    }
                    RaisePropertyChanged();
                }
            }
        }

        public ToolType CurrentTool
        {
            get
            {
                if (ToolManager != null)
                {
                    return ToolManager.CurrentTool.ToolMode;
                }
                return ToolType.e_pan;
            }
        }

        #endregion Properties


        #region Impl

        private ToolManager _CurrentSubscribedToolManager;

        private void CurrentSubscribedToolManager_NewToolCreated(ToolType toolType)
        {
            RaisePropertyChanged("CurrentTool");
            if (CurrentInkTool != null)
            {
                CurrentInkTool.MultiStrokeMode = true;
                CurrentInkTool.SaveInkOnClose = true;
                CurrentInkTool.RetainInkWhenScalingOrSizing = true;
                CurrentInkTool.StrokeAdded += CurrentInkTool_StrokeAdded;
                if (CurrentPresetLabel >= 100)
                {
                    SetEraser();
                }
            }
            RaisePropertyChanged("CanUndo");
            RaisePropertyChanged("CanRedo");
        }

        private void CurrentInkTool_StrokeAdded()
        {
            RaisePropertyChanged("CanUndo");
            RaisePropertyChanged("CanRedo");
        }

        private FreehandCreate CurrentInkTool
        {
            get
            {
                if (ToolManager != null && ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
                {
                    return ToolManager.CurrentTool as FreehandCreate;
                }
                return null;
            }
        }

        private void ReadyInkPresets()
        {
            for (int i = 0; i < 3; i++)
            {
                SettingsColor color = Settings.GetShapeColor(ToolType.e_ink_create, false, i);
                _PresetBrushes[i] = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
            }
        }

        private void SetInkPropertiesFromLabel()
        {
            SettingsColor color = Settings.GetShapeColor(ToolType.e_ink_create, false, CurrentPresetLabel);
            Settings.SetShapeColor(ToolType.e_ink_create, false, color);
            switch (CurrentPresetLabel)
            {
                case 0:
                    PresetBrush0 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
                case 1:
                    PresetBrush1 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
                case 2:
                    PresetBrush2 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
            }
            double thickness = Settings.GetThickness(ToolType.e_ink_create, _CurrentPresetLabel);
            Settings.SetThickness(ToolType.e_ink_create, thickness);
            double opacity = Settings.GetOpacity(ToolType.e_ink_create, _CurrentPresetLabel);
            Settings.SetOpacity(ToolType.e_ink_create, opacity);

            if (CurrentInkTool != null)
            {
                CurrentInkTool.SetInkProperties(thickness, opacity, UtilityFunctions.ConvertToColor(color));
            }
        }

        private Popup _Popup;
        private FrameworkElement _Target;
        private AnnotationPropertyDialog _Dialog;
        private void ShowPopup(FrameworkElement target)
        {
            _Target = target;

            _Dialog = new AnnotationPropertyDialog();
            UIRect rect = UtilityFunctions.GetElementRect(_Target);

            if (ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
            {
                _Dialog.ViewModel.Sequencenumber = _CurrentPresetLabel;
            }

            if (_CurrentPresetLabel >= 100)
            {
                _Dialog.ViewModel.AssociatedTool = ToolType.e_ink_eraser;
            }
            else
            {
                _Dialog.ViewModel.AssociatedTool = ToolManager.CurrentTool.ToolMode;
            }
            AnnotationPropertiesDisplay display = AnnotationPropertiesDisplay.CreateAnnotationPropertiesDisplay(rect, _Dialog);
            display.AnnotationPropertyDialogClosed += display_AnnotationPropertyDialogClosed;
        }

        void display_AnnotationPropertyDialogClosed(AnnotationPropertyDialog dialog)
        {
            if (ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
            {
                if (_CurrentPresetLabel >= 100)
                {
                    SetEraser();
                }
                else
                {
                    SetInkPropertiesFromLabel();
                }
            }
        }

        void _Popup_Closed(object sender, object e)
        {
            _Popup = null;
            _Dialog = null;

            if (ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
            {
                SetInkPropertiesFromLabel();
            }
        }

        private void SetEraser()
        {
            if (CurrentPresetLabel >= 100 && CurrentInkTool != null)
            {
                double thickness = Settings.GetThickness(ToolType.e_ink_eraser, CurrentPresetLabel);
                Windows.UI.Color eraserColor = Windows.UI.Colors.LightGray;
                double opacity = 0.5;
                CurrentInkTool.SetEraserProperties(true, thickness, opacity, eraserColor);
                CurrentInkTool.IsEraser = true;
            }
        }

        #endregion Impl
    }
}
