﻿using pdftron.Common;
using pdftron.Filters;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Multimedia;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using Windows.ApplicationModel;
using Windows.ApplicationModel.Core;
using Windows.ApplicationModel.Resources;
using Windows.Devices.Enumeration;
using Windows.Foundation;
using Windows.Media.Audio;
using Windows.Media.Capture;
using Windows.Media.Capture.Frames;
using Windows.Media.Core;
using Windows.Media.Devices;
using Windows.Media.MediaProperties;
using Windows.Media.Playback;
using Windows.Media.Transcoding;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

using PDFRect = pdftron.PDF.Rect;


namespace pdftron.PDF.Tools.Controls.ViewModels.Multimedia
{
    public class AudioControlViewModel : ViewModelBase, IDisposable
    {
        private readonly ResourceLoader _loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
        private const int SOUND_ANNOT_ICON_SIZE = 25;

        MediaCapture _mediaCapture; // capture audio to be used on creating Sound Annotation PDF spec
        MediaCapture _mediaCapturePreview; // capture MP3 audio for user to preview
        InMemoryRandomAccessStream _pdfAudioStream; // audio stream with Lossless PCM encoding settings
        InMemoryRandomAccessStream _previewAudioStream; // audio stream with MP3 format used only for previewing the recording
        MediaPlayer _mediaPlayerPlayback; // this control is used to play the recorded audio
        //AudioGraphPlayer _audioGraphPlayback; // TODO: used for playback, currenctly lives on AnnotEdit.cs
        DispatcherTimer _recTimerUpdateUI;

        private Windows.UI.Core.CoreWindow _coreWindow;

        private bool _isRecButtonVisible = true;
        private bool _isPlayButtonVisible = false;
        private bool _isPauseRecordingButtonVisible = false;
        private bool _isPlaybackButtonVisible = true;
        private bool _isRecording = false;
        private bool _isRecTimeVisible = true;
        private bool _isRecordingInProgress = false;
        private bool _isPlayButtonEnabled = false;        
        private string _elapsedTime = "00:00";
        private string _totalTime = "00:00";
        private TimeSpan _elpasedTimeSpan;
        private TimeSpan _totalTimeSpan;
        private double _sliderPosition = 0;

        // PDF annot related info
        private PDFViewCtrl mPDFViewCtrl;
        private Windows.Foundation.Point mTarget; // Sound Annot target
        private int mPageNumber;

        private OnCloseRequestedHandler _closeRequested;
        public delegate void OnCloseRequestedHandler();

        /// <summary>
        /// Raised when Audio Control is requesting to be closed
        /// </summary>
        public event OnCloseRequestedHandler CloseRequested
        {
            add
            {
                if (_closeRequested?.GetInvocationList().Contains(value) != true)
                    _closeRequested += value;
            }
            remove => _closeRequested -= value;
        }

        private OnRecordingCompletedHandler _recordingCompleted;
        public delegate void OnRecordingCompletedHandler(StorageFile file, pdftron.PDF.Rect bbox, int pageNumber);

        /// <summary>
        /// Raised when the recording is completed and audio file saved to temp folder
        /// </summary>
        public event OnRecordingCompletedHandler RecordingCompleted
        {
            add
            {
                if (_recordingCompleted?.GetInvocationList().Contains(value) != true)
                    _recordingCompleted += value;
            }
            remove => _recordingCompleted -= value;
        }

        /// <summary>
        /// <c>AudioControlViewModel</c> constructor
        /// </summary>
        /// <param name="mPDFView">Current instance of PDFViewCtrl</param>
        /// <param name="mode">Only play or Play/Record audio</param>
        public AudioControlViewModel(PDFViewCtrl mPDFView, 
            Windows.Foundation.Point target, 
            AudioPlayerMode mode = AudioPlayerMode.RecordAndPlay)
        {
            mPDFViewCtrl = mPDFView;
            mPageNumber = mPDFView.GetCurrentPage();
            mTarget = target;

            CMDRecordAudio = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(RecordAudioAsync);
                command.Execute(parameter);
            });

            CMDPauseRecording = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(PauseRecordingAsync);
                command.Execute(parameter);
            });

            CMDStopAudio = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(StopAndSaveRecAudioAsync);
                command.Execute(parameter);
            });

            CMDPausePlayback = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(PausePlaybackAsync);
                command.Execute(parameter);
            });

            CMDPlayPlayback = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(PlayPlaybackAsync);
                command.Execute(parameter);
            });

            CMDCloseAudio = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(CloseAudioAsync);
                command.Execute(parameter);
            });

            this.PropertyChanged += AudioControlViewModel_PropertyChanged;
        }

        public void Dispose()
        {
            if (_mediaCapture != null)
            {
                _mediaCapture.RecordLimitationExceeded -= _mediaCapture_RecordLimitationExceeded;
                _mediaCapture.Failed -= _mediaCapture_Failed;

                _mediaCapture.Dispose();
                _mediaCapture = null;
            }

            if (_mediaCapturePreview != null)
            {
                _mediaCapturePreview.Dispose();
                _mediaCapturePreview = null;
            }

            if (_mediaPlayerPlayback != null)
            {
                _mediaPlayerPlayback.MediaFailed -= _mediaPlayerPlayback_MediaFailed;
                _mediaPlayerPlayback.MediaEnded -= _mediaPlayer_MediaEnded;
                _mediaPlayerPlayback.PlaybackSession.NaturalDurationChanged -= PlaybackSession_NaturalDurationChanged;
                _mediaPlayerPlayback.PlaybackSession.PositionChanged -= PlaybackSession_PositionChanged;

                _mediaPlayerPlayback.Dispose();
                _mediaPlayerPlayback = null;
            }

            this.PropertyChanged -= AudioControlViewModel_PropertyChanged;
        }

        #region Public Properties

        public bool IsRecButtonVisible
        {
            get { return _isRecButtonVisible; }
            set { Set(ref _isRecButtonVisible, value); }
        }

        public bool IsRecTimeVisible
        {
            get { return _isRecTimeVisible; }
            set { Set(ref _isRecTimeVisible, value); }
        }

        public bool IsPlayButtonVisible
        {
            get { return _isPlayButtonVisible; }
            set { Set(ref _isPlayButtonVisible, value); }
        }

        public bool IsPauseRecordingButtonVisible
        {
            get { return _isPauseRecordingButtonVisible; }
            set { Set(ref _isPauseRecordingButtonVisible, value); }
        }

        public bool IsPlaybackButtonVisible
        {
            get { return _isPlaybackButtonVisible; }
            set { Set(ref _isPlaybackButtonVisible, value); }
        }

        /// <summary>
        /// Whether or not we can play the recording preview
        /// </summary>
        public bool CanPlayPreview
        {
            get { return _isPlayButtonEnabled; }
            set { Set(ref _isPlayButtonEnabled, value); }
        }

        /// <summary>
        /// Weather or not we are recording from microphone and it has not been completed yet
        /// NOTE: it might be paused but not finished
        /// </summary>
        public bool IsRecording
        {
            get { return _isRecording; }
            set { Set(ref _isRecording, value); }
        }

        /// <summary>
        /// Weather or not we are activaly recording/capturing audio now
        /// </summary>
        public bool IsRecordingInProgress
        {
            get { return _isRecordingInProgress; }
            set { Set(ref _isRecordingInProgress, value); }
        }

        /// <summary>
        /// Total time of the audio
        /// </summary>
        public string TotalTime
        {
            get { return _totalTime; }
            set { Set(ref _totalTime, value); }
        }

        /// <summary>
        /// Elapsed time of the currente playback preview
        /// </summary>
        public string ElapsedTime
        {
            get { return _elapsedTime; }
            set { Set(ref _elapsedTime, value); }
        }

        /// <summary>
        /// Preview playback slider position
        /// </summary>
        public double SliderPosition
        {
            get { return _sliderPosition; }
            set { Set(ref _sliderPosition, value); }
        }

        /// <summary>
        /// Audio Control is initialized and ready to be used
        /// </summary>
        public bool IsReady { get; set; }

        /// <summary>
        /// Get the last sucessfully recorded file
        /// </summary>
        public StorageFile RecordedFile { get; set; }

        #endregion

        #region Public Methods

        /// <summary>
        /// Initialize all realted sound/audio libraries to ensure we can start
        /// capturing audio
        /// </summary>
        public async Task<bool> InitAsync()
        {
            IsReady = await InitAudioControlAsync();
            return IsReady;
        }

        #endregion

        #region Public Commands

        public ICommand CMDRecordAudio { get; set; }

        public ICommand CMDPauseRecording { get; set; }

        public ICommand CMDCloseAudio { get; set; }

        public ICommand CMDPlayPlayback { get; set; }

        public ICommand CMDPausePlayback { get; set; }

        public ICommand CMDStopAudio { get; set; }
        #endregion

        #region Public Methods

        #endregion

        #region Operations

        /// <summary>
        /// Initialization of the audio capture device (Microphone)
        /// </summary>
        private async Task<bool> InitAudioControlAsync()
        {
            var audioId = MediaDevice.GetDefaultAudioCaptureId(Windows.Media.Devices.AudioDeviceRole.Default);

            if (string.IsNullOrEmpty(audioId))
            {
                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("pdftron_Controls_AudioControl_SoundAnnot_Title"),
                    Content = _loader.GetString("pdftron_Controls_AudioControl_MicNotFound"),
                    CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                return false;
            }

            if (_mediaCapture == null)
            {
                _mediaCapture = new MediaCapture();
                _mediaCapture.RecordLimitationExceeded += _mediaCapture_RecordLimitationExceeded;
                _mediaCapture.Failed += _mediaCapture_Failed;

                _mediaCapturePreview = new MediaCapture();

                var settings = new MediaCaptureInitializationSettings
                {
                    StreamingCaptureMode = StreamingCaptureMode.Audio,
                    AudioDeviceId = audioId                 
                };

                try
                {
                    await _mediaCapture.InitializeAsync(settings);
                    await _mediaCapturePreview.InitializeAsync(settings);
                }
                catch (Exception ex)
                {
                    ContentDialog dialog = new ContentDialog
                    {
                        Title = _loader.GetString("pdftron_Controls_AudioControl_RecError"),
                        Content = _loader.GetString("pdftron_Controls_AudioControl_AccessDenied"),
                        CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                    };

                    await DialogHelper.ShowContentDialogAsync(dialog);

                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                    return false;
                }
            }

            if (_recTimerUpdateUI == null)
            {
                _recTimerUpdateUI = new DispatcherTimer();
                _recTimerUpdateUI.Interval = new TimeSpan(0, 0, 1);
                _recTimerUpdateUI.Tick += _timerToUpdateUI_Tick;
            }

            if (_mediaPlayerPlayback == null)
            {
                _mediaPlayerPlayback = new MediaPlayer();
                _mediaPlayerPlayback.AudioCategory = MediaPlayerAudioCategory.Media;

                _mediaPlayerPlayback.MediaFailed += _mediaPlayerPlayback_MediaFailed;
                _mediaPlayerPlayback.MediaEnded += _mediaPlayer_MediaEnded;
                _mediaPlayerPlayback.PlaybackSession.NaturalDurationChanged += PlaybackSession_NaturalDurationChanged;
                _mediaPlayerPlayback.PlaybackSession.PositionChanged += PlaybackSession_PositionChanged;
            }

            return true;
        }

        private async Task RecordAudioAsync()
        {
            if (!IsReady)
                return;

            if (_mediaCapture == null)
                return;

            if (CheckIfMicMuted())
            {
                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("pdftron_Controls_AudioControl_RecError"),
                    Content = _loader.GetString("pdftron_Controls_AudioControl_MicMuted"),
                    CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                return;
            }

            if (_pdfAudioStream == null)
                _pdfAudioStream = new InMemoryRandomAccessStream();

            if (_previewAudioStream == null)
                _previewAudioStream = new InMemoryRandomAccessStream();

            try
            {
                if (IsRecording)
                {
                    await PausePlaybackAsync();

                    await _mediaCapture.ResumeRecordAsync();
                    await _mediaCapturePreview.ResumeRecordAsync();

                    IsRecordingInProgress = true;
                    CanPlayPreview = false;
                    IsRecTimeVisible = true;

                    _recTimerUpdateUI.Start();
                }
                else
                {
                    var encodingProfile = new MediaEncodingProfile();

                    encodingProfile.Audio = AudioEncodingProperties.CreatePcm(
                        UtilityFunctions.PCM_AUDIO_SAMPLE_RATE,
                        UtilityFunctions.PCM_AUDIO_NUM_CHANNELS,
                        UtilityFunctions.PCM_AUDIO_BITS_PER_SAMPLE);

                    encodingProfile.Video = null;
                    encodingProfile.Container = new ContainerEncodingProperties() { Subtype = MediaEncodingSubtypes.Wave };

                    await _mediaCapture.StartRecordToStreamAsync(encodingProfile, _pdfAudioStream);

                    await _mediaCapturePreview.StartRecordToStreamAsync(MediaEncodingProfile.CreateMp3(AudioEncodingQuality.Auto), _previewAudioStream);

                    IsRecording = true;
                    IsRecordingInProgress = true;
                    IsRecTimeVisible = true;
                    CanPlayPreview = false;

                    _recTimerUpdateUI.Start();
                }

                _elpasedTimeSpan = new TimeSpan(0, 0, 0);

            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private async Task PauseRecordingAsync()
        {
            if (!IsReady)
                return;

            if (_pdfAudioStream == null)
                return;

            await _mediaCapture.PauseRecordAsync(MediaCapturePauseBehavior.ReleaseHardwareResources);
            await _mediaCapturePreview.PauseRecordAsync(MediaCapturePauseBehavior.ReleaseHardwareResources);

            _mediaPlayerPlayback.Source = MediaSource.CreateFromStream(_previewAudioStream, "audio/mp3");

            CanPlayPreview = true;
            IsRecordingInProgress = false;
            IsRecTimeVisible = false;

            _elpasedTimeSpan = new TimeSpan(0, 0, 0);
            SliderPosition = 0;
            _recTimerUpdateUI.Stop();
        }

        /// <summary>
        /// Request close and perform audio file cleanup
        /// </summary>
        /// <returns></returns>
        private async Task CloseAudioAsync()
        {
            if (_pdfAudioStream == null)
            {
                _closeRequested?.Invoke();
            }
            else
            {
                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("pdftron_Controls_AudioControl_SoundAnnot_Title"),
                    Content = _loader.GetString("pdftron_Controls_AudioControl_DiscardRec"),
                    PrimaryButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK"),
                    CloseButtonText = _loader.GetString("pdftron_Tool_NoteDialog_Cancel")
                };

                dialog.PrimaryButtonCommand = new RelayCommand((_) =>
                {                    
                    _closeRequested?.Invoke();
                });

                await DialogHelper.ShowContentDialogAsync(dialog);
            }
        }

        /// <summary>
        /// Plays a preview of the recorded audio
        /// </summary>
        private async Task PlayPlaybackAsync()
        {
            if (!IsReady)
                return;

            if (_mediaCapture == null || _pdfAudioStream == null || _mediaPlayerPlayback == null)
                return;

            if (IsRecordingInProgress)
                return;

            try
            {
                if (IsRecording && CanPlayPreview)
                {
                    _mediaPlayerPlayback.PlaybackSession.Position = _elpasedTimeSpan;
                    _mediaPlayerPlayback.Play();
                }
                else
                {
                    // TODO: use it to play lossless PCM audio
                    //AudioGraphPlayer audioGraphPlayback = new AudioGraphPlayer();
                    //await audioGraphPlayback.InitAudioGraphAsync();
                    //await audioGraphPlayback.LoadFileAsync(file);
                    //audioGraphPlayback.Play();
                }

                IsPlaybackButtonVisible = false;
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private async Task PausePlaybackAsync()
        {
            if (!IsReady)
                return;

            if (_mediaCapture == null
                || _pdfAudioStream == null
                || _mediaPlayerPlayback == null)
                return;

            if (IsRecordingInProgress)
                return;

            try
            {
                if (IsRecording)
                {
                    if (_mediaPlayerPlayback.PlaybackSession.CanPause)
                    {
                        _mediaPlayerPlayback.Pause();

                        IsPlaybackButtonVisible = true;
                    }
                }
                else
                {
                    // TODO: for Play mode only
                    //_audioGraphPlayback.Stop();
                }
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        /// <summary>
        /// Stop and Save the recording making available as a StorageFile
        /// </summary>
        private async Task StopAndSaveRecAudioAsync()
        {
            if (_mediaCapture == null || _pdfAudioStream == null || _pdfAudioStream.Size == 0)
                return;

            try
            {
                await _mediaCapture.StopRecordAsync();
                await _mediaCapturePreview.StopRecordAsync();

                // Save a local copy of the recording to guarantee stream is properly built (playable using AudioGraph)
                var file = await DocumentManager.SaveAudioStreamToDiskAsync(_pdfAudioStream);

                /* 
                 * Open stream from saved file and swap the bytes which is needed for PDF specs
                 * then we save a new copy to disk which contains the reversed audio file (unplayable using AudioGraph) 
                 */
                var stream = await file.OpenAsync(FileAccessMode.Read);
                byte[] bytes = new byte[stream.Size];
                await stream.AsStream().ReadAsync(bytes, 0, bytes.Length);
                var bytesReversed = UtilityFunctions.SwapByteArray(bytes);

                MemoryStream memoryStream = new MemoryStream(bytesReversed);
                var memoryStreamClone = await UtilityFunctions.ConvertToRandomAccessStreamAsync(memoryStream);
                var fileReversed = await DocumentManager.SaveAudioStreamToDiskAsync(memoryStreamClone);

                // audio file is ready so raise rec complete event
                RecordedFile = fileReversed;

                // Calculate the target position for the sound annot so we can add it to the proper PDF page
                DoubleRef target_x = new DoubleRef(mTarget.X);
                DoubleRef target_y = new DoubleRef(mTarget.Y);

                mPDFViewCtrl.ConvScreenPtToPagePt(target_x, target_y, mPDFViewCtrl.GetCurrentPage());

                PDFRect rect = new PDFRect(target_x.Value,
                    target_y.Value,
                    target_x.Value + SOUND_ANNOT_ICON_SIZE,
                    target_y.Value + SOUND_ANNOT_ICON_SIZE);

                _recordingCompleted?.Invoke(fileReversed, rect, mPageNumber);

                // Dispose streams
                await stream.FlushAsync();
                stream.Dispose();

                await memoryStream.FlushAsync();
                memoryStream.Dispose();
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private bool CheckIfMicMuted()
        {
            if (_mediaCapture == null || _mediaCapture.AudioDeviceController == null)
                return false;

            return _mediaCapture.AudioDeviceController.Muted;
        }

        #endregion

        #region Events
        private void _mediaCapture_RecordLimitationExceeded(MediaCapture sender)
        {
            AppCenterHandlerBase.CURRENT.TrackError(new Exception("AudioControlViewModel - RecordLimitationExceeded error"));
        }

        private void _mediaCapture_Failed(MediaCapture sender, MediaCaptureFailedEventArgs errorEventArgs)
        {
            AppCenterHandlerBase.CURRENT.TrackError(new Exception(errorEventArgs.Message));
        }

        /// <summary>
        /// Media Element has finished playing the audio
        /// </summary>
        private async void _mediaPlayer_MediaEnded(MediaPlayer sender, object args)
        {
            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                _elpasedTimeSpan = new TimeSpan(0, 0, 0);
                _mediaPlayerPlayback.PlaybackSession.Position = _elpasedTimeSpan;
                ElapsedTime = string.Format("{0:mm\\:ss}", _elpasedTimeSpan);

                IsPlaybackButtonVisible = true;
            });
        }

        private void _mediaPlayerPlayback_MediaFailed(MediaPlayer sender, MediaPlayerFailedEventArgs args)
        {
            AppCenterHandlerBase.CURRENT.TrackError(args.ExtendedErrorCode);
        }

        private async void _timerToUpdateUI_Tick(object sender, object e)
        {
            if (_mediaCapture == null)
                return;

            if (IsRecordingInProgress)
            {
                if (CheckIfMicMuted())
                {
                    // TODO: MIC got muted half way
                }

                await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
                {
                    _totalTimeSpan += new TimeSpan(0, 0, 1); // since we don't have API from MSF to catch the current recorded time we use a helper timer
                    TotalTime = string.Format("{0:mm\\:ss}", _totalTimeSpan);
                });
            }
        }

        /// <summary>
        /// Updates the duration of the loaded audio stream
        /// </summary>
        private async void PlaybackSession_NaturalDurationChanged(MediaPlaybackSession sender, object args)
        {
            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                _totalTimeSpan = sender.NaturalDuration;
                TotalTime = string.Format("{0:mm\\:ss}", _totalTimeSpan);
            });
        }

        /// <summary>
        /// Updates the current position of the loaded audio stream
        /// </summary>
        private async void PlaybackSession_PositionChanged(MediaPlaybackSession sender, object args)
        {
            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                ElapsedTime = string.Format("{0:mm\\:ss}", sender.Position);
                SliderPosition = (sender.Position.TotalSeconds / sender.NaturalDuration.TotalSeconds) * 100;
            });
        }

        private async void AudioControlViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if(e.PropertyName == nameof(SliderPosition))
            {
                var mediaPlaybackSession = _mediaPlayerPlayback.PlaybackSession;
                var playbackPosition = (mediaPlaybackSession.Position.TotalSeconds / mediaPlaybackSession.NaturalDuration.TotalSeconds) * 100;
                if ((int)SliderPosition != (int)playbackPosition)
                {
                    await PausePlaybackAsync();

                    var positFractionInSec = mediaPlaybackSession.NaturalDuration.TotalSeconds / 100;
                    var newPositValue = positFractionInSec * SliderPosition;

                    if (newPositValue != _mediaPlayerPlayback.PlaybackSession.Position.TotalSeconds)
                    {
                        _elpasedTimeSpan = new TimeSpan(0, 0, (int)newPositValue);
                        ElapsedTime = string.Format("{0:mm\\:ss}", _elpasedTimeSpan);
                    }
                }
            }
        }

        #endregion

        /// <summary>
        /// Get cached CoreWindow
        /// </summary>
        private Windows.UI.Core.CoreWindow GetCoreWindow()
        {
            if (_coreWindow == null)
                _coreWindow = CoreApplication.MainView.CoreWindow;

            return _coreWindow;
        }
    }

    /// <summary>
    /// Type of the AudioControl mode
    /// </summary>
    public enum AudioPlayerMode
    {
        RecordAndPlay,
        PlayOnly
    }
}
