﻿using Microsoft.UI.Xaml.Controls;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Windows.Foundation;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    /// <summary>
    /// Represents a node within a TreeView.
    /// </summary>
    public class BookmarkNode : ViewModelBase
    {
        public enum ItemRequestUpdates
        {
            DeleteRequested = 0,
            FocusRequested,
            AddRequested,
        }

        private bool _isExpanded;
        private bool _isInvoked;
        private bool _isEditing;
        private bool _isIndeterminate;
        private bool _CollapsedGlyphVisibility = false;
        private bool _ExpandedGlyphVisibility = false;
        private bool _HasUnrealizedChildren = false;
        private bool _isViewModelEditing = false;

        private string _title;
        private Nullable<bool> _isChecked = false;

        private BookmarkNode _parent;

        #region Public Properties

        public bool IsNew { get; set; }

        public BookmarkNode Parent
        {
            get { return _parent; }
            set
            {
                if (_parent != value)
                {
                    _parent = value;
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Represents the Bookmark within the PDF doc that this BookmarkNode is linked to.
        /// </summary>
        public Bookmark SourceBookmark { get; set; }

        public bool HasChildren { get { return Children.Count > 0; } }

        /// <summary>
        /// This BookmarkNode's Children.
        /// </summary>
        public ObservableCollection<BookmarkNode> Children { get; set; }

        public string Title
        {
            get { return _title; }
            set
            {
                _title = value;
                RaisePropertyChanged();
            }
        }

        public bool HasCheckedChildren
        {
            get { return Children.Any(x => x.IsChecked != false); }
        }

        /// <summary>
        /// Automatically adjust children and parent node's IsChecked property.
        /// Used only in Edit mode for deciding which BookmarkNodes deleted.
        /// Only nodes that have IsChecked set to true will be delete.
        /// </summary>
        public Nullable<bool> IsChecked
        {
            get
            {
                return _isChecked;
            }
            set
            {
                if (_isChecked != value)
                {
                    _isChecked = value;
                    RaisePropertyChanged();
                    if (_parent != null)
                    {
                        if (_parent.HasCheckedChildren)
                        {
                            _parent.IsChecked = null;
                        }
                        else
                        {
                            _parent.IsChecked = false;
                        }
                    }
                    if (value != null)
                    {
                        foreach (BookmarkNode node in Children)
                        {
                            node.IsChecked = value;
                        }
                        IsIndeterminate = false;
                        _isChecked = value;
                        RaisePropertyChanged();
                    }
                    else
                    {
                        IsIndeterminate = true;
                    }

                }
            }
        }

        public bool? ChildrenLoadedFromDoc { get; set; }

        /// <summary>
        /// Displays an indeterminate style checkmark inside CheckBox.
        /// </summary>
        public bool IsIndeterminate
        {
            get { return _isIndeterminate; }
            set
            {
                if (_isIndeterminate != value)
                {
                    _isIndeterminate = value;
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Determines if this BookmarkNode's title is currently the one is considered for Editing.
        /// </summary>
        public bool IsInvoked
        {
            get { return _isInvoked; }
            set
            {
                if (_isInvoked != value)
                {
                    _isInvoked = value;
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// This node's title is currently being editted.
        /// </summary>
        public bool IsEditing
        {
            get { return _isEditing; }
            set
            {
                if (Set(ref _isEditing, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Controls if the BookmarkNode within the TreeView is expanded or collapsed.
        /// Also takes care of the visibilty of the expanded and collapsed chevron textboxes.
        /// </summary>
        public bool IsExpanded
        {
            get { return _isExpanded; }
            set
            {
                if (HasChildren || _HasUnrealizedChildren)
                {
                    if (_isExpanded != value)
                    {
                        _isExpanded = value;
                        RaisePropertyChanged();
                    }

                    if (value)
                    {
                        ExpandedGlyphVisibility = true;
                    }
                    else
                    {
                        CollapsedGlyphVisibility = true;
                    }
                }
                else
                {
                    if(_isExpanded != value)
                    {
                        _isExpanded = false;
                        RaisePropertyChanged();
                    }
                    ExpandedGlyphVisibility = false;
                    CollapsedGlyphVisibility = false;
                }
            }
        }

        /// <summary>
        /// Controls the visibility of the collapsed chevron textblock.
        /// </summary>
        public bool CollapsedGlyphVisibility
        {
            get { return _CollapsedGlyphVisibility; }
            set
            {
                if (Set(ref _CollapsedGlyphVisibility, value))
                {
                    RaisePropertyChanged();
                }
                if (value)
                {
                    ExpandedGlyphVisibility = false;
                }
            }
        }

        /// <summary>
        /// Controls the visibility of the expanded chevron textblock.
        /// </summary>
        public bool ExpandedGlyphVisibility
        {
            get { return _ExpandedGlyphVisibility; }
            set
            {
                if (Set(ref _ExpandedGlyphVisibility, value))
                {
                    RaisePropertyChanged();
                }
                if (value)
                {
                    CollapsedGlyphVisibility = false;
                }
            }
        }

        /// <summary>
        /// Controls the visibility of the Checkbox and Overflow button based on if the ViewModel is in edit mode or not.
        /// </summary>
        public bool IsViewModelEditing
        {
            get { return _isViewModelEditing; }
            set
            {
                if (Set(ref _isViewModelEditing, value))
                {
                    RaisePropertyChanged();
                    foreach (BookmarkNode node in Children)
                    {
                        node.IsViewModelEditing = value;
                    }
                }
            }
        }

        /// <summary>
        /// Determines if children of this BookmarkNode have Bookmarks to be loaded from the PDF document.
        /// </summary>
        public bool HasUnrealizedChildren
        {
            get { return _HasUnrealizedChildren; }
            set { _HasUnrealizedChildren = value; RaisePropertyChanged(); }
        }

        #endregion

        #region Commands
        public RelayCommand BookmarkNodeButtonClickedCommand { get; private set; }
        #endregion

        #region Events
        public delegate void ItemRequestDelegate(BookmarkNode node, ItemRequestUpdates updateType);
        public event ItemRequestDelegate ItemRequest;
        #endregion

        /// <summary>
        /// Create a BookmarkNode
        /// </summary>
        /// <param name="bookmark">The root bookmark that this BookMarkNode represents.</param>
        public BookmarkNode(Bookmark bookmark)
        {
            SourceBookmark = bookmark;
            if (SourceBookmark != null)
            {
                _title = SourceBookmark.GetTitle();
            }
            Children = new ObservableCollection<BookmarkNode>();
            InitCommands();
        }

        /// <summary>
        /// Controls which function to invoke based on which item within the overflow button's menu is selected.
        /// </summary>
        /// <param name="parameter">edit, add, delete, done</param>
        private void BookmarkNodeButtonClickedCommandImpl(object parameter)
        {
            string paramText = parameter as string;
            if (!string.IsNullOrEmpty(paramText))
            {
                if (paramText.Equals("delete", StringComparison.OrdinalIgnoreCase))
                {
                    ItemRequest?.Invoke(this, ItemRequestUpdates.DeleteRequested);
                }
                if (paramText.Equals("edit", StringComparison.OrdinalIgnoreCase))
                {
                    IsEditing = true;
                    ItemRequest?.Invoke(this, ItemRequestUpdates.FocusRequested);
                }
                if (paramText.Equals("add", StringComparison.OrdinalIgnoreCase))
                {
                    ItemRequest?.Invoke(this, ItemRequestUpdates.AddRequested);
                }
                if (paramText.Equals("done", StringComparison.OrdinalIgnoreCase))
                {
                    IsEditing = false;
                }
            }
        }

        private void InitCommands()
        {
            BookmarkNodeButtonClickedCommand = new RelayCommand(BookmarkNodeButtonClickedCommandImpl);
        }
    }

    public class OutlineViewModel : ViewModelBase, IDisposable
    {
        private SemaphoreSlim OutlineReadingSemaphore { get; set; }

        private PDFViewCtrl _PDFViewCtrl;

        private bool _IsEditing;
        private bool _IsEditEnabled = true;
        private bool _canReorder;
        private bool _hasBookmarkNodes;
        private bool _IsEditButtonVisible = true;

#if SUBSCRIPTIONS
        private bool _ShowLockEditButton = true;
#endif

        private int _checkedAmount;

        private BookmarkNode _lastInvokedNode;

        #region Events
        public delegate void DocumentModifiedEventHandler(PDFDoc doc);
        public DocumentModifiedEventHandler DocumentModified;

        public delegate void FocusOnSelectedItemEventHandler(BookmarkNode node);
        public event FocusOnSelectedItemEventHandler FocusOnSelectedItemRequested = delegate { };

#if SUBSCRIPTIONS
        private static OutlineSubscriptionsDelegate _subsEventCalled;
        public delegate void OutlineSubscriptionsDelegate (object data);

        /// <summary>
        /// Event used specific for subscription related notification
        /// </summary>
        public static event OutlineSubscriptionsDelegate SubscriptionEventCalled
        {
            add
            {
                if (_subsEventCalled?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _subsEventCalled += value;
            }
            remove
            {
                if (_subsEventCalled != null)
                {
                    OutlineSubscriptionsDelegate eventMethod = (OutlineSubscriptionsDelegate)Array.Find(
                        _subsEventCalled.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _subsEventCalled -= eventMethod;
                }
            }
        }
#endif

        #endregion

        #region Commands
        public RelayCommand AddOutlineCommand { get; private set; }
        public RelayCommand EnterEditingCommand { get; private set; }
        public RelayCommand ExitEditingCommand { get; private set; }
        #endregion

        public OutlineViewModel(PDFViewCtrl ctrl)
        {
            DataSource = new ObservableCollection<BookmarkNode>();
            OutlineReadingSemaphore = new SemaphoreSlim(1);
            _PDFViewCtrl = ctrl;
            InitCommands();

#if SUBSCRIPTIONS
            if (Controls.Viewer.ViewerSettings.IsSubscriptionActivated)
            {
                ShowLockEditButton = false;
            }
            else
            {
                IsEditButtonVisible = false;
            }

#endif
        }

        private void RaiseDocumentModified(PDFDoc doc)
        {
            if (DocumentModified != null)
            {
                DocumentModified(doc);
            }
            if (DataSource.Count == 0)
            {
                HasBookmarkNodes = false;
            }
            else
            {
                HasBookmarkNodes = true;
            }
        }

        #region Public Properties
        public BookmarkNode LastInvokedNode
        {
            get { return _lastInvokedNode; }
            set
            {
                if (_lastInvokedNode != null)
                {
                    _lastInvokedNode.ItemRequest -= BookmarkNode_ItemRequest;
                }
                _lastInvokedNode = value;
                if (value != null)
                {
                    _lastInvokedNode.ItemRequest += BookmarkNode_ItemRequest;
                }

            }
        }

        public int CheckedAmount
        {
            get { return _checkedAmount; }
            set
            {
                if (value < 0)
                    value = 0;
                _checkedAmount = value;
                if (_checkedAmount == 0)
                    CanReorder = true;
                else
                    CanReorder = false;
            }
        }

        public bool CanReorder
        {
            get { return _canReorder; }
            set
            {
                _canReorder = value;
                RaisePropertyChanged();
            }
        }

        public bool HasBookmarkNodes
        {
            get { return _hasBookmarkNodes; }
            set
            {
                if (_hasBookmarkNodes != value)
                {
                    _hasBookmarkNodes = value;
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// The state of the ViewModel. Must be set to true to add, remove, rename, and 
        /// reorder BookmarkNodes.
        /// </summary>
        public bool IsEditing
        {
            get { return _IsEditing; }
            set
            {
                _IsEditing = value;
                RaisePropertyChanged();
                if (_checkedAmount == 0)
                    CanReorder = true;
                else
                    CanReorder = false;

                IsEditButtonVisible = !_IsEditing;
            }
        }

        /// <summary>
        /// Controls the visibility of the main Edit Outline button
        /// </summary>
        public bool IsEditButtonVisible
        {
            get { return _IsEditButtonVisible; }
            set
            {
                _IsEditButtonVisible = value;
                RaisePropertyChanged();
            }
        }

        /// <summary>
        /// Enable or disable the ability to edit the outline
        /// </summary>
        public bool IsEditEnabled
        {
            get { return _IsEditEnabled; }
            set
            {
                _IsEditEnabled = value;
                RaisePropertyChanged();
            }
        }

#if SUBSCRIPTIONS
        /// <summary>
        /// Choose weather or not to lock the edit button
        /// </summary>
        /// <remarks>
        /// IsEditEnabled must be set to true
        /// </remarks>
        public bool ShowLockEditButton
        {
            get { return _ShowLockEditButton; }
            set
            {
                _ShowLockEditButton = value;
                RaisePropertyChanged();
            }
        }
#endif

        /// <summary>
        /// The root level nodes within the Outline's TreeView. These Root level nodes can have 0
        /// or more child level nodes. Each child node is also a root node for their own child node collection.
        /// This is used as the TreeView's ItemSource.
        /// </summary>
        public ObservableCollection<BookmarkNode> DataSource { get; internal set; }
        #endregion

        private void InitCommands()
        {
            AddOutlineCommand = new RelayCommand(AddOutlineCommandImpl);
            EnterEditingCommand = new RelayCommand(EnterEditingCommandImpl);
            ExitEditingCommand = new RelayCommand(ExitEditingCommandImpl);
        }

        public async Task WaitForOutlineToLoadAsync()
        {
            await OutlineReadingSemaphore.WaitAsync();
            OutlineReadingSemaphore.Release();
        }

        /// <summary>
        /// Gets a level of Bookmarks from the doc.
        /// </summary>
        /// <param name="root">The root bookmark</param>
        /// <param name="doc">The PDF document</param>
        /// <returns>A list of Bookmarks</returns>
        public IAsyncOperation<IList<Bookmark>> GetBookmarkListAsync(Bookmark root, PDFDoc doc)
        {
            Task<IList<Bookmark>> t = new Task<IList<Bookmark>>(() => GetBookmarkList(root, doc));
            t.Start();
            return t.AsAsyncOperation();
        }

        /// <summary>
        /// Gets a level of Bookmarks from the doc.
        /// </summary>
        /// <param name="node">The target node to adjust IsOpen</param>
        /// <returns>True if the target node's SourceBookmark was adjusted</returns>
        public IAsyncOperation<bool> AdjustIsOpen(BookmarkNode node)
        {
            Task<bool> t = new Task<bool>(() => SaveIsOpen(node));
            t.Start();
            return t.AsAsyncOperation();
        }

        private IList<Bookmark> GetBookmarkList(Bookmark root, PDFDoc doc, bool gotLock = false)
        {
            IList<Bookmark> items = new List<Bookmark>();

            try
            {
                doc?.LockRead();
                if (!gotLock)
                {
                    OutlineReadingSemaphore?.Wait();
                }

                Bookmark current = root ?? doc?.GetFirstBookmark();

                int numbookmarks = 0;

                while (current?.IsValid() == true)
                {
                    items.Add(current);
                    current = current.GetNext();
                    numbookmarks++;
                }

            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc?.UnlockRead();
                if (!gotLock)
                {
                    OutlineReadingSemaphore?.Release();
                }
            }
            return items;
        }

        /// <summary>
        /// Create the list of bookmarks present in the current document by loading the
        /// first depth of node
        /// </summary>
        public async Task CreateTreeAsync()
        {
            if(DataSource != null && DataSource.Count > 0)
            {
                Dispose();
            }
            if(DataSource == null)
            {
                DataSource = new ObservableCollection<BookmarkNode>();
            }
            await CreateRootTreeViewNodesAsync(null);
            DataSource.CollectionChanged += DataSource_CollectionChanged;
        }

        /// <summary>
        /// Load the root nodes (depth 0) for the selected document
        /// </summary>
        /// <param name="root">The root bookmark</param>
        /// <returns></returns>
        private async Task CreateRootTreeViewNodesAsync(Bookmark root)
        {
            IList<Bookmark> items;
            Bookmark bookmark;
            if (root != null)
            {
                bookmark = root;
                items = await GetBookmarkListAsync(bookmark.GetFirstChild(), _PDFViewCtrl.GetDoc());
            }
            else
            {
                items = await GetBookmarkListAsync(null, _PDFViewCtrl.GetDoc());
            }

            if (items == null)
            {
                HasBookmarkNodes = false;
                return;
            }

            foreach (Bookmark item in items.ToList())
            {
                if (item != null)
                {
                    BookmarkNode node = new BookmarkNode(item);

                    if (node != null)
                    {
                        if (item.HasChildren())
                        {
                            node.HasUnrealizedChildren = true;
                            node.CollapsedGlyphVisibility = true;
                            node.ChildrenLoadedFromDoc = false;
                            if (item.IsOpen())
                            {
                                node.IsExpanded = true;
                            }
                        }
                        node.Parent = null;

                        if(DataSource != null)
                            DataSource.Add(node);

                        if(node.Children != null)
                            node.Children.CollectionChanged += Children_CollectionChanged;
                    }
                }
            }

            if (DataSource != null)
            {
                HasBookmarkNodes = DataSource.Count > 0;
            }
            else
            {
                HasBookmarkNodes = false;
            }
        }

        /// <summary>
        /// Expand a specific parent node by loading it's children from it's source bookmark
        /// </summary>
        /// <param name="parentBookmarkNode">Parent node to be expanded</param>
        public async Task ExpandTreeViewNodeAsync(BookmarkNode parentBookmarkNode)
        {
            if (parentBookmarkNode == null)
                return;

            if(parentBookmarkNode.Children != null)
                parentBookmarkNode.Children.CollectionChanged -= Children_CollectionChanged;

            IList<Bookmark> items = await GetBookmarkListAsync(parentBookmarkNode.SourceBookmark.GetFirstChild(), _PDFViewCtrl.GetDoc());
            if (items == null)
                return;

            // if we are in edit mode and parent is checked, save state so we can set all children to checked
            bool? isParentChecked = parentBookmarkNode.IsChecked;
            if (isParentChecked == true)
                parentBookmarkNode.IsChecked = false;

            foreach (Bookmark item in items.ToList())
            {
                if(item != null)
                {
                    BookmarkNode node = new BookmarkNode(item);
                    if(node != null)
                    {
                        node.IsViewModelEditing = IsEditing;
                        node.Parent = parentBookmarkNode;

                        if (item.HasChildren())
                        {
                            node.HasUnrealizedChildren = true;
                            node.CollapsedGlyphVisibility = true;
                            node.ChildrenLoadedFromDoc = false;
                            if (item.IsOpen() && node.Parent.IsExpanded)
                            {
                                node.IsExpanded = true;
                            }
                        }
                    }

                    if (node.Children != null)
                        node.Children.CollectionChanged += Children_CollectionChanged;

                    if (parentBookmarkNode.Children != null)
                        parentBookmarkNode.Children.Add(node);
                }
            }

            if(parentBookmarkNode.Children != null)
                parentBookmarkNode.Children.CollectionChanged += Children_CollectionChanged;

            if(parentBookmarkNode != null)
            {
                if (isParentChecked == true)
                    parentBookmarkNode.IsChecked = true;

                parentBookmarkNode.HasUnrealizedChildren = false;
                parentBookmarkNode.ChildrenLoadedFromDoc = true;
            }
        }

        /// <summary>
        /// Adjusts the root level Bookmarks when nodes added to the root collection or reordered 
        /// within the child collection.
        /// </summary>
        private void DataSource_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Add)
            {
                BookmarkNode node = null;
                BookmarkNode droppedNode = DataSource.ElementAt(e.NewStartingIndex);

                PDFDoc doc = _PDFViewCtrl.GetDoc();
                try
                {
                    doc.Lock();
                    if (e.NewStartingIndex == 0 && DataSource.Count == 1)
                    {
                        doc.AddRootBookmark(droppedNode.SourceBookmark);    //No bookmarks found within the doc
                        string txt = doc.GetFirstBookmark().GetTitle();
                    }
                    else
                    {
                        if (!droppedNode.IsNew)         //cant unlink a bookmark that currently isnt found within the doc
                        {
                            droppedNode.SourceBookmark.Unlink();
                        }
                        if (e.NewStartingIndex >= 1)    //adding to the right of the BookmarkNode
                        {
                            node = DataSource.ElementAt(e.NewStartingIndex - 1);
                            node.SourceBookmark.AddNext(droppedNode.SourceBookmark);
                        }
                        else
                        {
                            node = DataSource.ElementAt(1);     //adding to the start of the rootnodes
                            if (node.SourceBookmark.IsValid())
                            {
                                node.SourceBookmark.AddPrev(droppedNode.SourceBookmark);
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                    RaiseDocumentModified(doc);

                    droppedNode.Parent = null;
                    if (droppedNode.IsNew)
                    {
                        droppedNode.IsNew = false;
                    }
                }
            }
            if (DataSource.Count == 0)
            {
                HasBookmarkNodes = false;
            }
        }

        /// <summary>
        /// Adjusts the child level Bookmarks when nodes added to the child collection or reordered 
        /// within the child collection.
        /// </summary>
        private void Children_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Add)
            {
                var collection = sender as ObservableCollection<BookmarkNode>;
                BookmarkNode node = null;
                BookmarkNode droppedNode = collection.ElementAt(e.NewStartingIndex);

                PDFDoc doc = _PDFViewCtrl.GetDoc();
                try
                {
                    doc.Lock();

                    if (!droppedNode.IsNew)         //cant unlink a bookmark that currently isnt found within the doc
                    {
                        droppedNode.SourceBookmark.Unlink();
                    }

                    if (e.NewStartingIndex >= 1)    //adding to the right of the BookmarkNode
                    {
                        node = collection.ElementAt(e.NewStartingIndex - 1);
                        node.SourceBookmark.AddNext(droppedNode.SourceBookmark);
                        droppedNode.Parent = node.Parent;
                    }
                    else
                    {
                        if (collection.Count > 1)   //adding to the left of a BookmarkNode
                        {
                            node = collection.ElementAt(1);
                            node.SourceBookmark.AddPrev(droppedNode.SourceBookmark);
                            droppedNode.Parent = node.Parent;
                        }
                        else        //no child nodes exist therefore we need to add a child
                        {
                            foreach (BookmarkNode root in DataSource)    //find the root node that currently contains the dropped node
                            {
                                node = FindRootNode(root, droppedNode);
                                if (node != null)
                                {
                                    break;
                                }
                            }

                            if (node != null)
                            {
                                droppedNode.Parent = node;
                                droppedNode.Parent.HasUnrealizedChildren = true;
                                droppedNode.Parent.ChildrenLoadedFromDoc = false;

                                node.SourceBookmark.AddChild(droppedNode.SourceBookmark);


                                //need to remove the BookmarkNode from the collection that was added then in finally load the children from the file
                                node.Children.CollectionChanged -= Children_CollectionChanged;
                                node.Children.Remove(droppedNode);
                                node.Children.CollectionChanged += Children_CollectionChanged;
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();

                    if (IsEditing)
                        RaiseDocumentModified(doc);

                    if (droppedNode.Parent != null && !droppedNode.Parent.IsExpanded)
                    {
                        droppedNode.Parent.IsExpanded = true;
                    }
                    if (droppedNode.IsNew)
                    {
                        droppedNode.IsNew = false;
                    }
                }
            }
        }

        private BookmarkNode FindRootNode(BookmarkNode root, BookmarkNode target)
        {
            if (root.Children.Contains(target))
                return root;
            else
            {
                foreach (BookmarkNode current in root.Children)
                {
                    if (current.HasChildren)
                        return FindRootNode(current, target);
                }
            }
            return null;
        }

        private void ExitEditingCommandImpl(object sender)
        {
            IsEditing = false;

            if (LastInvokedNode != null)
            {
                LastInvokedNode.IsInvoked = false;
                LastInvokedNode = null;
            }

            foreach (BookmarkNode node in DataSource)
            {
                node.IsChecked = false;
                node.IsViewModelEditing = false;
            }
        }

        private void EnterEditingCommandImpl(object sender)
        {
#if SUBSCRIPTIONS
            if (Controls.Viewer.ViewerSettings.IsSubscriptionActivated == false)
            {
                _subsEventCalled?.Invoke(null);
                return;
            }
#endif

            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Edit Outline" });

            IsEditing = true;

            foreach (BookmarkNode node in DataSource)
            {
                node.IsViewModelEditing = true;
            }
        }

        /// <summary>
        /// Adds a new BookmarkNode to the outline.
        /// </summary>
        /// <param name="sender">
        /// The sender is the Parent node that is calling this command. The newly created BookmarkNode 
        /// will be add to the Parent BookmarkNode's collection, otherwise it will be added to the 
        /// Root nodes list (DataSource).
        /// </param>
        public async void AddOutlineCommandImpl(object sender)
        {
            BookmarkNode node = null;
            BookmarkNode parentNode = sender as BookmarkNode;

            PDFDoc doc = _PDFViewCtrl.GetDoc();

            if (doc != null)
            {
                try
                {
                    doc.Lock();

                    int currentPage = _PDFViewCtrl.GetCurrentPage();

                    Page page = doc.GetPage(currentPage);

                    //Create the bookmark and attach an action
                    Bookmark bookmark = Bookmark.Create(doc, "Page " + _PDFViewCtrl.GetCurrentPage());
                    bookmark.SetAction(Action.CreateGoto(Destination.CreateFit(page)));

                    if (parentNode != null)
                    {
                        if (parentNode.SourceBookmark.IsValid())
                            parentNode.SourceBookmark.AddChild(bookmark);
                    }

                    //Create a new BookmarkNode
                    node = new BookmarkNode(bookmark);
                    node.Children.CollectionChanged += Children_CollectionChanged;
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                    RaiseDocumentModified(doc);

                    if (node != null)
                    {
                        node.IsViewModelEditing = IsEditing;
                        node.IsNew = true;

                        if (parentNode != null)     //Adding a node to an existing parent's chldren
                        {
                            node.Parent = parentNode;

                            if (parentNode != null)
                            {
                                if (parentNode.IsExpanded || parentNode.ChildrenLoadedFromDoc != null)
                                {
                                    if (parentNode.ChildrenLoadedFromDoc == false)
                                    {
                                        await ExpandTreeViewNodeAsync(parentNode);
                                    }
                                    else
                                    {
                                        if (parentNode != null)
                                        {
                                            if (parentNode.Children != null)
                                                parentNode.Children.Add(node);
                                        }
                                    }
                                    if (parentNode != null)
                                    {
                                        if (parentNode.IsExpanded == false)
                                            parentNode.IsExpanded = true;
                                    }
                                }
                                else
                                {
                                    await ExpandTreeViewNodeAsync(parentNode);
                                    if (parentNode != null)
                                        parentNode.IsExpanded = true;
                                }
                            }
                        }
                        else        //Adding a node to the root level of the bookmarks
                        {
                            if (DataSource != null)
                                DataSource.Add(node);
                        }
                    }                    
                }
            }
        }

        private void BookmarkNode_ItemRequest(BookmarkNode item, BookmarkNode.ItemRequestUpdates updateType)
        {
            if (updateType == BookmarkNode.ItemRequestUpdates.FocusRequested)
            {
                FocusOnSelectedItemRequested?.Invoke(item);
            }
            if (updateType == BookmarkNode.ItemRequestUpdates.DeleteRequested)
            {
                RemoveSingleOutline(item);
            }
            if (updateType == BookmarkNode.ItemRequestUpdates.AddRequested)
            {
                AddOutlineCommandImpl(item);
            }
        }

        /// <summary>
        /// Removes a single BookmarkNode
        /// </summary>
        /// <param name="node">
        /// The node to be removed
        /// </param>
        private void RemoveSingleOutline(BookmarkNode node)
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            if (doc != null)
            {
                try
                {
                    doc.Lock();
                    if (node != null)
                    {
                        node.SourceBookmark.Delete();
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc?.Unlock();
                    RaiseDocumentModified(doc);
                    node.Children.CollectionChanged -= Children_CollectionChanged;

                    if (DataSource.Contains(node))
                    {
                        DataSource.Remove(node);
                    }
                    else
                    {
                        foreach (BookmarkNode root in DataSource)
                        {
                            BookmarkNode parent = FindRootNode(root, node);

                            if (parent != null)
                            {
                                parent.Children.Remove(node);
                                parent.IsExpanded = parent.HasChildren;
                                break;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Removes all root BookmarkNodes that are Checked.
        /// </summary>
        public void RemoveBookmarkNodes()
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();

            if (doc != null)
            {
                try
                {
                    doc.Lock();

                    IList<BookmarkNode> removeList = new List<BookmarkNode>();

                    foreach (BookmarkNode node in DataSource)
                    {
                        if (node.IsChecked == true)     //Remove root level node, child Bookmarks will be deleted automatically
                        {
                            if (node.SourceBookmark.IsValid())
                            {
                                node.SourceBookmark.Delete();
                            }

                            removeList.Add(node);
                        }
                        else if (node.IsChecked == null)    //Indeterminate node detected, find which children need to be removed
                        {
                            RemoveIndeterminateNodes(node);
                        }
                    }

                    foreach (BookmarkNode node in removeList)
                    {
                        node.IsChecked = false;
                        DataSource.Remove(node);
                    }

                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                    RaiseDocumentModified(doc);
                }
            }
        }

        /// <summary>
        /// Removes all of the BookmarkNodes that are checked within a collection whose parent
        /// is Indeterminate
        /// </summary>
        /// <param name="root">
        /// The node that is considered Indeterminate. Indeterminate means that the root level node should
        /// not be deleted, however the root level node has children that have IsChecked set to true and 
        /// therefore those child nodes should be removed.
        /// </param>
        private void RemoveIndeterminateNodes(BookmarkNode root)
        {
            IList<BookmarkNode> removeList = new List<BookmarkNode>();
            foreach (BookmarkNode node in root.Children)
            {
                if (node.IsChecked == true)
                {
                    if (node.SourceBookmark.IsValid())
                        node.SourceBookmark.Delete();
                    removeList.Add(node);
                }
                else if (node.IsChecked == null)
                {
                    RemoveIndeterminateNodes(node);
                }
            }
            foreach (BookmarkNode node in removeList)
            {
                root.IsChecked = false;
                root.Children.Remove(node);
            }
        }

        public void FinishEditingText()
        {
            if (LastInvokedNode != null)
            {
                FinalizeText(LastInvokedNode);
            }
        }

        private void FinalizeText(BookmarkNode node)
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();

            if (doc != null)
            {
                try
                {
                    doc.Lock();
                    if (node.SourceBookmark != null)
                    {
                        node.SourceBookmark.SetTitle(node.Title);
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                    RaiseDocumentModified(doc);

                    node.IsEditing = false;
                }
            }
        }

        public void CancelEditingText(string text)
        {
            LastInvokedNode.Title = text;
        }

        /// <summary>
        /// Adjusts the Bookmark's IsOpen property
        /// </summary>
        /// <param name="root"></param>
        private bool SaveIsOpen(BookmarkNode node)
        {

            PDFDoc doc = _PDFViewCtrl.GetDoc();
            if (doc != null)
            {
                try
                {
                    doc.Lock();
                    if (node.SourceBookmark != null)
                    {
                        node.SourceBookmark.SetOpen(node.IsExpanded);
                        return true;
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                        Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                }
            }
            return false;
        }

        /// <summary>
        /// Unsubscribes the parameter's child nodes from events, 
        /// and removes them from the parameter's child collection.
        /// </summary>
        /// <param name="node">
        /// </param>
        private void UnloadChildrenCollection(BookmarkNode node)
        {
            if (node != null)
            {
                foreach (BookmarkNode child in node.Children)
                {
                    if (child != null)
                        UnloadChildrenCollection(child);
                }

                if (node.Children != null)
                {
                    node.Children.CollectionChanged -= Children_CollectionChanged;
                    node.Children.Clear();
                }
            }
        }

        /// <summary>
        /// Unsubscribes the DataSource's items and its children from events, 
        /// and removes them from the observable collection.
        /// </summary>
        private void UnloadDataSourceCollection()
        {
            if(DataSource != null)
            {
                DataSource.CollectionChanged -= DataSource_CollectionChanged;

                foreach (BookmarkNode node in DataSource)
                {
                    if (node != null)
                    {
                        UnloadChildrenCollection(node);     //Unsubscribe and clear the child observable collections
                    }
                }

                DataSource.Clear();
                DataSource = null;
            }
        }

        /// <summary>
        /// Unsubscribes the DataSource's items and its children from events, 
        /// and removes items from their observable collections.
        /// </summary>
        public void Dispose()
        {
            UnloadDataSourceCollection();
        }
    }
}
