﻿using pdftron.PDF;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Utilities;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System.Threading.Tasks;
using System.Windows.Input;
using Windows.ApplicationModel.Resources;
using Windows.Storage;
using Windows.UI.Popups;
using System.Linq;
using Windows.UI.Xaml.Input;
using Windows.Foundation;
using System.Security;
using Actions.Crypto;
using System;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Controls.ViewModels.Password
{
    public partial class PdfPasswordViewModel : ViewModelBase
    {
        private const int DEFAULT_ATTEMPTS_AT_PASSWORD = 3;

        private PasswordHandledDelegate _passwordHandled;
        public delegate void PasswordHandledDelegate(bool success, string password, Deferral deferral);
        public event PasswordHandledDelegate PasswordHandled
        {
            add
            {
                if (_passwordHandled?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _passwordHandled += value;
            }
            remove
            {
                if (_passwordHandled != null)
                {
                    PasswordHandledDelegate eventMethod = (PasswordHandledDelegate)Array.Find(
                        _passwordHandled.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _passwordHandled -= eventMethod;
                }
            }
        }

        private string _userPassword;
        private IStorageFile _fileToCheck;
        private int _attemptsAtPassword = DEFAULT_ATTEMPTS_AT_PASSWORD;
        private bool _IsIncorrectPasswordNotificationVisible;
        private bool _validPassword;

        public ICommand PasswordKeyDownCommand { get; }

        public ICommand PasswordKeyUpCommand { get; }

        public ICommand PasswordChangedCommand { get; }

        public ICommand PasswordEnteredCommand { get; }

        public ICommand PasswordCancelCommand { get; }

        public PdfPasswordViewModel()
        {
            PasswordKeyDownCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ExecutePasswordKeyDownAsync);
                command.Execute(parameter);
            });
            PasswordKeyUpCommand = new RelayCommand<KeyRoutedEventArgs>(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ExecutePasswordKeyUpAsync);
                command.Execute(parameter);
            });
            PasswordChangedCommand = new RelayCommand(ExecutePasswordChanged);
            PasswordEnteredCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ExecutePasswordEnteredAsync);
                command.Execute(parameter);
            });
            PasswordCancelCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ExecutePasswordCancelAsync);
                command.Execute(parameter);
            });
        }

        public string UserPassword
        {
            get => _userPassword;
            set
            {
                if (Set(ref _userPassword, value))
                {
                    RaisePropertyChanged(nameof(HasPasswordBoxGotContent));
                    IsIncorrectPasswordNotificationVisible = false;
                }
            }
        }

        public bool IsIncorrectPasswordNotificationVisible
        {
            get => _IsIncorrectPasswordNotificationVisible;
            set
            {
                Set(ref _IsIncorrectPasswordNotificationVisible, value);
                RaisePropertyChanged(nameof(HasPasswordBoxGotContent));
            }
        }

        public IStorageFile FileToCheck
        {
            get => _fileToCheck;
            set => Set(ref _fileToCheck, value);
        }

        public bool ValidPassword
        {
            get => _validPassword;
            set => Set(ref _validPassword, value);
        }

        public int AttemptsAtPassword => _attemptsAtPassword;

        public bool HasPasswordBoxGotContent => !string.IsNullOrEmpty(UserPassword) && UserPassword.Length > 0 && !IsIncorrectPasswordNotificationVisible;

        public string GetEncryptedPassword()
        {
            if (!string.IsNullOrEmpty(UserPassword))
                return DataProtection.Instance.Encrypt(UserPassword);

            return string.Empty;
        }

        private async Task CancelPasswordAsync()
        {
            if (_passwordHandled != null)
            {
                var tcs = new TaskCompletionSource<bool>();
                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                {
                    _passwordHandled?.Invoke(false, null, deferral);
                    await tcs.Task;
                }
            }
        }

        private async Task VerifyPasswordAsync()
        {
            if (string.IsNullOrEmpty(UserPassword))
                return;

            if (!GetPasswordRequired(UserPassword))
            {
                if (_passwordHandled != null)
                {
                    var tcs = new TaskCompletionSource<bool>();
                    using (var deferral = new Deferral(() => tcs.SetResult(true)))
                    {
                        _passwordHandled?.Invoke(true, UserPassword, deferral);
                        await tcs.Task;
                    }
                }
            }
            else
            {
                --_attemptsAtPassword;
                IsIncorrectPasswordNotificationVisible = true;
                ValidPassword = false;
                if (_attemptsAtPassword == 0)
                {
                    await Task.Delay(200);
                    ResourceLoader loader = ResourceLoader.GetForCurrentView();

                    ContentDialog dialog = new ContentDialog
                    {
                        Title = loader.GetString("PasswordDialog_ManyFails_Title"),
                        Content = string.Format(loader.GetString("PasswordDialog_ManyFails_Info"), DEFAULT_ATTEMPTS_AT_PASSWORD),
                        CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                    };

                    await DialogHelper.ShowContentDialogAsync(dialog);
                }
            }
        }

        public bool GetPasswordRequired(string password = "")
        {
            UserPassword = password;

            using (PDFDoc doc = new PDFDoc(FileToCheck))
            {
                if (!doc.InitStdSecurityHandler(password))
                    return true;
            }

            return false;
        }
    }
}
