﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Extensions;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using Windows.Foundation;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class ThumbnailItem : ViewModelBase
    {
        private ThumbnailsViewViewModel _thumbnailsViewViewModel;
        private int _pageNumber;
        private ImageSource _thumbnail;
        private Size _bitmapSize = new Size(0, 0);
        private bool _isCurrentPage;
        private bool _isDirty;
        private int _angle;
        private double _uniformScale = 1;
        private bool? _isLandScape;

        public RelayCommand InsertBlankPageBeforeCommand { get; }

        public RelayCommand InsertBlankPageAfterCommand { get; }

        public RelayCommand RotatePageCommand { get; }

        public RelayCommand DeletePageCommand { get; }

        public RelayCommand ExportPageCommand { get; }

        public RelayCommand DuplicatePageCommand { get; }

        public ThumbnailItem(int pageNumber, Color color)
        {
            InsertBlankPageBeforeCommand = new RelayCommand(InsertBlankPageBefore);
            InsertBlankPageAfterCommand = new RelayCommand(InsertBlankPageAfter);
            RotatePageCommand = new RelayCommand(RotatePage);
            DeletePageCommand = new RelayCommand(DeletePage);
            ExportPageCommand = new RelayCommand(ExportPage);
            DuplicatePageCommand = new RelayCommand(DuplicatePage);

            PageNumber = pageNumber;
            DefaultBackground = new SolidColorBrush(color);
        }

        public int PageNumber
        {
            get => _pageNumber;
            set => Set(ref _pageNumber, value);
        }

        public ImageSource Thumbnail
        {
            get => _thumbnail;
            set
            {
                Set(ref _thumbnail, value);
                if (_thumbnail != null)
                    IsDirty = false;

                RaisePropertyChanged(nameof(HasThumbnail));
            }
        }

        public Size BitmapSize
        {
            get => _bitmapSize;
            set => Set(ref _bitmapSize, value);
        }

        public bool IsCurrentPage
        {
            get => _isCurrentPage;
            set => Set(ref _isCurrentPage, value);
        }

        public bool HasThumbnail => Thumbnail != null;

        public bool IsDirty
        {
            get => _isDirty;
            set => Set(ref _isDirty, value);
        }

        public SolidColorBrush DefaultBackground { get; set; }

        public int Angle
        {
            get => _angle;
            set => Set(ref _angle, value);
        }

        public double UniformScale
        {
            get => _uniformScale;
            set => Set(ref _uniformScale, value);
        }

        public bool IsLandscape
        {
            get
            {
                if (_isLandScape != null)
                    return (bool)_isLandScape;

                _isLandScape = Thumbnail is BitmapSource bitmap && bitmap.PixelHeight < bitmap.PixelWidth;
                return (bool)_isLandScape;
            }
        }

        public bool IsDeleted { get; set; }

        public bool EnabledThumbnailOverlay => MainViewModel.EnabledThumbnailOverlay;

        public bool IsConverting => _thumbnailsViewViewModel.IsConverting;

        public ThumbnailsViewViewModel MainViewModel
        {
            get => _thumbnailsViewViewModel;
            set => Set(ref _thumbnailsViewViewModel, value);
        }

        private void InsertBlankPageBefore(object parameter) => MainViewModel.AddBlankPage(true, PageNumber);

        private void InsertBlankPageAfter(object parameter) => MainViewModel.AddBlankPage(false, PageNumber);

        private void RotatePage(object parameter) => MainViewModel.RotateCommand.Execute(parameter);

        private void DeletePage(object parameter) => MainViewModel.DeleteCommand.Execute(parameter);

        private void ExportPage(object parameter) => MainViewModel.ExportSelectedPagesCommand.Execute(parameter);

        private void DuplicatePage(object parameter) => MainViewModel.DuplicatePage(PageNumber);

        [SuppressMessage("Redundancy", "RCS1163:Unused parameter.", Justification = "<Pending>")]
        public void ThumbnailPointerEntered(object sender, PointerRoutedEventArgs e)
        {
            if (MainViewModel.EnabledThumbnailOverlay)
            {
                DependencyObject obj = (DependencyObject)sender;
                foreach (var button in obj.FindControls<Button>())
                    button.Visibility = Visibility.Visible;
            }
        }

        [SuppressMessage("Redundancy", "RCS1163:Unused parameter.", Justification = "<Pending>")]
        public void ThumbnailPointerExited(object sender, PointerRoutedEventArgs e)
        {
            if (MainViewModel.EnabledThumbnailOverlay)
            {
                DependencyObject obj = (DependencyObject)sender;
                foreach (var button in obj.FindControls<Button>())
                    button.Visibility = Visibility.Collapsed;
            }
        }

        public void ThumbnailImagePointerEntered(object sender, PointerRoutedEventArgs e)
        {
            DependencyObject obj = (DependencyObject)sender;
            Border border = obj.FindControls<Border>().LastOrDefault();
            if (border != null)
                border.Visibility = Visibility.Visible;
        }

        public void ThumbnailImagePointerExited(object sender, PointerRoutedEventArgs e)
        {
            DependencyObject obj = (DependencyObject)sender;
            Border border = obj.FindControls<Border>().LastOrDefault();
            if (border != null)
                border.Visibility = Visibility.Collapsed;
        }
    }
}
