﻿using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class AnnotationItem : ViewModelBase
    {
        private string _author;
        private string _content;
        private int _pageNumber;
        private string _additionalInformation;
        private AnnotType _annotationType;

        /// <summary>
        ///  Should not be relied upon to stay consistent.
        /// </summary>
        public IAnnot Annot { get; set; }

        public string Author
        {
            get => _author;
            set
            {
                if (Set(ref _author, value))
                    RaisePropertyChanged(nameof(AuthorAndContent));
            }
        }

        public string Content
        {
            get => _content;
            set
            {
                if (Set(ref _content, value))
                    RaisePropertyChanged(nameof(AuthorAndContent));
            }
        }

        public string AuthorAndContent
        {
            get
            {
                if (!string.IsNullOrWhiteSpace(Author))
                {
                    return Author + ": " + Content;
                }
                else if (!string.IsNullOrWhiteSpace(Content))
                {
                    return Content;
                }

                return string.Empty;
            }
        }

        public int PageNumber
        {
            get => _pageNumber;
            set => Set(ref _pageNumber, value);
        }

        public Rect AnnotRect { get; set; }

        public string AdditionalInformation
        {
            get => _additionalInformation;
            set
            {
                if (Set(ref _additionalInformation, value))
                    RaisePropertyChanged(nameof(AnnotationSymbol));
            }
        }

        public AnnotType AnnotationType
        {
            get => _annotationType;
            set
            {
                if (Set(ref _annotationType, value))
                    RaisePropertyChanged(nameof(AnnotationSymbol));
            }
        }
        public string AnnotationSymbol
        {
            get
            {
                string annotGlyph = new string((char)0x270F, 1); // default
                switch (AnnotationType)
                {
                    case AnnotType.e_Line:
                        annotGlyph = "\uE756";
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation) && AdditionalInformation.Equals("Arrow", StringComparison.OrdinalIgnoreCase))
                            annotGlyph = "\uE755";
                        break;
                    case AnnotType.e_Square:
                        annotGlyph = "\uE749";
                        break;
                    case AnnotType.e_Circle:
                        annotGlyph = "\uE747";
                        break;
                    case AnnotType.e_Ink:
                        annotGlyph = "\uE719";
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation) && AdditionalInformation.Equals("Marker", StringComparison.OrdinalIgnoreCase))
                            annotGlyph = new string((char)0x00AE, 1);
                        break;
                    case AnnotType.e_Polygon:
                        annotGlyph = "\uE751";
                        break;
                    case AnnotType.e_Polyline:
                        annotGlyph = "\uE757";
                        break;
                    case AnnotType.e_FreeText:
                        annotGlyph = "\uE725";
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation) && AdditionalInformation.Equals("Callout", StringComparison.OrdinalIgnoreCase))
                            annotGlyph = new string((char)0x0049, 1);
                        break;
                    case AnnotType.e_Text:
                        annotGlyph = "\uE744";
                        break;
                    case AnnotType.e_Highlight:
                        annotGlyph = "\uE726";
                        break;
                    case AnnotType.e_Underline:
                        annotGlyph = "\uE735";
                        break;
                    case AnnotType.e_StrikeOut:
                        annotGlyph = "\uE732";
                        break;
                    case AnnotType.e_Squiggly:
                        annotGlyph = "\uE729";
                        break;
                    case AnnotType.e_Caret:
                        annotGlyph = new string((char)0x004F, 1);
                        break;
                    case AnnotType.e_Stamp:
                        annotGlyph = "\uE645";
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation))
                        {
                            if (AdditionalInformation.Equals("Signature", StringComparison.OrdinalIgnoreCase))
                            {
                                annotGlyph = "\uE738";
                            }
                            else if (AdditionalInformation.Equals("Image", StringComparison.OrdinalIgnoreCase))
                            {
                                annotGlyph = "\uE645";
                            }
                        }
                        break;
                    case AnnotType.e_Redact:
                        annotGlyph = "\uE681";
                        break;
                }

                return annotGlyph;
            }
        }

        public AnnotationItem(string author, string content, AnnotType annotationSymbol, int pageNumber, Rect annotRect, string additionalInformation = null)
        {
            _author = author;
            _content = content;
            _annotationType = annotationSymbol;
            _pageNumber = pageNumber;
            AnnotRect = annotRect;
            _additionalInformation = additionalInformation;
        }

        public void CopyFrom(AnnotationItem other)
        {
            Author = other.Author;
            Content = other.Content;
            AnnotationType = other.AnnotationType;
            PageNumber = other.PageNumber;
            AnnotRect = other.AnnotRect;
            AdditionalInformation = other.AdditionalInformation;
        }
    }
}
