﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using PDFDouble = pdftron.Common.DoubleRef;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    /// <summary>
    /// Note that if the document is being modified while this list is being generated
    /// it could cause errors. Ensure that ClearViewModel() followed by awaiting
    /// WaitForAnnotationListToCancel() happens before the document is modified.
    /// </summary>
    public class AnnotationListViewModel : ViewModelBase
    {
        private static readonly List<AnnotType> WORKABLE_ANNOTATIONS = new List<AnnotType>()
        {
            AnnotType.e_Circle,
            AnnotType.e_Square,
            AnnotType.e_Text,
            AnnotType.e_Line,
            AnnotType.e_Polygon,
            AnnotType.e_Underline,
            AnnotType.e_StrikeOut,
            AnnotType.e_Ink,
            AnnotType.e_Highlight,
            AnnotType.e_FreeText,
            AnnotType.e_Squiggly,
            AnnotType.e_Stamp,
            AnnotType.e_Caret,
            AnnotType.e_Polyline,
            AnnotType.e_Redact
        };

        private PageChangedDelegate _pageChanged;
        public delegate void PageChangedDelegate(int prevPage, int newPage);
        public event PageChangedDelegate PageChanged
        {
            add
            {
                if (_pageChanged?.GetInvocationList().Contains(value) != true)
                    _pageChanged += value;
            }
            remove => _pageChanged -= value;
        }

        private bool _isFinishedLoading;
        private PDFViewCtrl _pdfViewCtrl;
        private ToolManager _toolManager;
        private bool _cancel;
        private SemaphoreSlim _annotListSemaphore;
        private bool _foundAnnotations = true;
        private Rectangle _flashingRectangle;
        private Brush _flashBrush = new SolidColorBrush(Windows.UI.Colors.Transparent);
        private Storyboard _flashAnimation;
        private int _nextPageToProcess;
        private SelectionChangeSources _selectionChangedSource;
        private AnnotationItem _selectedItem;
        private List<AnnotationItem> _selectedGroupItems;
        private List<Tuple<AnnotationModifications, IAnnot, int>> _modificationList;

        public bool IsFinishedLoading
        {
            get => _isFinishedLoading;
            private set => Set(ref _isFinishedLoading, value);
        }

        public PDFViewCtrl PDFViewCtrl
        {
            get => _pdfViewCtrl;
            set
            {
                _pdfViewCtrl = value;
            }
        }

        public ToolManager ToolManager
        {
            get => _toolManager;
            set
            {
                if (_toolManager != value)
                {
                    if (_toolManager != null)
                    {
                        _toolManager.SelectionChanged -= ToolManager_SelectionChanged;
                        _toolManager.AnnotationAdded -= ToolManager_AnnotationAdded;
                        _toolManager.AnnotationEdited -= ToolManager_AnnotationEdited;
                        _toolManager.AnnotationPreRemoved -= ToolManager_AnnotationPreRemoved;

                        _toolManager.SelectionGroupChanged -= ToolManager_SelectionGroupChanged;
                        _toolManager.AnnotationGroupAdded -= ToolManager_AnnotationGroupAdded;
                        _toolManager.AnnotationGroupEdited -= ToolManager_AnnotationGroupEdited;
                        _toolManager.AnnotationGroupPreRemoved -= ToolManager_AnnotationGroupPreRemoved;
                        _toolManager.AnnotationGroupPreEdited -= ToolManager_AnnotationGroupPreEdited;
                    }

                    _toolManager = value;
                    if (_toolManager != null)
                    {
                        _toolManager.SelectionChanged += ToolManager_SelectionChanged;
                        _toolManager.AnnotationAdded += ToolManager_AnnotationAdded;
                        _toolManager.AnnotationEdited += ToolManager_AnnotationEdited;
                        _toolManager.AnnotationPreRemoved += ToolManager_AnnotationPreRemoved;

                        _toolManager.SelectionGroupChanged += ToolManager_SelectionGroupChanged;
                        _toolManager.AnnotationGroupAdded += ToolManager_AnnotationGroupAdded;
                        _toolManager.AnnotationGroupEdited += ToolManager_AnnotationGroupEdited;
                        _toolManager.AnnotationGroupPreRemoved += ToolManager_AnnotationGroupPreRemoved;
                        _toolManager.AnnotationGroupPreEdited += ToolManager_AnnotationGroupPreEdited;
                    }
                }
            }
        }

        public NavigationStack NavigationStack { get; set; }

        public AnnotationListViewModel()
        {
            AnnotationGroups = new ObservableCollection<PageHeaderItem>();
            InitInteraction();
        }

        public AnnotationListViewModel(PDFViewCtrl ctrl)
        {
            AnnotationGroups = new ObservableCollection<PageHeaderItem>();
            InitInteraction();
            PDFViewCtrl = ctrl;
        }

        /// <summary>
        /// Load the Annotation list by iterating for each page of the
        /// PDF document and adding each annotation to the list
        /// </summary>
        public void LoadAnnotationList() => CreateAnnotationList();

        public void ClearViewModel()
        {
            _cancel = true;
            ToolManager = null;
        }

        /// <summary>
        /// Wait for this before disposing the PDFDoc, after you call ClearViewModel().
        /// This guarantees that annots are not being extracted.
        /// </summary>
        /// <returns></returns>
        public async Task WaitForAnnotationListToFinish()
        {
            Debug.WriteLine("Waiting for annot list to cancel");
            await AnnotListSemaphore.WaitAsync();
            Debug.WriteLine("Annot list is cancelled or finished");
            AnnotListSemaphore.Release();
        }

        public SemaphoreSlim AnnotListSemaphore => _annotListSemaphore ?? (_annotListSemaphore = new SemaphoreSlim(1));

        public ObservableCollection<PageHeaderItem> AnnotationGroups { get; }

        public bool FoundAnnotations
        {
            get => _foundAnnotations;
            set
            {
                if (_foundAnnotations != value)
                {
                    _foundAnnotations = value;
                    RaisePropertyChanged();
                }
            }
        }

        public Brush FlashBrush
        {
            get => _flashBrush;
            set
            {
                _flashBrush = value;
                if (_flashingRectangle != null)
                    _flashingRectangle.Fill = value;
            }
        }

        public Storyboard FlashAnimation
        {
            get => _flashAnimation;
            set
            {
                _flashAnimation = value;
                Storyboard.SetTarget(FlashAnimation, _flashingRectangle);
                Storyboard.SetTargetName(FlashAnimation, _flashingRectangle.Name);
                _flashAnimation.Completed += FlashAnimation_Completed;
            }
        }

        private async void CreateAnnotationList()
        {
            PDFDoc doc = _pdfViewCtrl.GetDoc();

            if (doc != null)
            {
                bool gotItems = false;
                _nextPageToProcess = 1;
                IList<AnnotationItem> items = await GetNextPageWithAnnotsAsync(_nextPageToProcess, doc);
                while (items != null && !_cancel)
                {
                    if (items.Count > 0)
                        _nextPageToProcess = items[0].PageNumber + 1;

                    gotItems = true;
                    foreach (AnnotationItem item in items)
                    {
                        item.AnnotRect = _pdfViewCtrl.GetPageRectForAnnot(item.Annot, item.PageNumber);
                        item.Annot = null;
                    }

                    PageHeaderItem header = new PageHeaderItem(items);
                    AnnotationGroups.Add(header);
                    items = await GetNextPageWithAnnotsAsync(_nextPageToProcess, doc);
                }

                _nextPageToProcess = 0;
                IsFinishedLoading = true;
                if (!gotItems)
                    FoundAnnotations = false;
            }
        }

        private IAsyncOperation<IList<AnnotationItem>> GetNextPageWithAnnotsAsync(int pageNumber, PDFDoc doc)
        {
            Task<IList<AnnotationItem>> t = new Task<IList<AnnotationItem>>(() => GetNextPageWithAnnots(pageNumber, doc));
            t.Start();

            return t.AsAsyncOperation();
        }

        private IList<AnnotationItem> GetNextPageWithAnnots(int pageNumber, PDFDoc doc)
        {
            try
            {
                AnnotListSemaphore.Wait();
                if (_cancel)
                    return null;

                try
                {
                    doc.LockRead();
                    PageIterator pageIter = doc.GetPageIterator(pageNumber);
                    while (pageIter.HasNext() && !_cancel)
                    {
                        Page page = pageIter.Current();
                        int numAnnots = page.GetNumAnnots();
                        if (numAnnots > 0)
                        {
                            IList<AnnotationItem> items = GetAnnotationItemsOnPage(page);
                            if (items.Count > 0)
                            {
                                pageIter.Next();
                                return items;
                            }
                        }

                        pageIter.Next();
                    }
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.UnlockRead();
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                AnnotListSemaphore.Release();
            }

            return null;
        }

        private IList<AnnotationItem> GetAnnotationItemsOnPage(Page page)
        {
            List<AnnotationItem> items = new List<AnnotationItem>();
            int numAnnots = page.GetNumAnnots();
            for (int i = 0; i < numAnnots; i++)
            {
                IAnnot annot = page.GetAnnot(i);
                if (annot?.IsValid() != true)
                    continue;

                AnnotType annotType = annot.GetAnnotType();
                if (WORKABLE_ANNOTATIONS.Contains(annotType) && annot.IsMarkup())
                    items.Add(MakeAnnotationItem(annot, annotType, page));
            }

            return items;
        }

        private AnnotationItem MakeAnnotationItem(IAnnot annot, AnnotType annotType, Page page)
        {
            string author = "";
            string contents = "";
            string additonalInfo = "";

            if (annot is Annots.IMarkup markup)
            {
                author = markup.GetTitle();

                Annots.Popup popup = markup.GetPopup();
                if (popup?.IsValid() == true)
                {
                    contents = popup.GetContents();
                }
                else
                {
                    contents = markup.GetContents();
                }

                if (string.IsNullOrWhiteSpace(contents) && annot is Annots.ITextMarkup)
                {
                    TextExtractor textExtractor = new TextExtractor();

                    try
                    {
                        int indexy = page.GetIndex();
                        textExtractor.Begin(page);
                        string text = textExtractor.GetTextUnderAnnot(annot);
                        if (!string.IsNullOrWhiteSpace(text))
                            contents = text;
                    }
                    catch (Exception ex)
                    {
                        AnalyticsHandlerBase.CURRENT.SendEvent(
                            AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                            AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                }
            }

            // if stamp, see if it's a signature
            if (annotType == AnnotType.e_Stamp)
            {
                Obj sigObj = annot.GetSDFObj();
                Obj stampObj = annot.GetSDFObj();
                sigObj = sigObj.FindObj(ToolManager.SignatureAnnotationIdentifyingString);
                stampObj = stampObj.FindObj(ToolManager.StamperAnnotationIdentifyingString);
                if (sigObj != null)
                {
                    additonalInfo = "Signature";
                }
                else if (stampObj != null)
                {
                    additonalInfo = "Image";
                }
            }
            // if a line, see if it is an arrow
            else if (annotType == AnnotType.e_Line)
            {
                if (MeasureUtils.IsRulerMeasure(annot))
                {
                    additonalInfo = "Ruler";
                }
                else
                {
                    Annots.Line line = (Annots.Line)annot;
                    if (UtilityFunctions.IsArrow(line))
                    {
                        additonalInfo = "arrow";
                    }
                }
            }
            // if a free text, see if it's a callout
            else if (annotType == AnnotType.e_FreeText)
            {
                Obj obj = annot.GetSDFObj();
                obj = obj.FindObj("CL");
                if (obj != null)
                    additonalInfo = "Callout";
            }
            else if (annotType == AnnotType.e_Ink)
            {
                Annots.Ink ink = (Annots.Ink)annot;
                // check if it is a marker
                if (ink.GetBlendMode() == GStateBlendMode.e_bl_multiply)
                {
                    additonalInfo = "Marker";
                }
                else
                {
                    additonalInfo = "Ink";
                }
            }
            else if (annotType == AnnotType.e_Redact)
            {
                additonalInfo = "Redaction";
            }
            else if (annotType == AnnotType.e_Polygon)
            {
                if (MeasureUtils.IsAreaMeasure(annot))
                {
                    additonalInfo = "Area";
                }
            }
            else if (annotType == AnnotType.e_Polyline)
            {
                if (MeasureUtils.IsPerimeterMeasure(annot))
                {
                    additonalInfo = "Perimeter";
                }
            }

            AnnotationItem item = new AnnotationItem(author, contents, annotType, page.GetIndex(), annot.GetRect(), additonalInfo)
            {
                Annot = annot
            };

            return item;
        }

        private void InitInteraction()
        {
            ItemClickCommand = new RelayCommand(ItemClickCommandImpl);
            _flashingRectangle = new Rectangle
            {
                Opacity = 0,
                Fill = FlashBrush
            };
        }

        public AnnotationItem SelectedItem
        {
            get => _selectedItem;
            set
            {
                if (Set(ref _selectedItem, value))
                {
                    if (_selectionChangedSource == SelectionChangeSources.None && ToolManager != null)
                    {
                        _selectionChangedSource = SelectionChangeSources.AnnotationList;
                        ToolManager.UnselectAllAnnotations();
                        _selectionChangedSource = SelectionChangeSources.None;
                    }

                    if (_selectionChangedSource != SelectionChangeSources.ToolManager && SelectedItem != null)
                        ShowAnnotation(SelectedItem);
                }
            }
        }

        public List<AnnotationItem> SelectedGroupItems
        {
            get => _selectedGroupItems;
            set
            {
                if (Set(ref _selectedGroupItems, value))
                {
                    if (_selectionChangedSource == SelectionChangeSources.None && ToolManager != null)
                    {
                        _selectionChangedSource = SelectionChangeSources.AnnotationList;
                        ToolManager.UnselectAllAnnotations();
                        _selectionChangedSource = SelectionChangeSources.None;
                    }

                    if (_selectionChangedSource != SelectionChangeSources.ToolManager && SelectedGroupItems != null)
                        ShowAnnotation(SelectedGroupItems[0]);
                }
            }
        }

        public RelayCommand ItemClickCommand { get; set; }

        private void ItemClickCommandImpl(object param)
        {
            if (param is ItemClickEventArgs args && args.ClickedItem is AnnotationItem item && item == SelectedItem)
                ShowAnnotation(item);
        }

        public void ToolManager_SelectionChanged(IAnnot annotation, int pageNumber)
        {
            try
            {
                if (_selectionChangedSource == SelectionChangeSources.None)
                {
                    _selectionChangedSource = SelectionChangeSources.ToolManager;
                    if (annotation == null)
                    {
                        SelectedItem = null;
                    }
                    else
                    {
                        SelectedItem = FindItemWithAnnot(annotation, pageNumber);
                    }
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _selectionChangedSource = SelectionChangeSources.None;
            }
        }

        public void ToolManager_SelectionGroupChanged(Dictionary<IAnnot, int> annotationGroup)
        {
            try
            {
                if (_selectionChangedSource == SelectionChangeSources.None)
                {
                    _selectionChangedSource = SelectionChangeSources.ToolManager;
                    if (annotationGroup == null)
                    {
                        SelectedGroupItems = null;
                    }
                    else
                    {
                        SelectedGroupItems.Clear();
                        foreach (var annotationItem in annotationGroup)
                        {
                            AnnotationItem item = FindItemWithAnnot(annotationItem.Key, annotationItem.Value);
                            SelectedGroupItems.Add(item);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _selectionChangedSource = SelectionChangeSources.None;
            }
        }

        private void ShowAnnotation(AnnotationItem item)
        {
            NavigationState beforeState = NavigationState.TakeSnapshot(PDFViewCtrl);
            SwitchPageIfNecessary(item.PageNumber);
            int nowPage = PDFViewCtrl.GetCurrentPage();

            Rect canvasRect = null;
            try
            {
                _pdfViewCtrl.DocLockRead();
                IAnnot annot = FindAnnotationFromItem(item);
                if (annot?.IsValid() == true)
                {
                    Rect annotScreenRect = _pdfViewCtrl.GetScreenRectForAnnot(annot, item.PageNumber);
                    canvasRect = UtilityFunctions.ConvertFromScreenRectToAnnotationCanvasRect(_pdfViewCtrl, annotScreenRect);
                    canvasRect.Normalize();

                    ScrollRectIntoView(item.PageNumber, canvasRect);

                    annotScreenRect = _pdfViewCtrl.GetScreenRectForAnnot(annot, item.PageNumber);
                    canvasRect = UtilityFunctions.ConvertFromScreenRectToAnnotationCanvasRect(_pdfViewCtrl, annotScreenRect);
                    canvasRect.Normalize();
                }
            }
            catch (Exception e)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                _pdfViewCtrl.DocUnlockRead();
            }

            if (canvasRect == null)
                return;

            NavigationState afterState = NavigationState.TakeSnapshot(PDFViewCtrl);
            NavigationStack?.RegisterJump(beforeState, afterState);

            if (FlashAnimation != null)
                FlashAnnotation(item.PageNumber, canvasRect);
        }

        /// <summary>
        /// Will switch page if the current page can not be scrolled into view.
        /// </summary>
        /// <param name="pageNumber"></param>
        public void SwitchPageIfNecessary(int pageNumber)
        {
            // If in a non-continuous mode we want to first go to the page.
            PDFViewCtrlPagePresentationMode mode = PDFViewCtrl.GetPagePresentationMode();
            if (mode == PDFViewCtrlPagePresentationMode.e_single_page ||
                mode == PDFViewCtrlPagePresentationMode.e_facing ||
                mode == PDFViewCtrlPagePresentationMode.e_facing_cover)
            {
                int currentPage = PDFViewCtrl.GetCurrentPage();
                if (currentPage == pageNumber)
                    return;

                if (mode == PDFViewCtrlPagePresentationMode.e_facing && ((currentPage + 1) / 2 == (pageNumber + 1) / 2))
                {
                    return;
                }
                else if (mode == PDFViewCtrlPagePresentationMode.e_facing_cover && (currentPage / 2 == pageNumber / 2))
                {
                    return;
                }

                _pageChanged?.Invoke(_pdfViewCtrl.GetCurrentPage(), pageNumber);
                PDFViewCtrl.SetCurrentPage(pageNumber);
            }
        }

        public void ScrollRectIntoView(int pageNumber, Rect rect)
        {
            // Now, compare the view's position on the canvas to that of the annotation

            PDFDouble px = new PDFDouble(0);
            PDFDouble py = new PDFDouble(0);
            _pdfViewCtrl.ConvScreenPtToAnnotationCanvasPt(px, py);

            Rect viewRect = new Rect(px.Value, py.Value, px.Value + PDFViewCtrl.ActualWidth, py.Value + PDFViewCtrl.ActualHeight);
            if (viewRect.Width() < 1 || viewRect.Height() < 1)
                return;

            Rect intersectRect = new Rect();
            bool intersect = intersectRect.IntersectRect(viewRect, rect);

            // If we don't intersect enough, scroll it into view.
            if (!intersect || (intersectRect.Width() < rect.Width() - 1) || (intersectRect.Height() < rect.Height() - 1))
            {
                double annotWidth = rect.Width();
                double annotHeight = rect.Height();
                double viewWidth = viewRect.Width();
                double viewHeight = viewRect.Height();
                double x = -1;
                double y = -1;
                double zoomToBeDone = 1.0;
                double MAX_MARGIN = 10.0;

                if (annotWidth > 0 && annotHeight > 0)
                {
                    if (viewWidth < annotWidth)
                    {
                        zoomToBeDone = viewWidth / annotWidth;
                        x = (rect.x1 + rect.x2 - viewWidth) / 2.0; // center
                    }
                    else
                    {
                        double margin = Math.Min(MAX_MARGIN, (viewWidth - annotWidth) / 2.0);
                        if (viewRect.x1 > rect.x1 - margin)
                        {
                            x = Math.Max(rect.x1 - margin, 0);
                        }
                        else if (viewRect.x2 < rect.x2 + margin)
                        {
                            x = Math.Max(rect.x2 + margin - viewWidth, 0);
                        }
                    }

                    if (viewHeight < annotHeight)
                    {
                        zoomToBeDone = Math.Min(viewHeight / annotHeight, zoomToBeDone);
                        y = (rect.y1 + rect.y2 - viewHeight) / 2.0;
                    }
                    else
                    {
                        double margin = Math.Min(MAX_MARGIN, (viewHeight - annotHeight) / 2.0);
                        if (viewRect.y1 > rect.y1 - margin)
                        {
                            y = Math.Max(rect.y1 - margin, 0);
                        }
                        else if (viewRect.y2 < rect.y2 + margin)
                        {
                            y = Math.Max(rect.y2 + margin - viewHeight, 0);
                        }
                    }
                }

                if (x >= 0)
                {
                    PDFViewCtrl.SetHScrollPos(x);
                }
                if (y >= 0)
                {
                    PDFViewCtrl.SetVScrollPos(y);
                }
                if (zoomToBeDone < 1.0)
                {
                    PDFViewCtrl.SetZoom(PDFViewCtrl.GetZoom() * zoomToBeDone);
                }
            }
        }

        private void FlashAnnotation(int pagenumber, Rect rect)
        {
            DetachFlashinfRectangle();

            Canvas canvas = _pdfViewCtrl.GetAnnotationCanvas();
            canvas.Children.Add(_flashingRectangle);
            _flashingRectangle.Tag = canvas;
            _flashingRectangle.Opacity = 0;
            _flashingRectangle.Width = rect.Width();
            _flashingRectangle.Height = rect.Height();
            var px = new PDFDouble(rect.x1);
            var py = new PDFDouble(rect.y1);

            var mode = PDFViewCtrl.GetPagePresentationMode();

            _flashingRectangle.SetValue(Canvas.LeftProperty, px.Value);
            _flashingRectangle.SetValue(Canvas.TopProperty, py.Value);

            FlashAnimation.Begin();
        }

        private void FlashingAnimationCompleted(object sender, object e) => DetachFlashinfRectangle();

        private void DetachFlashinfRectangle()
        {
            if (_flashingRectangle.Tag is Canvas parent)
                parent.Children.Remove(_flashingRectangle);
        }

        private void FlashAnimation_Completed(object sender, object e) => DetachFlashinfRectangle();

        private IAnnot FindAnnotationFromItem(AnnotationItem item)
        {
            foreach (PageHeaderItem header in AnnotationGroups)
            {
                if (header.PageNumber == item.PageNumber && header.Annotations.Contains(item))
                    return FindAnnotationFromIndex(header.Annotations.IndexOf(item), header);
            }
            return null;
        }

        private IAnnot FindAnnotationFromIndex(int index, PageHeaderItem header)
        {
            try
            {
                PDFViewCtrl.DocLockRead();
                PDFDoc doc = PDFViewCtrl.GetDoc();

                Page page = doc.GetPage(header.PageNumber);
                int numAnnots = page.GetNumAnnots();
                int validIndexes = 0;
                for (int i = 0; i < numAnnots; ++i)
                {
                    IAnnot pageAnnot = page.GetAnnot(i);
                    if (pageAnnot?.IsValid() != true)
                        continue;

                    if (WORKABLE_ANNOTATIONS.Contains(pageAnnot.GetAnnotType()) && pageAnnot.IsMarkup())
                    {
                        if (index == validIndexes)
                            return pageAnnot;

                        ++validIndexes;
                    }
                }
            }
            catch (Exception e)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                PDFViewCtrl.DocUnlockRead();
            }

            return null;
        }

        private AnnotationItem FindItemWithAnnot(IAnnot annot, int pageNumber)
        {
            foreach (PageHeaderItem header in AnnotationGroups)
            {
                if (header.PageNumber == pageNumber)
                    return FindItemWithAnnot(annot, header);
            }

            return null;
        }

        private AnnotationItem FindItemWithAnnot(IAnnot annot, PageHeaderItem header)
        {
            try
            {
                PDFViewCtrl.DocLockRead();
                PDFDoc doc = PDFViewCtrl.GetDoc();

                Page page = doc.GetPage(header.PageNumber);
                int numAnnots = page.GetNumAnnots();
                int index = 0;

                for (int i = 0; i < numAnnots; ++i)
                {
                    IAnnot pageAnnot = page.GetAnnot(i);
                    if (pageAnnot?.IsValid() != true)
                        continue;

                    if (WORKABLE_ANNOTATIONS.Contains(pageAnnot.GetAnnotType()) && pageAnnot.IsMarkup())
                    {
                        if (pageAnnot.GetSDFObj().IsEqual(annot.GetSDFObj()))
                        {
                            if (header.Annotations != null && header.Annotations.Count > index)
                                return header.Annotations[index];

                            return null; // something is wrong, and the annotation is not in the list.
                        }

                        ++index;
                    }
                }
            }
            catch (Exception e)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                PDFViewCtrl.DocUnlockRead();
            }

            return null;
        }

        private void ToolManager_AnnotationAdded(IAnnot annotation, int pageNumber)
        {
            if (_nextPageToProcess <= 1 || _nextPageToProcess > pageNumber)
            {
                ModificationList.Add(new Tuple<AnnotationModifications, IAnnot, int>(AnnotationModifications.Add, annotation, pageNumber));
                HandleModifications();
            }
        }

        private void ToolManager_AnnotationGroupAdded(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationAdded(annotationItem.Key, annotationItem.Value);
            }
        }

        private void ToolManager_AnnotationEdited(IAnnot annotation, int pageNumber)
        {
            if (_nextPageToProcess <= 1 || _nextPageToProcess > pageNumber)
            {
                ModificationList.Add(new Tuple<AnnotationModifications, IAnnot, int>(AnnotationModifications.Edit, annotation, pageNumber));
                HandleModifications();
            }
        }

        private void ToolManager_AnnotationGroupEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationEdited(annotationItem.Key, annotationItem.Value);
            }
        }

        private void ToolManager_AnnotationGroupPreEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationEdited(annotationItem.Key, annotationItem.Value);
            }
        }

        private void ToolManager_AnnotationPreRemoved(IAnnot annotation, int pageNumber)
        {
            if (_nextPageToProcess <= 1 || _nextPageToProcess > pageNumber)
            {
                ModificationList.Add(new Tuple<AnnotationModifications, IAnnot, int>(AnnotationModifications.Remove, annotation, pageNumber));
                HandleModifications();
            }
        }

        private void ToolManager_AnnotationGroupPreRemoved(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationPreRemoved(annotationItem.Key, annotationItem.Value);
            }
        }

        private List<Tuple<AnnotationModifications, IAnnot, int>> ModificationList => _modificationList ?? (_modificationList = new List<Tuple<AnnotationModifications, IAnnot, int>>());

        private void HandleModifications()
        {
            while (ModificationList.Count > 0)
            {
                _selectionChangedSource = SelectionChangeSources.ToolManager;
                switch (ModificationList[0].Item1)
                {
                    case AnnotationModifications.Add:
                        AddAnnotation(ModificationList[0].Item2, ModificationList[0].Item3);
                        break;
                    case AnnotationModifications.Edit:
                        EditAnnotation(ModificationList[0].Item2, ModificationList[0].Item3);
                        break;
                    case AnnotationModifications.Remove:
                        RemoveAnnotation(ModificationList[0].Item2, ModificationList[0].Item3);
                        break;
                }

                _selectionChangedSource = SelectionChangeSources.None;
                ModificationList.RemoveAt(0);
            }
        }

        private void AddAnnotation(IAnnot annot, int pageNumber)
        {
            AnnotType annotType = annot.GetAnnotType();
            if (!WORKABLE_ANNOTATIONS.Contains(annotType) || !annot.IsMarkup())
                return;

            PageHeaderItem header = null;
            foreach (PageHeaderItem head in AnnotationGroups)
            {
                if (head.PageNumber >= pageNumber)
                {
                    header = head;
                    break;
                }
            }
            try
            {
                PDFViewCtrl.DocLockRead();
                Page page = PDFViewCtrl.GetDoc().GetPage(pageNumber);

                AnnotationItem item = MakeAnnotationItem(annot, annotType, page);
                if (header == null || header.PageNumber != pageNumber)
                {
                    List<AnnotationItem> items = new List<AnnotationItem> { item };
                    PageHeaderItem newHeader = new PageHeaderItem(items);
                    if (header == null)
                    {
                        AnnotationGroups.Add(newHeader);
                        FoundAnnotations = true;
                    }
                    else
                    {
                        int insertionIndex = AnnotationGroups.IndexOf(header);
                        AnnotationGroups.Insert(insertionIndex, newHeader);
                    }
                }
                else
                {
                    int annotIndex = 0;
                    int numAnnots = page.GetNumAnnots();
                    for (int i = 0; i < numAnnots; ++i)
                    {
                        IAnnot pageAnnot = page.GetAnnot(i);
                        if (pageAnnot?.IsValid() != true) continue;

                        if (WORKABLE_ANNOTATIONS.Contains(pageAnnot.GetAnnotType()) && pageAnnot.IsMarkup())
                        {
                            if (pageAnnot.GetSDFObj().IsEqual(annot.GetSDFObj()))
                            {
                                if (header.Annotations != null && header.Annotations.Count > annotIndex)
                                {
                                    header.Annotations.Insert(annotIndex, item);
                                    return;
                                }

                                header.Annotations.Add(item);
                                return;
                            }

                            ++annotIndex;
                        }
                    }
                }
            }
            catch (Exception e)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(e));
        }
        finally
            {
                PDFViewCtrl.DocUnlockRead();
            }
        }

        private void RemoveAnnotation(IAnnot annot, int pageNumber)
        {
            AnnotationItem item = null;
            PageHeaderItem header = null;
            foreach (PageHeaderItem headerItem in AnnotationGroups)
            {
                if (headerItem.PageNumber == pageNumber)
                {
                    header = headerItem;
                    item = FindItemWithAnnot(annot, header);
                    break;
                }
            }

            if (header != null && item != null && header.Annotations.Contains(item))
            {
                header.Annotations.Remove(item);
                if (header.Annotations.Count == 0)
                    AnnotationGroups.Remove(header);
            }
        }

        private void EditAnnotation(IAnnot annot, int pageNumber)
        {
            AnnotationItem item = null;
            PageHeaderItem header = null;
            foreach (PageHeaderItem headerItem in AnnotationGroups)
            {
                if (headerItem.PageNumber == pageNumber)
                {
                    header = headerItem;
                    item = FindItemWithAnnot(annot, header);
                    break;
                }
            }

            if (header != null && item != null && header.Annotations.Contains(item))
            {
                try
                {
                    PDFViewCtrl.DocLockRead();
                    Page page = PDFViewCtrl.GetDoc().GetPage(pageNumber);
                    AnnotType annotType = annot.GetAnnotType();
                    AnnotationItem newItem = MakeAnnotationItem(annot, annotType, page);
                    item.CopyFrom(newItem);
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    PDFViewCtrl.DocUnlockRead();
                }
            }
        }

        public bool GoBack()
        {
            return false;
        }
    }
}
