﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Settings;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using Windows.UI;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.ColorMode
{
    public partial class CustomColorViewModel : ViewModelBase
    {
        private bool _isPopupOpen;
        private ViewModeType _viewMode;
        private EditModeType _editMode;
        private bool _isLightDismissable = true;
        private CustomColorIconViewModel _currSelectedIcon;
        private Color _currEditBrushGradient;

        // Signals to the view that a color icon is pressed
        private CustomColorSelectedDelegate _customColorSelected;
        public delegate void CustomColorSelectedDelegate(CustomColorIconViewModel icon);
        public event CustomColorSelectedDelegate CustomColorSelected
        {
            add
            {
                if (_customColorSelected?.GetInvocationList().Contains(value) != true)
                    _customColorSelected += value;
            }
            remove => _customColorSelected -= value;
        }

        // Signals to the ViewerViewModel to apply the color icon
        private CustomColorRequestedDelegate _customColorRequested;
        public delegate void CustomColorRequestedDelegate(CustomColorIconViewModel icon);
        public event CustomColorRequestedDelegate CustomColorRequested
        {
            add
            {
                if (_customColorRequested?.GetInvocationList().Contains(value) != true)
                    _customColorRequested += value;
            }
            remove => _customColorRequested -= value;
        }

        private PopupClosedDelegate _popupClosed;
        public delegate void PopupClosedDelegate();
        public event PopupClosedDelegate PopupClosed
        {
            add
            {
                if (_popupClosed?.GetInvocationList().Contains(value) != true)
                    _popupClosed += value;
            }
            remove => _popupClosed -= value;
        }

        public CustomColorViewModel()
        {
            Init();
        }

        public bool IsPopupOpen
        {
            get => _isPopupOpen;
            set
            {
                if (Set(ref _isPopupOpen, value))
                {
                    if (!value)
                    {
                        if (ViewMode == ViewModeType.Edit)
                            PopulateCustomColorCollection();

                        _popupClosed?.Invoke();
                    }
                    else
                    {
                        SetDefaultValues();
                    }
                }
            }
        }

        public ViewModeType ViewMode
        {
            get { return _viewMode; }
            set { Set(ref _viewMode, value); }
        }

        public EditModeType EditMode
        {
            get => _editMode;
            set
            {
                if (Set(ref _editMode, value))
                    UpdateOpacityGradient();
            }
        }

        public bool IsLightDismissable
        {
            get => _isLightDismissable;
            set => Set(ref _isLightDismissable, value);
        }

        public CustomColorIconViewModel CurrSelectedIcon
        {
            get => _currSelectedIcon;
            set
            {
                if (Set(ref _currSelectedIcon, value) && value != null)
                {
                    UpdateOpacityGradient();
                    if (CurrSelectedIcon != null)
                        _customColorSelected?.Invoke(CurrSelectedIcon);
                }
            }
        }

        /// <summary>
        /// Essentially, the RGB equivalent of the HSV value of the current color except with V set to 255
        /// since this is the gradient (topmost color) that the V grid will bind to.
        /// </summary>
        public Color CurrEditBrushGradient
        {
            get => _currEditBrushGradient;
            set => Set(ref _currEditBrushGradient, value);
        }

        public Color CurrEditBrush
        {
            get
            {
                if (CurrSelectedIcon == null)
                    return Colors.Transparent;

                if (EditMode == EditModeType.Background)
                {
                    Color color = CurrSelectedIcon.Background;
                    return Color.FromArgb(255, color.R, color.G, color.B);
                }
                else
                {
                    Color color = CurrSelectedIcon.Foreground;
                    return Color.FromArgb(255, color.R, color.G, color.B);
                }
            }
        }

        /// <summary>
        /// The current V value of the Background/Foreground grid
        /// </summary>
        public double CurrEditBrushOpacity
        {
            get
            {
                if (CurrSelectedIcon == null)
                    return 0;

                Color color = EditMode == EditModeType.Background ? CurrSelectedIcon.Background : CurrSelectedIcon.Foreground;
                return Math.Max(color.B, Math.Max(color.R, color.G)) / 255.0;
            }
        }

        public ObservableCollection<CustomColorIconViewModel> CustomColorIconCollection { get; private set; }

        private void Init()
        {
            CancelCommand = new RelayCommand(CancelCustomColor);
            OkCommand = new RelayCommand(SelectCustomColor);
            IconClickCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(IconClickAsync);
                command.Execute(parameter);
            });
            EditModeCommand = new RelayCommand(SelectEditMode);

            CustomColorIconCollection = new ObservableCollection<CustomColorIconViewModel>();
            PopulateCustomColorCollection();
            if (ViewerSettings.ColorMode == CustomColorModeType.Custom)
                CurrSelectedIcon = CustomColorIconCollection[ViewerSettings.CurrentCustomColorIcon];
        }

        /// <summary>
        /// Since this View/ViewModel is persistent with the ViewerViewModel,
        /// when user navigates back here again, set back to the default values
        /// </summary>
        private void SetDefaultValues()
        {
            ViewMode = ViewModeType.Icon;
            EditMode = EditModeType.Background;
            if (ViewerSettings.ColorMode == CustomColorModeType.Custom)
            {
                CurrSelectedIcon = CustomColorIconCollection[ViewerSettings.CurrentCustomColorIcon];
            }
            else if (CurrSelectedIcon != null)
            {
                CurrSelectedIcon.IsSelected = false;
                CurrSelectedIcon = null;
            }
        }

        public RelayCommand CancelCommand { get; private set; }

        public RelayCommand OkCommand { get; private set; }

        public RelayCommand IconClickCommand { get; private set; }

        public RelayCommand EditModeCommand { get; private set; }

        private void PopulateCustomColorCollection()
        {
            IList<Tuple<Color,Color>> colors = ViewerSettings.GetCustomColors();
            CustomColorIconCollection.Clear();
            foreach(var color in colors)
            {
                CustomColorIconCollection.Add(new CustomColorIconViewModel(color.Item1, color.Item2));
            }

            CustomColorIconViewModel defaultIcon = new CustomColorIconViewModel(Colors.Transparent, Colors.Transparent)
            {
                IsLoadDefault = true
            };

            CustomColorIconCollection.Add(defaultIcon);
        }

        /// <summary>
        /// This is used since the IsLightDismissable property of a popup only takes effect after
        /// the popup is closed then opened again
        /// </summary>
        /// <param name="isDismissable"></param>
        private void ChangeIsLightDismissable(bool isDismissable)
        {
            IsLightDismissable = isDismissable;
            _isPopupOpen = false;
            RaisePropertyChanged(nameof(IsPopupOpen));
            IsPopupOpen = true;
        }

        /// <summary>
        /// This method is used by the view to notify the ViewModel of the new color.
        /// </summary>
        /// <param name="color"></param>
        /// <param name="isHs">Refers to whether the color was taken from the HS grid or the V grid</param>
        public void UpdateColor(Color color, bool isHs)
        {
            if (EditMode == EditModeType.Background)
            {
                CurrSelectedIcon.Background = color;
            }
            else
            {
                CurrSelectedIcon.Foreground = color;
            }

            if (isHs)
            {
                var hsv = UtilityFunctions.GetHSVFromRGB(color);
                CurrEditBrushGradient = UtilityFunctions.GetRGBFromHSV(hsv.Item1, hsv.Item2, 100);
            }
            else
            {
                UpdateOpacityGradient();
            }
        }

        private void UpdateOpacityGradient() => RaisePropertyChanged(nameof(CurrEditBrushOpacity));
    }
}
