﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.ObjectModel;
using System.Linq;
using Windows.ApplicationModel.Resources;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Optimize
{
    public class OptimizeFileViewModel : ViewModelBase
    {
        private PopupClosedDelegate _popupClosed;
        public delegate void PopupClosedDelegate();
        public event PopupClosedDelegate PopupClosed
        {
            add
            {
                if (_popupClosed?.GetInvocationList().Contains(value) != true)
                    _popupClosed += value;
            }
            remove => _popupClosed -= value;
        }

        private OptimizeConfirmedDelegate _optimizeConfirmed;
        public delegate void OptimizeConfirmedDelegate();
        public event OptimizeConfirmedDelegate OptimizeConfirmed
        {
            add
            {
                if (_optimizeConfirmed?.GetInvocationList().Contains(value) != true)
                    _optimizeConfirmed += value;
            }
            remove => _optimizeConfirmed -= value;
        }

        private ViewType _currentView;
        private bool _isPopupOpen;
        private BasicFileSizeType _fileSizeOption = BasicFileSizeType.Balanced;
        private DpiOption _greaterThanDpiOption;
        private DpiOption _downSampleDpiOption;
        private CompressionColorItem _compressionColorOption;
        private CompressionMonochromeItem _compressionMonochromeOption;
        private CompressionQualityItem _compressionQualityOption;
        private ObservableCollection<CompressionColorItem> _compressionColorOptionCollection;
        private ObservableCollection<CompressionMonochromeItem> _compressionMonochromeOptionCollection;
        private ObservableCollection<CompressionQualityItem> _compressionQualityOptionCollection;
        private ObservableCollection<DpiOption> _greaterThanDpiCollection;

        public OptimizeFileViewModel()
        {
            AdvancedCommand = new RelayCommand(AdvancedCommandImpl);
            BasicCommand = new RelayCommand(BasicCommandImpl);
            BasicFileSizeCheckedCommand = new RelayCommand(BasicFileSizeCheckedCommandImpl);
            CancelCommand = new RelayCommand(CancelCommandImpl);
            OkCommand = new RelayCommand(OkCommandImpl);

            PrepareCompressionColorOptions();
            PrepareCompressionMonochromeOptions();
            PrepareCompressionQualityOptions();
            PrepareGreaterThanDpiOptions();
        }

        public RelayCommand AdvancedCommand { get; }

        public RelayCommand BasicCommand { get; }

        public RelayCommand BasicFileSizeCheckedCommand { get; }

        public RelayCommand CancelCommand { get; }

        public RelayCommand OkCommand { get; }

        public ViewType CurrentView
        {
            get => _currentView;
            set => Set(ref _currentView, value);
        }

        public bool IsPopupOpen
        {
            get => _isPopupOpen;
            set
            {
                if (Set(ref _isPopupOpen, value))
                {
                    if (!value)
                    {
                        _popupClosed?.Invoke();
                    }
                    else
                    {
                        SetDefaultValues();
                    }
                }
            }
        }

        public BasicFileSizeType FileSizeOption
        {
            get => _fileSizeOption;
            set => Set(ref _fileSizeOption, value);
        }

        public DpiOption GreaterThanDpiOption
        {
            get => _greaterThanDpiOption;
            set
            {
                Set(ref _greaterThanDpiOption, value);
                RaisePropertyChanged(nameof(DownSampleDpiCollection));
                if (DownSampleDpiOption == null)
                    DownSampleDpiOption = DownSampleDpiCollection.LastOrDefault();
            }
        }

        public DpiOption DownSampleDpiOption
        {
            get => _downSampleDpiOption;
            set => Set(ref _downSampleDpiOption, value);
        }

        public CompressionColorItem CompressionColorOption
        {
            get => _compressionColorOption;
            set
            {
                if (Set(ref _compressionColorOption, value))
                    RaisePropertyChanged(nameof(HasQuality));
            }
        }

        public CompressionMonochromeItem CompressionMonochromeOption
        {
            get => _compressionMonochromeOption;
            set => Set(ref _compressionMonochromeOption, value);
        }

        public CompressionQualityItem CompressionQualityOption
        {
            get => _compressionQualityOption;
            set => Set(ref _compressionQualityOption, value);
        }

        public bool HasQuality => CompressionColorOption != null &&
            (CompressionColorOption.ColorOption == CompressionColorType.JPEG
            || CompressionColorOption.ColorOption == CompressionColorType.JPEG2000);

        public ObservableCollection<CompressionColorItem> CompressionColorOptionCollection => _compressionColorOptionCollection;

        public ObservableCollection<CompressionMonochromeItem> CompressionMonochromeOptionCollection => _compressionMonochromeOptionCollection;

        public ObservableCollection<CompressionQualityItem> CompressionQualityOptionCollection => _compressionQualityOptionCollection;

        public ObservableCollection<DpiOption> GreaterThanDpiCollection => _greaterThanDpiCollection;

        public ObservableCollection<DpiOption> DownSampleDpiCollection => new ObservableCollection<DpiOption>(
            GreaterThanDpiCollection.Where(
                x => GreaterThanDpiOption != null
                && x.Dpi <= GreaterThanDpiOption.Dpi
                && x.Dpi <= 225));

        private void SetDefaultValues()
        {
            CurrentView = ViewType.Basic;
            CompressionColorOption = CompressionColorOptionCollection.FirstOrDefault(x => x.ColorOption == CompressionColorType.JPEG);
            CompressionMonochromeOption = CompressionMonochromeOptionCollection.FirstOrDefault(x => x.MonochromeOption == CompressionMonochromeType.JBIG2);
            CompressionQualityOption = CompressionQualityOptionCollection.FirstOrDefault(x => x.QualityOption == CompressionQualityType.High);
            GreaterThanDpiOption = GreaterThanDpiCollection.FirstOrDefault(x => x.Dpi == 225);
            DownSampleDpiOption = DownSampleDpiCollection.FirstOrDefault(x => x.Dpi == 150);
        }

        private void AdvancedCommandImpl(object command) => CurrentView = ViewType.Advanced;

        private void BasicCommandImpl(object command) => CurrentView = ViewType.Basic;

        private void BasicFileSizeCheckedCommandImpl(object command)
        {
            bool result = Enum.TryParse(command.ToString(), out BasicFileSizeType option);
            if (!result)
                return;

            switch (option)
            {
                case BasicFileSizeType.High:
                    FileSizeOption = BasicFileSizeType.High;
                    break;
                case BasicFileSizeType.Balanced:
                    FileSizeOption = BasicFileSizeType.Balanced;
                    break;
                case BasicFileSizeType.Small:
                    FileSizeOption = BasicFileSizeType.Small;
                    break;
            }
        }

        private void CancelCommandImpl(object command) => IsPopupOpen = false;

        private void OkCommandImpl(object command)
        {
            _isPopupOpen = false;
            RaisePropertyChanged(nameof(IsPopupOpen));
            _optimizeConfirmed?.Invoke();
        }

        public static string CompressionColorOptionToString(CompressionColorType option)
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            switch (option)
            {
                case CompressionColorType.None:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Color_None");
                case CompressionColorType.PNG:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Color_PNG");
                case CompressionColorType.JPEG:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Color_JPEG");
                default:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Color_JPEG2000");
            }
        }
        public static string CompressionMonochromeOptionToString(CompressionMonochromeType option)
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            switch (option)
            {
                case CompressionMonochromeType.PNG:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Monochrome_PNG");
                default:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Monochrome_JBIG2");
            }
        }
        public static string CompressionQualityOptionToString(CompressionQualityType option)
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            switch (option)
            {
                case CompressionQualityType.Low:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Quality_Low");
                case CompressionQualityType.Medium:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Quality_Medium");
                case CompressionQualityType.High:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Quality_High");
                default:
                    return loader.GetString("ViewerPage_Optimize_Dialog_Advanced_Compression_Quality_Max");
            }
        }

        private void PrepareCompressionColorOptions()
        {
            _compressionColorOptionCollection = new ObservableCollection<CompressionColorItem>
            {
                new CompressionColorItem(CompressionColorType.None),
                new CompressionColorItem(CompressionColorType.PNG)
            };

            CompressionColorOption = new CompressionColorItem(CompressionColorType.JPEG);
            _compressionColorOptionCollection.Add(CompressionColorOption);
            _compressionColorOptionCollection.Add(new CompressionColorItem(CompressionColorType.JPEG2000));
        }

        private void PrepareCompressionMonochromeOptions()
        {
            _compressionMonochromeOptionCollection = new ObservableCollection<CompressionMonochromeItem>
            {
                new CompressionMonochromeItem(CompressionMonochromeType.PNG)
            };

            CompressionMonochromeOption = new CompressionMonochromeItem(CompressionMonochromeType.JBIG2);
            _compressionMonochromeOptionCollection.Add(CompressionMonochromeOption);
        }

        private void PrepareCompressionQualityOptions()
        {
            _compressionQualityOptionCollection = new ObservableCollection<CompressionQualityItem>
            {
                new CompressionQualityItem(CompressionQualityType.Low),
                new CompressionQualityItem(CompressionQualityType.Medium)
            };

            CompressionQualityOption = new CompressionQualityItem(CompressionQualityType.High);
            _compressionQualityOptionCollection.Add(CompressionQualityOption);
            _compressionQualityOptionCollection.Add(new CompressionQualityItem(CompressionQualityType.Max));
        }

        private void PrepareGreaterThanDpiOptions()
        {
            _greaterThanDpiCollection = new ObservableCollection<DpiOption>
            {
                new DpiOption(50),
                new DpiOption(72),
                new DpiOption(96),
                new DpiOption(120)
            };

            DownSampleDpiOption = new DpiOption(150);
            _greaterThanDpiCollection.Add(DownSampleDpiOption);
            GreaterThanDpiOption = new DpiOption(225);
            _greaterThanDpiCollection.Add(GreaterThanDpiOption);
            _greaterThanDpiCollection.Add(new DpiOption(300));
            _greaterThanDpiCollection.Add(new DpiOption(600));
        }

        private OptimizerMonoImageSettingsCompressionMode GetPdfNetSettingFromCompressionMonochrome()
        {
            switch (CompressionMonochromeOption.MonochromeOption)
            {
                case CompressionMonochromeType.PNG:
                    return OptimizerMonoImageSettingsCompressionMode.e_flate;
                default:
                    return OptimizerMonoImageSettingsCompressionMode.e_jbig2;
            }
        }

        private OptimizerImageSettingsCompressionMode GetPdfNetSettingFromCompressionColor()
        {
            switch (CompressionColorOption.ColorOption)
            {
                case CompressionColorType.None:
                    return OptimizerImageSettingsCompressionMode.e_none;
                case CompressionColorType.PNG:
                    return OptimizerImageSettingsCompressionMode.e_flate;
                case CompressionColorType.JPEG:
                    return OptimizerImageSettingsCompressionMode.e_jpeg;
                default:
                    return OptimizerImageSettingsCompressionMode.e_jpeg2000;
            }
        }

        private int GetPdfNetSettingFromCompressionQuality()
        {
            switch (CompressionQualityOption.QualityOption)
            {
                case CompressionQualityType.Low:
                    return 4;
                case CompressionQualityType.Medium:
                    return 6;
                case CompressionQualityType.High:
                    return 8;
                default:
                    return 10;
            }
        }

        public void Optimize(PDFDoc doc)
        {
            if (doc == null)
                return;

            bool shouldUnlock = false;
            try
            {
                doc.Lock();
                shouldUnlock = true;

                OptimizerOptimizerSettings optimizerSettings = new OptimizerOptimizerSettings();
                OptimizerImageSettings imageSettings = new OptimizerImageSettings();
                OptimizerMonoImageSettings monoSettings = new OptimizerMonoImageSettings();

                OptimizerImageSettingsDownsampleMode imageDownSettingMode = OptimizerImageSettingsDownsampleMode.e_off;
                OptimizerImageSettingsCompressionMode imageCompressionSettingMode = OptimizerImageSettingsCompressionMode.e_none;
                OptimizerMonoImageSettingsDownsampleMode monoDownSettingMode = OptimizerMonoImageSettingsDownsampleMode.e_off;
                OptimizerMonoImageSettingsCompressionMode monoCompressionSettingMode = OptimizerMonoImageSettingsCompressionMode.e_none;

                double imageMaxDpi = 0;
                double imageResampleDpi = 0;
                double monoMaxDpi = 0;
                double monoResampleDpi = 0;
                int quality = 0;

                bool forceRecompression = false;

                if (CurrentView == ViewType.Basic)
                {
                    // set preset high, balanced, small values
                    switch (FileSizeOption)
                    {
                        case BasicFileSizeType.High:
                            imageDownSettingMode = OptimizerImageSettingsDownsampleMode.e_off;
                            imageCompressionSettingMode = OptimizerImageSettingsCompressionMode.e_retain;
                            quality = 10;
                            monoDownSettingMode = OptimizerMonoImageSettingsDownsampleMode.e_off;
                            monoCompressionSettingMode = OptimizerMonoImageSettingsCompressionMode.e_jbig2;
                            forceRecompression = false;
                            break;
                        case BasicFileSizeType.Balanced:
                            imageDownSettingMode = OptimizerImageSettingsDownsampleMode.e_default;
                            imageMaxDpi = 225.0;
                            imageResampleDpi = 150.0;
                            imageCompressionSettingMode = OptimizerImageSettingsCompressionMode.e_jpeg;
                            quality = 8;
                            monoDownSettingMode = OptimizerMonoImageSettingsDownsampleMode.e_default;
                            monoMaxDpi = imageMaxDpi * 2.0;
                            monoResampleDpi = imageResampleDpi * 2.0;
                            monoCompressionSettingMode = OptimizerMonoImageSettingsCompressionMode.e_jbig2;
                            forceRecompression = true;
                            break;
                        case BasicFileSizeType.Small:
                            imageDownSettingMode = OptimizerImageSettingsDownsampleMode.e_default;
                            imageMaxDpi = 120.0;
                            imageResampleDpi = 96.0;
                            imageCompressionSettingMode = OptimizerImageSettingsCompressionMode.e_jpeg;
                            quality = 6;
                            monoDownSettingMode = OptimizerMonoImageSettingsDownsampleMode.e_default;
                            monoMaxDpi = imageMaxDpi * 2.0;
                            monoResampleDpi = imageResampleDpi * 2.0;
                            monoCompressionSettingMode = OptimizerMonoImageSettingsCompressionMode.e_jbig2;
                            forceRecompression = true;
                            break;
                    }
                }
                // otherwise, get the user selected values from advanced
                else
                {
                    imageDownSettingMode = OptimizerImageSettingsDownsampleMode.e_default;
                    imageCompressionSettingMode = GetPdfNetSettingFromCompressionColor();
                    monoDownSettingMode = OptimizerMonoImageSettingsDownsampleMode.e_default;
                    monoCompressionSettingMode = GetPdfNetSettingFromCompressionMonochrome();
                    imageMaxDpi = GreaterThanDpiOption.Dpi;
                    imageResampleDpi = DownSampleDpiOption.Dpi;
                    quality = GetPdfNetSettingFromCompressionQuality();
                    forceRecompression = true;
                    monoMaxDpi = GreaterThanDpiOption.Dpi * 2.0;
                    monoResampleDpi = DownSampleDpiOption.Dpi * 2.0;
                }

                // apply the values
                imageSettings.SetDownsampleMode(imageDownSettingMode);
                imageSettings.SetImageDPI(imageMaxDpi, imageResampleDpi);
                imageSettings.SetCompressionMode(imageCompressionSettingMode);
                imageSettings.SetQuality(quality);
                imageSettings.ForceChanges(false);
                imageSettings.ForceRecompression(forceRecompression);

                monoSettings.SetDownsampleMode(monoDownSettingMode);
                monoSettings.SetImageDPI(monoMaxDpi, monoResampleDpi);
                monoSettings.SetCompressionMode(monoCompressionSettingMode);
                monoSettings.ForceChanges(false);
                monoSettings.ForceRecompression(forceRecompression);

                optimizerSettings.SetColorImageSettings(imageSettings);
                optimizerSettings.SetGrayscaleImageSettings(imageSettings);
                optimizerSettings.SetMonoImageSettings(monoSettings);

                Optimizer.Optimize(doc, optimizerSettings);
            }
            catch
            {
                // log error
            }

            finally
            {
                if (shouldUnlock)
                {
                    doc.Unlock();
                    shouldUnlock = false;
                }
            }
        }
    }
}
