﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline
{
    public class OutlineDialogViewModel : ViewModelBase, IDisposable
    {
        private OutlineDialogStateBundle _restoreState;
        private bool _hasRestoredUserBookmarks = true;
        private bool _hasRestoredOutline = true;
        private bool _hasRestoredAnnotationList = true;
        private NavigationStack _navigationStack;
        private bool _closeDialogOnItemClick;
        private bool _activatedSubViews;
        private bool _isConverting;
        private SubViews _subView;
        private int _subviewIndex;
        private ThumbnailViewer _thumbnails;
        private Controls.Outline _outline;
        private UserBookmarkControl _userBookmarks;
        private AnnotationList _annotationList;
        private Storyboard _fadeAnimation;
        private bool _isPivotItemThumbnailsEnabled = true;
        private bool _isPivotItemBookmarksEnabled = true;
        private bool _isPivotItemOutlineEnabled = true;
        private bool _isPivotItemAnnotationsEnabled = true;

        private Controls.Outline _oldOutline;
        private UserBookmarkControl _oldUserBookmarks;

        public PDFViewCtrl PDFViewCtrl { get; set; }

        public NavigationStack NavigationStack
        {
            get => _navigationStack;
            set
            {
                _navigationStack = value;
                if (AnnotationList != null)
                    AnnotationList.NavigationStack = value;

                if (Outline != null)
                    Outline.NavigationStack = value;

                if (UserBookmarks != null)
                    UserBookmarks.NavigationStack = value;

                if (Thumbnails != null)
                    Thumbnails.NavigationStack = value;
            }
        }

        /// <summary>
        /// Gets or sets whether clicking on an item should collapse  the outline.
        /// </summary>
        public bool CloseDialogOnItemClick
        {
            get => _closeDialogOnItemClick;
            set
            {
                if (Set(ref _closeDialogOnItemClick, value) && UserBookmarks != null)
                    UserBookmarks.ShowOverflow = _closeDialogOnItemClick;
            }
        }

        public string DocumentTag { get; set; }

        private EventHandler _requestClosing;
        public event EventHandler RequestClosing
        {
            add
            {
                if (_requestClosing?.GetInvocationList().Contains(value) != true)
                    _requestClosing += value;
            }
            remove => _requestClosing -= value;
        }

        private EventHandler _requestSwitchSide;
        public event EventHandler RequestSwitchSide
        {
            add
            {
                if (_requestSwitchSide?.GetInvocationList().Contains(value) != true)
                    _requestSwitchSide += value;
            }
            remove => _requestSwitchSide -= value;
        }

        private DocumentModifiedEventHandler _documentModified;
        public delegate void DocumentModifiedEventHandler(PDFDoc doc);
        public event DocumentModifiedEventHandler DocumentModified
        {
            add
            {
                if (_documentModified?.GetInvocationList().Contains(value) != true)
                    _documentModified += value;
            }
            remove => _documentModified -= value;
        }

        private UserBookmarksEditedEventHandler _userBookmarksEdited;
        public delegate void UserBookmarksEditedEventHandler(PDFDoc doc);
        public event UserBookmarksEditedEventHandler UserBookmarksEdited
        {
            add
            {
                if (_userBookmarksEdited?.GetInvocationList().Contains(value) != true)
                    _userBookmarksEdited += value;
            }
            remove => _userBookmarksEdited -= value;
        }

        private PageChangedDelegate _pageChanged;
        public delegate void PageChangedDelegate(int prevPage, int newPage);
        public event PageChangedDelegate PageChanged
        {
            add
            {
                if (_pageChanged?.GetInvocationList().Contains(value) != true)
                    _pageChanged += value;
            }
            remove => _pageChanged -= value;
        }

        private void RaiseDocumentModified(PDFDoc doc) => _documentModified?.Invoke(doc);

        public OutlineDialogViewModel() => InitCommands();

        public OutlineDialogViewModel(PDFViewCtrl ctrl)
        {
            InitCommands();
            PDFViewCtrl = ctrl;
        }

        public void ActivateSubView()
        {
            DocumentTag = "NULL";
            if (_activatedSubViews)
                return;

            _activatedSubViews = true;
            Thumbnails = new ThumbnailViewer(PDFViewCtrl, DocumentTag)
            {
                Background = new SolidColorBrush(Colors.Transparent),
                BlankPageDefaultColor = UtilityFunctions.GetPostProcessedColor(Colors.White, PDFViewCtrl),
                FitItemsToWidth = true,
                NavigationOnly = true,
                NumberOfColumns = 2,
                NavigationStack = NavigationStack,
                ItemPadding = new Thickness(1, 2, 6, 2)
            };

            Thumbnails.ViewModel.PageClicked += ViewModel_PageChanged;

            UserBookmarks = new UserBookmarkControl(PDFViewCtrl, DocumentTag)
            {
                ShowOverflow = _closeDialogOnItemClick,
                CanAwalysEdit = false,
                AddButtonVisibility = Visibility.Collapsed,
                NavigationStack = NavigationStack
            };

            UserBookmarks._ViewModel.PageChanged += ViewModel_PageChanged;
            UserBookmarks.BookmarksEdited += UserBookmarks_BookmarksEdited;

            Outline = new Controls.Outline(PDFViewCtrl)
            {
                NavigationStack = NavigationStack
            };

            Outline.PageChanged += ViewModel_PageChanged;

            AnnotationList = new AnnotationList(PDFViewCtrl)
            {
                NavigationStack = NavigationStack
            };

            AnnotationList.ItemClicked += AnnotationList_ItemClicked;
            AnnotationList.ViewModel.PageChanged += ViewModel_PageChanged;
            UserBookmarks.ItemClicked += ItemClickedHandler;
            Outline.ItemClicked += ItemClickedHandler;
            AnnotationList.ItemClicked += ItemClickedHandler;
            Thumbnails.ItemClicked += ItemClickedHandler;

            UpdateColumnNumber();
        }

        public OutlineDialogStateBundle GetState()
        {
            double userBookmarksState = 0;
            if (_restoreState != null && !_hasRestoredUserBookmarks)
            {
                userBookmarksState = _restoreState.UserBookmarksState;
            }
            else if (UserBookmarks != null)
            {
                userBookmarksState = UserBookmarks.GetState();
            }

            List<int> outlineState;
            if (_restoreState != null && !_hasRestoredOutline)
            {
                outlineState = _restoreState.OutlineState;
            }
            else
            {
                outlineState = Outline.GetState();
            }

            double annotationListState;
            if (_restoreState != null && !_hasRestoredAnnotationList)
            {
                annotationListState = _restoreState.AnnotationListState;
            }
            else
            {
                annotationListState = AnnotationList.GetState();
            }

            return new OutlineDialogStateBundle(SubView, userBookmarksState, outlineState, annotationListState);
        }

        public void RestoreState(OutlineDialogStateBundle stateBundle)
        {
            _restoreState = stateBundle;
            _hasRestoredUserBookmarks = false;
            _hasRestoredOutline = false;
            _hasRestoredAnnotationList = false;

            ActivateSubView();
            SubView = stateBundle.VisibleState;
            AnnotationList.ViewIsVisible = _subView == SubViews.AnnotationList;
            Thumbnails.ViewIsVisible = _subView == SubViews.Thumbnails;
            RestoreCurrentSubView();
        }

        private void RestoreCurrentSubView()
        {
            if (_restoreState != null)
            {
                if (SubView == SubViews.UserBookmarks && !_hasRestoredUserBookmarks)
                {
                    if (_restoreState.UserBookmarksState != 0)
                        UserBookmarks.RestoreState(_restoreState.UserBookmarksState);

                    _hasRestoredUserBookmarks = true;
                }
                else if (SubView == SubViews.Outline && !_hasRestoredOutline)
                {
                    Task.Run(async () => { await Outline.ShowAsync(); });

                    if (_restoreState.OutlineState != null)
                        Outline.RestoreState(_restoreState.OutlineState);

                    _hasRestoredOutline = true;
                }
                else if (SubView == SubViews.AnnotationList && !_hasRestoredAnnotationList)
                {
                    AnnotationList.ViewModel.LoadAnnotationList();

                    if (_restoreState.AnnotationListState != 0)
                        AnnotationList.RestoreState(_restoreState.AnnotationListState);

                    _hasRestoredAnnotationList = true;
                }

                if (_hasRestoredUserBookmarks && _hasRestoredOutline && _hasRestoredAnnotationList)
                    _restoreState = null;
            }
        }

        public async Task CancelAllViewModels()
        {
            await CleanUpSubViewsAsync();
        }

        public async Task CleanUpSubViewsAsync()
        {
            if (AnnotationList != null)
                await AnnotationList.WaitForAnnotationListToFinish();

            if (UserBookmarks != null)
                await UserBookmarks.WaitForBookmarkSavingAsync();

            if (Outline != null)
                await Outline.WaitForOutlineToLoadAsync();
        }

        /// <summary>
        /// Call this function to indicate that the user has interacted outside the Outline Dialog
        /// </summary>
        public void InteractionOutsideDialog()
        {
            SaveBookmarks();
            UserBookmarks?._ViewModel?.DeselectAll();
        }

        public void SaveBookmarks() => UserBookmarks?.SaveBookmarks();

        public bool Isconverting
        {
            get => _isConverting;
            set => Set(ref _isConverting, value);
        }

        public SubViews SubView
        {
            get => _subView;
            set
            {
                if (Set(ref _subView, value))
                {
                    SubViewIndex = (int)_subView;
                    AnnotationList.ViewIsVisible = _subView == SubViews.AnnotationList;
                    Thumbnails.ViewIsVisible = _subView == SubViews.Thumbnails;
                    RestoreCurrentSubView();
                }
            }
        }

        public bool HasUnsavedUserbookmarks
        {
            get
            {
                if (UserBookmarks?._ViewModel != null)
                    return UserBookmarks._ViewModel.HasUnsavedBookmarks;

                return false;
            }
        }

        public int SubViewIndex
        {
            get => _subviewIndex;
            set
            {
                if (Set(ref _subviewIndex, value) && SubView != (SubViews)_subviewIndex)
                {
                    SubView = (SubViews)_subviewIndex;
                    ViewerSettings.OutlineDefaultView = _subviewIndex;
                }
            }
        }

        public ThumbnailViewer Thumbnails
        {
            get => _thumbnails;
            set => Set(ref _thumbnails, value);
        }

        public Controls.Outline Outline
        {
            get => _outline;
            set {
                _oldOutline = _outline;
                if(Set(ref _outline, value))
                {
                    if(_oldOutline != null)
                        _oldOutline.DocumentModified -= OutlinesDocumentModified;
                    if(_outline != null)
                        _outline.DocumentModified += OutlinesDocumentModified;
                }
            }
        }

        /// <summary>
        /// Enable weather the ability to edit the outline is enabled or disabled
        /// </summary>
        public bool IsOutlineEditingEnabled
        {
            get
            {
                if (Outline != null)
                    return Outline.IsEditEnabled;
                else
                {
                    return false;
                }
            }
            set
            {
                if (Outline != null)
                    Outline.IsEditEnabled = value;
            }
        }

        public UserBookmarkControl UserBookmarks
        {
            get => _userBookmarks;
            set
            {
                _oldUserBookmarks = _userBookmarks;
                if (Set(ref _userBookmarks, value))
                {
                    if (_oldUserBookmarks != null)
                        _oldUserBookmarks.DocumentModified -= UserBookmarksDocumentModified;

                    if (_userBookmarks != null)
                        _userBookmarks.DocumentModified += UserBookmarksDocumentModified;
                }
            }
        }

        private void OutlinesDocumentModified(PDFDoc doc) => RaiseDocumentModified(doc);

        private void UserBookmarksDocumentModified(PDFDoc doc) => RaiseDocumentModified(doc);

        public AnnotationList AnnotationList
        {
            get => _annotationList;
            set => Set(ref _annotationList, value);
        }

        public Storyboard FadeAnimation
        {
            get => _fadeAnimation;
            set => Set(ref _fadeAnimation, value);
        }

        public int NumberOfColumns
        {
            get
            {
                if (Thumbnails != null)
                    return Thumbnails.NumberOfColumns;

                return 0;
            }
        }

        private void AnnotationList_ItemClicked(object sender, ItemClickEventArgs e) => _fadeAnimation?.Begin();

        private void ViewModel_PageChanged(int prevPage, int newPage) => _pageChanged?.Invoke(prevPage, newPage);

        private void UserBookmarks_BookmarksEdited(PDFDoc doc) => _userBookmarksEdited?.Invoke(doc);

        private void ItemClickedHandler(object sender, ItemClickEventArgs e)
        {
            if (CloseDialogOnItemClick)
                _requestClosing?.Invoke(this, EventArgs.Empty);
        }

        private void UpdateColumnNumber()
        {
            if (Thumbnails != null)
            {
                Thumbnails.NumberOfColumns = ViewerSettings.OutlineDialogNumThumbnailColumns;
                RaisePropertyChanged(nameof(NumberOfColumns));
            }
        }

        private void InitCommands()
        {
            SubViewSelectionCommand = new RelayCommand(SubViewSelectionCommandImpl);
            SelectionChangedCommand = new RelayCommand(SelectionChangedCommandImpl);
            CloseCommand = new RelayCommand(CloseCommandImpl);
            SwitchSidesCommand = new RelayCommand(SwitchSidesCommandImpl);
            AddBookmarkCommand = new RelayCommand(AddBookmarkCommandImpl);
            SetNumColumnsCommand = new RelayCommand(SetNumColumnsCommandImpl);            
        }

        public RelayCommand SubViewSelectionCommand { get; private set; }

        public RelayCommand SelectionChangedCommand { get; private set; }

        public RelayCommand CloseCommand { get; private set; }

        public RelayCommand SwitchSidesCommand { get; private set; }

        public RelayCommand AddBookmarkCommand { get; private set; }

        public RelayCommand SetNumColumnsCommand { get; private set; }

        private void SubViewSelectionCommandImpl(object commandParam)
        {
            string commandName = commandParam as string;
            if (!string.IsNullOrWhiteSpace(commandName))
            {
                if (commandName.Equals("UserBookmarks", StringComparison.OrdinalIgnoreCase))
                {
                    SubView = SubViews.UserBookmarks;
                }
                else if (commandName.Equals("Outline", StringComparison.OrdinalIgnoreCase))
                {
                    SubView = SubViews.Outline;
                }
                else if (commandName.Equals("AnnotationList", StringComparison.OrdinalIgnoreCase))
                {
                    SubView = SubViews.AnnotationList;
                }
            }
        }

        /// <summary>
        /// We just want to make sure it's in it's default state when opened
        /// </summary>
        /// <param name="parameters"></param>
        private void SelectionChangedCommandImpl(object parameters) => UserBookmarks?.GoBack();

        private void CloseCommandImpl(object parameters) => _requestClosing?.Invoke(this, EventArgs.Empty);

        private void SwitchSidesCommandImpl(object parameters) => _requestSwitchSide?.Invoke(this, EventArgs.Empty);

        private void AddBookmarkCommandImpl(object parameter)
        {
            if (!Isconverting)
                UserBookmarks?._ViewModel?.AddBookmarkCommand?.Execute(parameter);
        }

        private void SetNumColumnsCommandImpl(object parameter)
        {
            string numString = parameter as string;
            if (!string.IsNullOrEmpty(numString) && int.TryParse(numString, out int columnNumber))
            {
                ViewerSettings.OutlineDialogNumThumbnailColumns = columnNumber;
                UpdateColumnNumber();
            }
        }

        public override bool GoBack()
        {
            if (UserBookmarks?.GoBack() == true)
                return true;

            if (Outline != null)
            {
                //if (Outline.GoBack())
                //{
                //    return true;
                //}
            }
            if (AnnotationList?.GoBack() == true)
                return true;

            SaveBookmarks();
            return false;
        }

        public bool IsPivotItemThumbnailsEnabled
        {
            get => _isPivotItemThumbnailsEnabled;
            set => Set(ref _isPivotItemThumbnailsEnabled, value);
        }

        public bool IsPivotItemBookmarksEnabled
        {
            get => _isPivotItemBookmarksEnabled;
            set => Set(ref _isPivotItemBookmarksEnabled, value);
        }

        public bool IsPivotItemOutlineEnabled
        {
            get => _isPivotItemOutlineEnabled;
            set => Set(ref _isPivotItemOutlineEnabled, value);
        }

        public bool IsPivotItemAnnotationsEnabled
        {
            get => _isPivotItemAnnotationsEnabled;
            set => Set(ref _isPivotItemAnnotationsEnabled, value);
        }

        /// <summary>
        /// Unsubscribes the objects within this class from events and sets them to null.
        /// </summary>
        public void Dispose()
        {
            if (Thumbnails != null)
            {
                Thumbnails.ViewModel.PageClicked -= ViewModel_PageChanged;
                Thumbnails.ItemClicked -= ItemClickedHandler;
                Thumbnails = null;
            }

            if (UserBookmarks != null)
            {
                UserBookmarks._ViewModel.PageChanged -= ViewModel_PageChanged;
                UserBookmarks.BookmarksEdited -= UserBookmarks_BookmarksEdited;
                UserBookmarks.ItemClicked -= ItemClickedHandler;
                UserBookmarks = null;
            }

            if (AnnotationList != null)
            {
                AnnotationList.ItemClicked -= AnnotationList_ItemClicked;
                AnnotationList.ViewModel.PageChanged -= ViewModel_PageChanged;
                AnnotationList.ItemClicked -= ItemClickedHandler;
                AnnotationList = null;
            }

            if (Outline != null)
            {
                Outline.ItemClicked -= ItemClickedHandler;
                Outline.PageChanged -= ViewModel_PageChanged;
                Outline = null;
            }

            if (_oldOutline != null)
            {
                _oldOutline.Dispose();
                _oldOutline.DocumentModified -= OutlinesDocumentModified;
                _oldOutline = null;
            }

            if (_oldUserBookmarks != null)
            {
                _oldUserBookmarks.DocumentModified -= UserBookmarksDocumentModified;
                _oldUserBookmarks = null;
            }
        }
    }
}
