﻿using pdftron.Common;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Settings;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.Foundation;
using Windows.System;
using Windows.UI.Core;
using Windows.UI.Popups;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;
using UIRect = Windows.Foundation.Rect;

// TODO: check it due some FindText related crashes
namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Search
{
    public class FindTextViewModel : ViewModelBase
    {
        private readonly PDFViewCtrl _pdfViewCtrl;
        private TextHighlighter _textHighlighter;
        private readonly SolidColorBrush _nightModecolor = new SolidColorBrush(Windows.UI.Color.FromArgb(100, 200, 0, 0));
        private readonly SolidColorBrush _normalColor = new SolidColorBrush(Windows.UI.Color.FromArgb(100, 100, 0, 255));
        private readonly SolidColorBrush _currentColor;
        private bool _isPrevNextAvailable;
        private double _searchProgress;
        private bool _isSearchProgessVisible;
        private string _searchTextTerm = string.Empty;
        private string _prevSearchTextTerm = string.Empty;
        private bool _isInSearchMode;
        private bool _isTermNotFound;
        private bool _isSearchBoxEnabled = true;
        private FullSearchItem _currentFullSearchItem;
        private IAsyncOperation<PDFViewCtrlSelection> _searchResult;
        private List<Rectangle> _onScreenSelection = new List<Rectangle>();
        private string _searchString;
        private bool _cancelSearch;
        private bool _isSearching;
        private bool _foundFullSearchItems;
        private int _currentFullSearchPage = 1;
        private PDFViewCtrlSelection _currentSearchSelection;
        // search progress
        private DispatcherTimer _searchProgressTimer;
        private int _ticksBeforeProgress;
        private FullTextSearchCanceller _currentTextSearchCanceller;
        private SemaphoreSlim _textSearchSemaphore;
        private FindTextClosedDelegate _findTextClosed;
        private FindTextResultFoundDelegate _findTextResultFound;

        public delegate void FindTextClosedDelegate(Deferral deferral);
        public event FindTextClosedDelegate FindTextClosed
        {
            add
            {
                if (_findTextClosed?.GetInvocationList().Contains(value) != true)
                    _findTextClosed += value;
            }
            remove => _findTextClosed -= value;
        }

        public delegate void PageChangedDelegate(int prevPage, int newPage, Deferral deferral);
        public event PageChangedDelegate PageChanged;

        public delegate void FindTextResultFoundDelegate(Deferral deferral);
        public event FindTextResultFoundDelegate FindTextResultFound
        {
            add
            {
                if (_findTextResultFound?.GetInvocationList().Contains(value) != true)
                    _findTextResultFound += value;
            }
            remove => _findTextResultFound -= value;
        }

        public event EventHandler<Deferral> FocusRequested;

        public RelayCommand SearchCommand { get; private set; }

        public RelayCommand SearchFullCommand { get; private set; }

        public RelayCommand FullSearchItemClickedCommand { get; private set; }

        public RelayCommand PrevNextCommand { get; private set; }

        public RelayCommand CloseCommand { get; private set; }

        public RelayCommand SearchTermChangedCommand { get; private set; }

        public RelayCommand SearchTermKeyUpCommand { get; private set; }

        public FindTextViewModel(PDFViewCtrl ctrl, CustomColorModeType colorMode)
        {
            if (ctrl != null)
            {
                _pdfViewCtrl = ctrl;
                _pdfViewCtrl.OnScale += PDFViewCtrl_OnScale;
                _pdfViewCtrl.OnPageNumberChanged += PDFViewCtrl_OnPageNumberChanged;
            }

            InitCommands();

            if (colorMode == CustomColorModeType.Night)
            //if (SharedSettings.ColorMode == CustomColorModeType.Night)
            {
                _currentColor = _nightModecolor;
            }
            else
            {
                _currentColor = _normalColor;
            }

            FullSearchItems = new ObservableCollection<FullSearchItem>();
        }

        public void ClearViewModel()
        {
            _cancelSearch = true;
            CancelCurrentSearch();
            if (_pdfViewCtrl != null)
            {
                Canvas currentCanvas = _pdfViewCtrl.GetAnnotationCanvas();

                if (currentCanvas != null)
                    currentCanvas.Children.Clear();

                _pdfViewCtrl.OnScale -= PDFViewCtrl_OnScale;
                _pdfViewCtrl.OnPageNumberChanged -= PDFViewCtrl_OnPageNumberChanged;
            }
        }

        /// <summary>
        /// Wait for this before disposing the PDFDoc, after you call ClearViewModel().
        /// This guarantees that no searching is active.
        /// </summary>
        /// <returns></returns>
        public async Task WaitForTextSearchToCancel()
        {
            await TextSearchSemaphore.WaitAsync();
            TextSearchSemaphore.Release();
        }

        public async Task<bool> HandleKeyboardEventAsync(VirtualKey key, bool isCtrlDown, bool isAltDown, bool isShiftDown)
        {
            if (IsPrevNextAvailable && key == VirtualKey.F3)
            {
                if (!isCtrlDown && !isAltDown && !isShiftDown)
                {
                    await PrevNextImplAsync(false);
                    return true;
                }
                else if (!isCtrlDown && !isAltDown && isShiftDown)
                {
                    await PrevNextImplAsync(true);
                    return true;
                }
            }

            if (IsPrevNextAvailable && key == VirtualKey.Left)                    
            {
                await PrevNextAsync("prev");
            }

            if (IsPrevNextAvailable && key == VirtualKey.Right)
            {
                await PrevNextAsync("next");
            }

            return false;
        }

        private void InitCommands()
        {
            SearchCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(SearchAsync);
                command.Execute(parameter);
            });
            SearchFullCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(CreateSearchListAsync);
                command.Execute(parameter);
            });
            FullSearchItemClickedCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(FullSearchItemClickedAsync);
                command.Execute(parameter);
            });
            PrevNextCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PrevNextAsync);
                command.Execute(parameter);
            });
            CloseCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(CloseAsync);
                command.Execute(parameter);
            });

            SearchTermChangedCommand = new RelayCommand(SearchTermChangedCommandImpl);
            SearchTermKeyUpCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(SearchTermKeyUpAsync);
                command.Execute(parameter);
            });
        }

        private async Task SearchAsync(object ignore) => await BeginSearchAsync();

        private async Task FullSearchItemClickedAsync(object eventArgs)
        {
            if (eventArgs is ItemClickEventArgs args && args.ClickedItem is FullSearchItem item)
            {
                if (CurrentFullSearchItem != item)
                {
                    CurrentFullSearchItem = item;
                    _pdfViewCtrl.ClearSelection();
                }

                await HighlightFullSearchItemAsync();
            }
        }

        private async Task PrevNextAsync(object direction)
        {
            string dir = direction as string;
            if (!string.IsNullOrWhiteSpace(dir)) {
                bool backwards = dir.Equals("prev", StringComparison.OrdinalIgnoreCase);
                bool forwards = dir.Equals("next", StringComparison.OrdinalIgnoreCase);
                if (backwards || forwards)
                    await PrevNextImplAsync(backwards);
            }
        }

        private async Task PrevNextImplAsync(bool goBackwards)
        {
            if (CurrentFullSearchItem != null)
            {
                if (goBackwards)
                {
                    int index = FullSearchItems.IndexOf(CurrentFullSearchItem) - 1;
                    // Check if at first search item and moving back (need to wrap around to end page)
                    if (index < 0)
                    {
                        if (FoundFullSearchItems)
                        {
                            index = FullSearchItems.Count - 1;
                        }
                        // If full search is not completed, use FindTextAsync to get the proper search item in last page
                        else
                        {
                            CurrentFullSearchItem = null;
                            FindSameTerm(true, _pdfViewCtrl.GetPageCount());
                            return;
                        }
                    }

                    if (index >= 0 && index < FullSearchItems.Count)
                        CurrentFullSearchItem = FullSearchItems[index];
                }
                else
                {
                    // Check if at last full search item and trying to move to first page (wrap around)
                    int index = FullSearchItems.IndexOf(CurrentFullSearchItem) + 1;
                    if (index > FullSearchItems.Count - 1)
                    {
                        if (FoundFullSearchItems)
                        {
                            index = 0;
                        }
                        // If full search is not completed, use FindTextAsync incase we are not yet truly at the last item yet
                        else
                        {
                            FindSameTerm(false, CurrentFullSearchItem.PageNumber);
                            CurrentFullSearchItem = null;
                            return;
                        }
                    }

                    if (index >= 0 && index < FullSearchItems.Count)
                        CurrentFullSearchItem = FullSearchItems[index];
                }

                await HighlightFullSearchItemAsync();
            }
            // Handle case when using prev/next from normal search
            else
            {
                if (goBackwards)
                {
                    FindSameTerm(true);
                }
                else
                {
                    FindSameTerm(false);
                }
            }
        }

        private async Task CloseAsync(object ignore) => await CloseDialogAsync();

        private void SearchTermChangedCommandImpl(object newText)
        {
            IsTermNotFound = false;
            string text = newText as string;
            IsSearchProgessVisible = false;
            SearchTextTerm = text;
            CancelCurrentSearch();
            IsPrevNextAvailable = false;
        }

        private async Task SearchTermKeyUpAsync(object keyArgs)
        {
            KeyRoutedEventArgs args = keyArgs as KeyRoutedEventArgs;
            if (args.Key == VirtualKey.Enter)
            {
                await BeginSearchAsync();
            }
            else if (args.Key == VirtualKey.Escape)
            {
                await CloseDialogAsync();
            }
        }

        private void PDFViewCtrl_OnScale() => HighlightSelection(_currentSearchSelection);

        private void PDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages) => HighlightSelection(_currentSearchSelection);

        public bool IsPrevNextAvailable
        {
            get => _isPrevNextAvailable;
            set => Set(ref _isPrevNextAvailable, value);
        }

        public double SearchProgress
        {
            get => _searchProgress;
            set => Set(ref _searchProgress, value);
        }

        public bool IsSearchProgessVisible
        {
            get => _isSearchProgessVisible;
            set => Set(ref _isSearchProgessVisible, value);
        }

        public string SearchTextTerm
        {
            get => _searchTextTerm;
            set
            {
                if (Set(ref _searchTextTerm, value))
                    RaisePropertyChanged(nameof(IsFullSearchEnabled));
            }
        }

        public string PrevSearchTextTerm
        {
            get => _prevSearchTextTerm;
            set => Set(ref _prevSearchTextTerm, value);
        }

        public bool IsInSearchMode
        {
            get => _pdfViewCtrl != null && _isInSearchMode;
            set => Set(ref _isInSearchMode, value);
        }

        public bool IsTermNotFound
        {
            get => _isTermNotFound;
            set => Set(ref _isTermNotFound, value);
        }

        public bool IsSearchBoxEnabled
        {
            get => _isSearchBoxEnabled;
            set => Set(ref _isSearchBoxEnabled, value);
        }

        public bool IsFullSearchEnabled => !string.IsNullOrEmpty(SearchTextTerm) && _pdfViewCtrl != null;

        public ObservableCollection<FullSearchItem> FullSearchItems { get; }

        public FullSearchItem CurrentFullSearchItem
        {
            get => _currentFullSearchItem;
            set => Set(ref _currentFullSearchItem, value);
        }

        private bool IsSearching
        {
            get => _isSearching;
            set
            {
                DateTime time = DateTime.Now;
                _isSearching = value;
            }
        }

        public bool FoundFullSearchItems
        {
            get => _foundFullSearchItems;
            set => Set(ref _foundFullSearchItems, value);
        }

        public string CurrentFullSearchStatusText
        {
            get
            {
                var loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                return string.Format(loader.GetString("FindTextDialog_FullSearch_CurrentPageText"), _currentFullSearchPage, _pdfViewCtrl.GetPageCount());
            }
        }

        public string FullSearchResultText
        {
            get
            {
                var loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                return string.Format(loader.GetString("FindTextDialog_FullSearch_Results"), FullSearchItems.Count);
            }
        }

        public SemaphoreSlim TextSearchSemaphore => _textSearchSemaphore ?? (_textSearchSemaphore = new SemaphoreSlim(1));

        private async Task CreateSearchListAsync()
        {
            // Check if searching same text to prevent searching again
            if (PrevSearchTextTerm != SearchTextTerm)
            {
                FullSearchItems.Clear();
                PrevSearchTextTerm = SearchTextTerm;
                FoundFullSearchItems = false;
            }
            else
            {
                return;
            }

            FoundFullSearchItems = false;
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            if (doc != null)
            {
                if (_currentTextSearchCanceller != null)
                {
                    _currentTextSearchCanceller.Cancel = true;
                    _currentTextSearchCanceller = null;
                }

                await WaitForTextSearchToCancel();

                FullTextSearchCanceller textSearchCanceller = new FullTextSearchCanceller();
                _currentTextSearchCanceller = textSearchCanceller;
                PageIterator pageIter = null;

                bool isDocLocked = false;
                try
                {
                    _pdfViewCtrl.DocLockRead();
                    isDocLocked = true;
                    pageIter = doc.GetPageIterator();
                }
                finally
                {
                    if (isDocLocked)
                        _pdfViewCtrl.DocUnlockRead();
                }

                _currentFullSearchPage = 1;
                IList<FullSearchItem> items = await GetNextPageWithSearchItemsAsync(pageIter, doc, textSearchCanceller);
                while (items != null)
                {
                    if (textSearchCanceller.Cancel)
                        break;

                    foreach (var item in items)
                    {
                        FullSearchItems.Add(item);
                    }

                    _currentFullSearchPage++;
                    RaisePropertyChanged(nameof(CurrentFullSearchStatusText));
                    items = await GetNextPageWithSearchItemsAsync(pageIter, doc, textSearchCanceller);
                }

                RaisePropertyChanged("FullSearchResultText");
                FoundFullSearchItems = true;
            }
        }

        private IAsyncOperation<IList<FullSearchItem>> GetNextPageWithSearchItemsAsync(PageIterator pageIter, PDFDoc doc, FullTextSearchCanceller canceller)
        {
            Task<IList<FullSearchItem>> t = new Task<IList<FullSearchItem>>(() => GetNextPageWithSearchItems(pageIter, doc, canceller));
            t.Start();

            return t.AsAsyncOperation();
        }

        private IList<FullSearchItem> GetNextPageWithSearchItems(PageIterator pageIter, PDFDoc doc, FullTextSearchCanceller canceller)
        {
            bool gotSema = false;
            bool gotDocLock = false;
            try
            {
                TextSearchSemaphore.Wait();
                gotSema = true;
                doc.LockRead();
                gotDocLock = true;

                IList<FullSearchItem> items = new List<FullSearchItem>();
                while (pageIter.HasNext() && !canceller.Cancel)
                {
                    TextSearch textSearch = new TextSearch();
                    textSearch.SetMode((int)(TextSearchSearchMode.e_page_stop | TextSearchSearchMode.e_highlight | TextSearchSearchMode.e_ambient_string));

                    Int32Ref pageNum = new Int32Ref(0);
                    StringRef resultStr = new StringRef();
                    StringRef ambientStr = new StringRef();
                    Highlights hLts = new Highlights();

                    textSearch.Begin(_pdfViewCtrl.GetDoc(), SearchTextTerm, textSearch.GetMode(), pageIter.GetPageNumber(), pageIter.GetPageNumber());
                    TextSearchResultCode result = textSearch.Run(pageNum, resultStr, ambientStr, hLts);

                    while (result == TextSearchResultCode.e_found)
                    {
                        hLts.Begin(_pdfViewCtrl.GetDoc());
                        var test = hLts.GetCurrentQuads();
                        FullSearchItem item = new FullSearchItem(pageIter.GetPageNumber(), resultStr.Value, ambientStr.Value, "", hLts, test);
                        items.Add(item);
                        hLts = new Highlights();
                        result = textSearch.Run(pageNum, resultStr, ambientStr, hLts);
                    }
                    pageIter.Next();
                    return items;
                }
            }
            finally
            {
                if (gotDocLock)
                {
                    doc.UnlockRead();
                }
                if (gotSema)
                {
                    TextSearchSemaphore.Release();
                }
            }

            return null;
        }

        private async Task BeginSearchAsync()
        {
            // Don't search again if text has not changed
            if (IsSearching && (_searchString == SearchTextTerm))
                return;

            _cancelSearch = false;
            _searchString = SearchTextTerm;
            if (_currentFullSearchItem != null)
            {
                await PrevNextAsync("next");
                return;
            }

            if (CurrentFullSearchItem != null)
                CurrentFullSearchItem = null;  // Not in full search mode, prev/next command needs to account for this

            if (_searchString.Length > 0)
            {
                _searchResult = StartSearchAsync(_searchString, false, false, false, false);

                if (_searchResult != null)
                    _searchResult.Completed = SearchDelegate;

                if (_textHighlighter == null)
                    _textHighlighter = new TextHighlighter(_pdfViewCtrl, _searchString);
            }
        }

        private void FindSameTerm(bool goBackwards, int pageNum = -1)
        {
            if (IsSearching)
                return;
            _searchResult = StartSearchAsync(_searchString, false, false, goBackwards, false, pageNum);
            if (_searchResult != null)
                _searchResult.Completed = SearchDelegate;
        }

        private IAsyncOperation<PDFViewCtrlSelection> StartSearchAsync(string searchString, bool matchCase, bool matchWholeWord, bool searchUp, bool regExp, int pageNum = -1)
        {
            IsSearching = true;
            _ticksBeforeProgress = 20; // will be 1 second
            _searchProgressTimer?.Stop();
            _searchProgressTimer = new DispatcherTimer();
            _searchProgressTimer.Tick += SearchProgressTimer_Tick;
            _searchProgressTimer.Interval = TimeSpan.FromMilliseconds(50);
            _searchProgressTimer.Start();
            IsTermNotFound = false;

            return _pdfViewCtrl.FindTextAsync(searchString, matchCase, matchWholeWord, searchUp, regExp, pageNum);
        }

        void SearchProgressTimer_Tick(object sender, object e)
        {
            if (_ticksBeforeProgress > 0)
            {
                _ticksBeforeProgress--;
            }
            else
            {
                IsSearchProgessVisible = true;
                SearchProgress = 100.0 * _pdfViewCtrl.GetFindTextProgressAsFactor();
            }
        }

        public async void SearchDelegate(IAsyncOperation<PDFViewCtrlSelection> asyncAction, AsyncStatus asyncStatus)
        {
            double searchProgress = SearchProgress;
            SearchProgress = 0;
            await Window.Current.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, async delegate
            {
                IsSearching = false;
                if (_cancelSearch)
                {
                    _cancelSearch = false;
                    return;
                }

                ResetSearchProgress();
                switch (asyncStatus)
                {
                    case AsyncStatus.Error:
                        //SearchResultOutput.Text = "Error";
                        break;
                    case AsyncStatus.Completed:
                        if (_cancelSearch)
                            return;

                        _currentSearchSelection = asyncAction.GetResults();
                        if (HighlightSelection(_currentSearchSelection))
                        {
                            IsPrevNextAvailable = true;
                            if (_findTextResultFound != null)
                            {
                                var tcs = new TaskCompletionSource<bool>();
                                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                                {
                                    _findTextResultFound.Invoke(deferral);
                                    await tcs.Task;
                                }
                            }
                        }
                        else
                        {
                            IsPrevNextAvailable = false;
                            IsTermNotFound = true;
                            if (_pdfViewCtrl.GetFindTextProgressAsFactor() == 1)
                                ShowTextNotFoundMessage();
                        }
                        break;
                }
            });
        }

        private async Task HighlightFullSearchItemAsync()
        {
            _searchString = SearchTextTerm;

            if (PageChanged != null)
            {
                var tcs = new TaskCompletionSource<bool>();
                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                {
                    PageChanged.Invoke(_pdfViewCtrl.GetCurrentPage(), CurrentFullSearchItem.PageNumber, deferral);
                    await tcs.Task;
                }
            }

            _pdfViewCtrl.SetCurrentPage(CurrentFullSearchItem.PageNumber);
            _pdfViewCtrl.Select(CurrentFullSearchItem.Highlights);

            if (_textHighlighter == null)
                _textHighlighter = new TextHighlighter(_pdfViewCtrl, _searchString);

            if (HighlightSelection(CurrentFullSearchItem.Quads, CurrentFullSearchItem.Quads.Length / 8, CurrentFullSearchItem.PageNumber))
            {
                _pdfViewCtrl.ClearSelection();
                IsPrevNextAvailable = true;
            }
        }

        private void ClearSelection()
        {
            if (_onScreenSelection == null)
                return;

            foreach (Rectangle rect in _onScreenSelection)
            {
                if (rect.Tag is Canvas parent)
                    parent.Children.Remove(rect);
            }

            _onScreenSelection.Clear();
        }

        private bool HighlightSelection(double[] quads, int numQuads, int page)
        {
            int quadNumber = 0;
            List<UIRect> rects = new List<UIRect>();

            PDFViewCtrlPagePresentationMode mode = _pdfViewCtrl.GetPagePresentationMode();
            int currentPage = _pdfViewCtrl.GetCurrentPage();
            if ((mode == PDFViewCtrlPagePresentationMode.e_single_page && page != currentPage) ||
                (mode == PDFViewCtrlPagePresentationMode.e_facing && (((page + 1) / 2) != ((currentPage + 1) / 2)))  ||
                (mode == PDFViewCtrlPagePresentationMode.e_facing_cover && ((page / 2) != (currentPage / 2))) )
            {
                return false;
            }

            // get highlights in control (screen) space
            for (int i = 0; i < numQuads; i++)
            {
                quadNumber = i * 8;

                DoubleRef x1 = new DoubleRef(quads[quadNumber + 0]);
                DoubleRef y1 = new DoubleRef(quads[quadNumber + 1]);

                DoubleRef x2 = new DoubleRef(quads[quadNumber + 2]);
                DoubleRef y2 = new DoubleRef(quads[quadNumber + 3]);

                DoubleRef x3 = new DoubleRef(quads[quadNumber + 4]);
                DoubleRef y3 = new DoubleRef(quads[quadNumber + 5]);

                DoubleRef x4 = new DoubleRef(quads[quadNumber + 6]);
                DoubleRef y4 = new DoubleRef(quads[quadNumber + 7]);

                _pdfViewCtrl.ConvPagePtToScreenPt(x1, y1, page);
                _pdfViewCtrl.ConvPagePtToScreenPt(x2, y2, page);
                _pdfViewCtrl.ConvPagePtToScreenPt(x3, y3, page);
                _pdfViewCtrl.ConvPagePtToScreenPt(x4, y4, page);

                double left, right, top, bottom;

                left = Math.Min(x1.Value, Math.Min(x2.Value, Math.Min(x3.Value, x4.Value)));
                right = Math.Max(x1.Value, Math.Max(x2.Value, Math.Max(x3.Value, x4.Value)));
                top = Math.Min(y1.Value, Math.Min(y2.Value, Math.Min(y3.Value, y4.Value)));
                bottom = Math.Max(y1.Value, Math.Max(y2.Value, Math.Max(y3.Value, y4.Value)));

                rects.Add(new UIRect(left, top, right - left, bottom - top));
            }

            Canvas annotCanvas = _pdfViewCtrl.GetAnnotationCanvas();

            ClearSelection();

            // add highlight(s) to annotation canvas
            foreach (UIRect rect in rects)
            {
                Rectangle highlight = new Rectangle();
                highlight.Fill = _currentColor;
                highlight.Width = rect.Width;
                highlight.Height = rect.Height;

                var rLeft = new DoubleRef(rect.Left);
                var rTop = new DoubleRef(rect.Top);
                _pdfViewCtrl.ConvScreenPtToAnnotationCanvasPt(rLeft, rTop);
                Canvas.SetLeft(highlight, rLeft.Value);
                Canvas.SetTop(highlight, rTop.Value);
                annotCanvas.Children.Add(highlight);
                _onScreenSelection.Add(highlight);
                highlight.Tag = annotCanvas;
            }

            return numQuads > 0;
        }

        /// <summary>
        /// Highlights the search result if any
        /// </summary>
        /// <param name="result">A text selection acquired by mPDFView.FindText</param>
        /// <returns>true if and only if the selections contains at least one highlight</returns>
        private bool HighlightSelection(PDFViewCtrlSelection result)
        {
            if (result == null)
                return false;

            double[] quads = result.GetQuads();
            int numQuads = result.GetQuadArrayLength() / 8;

            return HighlightSelection(quads, numQuads, result.GetPageNum());
        }

        private void ResetSearchProgress()
        {
            IsSearchProgessVisible = false;
            SearchProgress = 0;
            _searchProgressTimer?.Stop();
        }

        private async Task CloseDialogAsync()
        {
            if (_findTextClosed != null)
            {
                var tcs = new TaskCompletionSource<bool>();
                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                {
                    _findTextClosed?.Invoke(deferral);
                    await tcs.Task;
                }
            }

            CancelCurrentSearch();
        }

        private void CancelCurrentSearch()
        {
            _pdfViewCtrl?.CancelFindText();
            IsPrevNextAvailable = false;
            ClearSelection();
            ResetSearchProgress();
            if (_textHighlighter != null)
            {
                _textHighlighter.Detach();
                _textHighlighter = null;
            }

            if (_currentTextSearchCanceller != null)
                _currentTextSearchCanceller.Cancel = true;
        }

        private async void ShowTextNotFoundMessage()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            ContentDialog dialog = new ContentDialog
            {
                Content = loader.GetString("FindTextDialog_NotFoundText"),
                CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
            };

            await DialogHelper.ShowContentDialogAsync(dialog);
        }
    }
}
