﻿using pdftron.Common;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.UI.ViewManagement;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Settings
{
    /// <summary>
    /// All the settings regarding the current ViewerPage are centralized here.
    /// </summary>
    public class ViewerPageSettingsViewModel : ViewModelBase
    {
        public static Windows.UI.Color SEPIA_BLACK = Windows.UI.Colors.Black;
        public static Windows.UI.Color SEPIA_WHITE = Windows.UI.Color.FromArgb(255, 255, 232, 206);

        private ViewRequestedDelegate _viewRequestedHandler;
        public delegate void ViewRequestedDelegate(RequestableViewType view);
        public event ViewRequestedDelegate ViewRequestedHandler
        {
            add
            {
                if (_viewRequestedHandler?.GetInvocationList().Contains(value) != true)
                    _viewRequestedHandler += value;
            }
            remove => _viewRequestedHandler -= value;
        }

        private TypedEventHandler<ApplicationView, object> _visibleBoundsChanged;
        public event TypedEventHandler<ApplicationView, object> VisibleBoundsChanged
        {
            add
            {
                if (_visibleBoundsChanged?.GetInvocationList().Contains(value) != true)
                    _visibleBoundsChanged += value;
            }
            remove => _visibleBoundsChanged -= value;
        }

        private readonly PdfViewCtrlTabInfo _tab;
        private PDFViewCtrl _pdfViewCtrl;
        private bool _isContinuousOption;
        private bool _isRightToLeftVisible;

        public RelayCommand CustomColorCommand { get; private set; }

        public RelayCommand PagePresentationModeCommand { get; private set; }

        public RelayCommand ReflowCommand { get; private set; }

        public RelayCommand RotateCommand { get; private set; }

        public RelayCommand ThumbnailViewCommand { get; private set; }

        public RelayCommand CropViewCommand { get; private set; }

        public RelayCommand FullScreenCommand { get; private set; }

        public RelayCommand ExitFullScreenCommand { get; private set; }

        public ViewerPageSettingsViewModel()
        {
            Init();
        }

        public ViewerPageSettingsViewModel(PdfViewCtrlTabInfo tab)
        {
            _tab = tab;
            _pdfViewCtrl = tab.PDFViewCtrl;
            Init();
        }

        private void Init()
        {
            CustomColorCommand = new RelayCommand(CustomColorCommandImpl);
            PagePresentationModeCommand = new RelayCommand(PagePresentationModeCommandImpl);
            ReflowCommand = new RelayCommand(ReflowCommandImpl);
            RotateCommand = new RelayCommand(RotateCommandImpl);
            ThumbnailViewCommand = new RelayCommand(ThumbnailViewCommandImpl);
            CropViewCommand = new RelayCommand(CropViewCommandImpl);
            FullScreenCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(ToggleFullScreenAsync);
                command.Execute(parameter);
            });
            ExitFullScreenCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(ToggleFullScreenAsync);
                command.Execute(parameter);
            });

            VisibleBoundsChanged += ViewerPageSettingsViewModel_VisibleBoundsChanged;
            ApplicationView.GetForCurrentView().VisibleBoundsChanged += ViewerPageSettingsViewModel_VisibleBoundsChanged;
        }

        /// <summary>
        /// Important to call this when done, as the subscription to
        /// ApplicationView.GetForCurrentView().VisibleBoundsChanged prevents it from being GCd
        /// </summary>
        public void CleanUp()
        {
            if (_visibleBoundsChanged != null)
            {
                ApplicationView.GetForCurrentView().VisibleBoundsChanged -= ViewerPageSettingsViewModel_VisibleBoundsChanged;
                VisibleBoundsChanged -= ViewerPageSettingsViewModel_VisibleBoundsChanged;
            }
        }

        private void CustomColorCommandImpl(object modeName) => ResolveCustomColorMode(modeName);

        private void ResolveCustomColorMode(object modeName)
        {
            if (modeName.ToString().Equals("D", StringComparison.OrdinalIgnoreCase))
            {
                ColorMode = CustomColorModeType.none;
            }
            else if (modeName.ToString().Equals("N", StringComparison.OrdinalIgnoreCase))
            {
                ColorMode = CustomColorModeType.Night;
            }
            else if (modeName.ToString().Equals("S", StringComparison.OrdinalIgnoreCase))
            {
                ColorMode = CustomColorModeType.Sepia;
            }
            else if (modeName.ToString().Equals("C", StringComparison.OrdinalIgnoreCase))
            {
                RaisePropertyChanged("CustomColor");
            }
        }

        private void PagePresentationModeCommandImpl(object modeName) => ResolvePagePresentationMode(modeName);

        private void ReflowCommandImpl(object param)
        {
            _tab.IsReflow = !_tab.IsReflow;
            RaiseCurrentPageModePropertyChanged();
        }

        private void ResolvePagePresentationMode(object modeName)
        {
            if (_pdfViewCtrl == null)
                return;

            string selectedMode = modeName as string;
            PDFViewCtrlPagePresentationMode oldMode = _pdfViewCtrl.GetPagePresentationMode();
            PDFViewCtrlPagePresentationMode newMode = oldMode;
            if (selectedMode == "con")
            {
                if (newMode == PDFViewCtrlPagePresentationMode.e_single_page || newMode == PDFViewCtrlPagePresentationMode.e_single_continuous)
                    selectedMode = "s";
                else if (newMode == PDFViewCtrlPagePresentationMode.e_facing || newMode == PDFViewCtrlPagePresentationMode.e_facing_continuous)
                    selectedMode = "f";
                else
                    selectedMode = "fcov";
            }

            if (!string.IsNullOrWhiteSpace(selectedMode))
            {
                if (selectedMode.Equals("s", StringComparison.OrdinalIgnoreCase))
                {
                    newMode = IsContinuousOption ? PDFViewCtrlPagePresentationMode.e_single_continuous : PDFViewCtrlPagePresentationMode.e_single_page;
                }
                else if (selectedMode.Equals("f", StringComparison.OrdinalIgnoreCase))
                {
                    newMode = IsContinuousOption ? PDFViewCtrlPagePresentationMode.e_facing_continuous : PDFViewCtrlPagePresentationMode.e_facing;
                }
                else if (selectedMode.Equals("fcov", StringComparison.OrdinalIgnoreCase))
                {
                    newMode = IsContinuousOption ? PDFViewCtrlPagePresentationMode.e_facing_continuous_cover : PDFViewCtrlPagePresentationMode.e_facing_cover;
                }
            }
            if (oldMode != newMode || IsReflow)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, UtilityFunctions.GetPagePresentationModeName(newMode) + " selected");
                IsReflow = false;
                RaiseCurrentPageModePropertyChanged();
                PagePresentationMode = newMode;
                ViewerSettings.PagePresentationMode = newMode;

                if (!IsContinuous(oldMode) && IsContinuous(newMode))
                {
                    _pdfViewCtrl.SetPageViewMode(PDFViewCtrlPageViewMode.e_fit_width);
                }
                else if (!IsContinuous(newMode))
                {
                    _pdfViewCtrl.SetPageViewMode(_pdfViewCtrl.GetPageRefViewMode());
                }
            }
        }

        private bool IsContinuous(PDFViewCtrlPagePresentationMode mode)
        {
            return mode == PDFViewCtrlPagePresentationMode.e_single_continuous || mode == PDFViewCtrlPagePresentationMode.e_facing_continuous || mode == PDFViewCtrlPagePresentationMode.e_facing_continuous_cover;
        }

        private void RotateCommandImpl(object button)
        {
            if (_pdfViewCtrl != null)
            {
                try
                {
                    _pdfViewCtrl.RotateClockwise();
                }
                catch (Exception e)
                {
                    string message = string.Empty;
                    PDFNetException pdfEx = new PDFNetException(e.HResult);
                    if (pdfEx.IsPDFNetException)
                    {
                        message = pdfEx.ToString();
                    }
                    else
                    {
                        message = e.ToString();
                    }

                    Debug.WriteLine("Error rotating: " + message);
                }

                RaisePropertyChanged("Rotation");
            }
        }

        private void ThumbnailViewCommandImpl(object button) => _viewRequestedHandler?.Invoke(RequestableViewType.Thumbnails);

        private void CropViewCommandImpl(object button) => _viewRequestedHandler?.Invoke(RequestableViewType.Crop);

        private async Task ToggleFullScreenAsync()
        {
            bool fullScreen = !IsFullScreen;
            await UtilityFunctions.SetFullScreenModeAsync(fullScreen);
            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                int fullScreenInt = 0;
                if (fullScreen)
                    fullScreenInt = 1;

                ViewerSettings.PhoneFullScreen = fullScreenInt;
            }

            RaisePropertyChanged("FullScreen");
            RaisePropertyChanged(nameof(IsFullScreen));
        }

        public PDFViewCtrlPagePresentationMode PagePresentationMode
        {
            get
            {
                if (_pdfViewCtrl != null)
                    return _pdfViewCtrl.GetPagePresentationMode();

                return PDFViewCtrlPagePresentationMode.e_single_continuous;
            }
            set
            {
                if (_pdfViewCtrl != null && value != _pdfViewCtrl.GetPagePresentationMode())
                {
                    IsReflow = false;
                    _pdfViewCtrl.SetPagePresentationMode(value);
                    RaisePropertyChanged();
                    RaiseCurrentPageModePropertyChanged();
                }
            }
        }

        public void UpdateSettings()
        {
            _isContinuousOption = IsContinuous(_pdfViewCtrl.GetPagePresentationMode());
            RaisePropertyChanged(nameof(IsContinuousOption));
            PagePresentationMode = _pdfViewCtrl.GetPagePresentationMode();
            RaisePropertyChanged(nameof(IsFullScreen));
            RaisePropertyChanged(nameof(IsConverting));
            RaisePropertyChanged(nameof(CanCrop));
        }

        public bool IsCurrentPagePresentationModeSinglePage
        {
            get
            {
                if (_pdfViewCtrl == null)
                    return true;

                if (IsReflow)
                    return false;

                return _pdfViewCtrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_single_page
                    || _pdfViewCtrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_single_continuous;
            }
        }

        public bool IsCurrentPagePresentationModeFacing
        {
            get
            {
                if (_pdfViewCtrl == null)
                    return false;

                if (IsReflow)
                    return false;

                return _pdfViewCtrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_facing
                    || _pdfViewCtrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_facing_continuous;
            }
        }
        public bool IsCurrentPagePresentationModeFacingCover
        {
            get
            {
                if (_pdfViewCtrl == null)
                    return false;

                if (IsReflow)
                    return false;

                return _pdfViewCtrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_facing_cover
                    || _pdfViewCtrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_facing_continuous_cover;
            }
        }

        public bool IsReflow
        {
            get
            {
                if (_tab != null)
                    return _tab.IsReflow;
                else
                    return false;
            }
            set
            {
                if (_tab != null && _tab.IsReflow != value)
                {
                    _tab.IsReflow = value;
                    RaisePropertyChanged();
                }
            }
        }

        public bool IsConverting => _tab?.IsConverting != false;

        public bool CanCrop => !IsConverting && !IsReflow;

        public CustomColorModeType ColorMode
        {
            get => ViewerSettings.ColorMode;
            set
            {
                if (value != ViewerSettings.ColorMode)
                {
                    ViewerSettings.ColorMode = value;
                    RaisePropertyChanged();
                }
            }
        }

        public PDFViewCtrl PDFViewCtrl
        {
            set => _pdfViewCtrl = value;
        }

        public bool IsContinuousOption
        {
            get => _isContinuousOption;
            set
            {
                if (_isContinuousOption != value)
                {
                    _isContinuousOption = value;
                    RaisePropertyChanged(nameof(IsContinuousOption));
                    ResolvePagePresentationMode("con");
                }
            }
        }

        /// <summary>
        /// Change the visibility of the Right to Left reading direction option
        /// </summary>
        public bool IsRightToLeftVisible
        {
            get
            {
                return _isRightToLeftVisible;
            }

            set
            {
                if (_isRightToLeftVisible != value)
                {
                    _isRightToLeftVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Enable Right To Left reading direction for the current selected tab
        /// </summary>
        public bool IsRightToLeftEnabled
        {
            get
            {
                if (_tab != null)
                    return _tab.IsRightToLeftEnabled;
                else
                    return false;
            }
            set
            {
                if (_tab != null && _tab.IsRightToLeftEnabled != value)
                {
                    _tab.IsRightToLeftEnabled = value;
                    RaisePropertyChanged(nameof(IsRightToLeftEnabled));
                }
            }
        }

        private bool _isThumbnailViewVisible = true;
        /// <summary>
        /// Change the visbility of the ThumbnailView button
        /// </summary>
        public bool IsThumbnailViewVisible
        {
            get { return _isThumbnailViewVisible; }
            set
            {
                _isThumbnailViewVisible = value;
                RaisePropertyChanged(); 
            }
        }

        private bool _isCropViewVisible = true;
        /// <summary>
        /// Change the visbility of the Crop View button
        /// </summary>
        public bool IsCropViewVisible
        {
            get { return _isCropViewVisible; }
            set
            {
                _isCropViewVisible = value;
                RaisePropertyChanged();
            }
        }

        public bool IsFullScreen => UtilityFunctions.IsFullScreen();

        private void ViewerPageSettingsViewModel_VisibleBoundsChanged(ApplicationView sender, object args) => RaisePropertyChanged(nameof(IsFullScreen));

        private void RaiseCurrentPageModePropertyChanged()
        {
            RaisePropertyChanged(nameof(IsReflow));
            RaisePropertyChanged(nameof(CanCrop));
            RaisePropertyChanged("IsCurrentPagePresentationModeContinuous");
            RaisePropertyChanged(nameof(IsCurrentPagePresentationModeSinglePage));
            RaisePropertyChanged(nameof(IsCurrentPagePresentationModeFacing));
            RaisePropertyChanged(nameof(IsCurrentPagePresentationModeFacingCover));
        }
    }
}
