﻿using Microsoft.UI.Xaml.Controls;
using pdftron.PDF.Tools.Controls.ViewModels;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Menu;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Search;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.ApplicationModel;
using Windows.ApplicationModel.Resources;
using Windows.Devices.Input;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Storage;
using Windows.System.Profile;
using Windows.UI.Core;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;

// The User Control item template is documented at https://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls.Viewer
{
    public sealed partial class ViewerControl : UserControl, IDisposable
    {
        private ViewerViewModel _viewModel;
        private bool _smallScreen;
        private double _commandBarCompactHeight = 40;
        private double _commandbarHeight = 60;
        private AnchorSides _oldOutlineAnchorSide;
        private double _oldWidth;
        private static bool mHasResourcesInitialized = false; // Only initialize resources once per application instance

        private MenuFlyout _ActionsMenuFlyout = new MenuFlyout
        {
            Placement = FlyoutPlacementMode.BottomEdgeAlignedLeft
        };

        private RoutedEventHandler _outlineResizer_PositionUpdated;
        private ViewerSettings.SettingUpdatedDelegate _settingsUpdatedHandler;

        private ToolManager.ShowingPopupMenuHandler _showToolPopupMenu;
        public event ToolManager.ShowingPopupMenuHandler ShowToolPopupMenu
        {
            add
            {
                if (_showToolPopupMenu?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _showToolPopupMenu += value;
            }
            remove
            {
                if (_showToolPopupMenu != null)
                {
                    ToolManager.ShowingPopupMenuHandler eventMethod = (ToolManager.ShowingPopupMenuHandler)Array.Find(
                        _showToolPopupMenu.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _showToolPopupMenu -= eventMethod;
                }
            }
        }

        private ToolManager.ShowingContextMenuHandler _showToolContextMenu;
        public event ToolManager.ShowingContextMenuHandler ShowToolContextMenu
        {
            add
            {
                if (_showToolContextMenu?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _showToolContextMenu += value;
            }
            remove
            {
                if (_showToolPopupMenu != null)
                {
                    ToolManager.ShowingContextMenuHandler eventMethod = (ToolManager.ShowingContextMenuHandler)Array.Find(
                        _showToolContextMenu.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _showToolContextMenu -= eventMethod;
                }
            }
        }

        private DocumentModifiedHandler _documentModified;
        public delegate void DocumentModifiedHandler();

        /// <summary>
        /// Raised when a document has been modified
        /// </summary>
        /// <remarks>
        /// Types of modification: Annotation added, removed, or edited 
        /// </remarks>
        public event DocumentModifiedHandler DocumentModified
        {
            add
            {
                if (_documentModified?.GetInvocationList().Contains(value) != true)
                    _documentModified += value;
            }
            remove => _documentModified -= value;
        }

        /// <summary>
        /// Get the Annotation Toolbar used by all opened documents
        /// </summary>
        public AnnotationCommandBar AnnotationToolbar
        {
            get { return ViewModel.AnnotationToolbar; }
        }

        /// <summary>
        /// Allows user to externally control whether the Thumbnail button is visible or collapsed
        /// </summary>
        public bool ShowThumbnailView
        {
            get { return ViewModel.IsThumbnailViewVisible; }
            set { ViewModel.IsThumbnailViewVisible = value; }
        }

        /// <summary>
        /// Get the Thumbnail Viewer used to display all document pages and perform insert, merging, delete, and duplicate of PDF pages.
        /// </summary>
        public ThumbnailViewer ThumbnailViewer
        {
            get { return ViewModel.ThumbnailViewer; }
        }

        /// <summary>
        /// Allows user to externally control whether the Crop View button is visible or collapsed
        /// </summary>
        public bool ShowCropView
        {
            get { return ViewModel.IsCropViewVisible; }
            set { ViewModel.IsCropViewVisible = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Actions button is visible or collapsed
        /// </summary>
        public bool ShowActionsOption
        {
            get { return ViewModel.IsActionsEnabled; }
            set { ViewModel.IsActionsEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Share document is visible or collapsed
        /// </summary>
        public bool ShowShareOption
        {
            get { return ViewModel.IsShareVisible; }
            set { ViewModel.IsShareVisible = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Print document is visible or collapsed
        /// </summary>
        public bool ShowPrintOption
        {
            get { return ViewModel.IsPrintVisible; }
            set { ViewModel.IsPrintVisible = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Export XFDF button is visible or collapsed
        /// </summary>
        public bool ShowExportXFDF
        {
            get { return ViewModel.IsExportXFDFVisible; }
            set { ViewModel.IsExportXFDFVisible = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Text Search button is visible or collapsed
        /// </summary>
        public bool ShowTextSearchOption
        {
            get { return ViewModel.IsTextSearchButtonEnabled; }
            set { ViewModel.IsTextSearchButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Save document button is visible or collapsed
        /// </summary>
        public bool ShowSaveOption
        {
            get { return ViewModel.IsSaveButtonEnabled; }
            set { ViewModel.IsSaveButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Save As document button is visible or collapsed
        /// </summary>
        public bool ShowSaveAsOption
        {
            get { return ViewModel.IsSaveAsButtonEnabled; }
            set { ViewModel.IsSaveAsButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the PDFCompare button is visible or collapsed
        /// </summary>
        public bool ShowPDFCompare
        {
            get { return ViewModel.IsPdfCompareEnabled; }
            set { ViewModel.IsPdfCompareEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Multi Window button is visible or collapsed
        /// </summary>
        public bool ShowMultiWindow
        {
            get { return ViewModel.IsMultiWindowEnabled; }
            set { ViewModel.IsMultiWindowEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Side Panel Thumbnails is visible or collapsed
        /// </summary>
        public bool ShowSidePanelThumbnails
        {
            get { return ViewModel.IsPivotItemThumbnailsEnabled; }
            set { ViewModel.IsPivotItemThumbnailsEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Side Panel Bookmarks is visible or collapsed
        /// </summary>
        public bool ShowSidePanelBookmarks
        {
            get { return ViewModel.IsPivotItemBookmarksEnabled; }
            set { ViewModel.IsPivotItemBookmarksEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Side Panel Outline is visible or collapsed
        /// </summary>
        public bool ShowSidePanelOutline
        {
            get { return ViewModel.IsPivotItemOutlineEnabled; }
            set { ViewModel.IsPivotItemOutlineEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Side Panel Annotations is visible or collapsed
        /// </summary>
        public bool ShowSidePanelAnnotations
        {
            get { return ViewModel.IsPivotItemAnnotationsEnabled; }
            set { ViewModel.IsPivotItemAnnotationsEnabled = value; }
        }

        /// Allows user to externally control whether the Close tab button is visible or collapsed
        /// NOTE: when close tab is disable the tab won't be able to get closed even on the wheel-clock of the mouse
        /// </summary>
        public bool ShowCloseTabButton
        {
            get { return ViewModel.IsCloseTabButtonEnabled; }
            set { ViewModel.IsCloseTabButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Add tab button is visible or collapsed
        /// </summary>
        public bool ShowAddTabButton
        {
            get { return ViewModel.IsAddTabButtonEnabled; }
            set { ViewModel.IsAddTabButtonEnabled = value; }
        }
        
        /// Allows user to externally control whether the Edit button (open annotation toolbar) is visible or collapsed
        /// </summary>
        public bool ShowEditOption
        {
            get { return ViewModel.IsEditButtonEnabled; }
            set { ViewModel.IsEditButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the Outline button is visible or collapsed
        /// </summary>
		/// <para> Note: If all SidePanel options are hidden, then the Outline button is collapsed</para>
        public bool ShowOutlineOption
        {
            get { return ViewModel.IsOutlineButtonEnabled; }
            set { ViewModel.IsOutlineButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the ViewMode button is visible or collapsed
        /// </summary>
        public bool ShowViewModeOption
        {
            get { return ViewModel.IsViewModeButtonEnabled; }
            set { ViewModel.IsViewModeButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the SeeMore button is visible or collapsed
        /// </summary>
        public bool ShowSeeMoreOption
        {
            get { return ViewModel.IsSeeMoreButtonEnabled; }
            set { ViewModel.IsSeeMoreButtonEnabled = value; }
        }

        /// <summary>
        /// Allows user to externally control whether the ViewerControl operates in View/Read only mode
        /// which won't allow the PDF documents to be edited (annotated).
        /// </summary>
        /// <remarks>Note1: It is still possible to select and copy text to the clipboard.
        /// <para /> Note2: There are still some compoenents that allow editing which need to be manually disabled.</remarks>
        public bool ReadOnlyMode
        {
            get { return ViewModel.IsReadOnlyModeEnabled; }
            set { ViewModel.IsReadOnlyModeEnabled = value; }
        }

        /// <summary>
        /// Allows user to navigate back from the ViewerControl
        /// </summary>
        public bool ShowBackButton
        {
            get { return ViewModel.ShowBackButton; }
            set { ViewModel.ShowBackButton = value; }
        }

        /// <summary>
        /// Ensure to deactivate all opened tabs when navigating away from the page that hosts
        /// the ViewerControl
        /// </summary>
        public bool IsDeactivationOnNavigationEnabled
        {
            get { return ViewModel.IsDeactivationOnBackNavigationEnabled; }
            set { ViewModel.IsDeactivationOnBackNavigationEnabled = value; }
        }

        /// <summary>
        /// When enabled it will force the ViewerControl to navigate back to the previous Frame page
        /// listening to the SystemNavigationManager.GetForCurrentView().BackRequested event
        /// </summary>
        public bool IsForceGoBackAllowed
        {
            get { return ViewModel.IsForceGoBackAllowed; }
            set { ViewModel.IsForceGoBackAllowed = value; }
        }

        /// <summary>
        /// Choose whether or not editing outlines is enabled
        /// </summary>
        public bool IsOutlineEditingEnabled
        {
            get { return ViewModel.IsOutlineEditingEnabled; }
            set { ViewModel.IsOutlineEditingEnabled = value; }
        }

        /// <summary>
        /// Choose whether or not Tab Reordering is enabled
        /// </summary>
        public bool IsTabReorderEnabled
        {
            get { return ViewModel.IsTabReorderEnabled; }
            set { ViewModel.IsTabReorderEnabled = value; }
        }

        private static readonly DependencyProperty MaximumTabsProperty = DependencyProperty.Register(
            "MaximumTabs",
            typeof(int),
            typeof(ViewerControl),
            new PropertyMetadata(20));

        public int MaximumTabs
        {
            get => (int)GetValue(MaximumTabsProperty);
            set => SetValue(MaximumTabsProperty, value);
        }

        private static readonly DependencyProperty InitializeOnCreationProperty = DependencyProperty.Register(
            "InitializeOnCreation",
            typeof(bool),
            typeof(ViewerControl),
            new PropertyMetadata(true));

        /// <summary>
        /// Perform default initialization on ViewerControl creation
        /// </summary>
        public bool InitializeOnCreation
        {
            get => (bool)GetValue(InitializeOnCreationProperty);
            set => SetValue(InitializeOnCreationProperty, value);
        }

        /// <summary>
        /// Whether or not the current document has been modified since the last time it got saved
        /// </summary>
        /// <remarks>
        /// Return false if no tabs are opened or an internal error ocurred
        /// </remarks>
        public bool IscurrentDocumentModified
        {
            get 
            {
                if (ViewModel != null)
                    return ViewModel.IsDocumentModifiedSinceLastSave;

                return false;
            }
        }

        public bool EnableTracking { get; set; }

        public static bool IsCustomViewModel { get; set; }

        public ViewerControl()
        {
            InitializeComponent();

            // NOTE: set DataContext here to avoid binding error when a Page is hosting ViewerControl and set its own DataContext before it
            if (!IsCustomViewModel)
                DataContext = new ViewerViewModel();

            ViewerTopCommandBar.Opening += ViewerTopCommandBar_Opening;
            ViewerTopCommandBar.Closing += ViewerTopCommandBar_Closing;

            // Initialize utility class to get current OS information to be used in Version Adaptive Code
            WindowsInfo.RetrieveApiInfo();

            // Version Adaptive Code to enforce closing the MenuFlyout on Windows 10 v1809
            if (WindowsInfo.Build <= Build.V1809)
            {
                this.PointerPressed += CloseFlyout_PointerPressed;
                PDFViewCtrlGrid.PointerPressed += CloseFlyout_PointerPressed;
            }

            if (Resources.ContainsKey("AppBarThemeCompactHeight"))
            {
                _commandbarHeight = (double)Resources["AppBarThemeCompactHeight"];
                _commandBarCompactHeight = _commandbarHeight;
            }

            PlaceOutlineDialog();

            string deviceFamily = AnalyticsInfo.VersionInfo.DeviceFamily;
            if (!string.IsNullOrWhiteSpace(deviceFamily) && deviceFamily.Equals("Windows.Mobile", StringComparison.OrdinalIgnoreCase))
                AppBarButton_Files.Visibility = Visibility.Collapsed;

            double tabButtonHeight = (double)this.Resources["OutlineTabButtonSize"];

            _ActionsMenuFlyout.MenuFlyoutPresenterStyle = (Style)this.Resources["MenuFlyoutPresenterDefaultStyle"];
            // NOTE: DO NOT set it to Auto. It might cause exception on earlier version of Windows (Like 10.0.1803)
            _ActionsMenuFlyout.Placement = FlyoutPlacementMode.Bottom;

            // Load Viewer resources so we can access it from View Model
            ResourceHandler.ViewerResources = this.Resources;


        }

        public void Dispose()
        {
            if (WindowsInfo.Build <= Build.V1809)
            {
                this.PointerPressed -= CloseFlyout_PointerPressed;

                if (PDFViewCtrlGrid != null)
                    PDFViewCtrlGrid.PointerPressed -= CloseFlyout_PointerPressed;
            }

            if (ViewerTopCommandBar != null)
            {
                ViewerTopCommandBar.Opening -= ViewerTopCommandBar_Opening;
                ViewerTopCommandBar.Closing -= ViewerTopCommandBar_Closing;
                // TODO: ViewerTopCommandBar.Dispose();
                ViewerTopCommandBar = null;
            }

            if (_settingsUpdatedHandler != null)
            {
                ViewerSettings.SettingUpdated -= _settingsUpdatedHandler;
                _settingsUpdatedHandler = null;
            }

            if (_outlineResizer_PositionUpdated != null)
            {
                OutlineResizer.PositionUpdated -= _outlineResizer_PositionUpdated;
                _outlineResizer_PositionUpdated = null;
            }

            if (ViewModel != null)
            {
                ViewModel.FindTextResultFound -= ViewModel_FindTextResultFound;
                ViewModel.DocumentModified -= ViewModel_DocumentModified;
                ViewModel.OpenSearch -= ViewModel_OpenSearch;
                ViewModel.PropertyChanged -= ViewModel_PropertyChanged;

                if (ViewModel.MediaElement != null)
                    ViewModel.MediaElement.MediaEnded -= Media_MediaEnded;

                ViewModel.AnnotationToolbar.AppBarChanged -= AnnotationToolbar_AppBarChanged;
                ViewModel.TabControlViewModel.ShowToolContextMenu -= TabControlViewModel_ShowToolContextMenu;
                ViewModel.Dispose();
                ViewModel = null;
            }
        }

        private void CloseFlyout_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            if (_ActionsMenuFlyout.IsOpen)
                _ActionsMenuFlyout.Hide();

            e.Handled = true;
        }

        public bool IsSaveCopyEnabled { get; set; } = true;

        /// <summary>
        /// Change the visibility of the Right To Left reading direction option
        /// </summary>
        public bool IsRightToLeftVisible
        { 
            get 
            {
                return ViewModel.IsRightToLeftVisibile;
            }
            set 
            {
                ViewModel.IsRightToLeftVisibile = value;
            }
        }

        /// <summary>
        /// Initialize the default menu flyouts on code behind for the Actions
        /// </summary>
        private void InitializeDefaultSaveCopyMenuFlyout()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            MenuFlyoutItem optimize = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)this.Resources["CommandBarDefaultForeground"],
                Text = loader.GetString("ViewerPage_SecondaryOptions_Optimize"),
                Command = ViewModel.SaveReducedFileSizeCopyCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-compress.png"),
                    ShowAsMonochrome = false
                }
            };

            MenuFlyoutItem encrypt = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)this.Resources["CommandBarDefaultForeground"],
                Command = ViewModel.SavePasswordCopyCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-encrypt.png"),
                    ShowAsMonochrome = false
                }
            };

            Binding encryptBinding = new Binding
            {
                Source = ViewModel,
                Path = new PropertyPath("SaveAPasswordCopyText"),
                Mode = BindingMode.TwoWay,
                UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
            };

            BindingOperations.SetBinding(encrypt, MenuFlyoutItem.TextProperty, encryptBinding);

            MenuFlyoutItem crop = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)this.Resources["CommandBarDefaultForeground"],
                Text = loader.GetString("ViewerPage_SecondaryOptions_PermCroppedCopy"),
                Command = ViewModel.SavePermCroppedCopyCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-crop-pdf.png"),
                    ShowAsMonochrome = false
                }
            };

            Binding cropVisBinding = new Binding
            {
                Source = ViewModel,                
                Path = new PropertyPath("IsCropViewVisible"),
                Mode = BindingMode.TwoWay,
                UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
            };

            BindingOperations.SetBinding(crop, MenuFlyoutItem.VisibilityProperty, cropVisBinding);

            MenuFlyoutItem flatten = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)this.Resources["CommandBarDefaultForeground"],
                Text = loader.GetString("MenuFlyoutItem_Flattened"),
                Command = ViewModel.SaveAndFlattenCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-flatten.png"),
                    ShowAsMonochrome = false
                }
            };

            _ActionsMenuFlyout.Items.Add(optimize);
            _ActionsMenuFlyout.Items.Add(encrypt);
            _ActionsMenuFlyout.Items.Add(crop);
            _ActionsMenuFlyout.Items.Add(flatten);
            AppBarButton_Actions.Flyout = _ActionsMenuFlyout;
            AppBarButton_Actions2.Flyout = _ActionsMenuFlyout;
        }

        private void InitializeCustomSaveCopyMenuFlyout()
        {
#if SUBSCRIPTIONS
            bool subscribed = ViewModel.GetSubscribed();
#else
            bool subscribed = true;
#endif
            foreach (var item in SaveCopyMenuItems)
            {
                MenuFlyoutItemBase menuItem = GetMenuItem(item, subscribed);
                if (menuItem is MenuFlyoutSubItem subMenuItem)
                {
                    foreach (var subItem in item.SubMenuItems)
                    {
                        MenuFlyoutItemBase menuSubItem = GetMenuItem(subItem, subscribed);
                        subMenuItem.Items.Add(menuSubItem);
                    }
                }

                _ActionsMenuFlyout.Items.Add(menuItem);
            }

#if SUBSCRIPTIONS
            //_saveCopyMenuFlyout.Opened += SaveCopyMenuFlyout_Opened;

            if (!subscribed)
            {
                //Add the upgrade menu item
                MenuFlyoutItem upgradeItem = new MenuFlyoutItem
                {
                    Text = "Unlock all actions",
                    Style = (Style)Application.Current.Resources["UpgradeMenuFlyoutItemStyle"]
                };

                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath("UpgradeCommand")
                };

                BindingOperations.SetBinding(upgradeItem, MenuFlyoutItem.CommandProperty, binding);

                _ActionsMenuFlyout.Items.Add(upgradeItem);
            }
#endif
            AppBarButton_Actions.Flyout = _ActionsMenuFlyout;
            AppBarButton_Actions2.Flyout = _ActionsMenuFlyout;
        }

        //private void SaveCopyMenuFlyout_Opened(object sender, object e)
        //{
        //    if (SaveCopyTip.IsOpen)
        //        SaveCopyTip.IsOpen = false;
        //}

        private MenuFlyoutItemBase GetMenuItem(IMenuItem item, bool subscribed)
        {
            MenuFlyoutItemBase menuItem;
            if (!item.IsSubMenuItem)
            {
                MenuFlyoutItem mItem = new MenuFlyoutItem
                {
                    Text = item.Text,
                    Command = item.Command,
                    CommandParameter = item.CommandParameter
                };

                if (!string.IsNullOrEmpty(item.Style))
                {
                    if (subscribed)
                    {
                        mItem.Style = (Style)Application.Current.Resources["MenuFlyoutItemRevealStyle"];
                    }
                    else
                    {
                        mItem.Style = (Style)Application.Current.Resources[item.Style];
                    }
                }

                if (item.IconType == MenuIconType.Font)
                {
                    mItem.Icon = new FontIcon
                    {
                        Glyph = item.Glyph
                    };
                }
                else
                {
                    mItem.Icon = new BitmapIcon
                    {
                        UriSource = new Uri(item.ImageUri),
                        ShowAsMonochrome = false
                    };
                }

                menuItem = mItem;
            }
            else
            {
                MenuFlyoutSubItem sItem = new MenuFlyoutSubItem
                {
                    Text = item.Text
                };

                if (!string.IsNullOrEmpty(item.Style))
                {
                    if (subscribed)
                    {
                        sItem.Style = (Style)Application.Current.Resources["MenuFlyoutSubItemRevealStyle"];
                    }
                    else
                    {
                        sItem.Style = (Style)Application.Current.Resources[item.Style];
                    }
                }

                if (item.IconType == MenuIconType.Font)
                {
                    sItem.Icon = new FontIcon
                    {
                        Glyph = item.Glyph
                    };
                }
                else
                {
                    sItem.Icon = new BitmapIcon
                    {
                        UriSource = new Uri(item.ImageUri),
                        ShowAsMonochrome = false
                    };
                }

                menuItem = sItem;
            }

            menuItem.Foreground = (SolidColorBrush)Resources["CommandBarDefaultForeground"];
            menuItem.FontSize = 13;

            if (item.IsCommandBinding)
            {
                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath(item.CommandBindingPath)
                };

                BindingOperations.SetBinding(menuItem, MenuFlyoutItem.CommandProperty, binding);
            }

            if (item.IsTextBinding)
            {
                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath(item.TextBindingPath),
                    Mode = BindingMode.TwoWay,
                    UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
                };

                BindingOperations.SetBinding(menuItem, MenuFlyoutItem.TextProperty, binding);
            }

            if (item.IsEnabledBinding)
            {
                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath(item.IsEnabledBindingPath),
                    Mode = BindingMode.TwoWay,
                    UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
                };

                BindingOperations.SetBinding(menuItem, MenuFlyoutItem.IsEnabledProperty, binding);
            }

            return menuItem;
        }

        private void ViewerTopCommandBar_Closing(object sender, object e) => HandleClosingCommandBar();

        private void HandleClosingCommandBar()
        {
            CommandbarHeight = _commandBarCompactHeight;

            if (ViewModel.IsAnnotationToolbarOpen)
            {
                //TabBorder.Margin = new Thickness(0, 40, 0, 0); // Note: Update margin to the TabBorder once the Annotation Toolbar is closed
                TabStoryBoardClose.Begin();
            }
            else
            {
                TabStoryBoardCollapse.Begin();
            }
        }

        private void ViewerTopCommandBar_Opening(object sender, object e) => HandleOpeningCommandBar();

        private void HandleOpeningCommandBar()
        {
            if (ViewModel?.IsAnnotationToolbarOpen == true)
            {
                //TabBorder.Margin = new Thickness(0, 48, 0, 0); // Note: Update margin to the TabBorder once the Annotation Toolbar is opened
                CommandbarHeight = 60; // Magic number, height AppBar expands to if all buttons have only 1 line of text
            }
            else
            {
                CommandbarHeight = 76; // Magic number, height AppBar expands to if at least one button has 2 lines of text
            }

            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                TabStoryBoardCollapse.Begin();
            }
            else
            {
                TabStoryBoardExpand.Begin();
            }
        }

        //private void ViewerTopAppBar_Closing(object sender, object e)
        //{
        //    if (ViewerSettings.PinCommandBar)
        //    {
        //        ViewerTopAppBar.IsOpen = true;
        //        ViewModel.UpdateUIVisiblity();
        //    }
        //}

        private void BackgroundGrid_Loaded(object sender, RoutedEventArgs e)
        {
            if (ViewerSettings.PinCommandBar)
            {
                SetViewerPageDown();
            }
        }

        private void SetViewerPageDown() => BackgroundGrid.Margin = new Thickness(0, 84, 0, 0);

        /// <summary>
        /// Handle right click event
        /// </summary>
        private void ViewerControl_RightTapped(object sender, RightTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType == PointerDeviceType.Mouse)
            {
                HandleCommandBarInteraction(e);
                e.Handled = true;
            }
        }

        /// <summary>
        /// Handle right click event
        /// </summary>
        private void HandleCommandBarInteraction(RightTappedRoutedEventArgs e)
        {
            // Make sure AnnotationBorder has been loaded
            DeferLoadAnnotationToolbar();

            if (AnnotationBorder == null || ViewModel == null)
                return;

            if (AnnotationBorder.Child is AnnotationCommandBar annotationCommandBar)
            {
                if (!ViewerSettings.PinCommandBar)
                {
                    if (!ViewModel.IsEntireAppBarOpen)
                    {
                        ViewModel.IsEntireAppBarOpen = true;
                    }
                    else if (!ViewModel.IsAnnotationToolbarOpen)
                    {
                        ViewModel.IsAppBarOpen = false;
                        ViewModel.IsEntireAppBarOpen = !ViewModel.IsEntireAppBarOpen;
                        annotationCommandBar.IsAppBarOpen = false;
                    }
                    else
                    {
                        annotationCommandBar.IsAppBarOpen = !annotationCommandBar.IsAppBarOpen;
                        ViewModel.UpdateUIVisiblity();
                    }
                }
                else
                {
                    if (ViewModel.PDFViewCtrl == null)
                        return;

                    var target = e.GetPosition(ViewModel.PDFViewCtrl);
                    if (ViewModel.PDFViewCtrl.GetPageNumberFromScreenPoint(target.X, target.Y) > 0)
                        return;

                    if (!ViewModel.IsAnnotationToolbarOpen)
                    {
                        ViewModel.IsAppBarOpen = !ViewModel.IsAppBarOpen;
                    }
                    else
                    {
                        if (!annotationCommandBar.IsInkOpen && !annotationCommandBar.IsPolygonSaveOpen && !annotationCommandBar.IsMeasureOpen)
                        {
                            annotationCommandBar.IsAppBarOpen = !annotationCommandBar.IsAppBarOpen;
                            ViewModel.UpdateUIVisiblity();
                        }
                    }
                }
            }
        }

        private void Settings_SettingUpdated(string settingName)
        {
            if (settingName == "OutlineDialogAnchorSide")
                PlaceOutlineDialog();

            if (settingName == "OutlineDialogWidth")
                PlaceOutlineDialog();
        }

        private void OutlineResizer_PositionUpdated(object sender, RoutedEventArgs e)
        {
            int position = OutlineResizer.Position;
            if (OutlineDialog.AnchorSide == AnchorSides.Right)
                position = 2 - position;

            ViewerSettings.OutlineDialogWidth = position;
            UpdateOutlineWidthPosition();
        }

        private bool SmallScreen
        {
            get => _smallScreen;
            set
            {
                if (value != _smallScreen)
                {
                    _smallScreen = value;
                    ViewModel.OutlineDialogNavigationCloses = _smallScreen;
                }
            }
        }

        private double CommandbarHeight
        {
            get => _commandbarHeight;
            set
            {
                if (_commandbarHeight != value)
                {
                    _commandbarHeight = value;

                    // NOTE: Once the Commandbar opens it adds margin to the tabbded control
                    //if (ViewModel.IsAnnotationToolbarOpen)
                    //    TabButtonControl.Margin = new Thickness(0, Math.Max(0, _commandbarHeight - _commandBarCompactHeight), 0, 0);
                }
            }
        }

        public static IList<IMenuItem> SaveCopyMenuItems { get; set; }

        public ViewerViewModel ViewModel
        {
            get
            {
                if (_viewModel != null)
                    return _viewModel;

                if (DataContext is ViewerViewModel viewModel)
                {
                    _viewModel = viewModel;
                }
                else
                {
                    _viewModel = new ViewerViewModel();
                    DataContext = _viewModel;
                }

                return _viewModel;
            }
            set => _viewModel = value;
        }

        /// <summary>
        /// This must be called in order to initialize all resources used by ViewerControl including:
        /// Reflow, conversion resources, DocumentPreviewCache
        /// </summary>
        private void Initialize()
        {
            if (mHasResourcesInitialized)
                return;

            mHasResourcesInitialized = true;

            try
            {
                string res_path = ApplicationData.Current.LocalFolder.Path;

                PDFNet.SetResourcesPath(res_path);
                PDFNet.SetPersistentCachePath(res_path);
                string additionalResourcePath = System.IO.Path.Combine(Package.Current.InstalledLocation.Path, "Resources");
                PDFNet.AddResourceSearchPath(additionalResourcePath);

                PDFViewCtrl.RegisterDependencyProperties();

                DocumentPreviewCache.Initialize(100 * 1024 * 1024, 0.1);
                ReflowProcessor.Initialize();
                DocumentPreviewCache.ClearCache(); // attempt to prevent startup crash
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        /// <summary>
        /// Initialize the viewer control internal components
        /// </summary>
        private async Task InitializeControlAsync()
        {
#if !ACTIONS
            IsSaveCopyEnabled = false;
#endif

            if (DataContext is ViewerViewModel viewModel)
            {
                ViewModel = viewModel;
            }
            else
            {
                ViewModel = new ViewerViewModel();
                DataContext = ViewModel;
            }

            if (InitializeOnCreation)
            {
                Initialize();
                await TabControlViewModel.Instance.InitializeInstanceAsync(EnableTracking);
            }

            if (_settingsUpdatedHandler == null)
            {
                _settingsUpdatedHandler = new ViewerSettings.SettingUpdatedDelegate(Settings_SettingUpdated);
                ViewerSettings.SettingUpdated += _settingsUpdatedHandler;
            }

            if (InitializeOnCreation)
            {
                ViewModel.Activate(null);
                InitializeDefaultSaveCopyMenuFlyout();
            }
            else
            {
                if (IsSaveCopyEnabled && SaveCopyMenuItems != null)
                {
                    InitializeCustomSaveCopyMenuFlyout();
                }
                else
                {
                    InitializeDefaultSaveCopyMenuFlyout();
                }
            }

            ViewModel.OpenSearch += ViewModel_OpenSearch;

            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                ViewModel.IsUiPinned = false;
                AppBarButton_PinUI.Visibility = Visibility.Collapsed;
                AppBarButton_UnpinUI.Visibility = Visibility.Collapsed;
            }

            ViewModel.OutlineDialogNavigationCloses = SmallScreen;
            ViewModel.PropertyChanged += ViewModel_PropertyChanged;
            ViewModel.AnnotationToolbar.AppBarChanged += AnnotationToolbar_AppBarChanged;
            ViewModel.FindTextResultFound += ViewModel_FindTextResultFound;
            ViewModel.DocumentModified += ViewModel_DocumentModified;

            _outlineResizer_PositionUpdated = new RoutedEventHandler(OutlineResizer_PositionUpdated);
            OutlineResizer.PositionUpdated += _outlineResizer_PositionUpdated;

            ViewModel.TabControlViewModel.MaximumItems = MaximumTabs;
            ViewModel.TabControlViewModel.ShowToolContextMenu += TabControlViewModel_ShowToolContextMenu;

            if (AnnotationToolbar != null)
            {
                ViewModel.AnnotationToolbar.IsStickyNoteToolButtonVisible = AnnotationToolbar.IsStickyNoteToolButtonVisible;
                ViewModel.AnnotationToolbar.IsHighlightToolButtonVisible = AnnotationToolbar.IsHighlightToolButtonVisible;
                ViewModel.AnnotationToolbar.IsUnderlineToolButtonVisible = AnnotationToolbar.IsUnderlineToolButtonVisible;
                ViewModel.AnnotationToolbar.IsStrikeoutToolButtonVisible = AnnotationToolbar.IsStrikeoutToolButtonVisible;
                ViewModel.AnnotationToolbar.IsSquigglyToolButtonVisible = AnnotationToolbar.IsSquigglyToolButtonVisible;
                ViewModel.AnnotationToolbar.IsStamperToolButtonVisible = AnnotationToolbar.IsStamperToolButtonVisible;
                ViewModel.AnnotationToolbar.IsSignatureToolButtonVisible = AnnotationToolbar.IsSignatureToolButtonVisible;
                ViewModel.AnnotationToolbar.IsInkToolButtonVisible = AnnotationToolbar.IsInkToolButtonVisible;
                ViewModel.AnnotationToolbar.IsEraserToolButtonVisible = AnnotationToolbar.IsEraserToolButtonVisible;
                ViewModel.AnnotationToolbar.IsMarkerToolButtonVisible = AnnotationToolbar.IsMarkerToolButtonVisible;
                ViewModel.AnnotationToolbar.IsTextBoxToolButtonVisible = AnnotationToolbar.IsTextBoxToolButtonVisible;
                ViewModel.AnnotationToolbar.IsShapeToolsButtonVisible = AnnotationToolbar.IsShapeToolsButtonVisible;
                ViewModel.AnnotationToolbar.IsMultiSelectToolButtonVisible = AnnotationToolbar.IsMultiSelectToolButtonVisible;
                ViewModel.AnnotationToolbar.IsPanToolsButtonVisible = AnnotationToolbar.IsPanToolsButtonVisible;
                ViewModel.AnnotationToolbar.IsUndoRedoToolButtonVisible = AnnotationToolbar.IsUndoRedoToolButtonVisible;
                ViewModel.AnnotationToolbar.IsSquareToolButtonVisible = AnnotationToolbar.IsSquareToolButtonVisible;
                ViewModel.AnnotationToolbar.IsCircleToolButtonVisible = AnnotationToolbar.IsCircleToolButtonVisible;
                ViewModel.AnnotationToolbar.IsLineToolButtonVisible = AnnotationToolbar.IsLineToolButtonVisible;
                ViewModel.AnnotationToolbar.IsArrowToolButtonVisible = AnnotationToolbar.IsArrowToolButtonVisible;
                ViewModel.AnnotationToolbar.IsRulerToolButtonVisible = AnnotationToolbar.IsRulerToolButtonVisible;
                ViewModel.AnnotationToolbar.IsMarqueeZoomToolButtonVisible = AnnotationToolbar.IsMarqueeZoomToolButtonVisible;
                ViewModel.AnnotationToolbar.IsRedactionToolButtonVisible = AnnotationToolbar.IsRedactionToolButtonVisible;
                ViewModel.AnnotationToolbar.IsCloseToolbarButtonVisible = AnnotationToolbar.IsCloseToolbarButtonVisible;
                ViewModel.AnnotationToolbar.IsExpandButtonVisible = AnnotationToolbar.IsExpandButtonVisible;
            }

#if !ACTIONS
            ViewModel.AnnotationToolbar.IsRedactionToolButtonVisible = false;
#endif

#if SUBSCRIPTIONS
            if (!ViewerSettings.SaveCopyExplained)
            {
                SaveCopyTip.IsOpen = true;
                ViewerSettings.SaveCopyExplained = true;
            }
#endif
        }

        private void TabControlViewModel_ShowToolContextMenu(ContextMenuControl menu, ShowingContextMenuEventArgs eventArgs)
        {
            _showToolContextMenu?.Invoke(menu, eventArgs);
        }

        public void Refresh(ViewerViewModel viewModel)
        {
            ViewModel.PropertyChanged -= ViewModel_PropertyChanged;

            ViewModel = viewModel;
            DataContext = viewModel;
            ViewModel.PropertyChanged += ViewModel_PropertyChanged;
            ViewModel.OpenSearch += ViewModel_OpenSearch;
            ResolveEditButtonFlyout();
        }

        private void ViewModel_DocumentModified()
        {
            _documentModified?.Invoke();
        }

        private void ViewModel_FindTextResultFound(Deferral deferral)
        {
            try
            {
                if (FindTextDialog.DoesSearchBoxHaveFocus)
                {
                    //AppBarButton_SearchOpen.Focus(FocusState.Keyboard);
                }
            }
            finally
            {
                deferral.Complete();
            }
        }

        private void AnnotationToolbar_AppBarChanged(bool isOpen)
        {
            if (isOpen)
            {
                HandleOpeningCommandBar();
            }
            else
            {
                HandleClosingCommandBar();
            }

            ViewModel.UpdateUIVisiblity();
        }

        private async void ViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            // hide the ViewerPage app bar in thumbnails/crop view
            if (e.PropertyName.Equals("IsThumbnailsViewOpen") 
                || e.PropertyName.Equals("IsCropViewOpen") 
                || e.PropertyName.Equals("IsPdfCompareViewOpen"))
            {
                if (ViewModel.IsThumbnailsViewOpen
                    || ViewModel.IsCropViewOpen
                    || ViewModel.IsPdfCompareViewOpen)
                {
                    //ViewModel.IsAppBarOpen = false;
                    ViewerTopAppBar.Visibility = Visibility.Collapsed;
                    AppBarButton_ViewMode.Flyout.Hide();
                    SetViewerPageUp();
                }
                else
                {
                    ViewModel.IsAppBarOpen = false;
                    ViewerTopAppBar.Visibility = Visibility.Visible;
                    if (ViewerSettings.PinCommandBar)
                        SetViewerPageDown();
                }
            }

            if (e.PropertyName.Equals("IsFullScreen"))
                AppBarButton_ViewMode.Flyout.Hide();

            if (e.PropertyName.Equals("MediaElement"))
            {
                // remove any previous text to speech elements placed
                foreach (var child in BackgroundGrid.Children)
                {
                    if (child is MediaElement element && element.Name == "TextToSpeechMedia")
                        BackgroundGrid.Children.Remove(child);
                }

                ViewModel.MediaElement.Name = "TextToSpeechMedia";

                if (ViewModel.MediaElement != null)
                {
                    // need to add MediaElement to UI tree in order to access its events
                    BackgroundGrid.Children.Add(ViewModel.MediaElement);
                    ViewModel.MediaElement.MediaEnded += Media_MediaEnded;
                }
            }

            if (e.PropertyName.Equals("IsQuickSettingsOpen") && !ViewModel.IsQuickSettingsOpen)
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () => AppBarButton_ViewMode.Flyout.Hide());

            if (e.PropertyName.Equals("IsAnnotationToolbarOpen"))
            {
                if (ViewModel.IsAnnotationToolbarOpen)
                {
                    SetViewerPageSlightlyDown();
                    TabStoryBoardClose.Begin();
                }
                else
                {
                    if (!ViewerSettings.PinCommandBar)
                    {
                        SetViewerPageUp();
                    }
                    else
                    {
                        SetViewerPageDown();
                    }

                    TabStoryBoardCollapse.Begin();
                }
            }

            if (e.PropertyName.Equals("IsUiPinned"))
            {
                if (ViewModel.IsUiPinned)
                {
                    SetViewerPageDown();
                }
                else
                {
                    SetViewerPageUp();
                }
            }

            if (e.PropertyName.Equals("IsConverting"))
                ResolveEditButtonFlyout();
        }

        private void ResolveEditButtonFlyout()
        {
            if (ViewModel?.IsConverting == true)
            {
                if (AppBarButton_Edit.Flyout == null)
                {
                    AppBarButton_Edit.Flyout = AppBarButton_Edit.Resources["EditButtonFlyout"] as Flyout;
                }
            }
            else
            {
                AppBarButton_Edit.Flyout = null;
            }
        }

        private void SetViewerPageUp() => BackgroundGrid.Margin = new Thickness(0, 0, 0, 0);

        private void SetViewerPageSlightlyDown()
        {
            BackgroundGrid.Margin = new Thickness(0, 40, 0, 0);
        }

        private void ViewModel_OpenSearch(object sender, Deferral e)
        {
            try
            {
                AppBarButton_Search.Flyout.ShowAt(AppBarButton_Search);
            }
            finally
            {
                e.Complete();
            }
        }

        private void FullSearchBox_Unloaded(object sender, RoutedEventArgs e)
        {
            if (sender is TextBlock control)
            {
                FindTextViewModel viewModel = (FindTextViewModel)control.DataContext;

                if (viewModel.IsPrevNextAvailable)
                    return;

                viewModel.ClearViewModel();
            }
        }

        private void ViewerControl_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (ViewModel == null)
                return;

            if (ViewModel.ToolManager != null)
                ViewModel.ToolManager.CloseOpenDialog(true);

            SmallScreen = (e.NewSize.Width < 500 || (e.NewSize.Height < 500 && e.NewSize.Width < 700));
            PlaceOutlineDialog();
            HandleSearchUi(e.NewSize.Width);

            if (e.NewSize.Width < 700)
            {
                Binding binding = AppBarButton_Save.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding != null)
                {
                    AppBarButton_Save2.SetBinding(VisibilityProperty, binding);
                    AppBarButton_Save.Visibility = Visibility.Collapsed;
                }

                Binding binding2 = AppBarButton_SaveAs.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding2 != null)
                {
                    AppBarButton_SaveAs2.SetBinding(VisibilityProperty, binding2);
                    AppBarButton_SaveAs.Visibility = Visibility.Collapsed;
                }

                if (ShowActionsOption)
                {
                    Binding binding3 = AppBarButton_Actions.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                    if (binding3 != null)
                    {
                        AppBarButton_Actions2.SetBinding(VisibilityProperty, binding3);
                        AppBarButton_Actions.Visibility = Visibility.Collapsed;
                    }
                }
            }
            else
            {
                Binding binding = AppBarButton_Save2.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding != null)
                {
                    AppBarButton_Save.SetBinding(VisibilityProperty, binding);
                    AppBarButton_Save2.Visibility = Visibility.Collapsed;
                }

                Binding binding2 = AppBarButton_SaveAs2.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding != null)
                {
                    AppBarButton_SaveAs.SetBinding(VisibilityProperty, binding2);
                    AppBarButton_SaveAs2.Visibility = Visibility.Collapsed;
                }

                if (ShowActionsOption)
                {
                    Binding binding3 = AppBarButton_Actions2.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                    if (binding3 != null)
                    {
                        AppBarButton_Actions.SetBinding(VisibilityProperty, binding3);
                        AppBarButton_Actions2.Visibility = Visibility.Collapsed;
                    }
                }
            }
        }

        private void HandleSearchUi(double width)
        {
            if (width < 450)
            {
                AppBarButton_FindClose.Width = 40;
                AppBarButton_FindClose.Visibility = Visibility.Collapsed;
                FullTextSearchListView.ItemTemplate = (DataTemplate)Resources["SmallFullTextSearchListViewDataTemplate"];
            }
            else
            {
                AppBarButton_FindClose.Width = 68;
                AppBarButton_FindClose.Visibility = Visibility.Visible;
                FullTextSearchListView.ItemTemplate = (DataTemplate)Resources["FullTextSearchListViewDataTemplate"];
            }
        }

        private void PlaceOutlineDialog()
        {
            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                OutlineDialog.AnchorSide = AnchorSides.None;
            }
            else
            {
                OutlineDialog.AnchorSide = ViewerSettings.OutlineDialogAnchorSide;
            }

            double newWidth = Math.Min(ActualWidth, 400);
            if (OutlineDialog.AnchorSide == _oldOutlineAnchorSide && newWidth == _oldWidth)
                return;

            _oldWidth = newWidth;
            if (OutlineDialog.AnchorSide == AnchorSides.None)
            {
                NonModalDisplayGrid.BorderThickness = new Thickness(0, 1, 0, 0);
                OutlineHostGrid.HorizontalAlignment = HorizontalAlignment.Stretch;
                OutlineSpaceHolderGrid.SetValue(Grid.ColumnProperty, 1);
                SnapResizeHost.Visibility = Visibility.Collapsed;
                OutlineHostEntranceControl.EntranceAnimation = FadeInOutline;
                OutlineHostEntranceControl.ExitAnimation = FadeOutOutline;
                if (!OutlineHostEntranceControl.IsOpen)
                {
                    OutlineHostGridTranslation.X = 0;
                    OutlineHostEntranceControl.Opacity = 0;
                }

                UpdateOutlineWidthPosition();
            }
            else if (OutlineDialog.AnchorSide == AnchorSides.Left)
            {
                NonModalDisplayGrid.BorderThickness = new Thickness(0, 1, 1, 0);
                OutlineSpaceHolderGrid.SetValue(Grid.ColumnProperty, 0);
                OutlineHostGrid.HorizontalAlignment = HorizontalAlignment.Left;
                SnapResizeHost.HorizontalAlignment = HorizontalAlignment.Left;
                OutlineHostEntranceControl.EntranceAnimation = SlideInOutline;
                OutlineHostEntranceControl.ExitAnimation = SlideOutOutlineToLeft;
                if (!OutlineHostEntranceControl.IsOpen)
                {
                    OutlineHostEntranceControl.Opacity = 1;
                    OutlineHostGridTranslation.X = -OutlineDialog.Width;
                }

                OutlineResizer.Offsets = GetOutlineOffsets();
                OutlineResizer.Position = ViewerSettings.OutlineDialogWidth;
                UpdateOutlineWidthPosition();
            }
            else if (OutlineDialog.AnchorSide == AnchorSides.Right)
            {
                NonModalDisplayGrid.BorderThickness = new Thickness(1, 1, 0, 0);
                OutlineSpaceHolderGrid.SetValue(Grid.ColumnProperty, 2);
                OutlineHostGrid.HorizontalAlignment = HorizontalAlignment.Right;
                SnapResizeHost.HorizontalAlignment = HorizontalAlignment.Right;
                OutlineHostEntranceControl.EntranceAnimation = SlideInOutline;
                OutlineHostEntranceControl.ExitAnimation = SlideOutOutlineToRight;
                if (!OutlineHostEntranceControl.IsOpen)
                {
                    OutlineHostEntranceControl.Opacity = 1;
                    OutlineHostGridTranslation.X = OutlineDialog.Width;
                }

                OutlineResizer.Offsets = GetOutlineOffsets();
                OutlineResizer.Position = 2 - ViewerSettings.OutlineDialogWidth;
                UpdateOutlineWidthPosition();
            }
        }

        private List<double> GetOutlineOffsets()
        {
            List<double> offsets = new List<double>();
            if (OutlineDialog.AnchorSide == AnchorSides.Left)
            {
                offsets.Add(OutlineDialog.LayoutMinWidths[WidthLayouts.Narrow]);
                offsets.Add(OutlineDialog.LayoutMinWidths[WidthLayouts.Full]);
                offsets.Add(Math.Min(ActualWidth, 400));
            }
            else if (OutlineDialog.AnchorSide == AnchorSides.Right)
            {
                offsets.Add(500 - Math.Min(ActualWidth, 400));
                offsets.Add(500 - OutlineDialog.LayoutMinWidths[WidthLayouts.Full]);
                offsets.Add(500 - OutlineDialog.LayoutMinWidths[WidthLayouts.Narrow]);
            }

            return offsets;
        }

        private void UpdateOutlineWidthPosition()
        {
            int position = ViewerSettings.OutlineDialogWidth;
            if (position == 0)
            {
                OutlineDialog.Width = OutlineDialog.LayoutMinWidths[WidthLayouts.Narrow];
            }
            else if (position == 1)
            {
                OutlineDialog.Width = OutlineDialog.LayoutMinWidths[WidthLayouts.Full];
            }
            else
            {
                OutlineDialog.Width = Math.Min(ActualWidth, 400.0);
            }
        }

        private async void ViewerControl_Loaded(object sender, RoutedEventArgs e)
        {
            await InitializeControlAsync();
        }

        /// <summary>
        /// Activate the ViewerControl with a file
        /// </summary>
        /// <remarks>
        /// Supported file types:
        /// .pdf, .doc, .docx, .ppt, .pptx, .xls, .xlsx, .oxps, .xps, .jpeg, .jpg, .png, .bmp, .cbz    
        ///  
        /// </remarks>
        /// <param name="file">StorageFile</param>
        /// <param name="properties">Optional properties</param>
        /// <returns></returns>
        public async Task ActivateWithFileAsync(StorageFile file, NewDocumentProperties properties = null)
        {
            if (DataContext is ViewerViewModel viewModel)
            {
                await viewModel.ActivateWithFileAsync(file, properties);
                viewModel.IsThumbnailsViewOpen = false;
            }
        }

        /// <summary>
        /// Waits until the data context containing the appropriate full search item is ready and
        /// then positions the highlight rectangle according to the information
        /// </summary>
        public void Highlight_Text(FrameworkElement sender, DataContextChangedEventArgs args)
        {
            Rectangle rect = sender as Rectangle;

            // Only care when data context becomes non-null
            if (!(rect.DataContext is FullSearchItem dataContext))
                return;

            // Use a textblock to hold text to measure where highlight should be positioned
            TextBlock textBlock = new TextBlock();
            string contextText = dataContext.ContextText;

            string leftText = contextText.Substring(0, contextText.IndexOf(dataContext.SearchText));
            textBlock.Text = leftText;
            // Trailing empty spaces won't count for width, so use a placeholder character
            if (textBlock.Text.EndsWith(" "))
                textBlock.Text = textBlock.Text.Substring(0, textBlock.Text.Length - 1) + "i";

            textBlock.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            textBlock.Arrange(new Windows.Foundation.Rect(new Windows.Foundation.Point(0, 0), textBlock.DesiredSize));
            double leftOffset = textBlock.DesiredSize.Width;

            string rightText = leftText + dataContext.SearchText;
            if (textBlock.Text.EndsWith(" "))
                textBlock.Text = textBlock.Text.Substring(0, textBlock.Text.Length - 1) + "i";

            textBlock.Text = rightText;
            textBlock.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            textBlock.Arrange(new Windows.Foundation.Rect(new Windows.Foundation.Point(0, 0), textBlock.DesiredSize));
            double rightOffset = textBlock.DesiredSize.Width;

            // Apply calculations to position highlight for text
            rect.Margin = new Thickness(leftOffset, 0, 0, 0);
            rect.Height = 17;
            rect.Width = rightOffset - leftOffset;
        }

        public void SearchFlyout_OnOpened(object sender, object e)
        {
            if (FindTextDialog == null)
                DeferLoadFindDialog();

            FindTextDialog.SetFocus();
            InputPane.GetForCurrentView().TryShow();
        }

        public void FindText_ContentIsAvailable(EntranceAnimationContentControlEx control)
        {
            DeferLoadFindDialog();
            FindTextDialog.SetFocus();
        }

        private void Media_MediaEnded(object sender, RoutedEventArgs e)
        {
            ViewModel.AreAudioButtonsVisible = false;
        }

        /*
         * NOTE: Using DeferLoadStrategy in order to boost ViewerControl startup time by lazy loading a few UI Elements
         * This method adds ~600 bytes to the memory for each element affected. Need to be carefull with overusage
         */
        #region Realize DeferLoadStrategy

        private void AppBarButton_Edit_Click(object sender, RoutedEventArgs e)
        {
            DeferLoadAnnotationToolbar();
        }

        private void AppBarButton_ViewMode_Click(object sender, RoutedEventArgs e)
        {
            this.FindName("QuickSettings");
            this.FindName("customColorPopupControl");
            this.FindName("CropViewerHost");
            this.FindName("cropPopupControl");
        }

        private void AppBarButton_SaveCopy_Click(object sender, RoutedEventArgs e)
        {
            this.FindName("CropViewerHost");
            this.FindName("cropPopupControl");
            this.FindName("optimizePopupControl");
            this.FindName("passwordPopupControl");
        }

        private void AppBarButton_Search_Click(object sender, RoutedEventArgs e)
        {
            DeferLoadFindDialog();
        }

        private void DeferLoadAnnotationToolbar()
        {
            this.FindName("AnnotationBorder");
        }

        private void DeferLoadFindDialog()
        {
            this.FindName("FindTextDialog");
        }

        #endregion
    }
}
