﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Media;


using pdftron.PDF;
using pdftron.SDF;

using PDFPoint = pdftron.PDF.Point;
using UIPoint = Windows.Foundation.Point;
using PDFRect = pdftron.PDF.Rect;
using UIRect = Windows.Foundation.Rect;
using Windows.Storage;
using Windows.Storage.Pickers;
using System.IO;
using pdftron.Common;
using Windows.Media.Capture;
using pdftron.PDF.Tools.Utilities;
using Windows.ApplicationModel.Resources;

namespace pdftron.PDF.Tools
{
    class ImageStamper : Tool
    {
        // Floating ImageStamper
        protected Utilities.StampManager mStampManager;

        protected UIPoint mTargetPoint;

        // Form filling ImageStamper
        protected Annots.Widget mWidget;
        protected Dictionary<string, string> mMenuTitles;
        protected int mPageNumber = -1;

        protected bool mShouldAddNote = false;

        protected readonly IList<string> Valid_Image_Types = new List<string>() { ".jpg", ".bmp", ".jpeg", ".tif", ".tiff", ".png", ".gif", ".ico", ".jpeg-xr" };

        public ImageStamper(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_stamper;
            mToolMode = ToolType.e_stamper;
            mStampManager = new Utilities.StampManager();
        }

        public void SetTargetPoint(UIPoint point)
        {
            mTargetPoint = point;
            CreateMenuIfNeeded();
        }

        internal override void OnCreate()
        {
            if (mAnnot != null)
            {

            }
        }

        internal override void OnClose()
        {
            base.OnClose();
            EnableScrolling();
            mPDFView.RequestRendering();
        }

        protected void CreateMenuIfNeeded()
        {
            Dictionary<string, string> createSignatureOptionsDict = new Dictionary<string, string>();
            createSignatureOptionsDict["Stamper FilePicker"] = ResourceHandler.GetString("CreateOption_Stamper_FilePicker");
            createSignatureOptionsDict["Stamper Camera"] = ResourceHandler.GetString("CreateOption_Stamper_Camera");
            mContextMenu = new Controls.ContextMenuControl(mPDFView, createSignatureOptionsDict);
            mContextMenu.ItemClicked += FloatingOnCommandMenuClicked;
            mContextMenu.TargetPoint(mTargetPoint);
            mIsShowingCommandMenu = true;

            var eventArgs = new ShowingContextMenuEventArgs
            {
                ToolType = ToolType.e_stamper,
                AnnotType = AnnotType.e_Stamp,
                Annotation = mAnnot,
                PageNumber = mPDFView.GetCurrentPage()
            };
            mToolManager.RaiseOnShowContextMenu(mContextMenu, eventArgs);

            mContextMenu.Show();
        }

        internal async void FloatingOnCommandMenuClicked(string title)
        {
            HideContextMenu();
            if (title.Equals("Stamper FilePicker", StringComparison.OrdinalIgnoreCase))
            {
                StorageFile file = await GetOpenFileFromPicker(Valid_Image_Types);
                if (file == null)
                    return;

                Page page = await CreatePDFPageFromImage(file);
                if (page != null)
                    AddStamp(page);
            }

            else if (title.Equals("Stamper Camera", StringComparison.OrdinalIgnoreCase))
            {
                CameraCaptureUI cameraUI = new CameraCaptureUI();
                cameraUI.PhotoSettings.AllowCropping = false;
                cameraUI.PhotoSettings.MaxResolution = CameraCaptureUIMaxPhotoResolution.HighestAvailable;
                cameraUI.PhotoSettings.AllowCropping = true;

                StorageFile capturedMedia = await cameraUI.CaptureFileAsync(CameraCaptureUIMode.Photo);
                if (capturedMedia == null)
                    return;

                Page page = await CreatePDFPageFromImage(capturedMedia);
                if (page != null)
                    AddStamp(page);
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }
        }

        private async Task<Page> CreatePDFPageFromImage(StorageFile file)
        {
            try
            {
                PDFDoc doc = new PDFDoc();
                Page page = doc.PageCreate();

                Stream stream = await file.OpenStreamForReadAsync();
                byte[] bytes = new byte[(int)stream.Length];
                stream.Read(bytes, 0, (int)stream.Length);

                Image img = Image.Create(doc.GetSDFDoc(), bytes);
                ElementBuilder f = new ElementBuilder();
                Element element = f.CreateImage(img, new Matrix2D(img.GetImageWidth(), 0, 0, img.GetImageHeight(), 0, 0));

                page.SetMediaBox(new PDFRect(0, 0, img.GetImageWidth(), img.GetImageHeight()));
                page.SetCropBox(new PDFRect(0, 0, img.GetImageWidth(), img.GetImageHeight()));

                ElementWriter writer = new ElementWriter();
                writer.Begin(page);
                writer.WritePlacedElement(element);
                writer.End();

                return page;
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendException(ex);
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                ContentDialog dialog = new ContentDialog
                {
                    PrimaryButtonText = loader.GetString("pdftron_Controls_ThumbnailsView_Delete_Message_Confirm")
                };
#if DEBUG
                dialog.Content = $"{loader.GetString("ContentDialog_Read_Image_Failed_Msg")}\n\n{ex.Message}";
#else
                dialog.Content = loader.GetString("ContentDialog_Read_Image_Failed_Msg");
#endif

                await dialog.ShowAsync(ContentDialogPlacement.Popup);
            }

            return null;
        }

        protected void AddStamp(Page stampPage)
        {
            int pageNumber = mPDFView.GetPageNumberFromScreenPoint(mTargetPoint.X, mTargetPoint.Y);
            if (pageNumber <= 0)
            {
                return;
            }

            try
            {
                mPDFView.DocLock(true);
                PDFDoc doc = mPDFView.GetDoc();
                Page page = doc.GetPage(pageNumber);

                PageRotate viewRotation = mPDFView.GetRotation();
                PageRotate pageRotation = page.GetRotation();

                // If the page itself is rotated, we want to "rotate" width, height and re-calculate edges as well
                PDFRect pageViewBox = page.GetBox(mPDFView.GetPageBox());
                Rect pageCropRect = page.GetCropBox();

                // Note, stamper stamps relative to the CropBox,
                // i.e. (0, 0) is the bottom left corner of crop box.
                double pageWidth = pageViewBox.Width();
                double pageHeight = pageViewBox.Height();
                double leftEdge = pageViewBox.x1 - pageCropRect.x1;
                double bottomEdge = pageViewBox.y1 - pageCropRect.y1;
                if (pageRotation == PageRotate.e_90 || pageRotation == PageRotate.e_270)
                {
                    pageWidth = pageViewBox.Height();
                    pageHeight = pageViewBox.Width();

                    if (pageRotation == PageRotate.e_90)
                    {
                        leftEdge = pageViewBox.y1 - pageCropRect.y1;
                        bottomEdge = pageCropRect.x2 - pageViewBox.x2;
                    }
                    else
                    {
                        leftEdge = pageCropRect.y2 - pageViewBox.y2;
                        bottomEdge = pageViewBox.x1 - pageCropRect.x1;
                    }
                }
                if (pageRotation == PageRotate.e_180)
                {
                    leftEdge = pageCropRect.x2 - pageViewBox.x2;
                    bottomEdge = pageCropRect.y2 - pageViewBox.y2;
                }

                PDFRect stampRect = stampPage.GetCropBox();
                double maxWidth = 200;
                double maxHeight = 200;

                if (pageWidth < maxWidth)
                {
                    maxWidth = pageWidth;
                }
                if (pageHeight < maxHeight)
                {
                    maxHeight = pageHeight;
                }

                double stampWidth = stampRect.Width();
                double stampHeight = stampRect.Height();

                // if the viewer rotates pages, we want to treat it as if it's the stamp that's rotated
                if (viewRotation == PageRotate.e_90 || viewRotation == PageRotate.e_270)
                {
                    double temp = stampWidth;
                    stampWidth = stampHeight;
                    stampHeight = temp;
                }

                double scaleFactor = Math.Min(maxWidth / stampWidth, maxHeight / stampHeight);
                stampWidth *= scaleFactor;
                stampHeight *= scaleFactor;

                Stamper stamper = new Stamper(StamperSizeType.e_absolute_size, stampWidth, stampHeight);
                stamper.SetAlignment(StamperHorizontalAlignment.e_horizontal_left, StamperVerticalAlignment.e_vertical_bottom);
                stamper.SetAsAnnotation(true);

                pdftron.Common.DoubleRef x = new Common.DoubleRef(mTargetPoint.X);
                pdftron.Common.DoubleRef y = new Common.DoubleRef(mTargetPoint.Y);
                mPDFView.ConvScreenPtToPagePt(x, y, pageNumber);

                int pgX = (int)x.Value;
                int pgY = (int)y.Value;
                pdftron.Common.Matrix2D mtx = page.GetDefaultMatrix();
                mtx.Mult(x, y);

                int postX = (int)x.Value;
                int postY = (int)y.Value;

                System.Diagnostics.Debug.WriteLine(string.Format("Page {0}, {1} -> {2}, {3}", pgX, pgY, postX, postY));

                double xPos = x.Value - (stampWidth / 2);
                double yPos = y.Value - (stampHeight / 2);

                if (xPos > leftEdge + pageWidth - stampWidth)
                {
                    xPos = leftEdge + pageWidth - stampWidth;
                }
                if (xPos < leftEdge)
                {
                    xPos = leftEdge;
                }
                if (yPos > bottomEdge + pageHeight - stampHeight)
                {
                    yPos = bottomEdge + pageHeight - stampHeight;
                }
                if (yPos < bottomEdge)
                {
                    yPos = bottomEdge;
                }

                stamper.SetPosition(xPos, yPos);

                int stampRotation = (4 - (int)viewRotation) % 4; // 0 = 0, 90 = 1; 180 = 2, and 270 = 3
                stamper.SetRotation(stampRotation * 90.0);

                stamper.StampPage(doc, stampPage, new PageSet(pageNumber));

                int numAnnots = page.GetNumAnnots();

                IAnnot annot = page.GetAnnot(numAnnots - 1);
                pdftron.SDF.Obj obj = annot.GetSDFObj();
                obj.PutString(ToolManager.StamperAnnotationIdentifyingString, "");
                AnnotType theType = annot.GetAnnotType();
                Annots.RubberStamp stamp = (Annots.RubberStamp)annot;

                UtilityFunctions.SetUniqueId(stamp, Guid.NewGuid().ToString());

                mPDFView.UpdateWithAnnot(annot, pageNumber);
                mToolManager.RaiseAnnotationAddedEvent(annot, pageNumber);

                // Set up to transfer to AnnotEditTool
                mAnnot = annot;
                BuildAnnotBBox();
                mAnnotPageNum = pageNumber;

                mNextToolMode = ToolType.e_annot_edit;
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                mNextToolMode = ToolType.e_pan;
            }
            finally
            {
                mPDFView.DocUnlock();
            }

            mToolManager.CreateTool(mNextToolMode, this);
            if (mNextToolMode == ToolType.e_annot_edit)
            {
                AnnotEdit annotEdit = mToolManager.CurrentTool as AnnotEdit;
                annotEdit.CreateAppearance();
            }

        }

#region Event Handlers

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            mShouldAddNote = false;
            AddPointer(e.Pointer);
            if (mToolManager.ContactPoints.Count > 1)
            {
                mToolManager.EnableScrollByManipulation = true;
                return false;
            }

            UIPoint sc_DownPoint = e.GetCurrentPoint(mPDFView).Position;
            int downPageNumber = mPDFView.GetPageNumberFromScreenPoint(sc_DownPoint.X, sc_DownPoint.Y);
            if (downPageNumber < 1)
            {
                if (mIsShowingCommandMenu)
                {
                    HideContextMenu();
                }
                mToolManager.EnableOneFingerScroll = true;
                mToolManager.EnableScrollByManipulation = true;
                return false;
            }

            if (mIsShowingCommandMenu)
            {
                EndCurrentTool(ToolType.e_pan);
            }
            else // this means we haven't gotten a point yet, we should do so now
            {
                DisableScrolling();
            }
            mShouldAddNote = true;

            return false;
        }


        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            RemovePointer(e.Pointer);
            if (mToolManager.ContactPoints.Count == 0 && mShouldAddNote)
            {
                UIPoint upPoint = e.GetCurrentPoint(mPDFView).Position;
                int pageNumber = mPDFView.GetPageNumberFromScreenPoint(upPoint.X, upPoint.Y);
                if (pageNumber > 0)
                {
                    SetTargetPoint(upPoint);
                }
                else
                {
                    EndCurrentTool(ToolType.e_pan);
                }
            }
            return false;
        }

#endregion Event Handlers

        //////////////////////////////////////////////////////////////////////////
        // Form ImageStamper
        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            if (mAnnot != null)
            {
                if (mJustSwitchedFromAnotherTool)
                {
                    mJustSwitchedFromAnotherTool = false;

                    // We know we're looking at a form widget with a digital ImageStamper
                    mPDFView.DocLockRead();
                    mWidget = (pdftron.PDF.Annots.Widget)mAnnot;
                    mPDFView.DocUnlockRead();

                    UIPoint screenPoint = e.GetPosition(mPDFView);
                    mPageNumber = mPDFView.GetPageNumberFromScreenPoint(screenPoint.X, screenPoint.Y);

                    GetCommandMenuOptions();
                    if (mToolManager.IsPopupMenuEnabled)
                    {
                        CreateCommandMenu();
                    }
                }
                else
                {
                    mNextToolMode = ToolType.e_pan;
                }
            }

            return true;
        }

        //////////////////////////////////////////////////////////////////////////
        // Command menu
        protected void GetCommandMenuOptions()
        {
            // in case an old one is somehow up
            if (mIsShowingCommandMenu)
            {
                HideContextMenu();
            }

            mMenuTitles = new Dictionary<string, string>();
        }

        protected void CreateCommandMenu()
        {
            mContextMenu = new Controls.ContextMenuControl(mPDFView, mMenuTitles);
            mContextMenu.ItemClicked += WidgetOnCommandMenuClicked;
            PositionMenu();
            ShowMenu();
        }

        internal void WidgetOnCommandMenuClicked(string title)
        {
            if (title.Equals("delete", StringComparison.OrdinalIgnoreCase))
            {
                HideMenu();
                try
                {
                    mPDFView.DocLock(true);
                    mWidget.GetSDFObj().Erase("AP");
                    mWidget.RefreshAppearance();
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlock();
                }

                mPDFView.UpdateWithAnnot(mWidget, mAnnotPageNum);
                mToolManager.RaiseAnnotationEditedEvent(mWidget, mAnnotPageNum);
                mToolManager.CreateTool(ToolType.e_pan, this);
            }
        }

        //////////////////////////////////////////////////////////////////////////
        // Utility Functions
        protected void ShowMenu()
        {
            if (mContextMenu != null && !mIsShowingCommandMenu)
            {
                mIsShowingCommandMenu = true;

                var eventArgs = new ShowingContextMenuEventArgs
                {
                    ToolType = ToolType.e_stamper,
                    AnnotType = mAnnot.GetAnnotType(),
                    Annotation = mAnnot,
                    PageNumber = mPDFView.GetCurrentPage()
                };
                mToolManager.RaiseOnShowContextMenu(mContextMenu, eventArgs);

                mContextMenu.Show();
                DisableScrolling();
            }
        }

        protected void HideMenu()
        {
            if (mContextMenu != null && mIsShowingCommandMenu)
            {
                HideContextMenu();
            }
        }
        protected virtual void PositionMenu()
        {
            if (mContextMenu != null)
            {
                PDFRect rect = mWidget.GetRect();
                rect = ConvertFromPageRectToScreenRect(rect, mAnnotPageNum);
                mContextMenu.TargetSquare(rect.x1, rect.y1, rect.x2, rect.y2);
            }
        }

        private async Task<StorageFile> GetOpenFileFromPicker(IList<string> fileTypes)
        {
            StorageFile theFile = null;

            FileOpenPicker fileOpenPicker = new FileOpenPicker();
            fileOpenPicker.ViewMode = PickerViewMode.Thumbnail;

            foreach (string filetype in fileTypes)
            {
                fileOpenPicker.FileTypeFilter.Add(filetype);
            }
            theFile = await fileOpenPicker.PickSingleFileAsync();
            return theFile;
        }

        private void HideContextMenu()
        {
            if (mContextMenu != null)
            {
                ResetContextMenuEvents();
                mContextMenu.Hide();
            }
            mIsShowingCommandMenu = false;
        }

        private void ResetContextMenuEvents()
        {
            if (mContextMenu != null)
            {
                mContextMenu.ItemClicked -= FloatingOnCommandMenuClicked;
                mContextMenu.ItemClicked -= WidgetOnCommandMenuClicked;
            }
        }
    }
}
