﻿using Microsoft.Graphics.Canvas;
using Microsoft.Graphics.Canvas.Geometry;
using pdftron.PDF.Annots;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using Windows.UI;
using Windows.UI.Input.Inking;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;
using PDFDouble = pdftron.Common.DoubleRef;
using PDFRect = pdftron.PDF.Rect;
using UIPoint = Windows.Foundation.Point;

namespace pdftron.PDF.Tools
{
    public class LassoCreate : AnnotEditRectGroup
    {
        private bool mShapeHasBeenCreated = false;
        Path mPath;
        Windows.UI.Xaml.Shapes.Line mLine;

        protected int START_DRAWING_THRESHOLD = 5;

        protected bool mStartDrawing = false;
        protected bool mPointerDown = false;
        protected Point m_cnPt1 = null;
        protected Point m_cnPt2 = null;
        protected bool mMirrorX = false;
        protected bool mMirrorY = false;
        protected CompositeTransform mRenderTransform;
        protected int mDownPageNumber = -1;

        protected bool mCancelledDrawing = false;

        // Visuals for shapes
        protected double DEFAULT_OPACITY = 0.3;
        protected double mStrokeThickness = 1;
        protected double mZoomLevel = 1;
        protected double mDrawThickness;
        protected SolidColorBrush mStrokeBrush = new SolidColorBrush(Colors.Black);
        protected bool mUseFill = true;
        protected bool mUseStroke = true;
        protected SolidColorBrush mFillBrush = new SolidColorBrush(Color.FromArgb(100, 255, 0, 0));
        protected bool mSimpleShapeCreateSubclassHandlesProperties = false;
        protected bool mHasFirstPoint = false;
        PathFigure mPathFigure;

        public LassoCreate(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_lasso_create;
            mToolMode = ToolType.e_lasso_create;

            mPathFigure = new PathFigure();
        }

        #region Interaction Events
        internal override bool PointerPressedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (!mCreated)
            {
                AddPointer(e.Pointer);
                DisableScrolling();
                mToolManager.ToolModeAfterEditing = ToolType.e_none;

                if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse && !e.GetCurrentPoint(null).Properties.IsLeftButtonPressed)
                {
                    return false; // if mouse, only allow left button
                }

                mPointerID = (int)e.Pointer.PointerId;
                UIPoint screenPoint = e.GetCurrentPoint(mPDFView).Position;
                PDFDoc doc = mPDFView.GetDoc();
                if (doc == null)
                {
                    return false;
                }

                mViewerCanvas = mPDFView.GetAnnotationCanvas();

                // Ensure we're on a page
                mToolManager.EnableOneFingerScroll = false;

                mDownPageNumber = mPDFView.GetPageNumberFromScreenPoint(screenPoint.X, screenPoint.Y);
                if (mDownPageNumber < 1)
                {
                    mDownPageNumber = mPDFView.GetCurrentPage();
                }

                if (mDownPageNumber < 1)
                {
                    mPointerID = -1;
                    mToolManager.EnableOneFingerScroll = true;
                    mToolManager.EnableScrollByManipulation = true;
                    mDownPageNumber = -1;
                    return false;
                }

                // Get this from mPDFViewCtrl
                mPage = doc.GetPage(mDownPageNumber);

                // Get the page's bounding box in canvas space
                mPageCropOnClient = BuildPageBoundBoxOnClient(mDownPageNumber);

                mPointerDown = true;

                // Drawing Setup
                mUseStroke = true;
                mStrokeThickness = Settings.GetThickness(this.ToolMode);
                mOpacity = Settings.GetOpacity(this.ToolMode);

                mZoomLevel = mPDFView.GetZoom();
                mDrawThickness = mStrokeThickness * mZoomLevel;

                UIPoint pp = e.GetCurrentPoint(mViewerCanvas).Position;
                m_cnPt1 = new Point(pp.X, pp.Y);
                if (mPageCropOnClient != null)
                {
                    if (m_cnPt1.x < mPageCropOnClient.x1)
                    {
                        m_cnPt1.x = mPageCropOnClient.x1;
                    }
                    if (m_cnPt1.x > mPageCropOnClient.x2)
                    {
                        m_cnPt1.x = mPageCropOnClient.x2;
                    }
                    if (m_cnPt1.y < mPageCropOnClient.y1)
                    {
                        m_cnPt1.y = mPageCropOnClient.y1;
                    }
                    if (m_cnPt1.y > mPageCropOnClient.y2)
                    {
                        m_cnPt1.y = mPageCropOnClient.y2;
                    }
                }

                m_cnPt2 = new Point(m_cnPt1.x, m_cnPt1.y);
                mViewerCanvas.CapturePointer(e.Pointer);

                mShapeHasBeenCreated = false;
            }
            else
            {
                base.PointerPressedHandler(sender, e);
            }

            return false;
        }

        internal override bool PointerMovedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (!mCreated)
            {
                // Canvas takes care of mirroring
                if (mPointerDown)
                {
                    if (e.Pointer.PointerId != mPointerID)
                    {
                        return false;
                    }
                    UIPoint pp = e.GetCurrentPoint(mViewerCanvas).Position;
                    m_cnPt2.x = pp.X;
                    m_cnPt2.y = pp.Y;

                    // want to make sure we're within our bounding box.
                    if (mPageCropOnClient != null)
                    {
                        if (m_cnPt2.x < mPageCropOnClient.x1)
                        {
                            m_cnPt2.x = mPageCropOnClient.x1;
                        }
                        if (m_cnPt2.x > mPageCropOnClient.x2)
                        {
                            m_cnPt2.x = mPageCropOnClient.x2;
                        }
                        if (m_cnPt2.y < mPageCropOnClient.y1)
                        {
                            m_cnPt2.y = mPageCropOnClient.y1;
                        }
                        if (m_cnPt2.y > mPageCropOnClient.y2)
                        {
                            m_cnPt2.y = mPageCropOnClient.y2;
                        }
                    }

                    // We don't want to start drawing until we have a bit of a distance
                    if (!mStartDrawing)
                    {
                        // wait until you're at least START_DRAWING_THRESHOLD distance from the start
                        if ((Math.Abs(m_cnPt2.x - m_cnPt1.x) > START_DRAWING_THRESHOLD)
                            || (Math.Abs(m_cnPt2.y - m_cnPt1.y) > START_DRAWING_THRESHOLD))
                        {
                            mStartDrawing = true;
                            mViewerCanvas.Children.Add(this);
                        }
                    }

                    // Size the canvas according to the two points
                    if (mStartDrawing)
                    {
                        this.Width = Math.Abs(m_cnPt2.x - m_cnPt1.x);
                        this.Height = Math.Abs(m_cnPt2.y - m_cnPt1.y);
                    }
                }

                if (mStartDrawing) // Simple Shape is ready to draw
                {
                    var pt = new UIPoint(m_cnPt2.x, m_cnPt2.y);

                    if (!mShapeHasBeenCreated)
                    {
                        mShapeHasBeenCreated = true;

                        var pathGeometry = new PathGeometry();
                        var pathFigureCollection = new PathFigureCollection();

                        pathFigureCollection.Add(mPathFigure);
                        pathGeometry.Figures = pathFigureCollection;

                        mPath = new Path();
                        mPath.Data = pathGeometry;

                        if (!mHasFirstPoint)
                        {
                            mHasFirstPoint = true;

                            mPathFigure.StartPoint = pt;
                            mPath.Fill = new SolidColorBrush(Colors.Red);    
                            mPath.Stroke = new SolidColorBrush(Colors.Red);
                            mPath.StrokeThickness = 2;
                            mPath.Opacity = DEFAULT_OPACITY;
                            mPath.StrokeDashArray = new DoubleCollection { 4, 2 };
                            mPath.StrokeDashOffset = 1.5;
                            mPath.StrokeEndLineCap = PenLineCap.Triangle;

                            mLine = new Windows.UI.Xaml.Shapes.Line();
                            mLine.Stroke = new SolidColorBrush(Colors.Red);
                            mLine.StrokeThickness = 1;
                            mLine.Opacity = DEFAULT_OPACITY;

                            mLine.X1 = m_cnPt2.x;
                            mLine.Y1 = m_cnPt2.y;
                        }

                        this.Children.Add(mPath);
                        this.Children.Add(mLine);
                    }

                    LineSegment segment = new LineSegment();
                    segment.Point = pt;
                    mPathFigure.Segments.Add(segment);

                    mLine.X2 = m_cnPt2.x;
                    mLine.Y2 = m_cnPt2.y;
                }
            }
            else
            {
                base.PointerMovedHandler(sender, e);
            }
            return true;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (!mCreated)
            {
                mSelectedAnnotations.Clear();

                // Remove contact from dictionary.
                RemovePointer(e.Pointer);

                if (mCancelledDrawing)
                    return false; // ignore any secondary pointer
                
                if (mPath == null)
                {
                    mPointerDown = false;
                    return false;
                }

                ProcessLassoFindAndAnnotSelection();

                // TOOD: this can stay OnRelease
                if (mSelectedRectInPage != null)
                {
                    if (mAnnot != null)
                    {
                        mAnnotPageNum = mAnnot.GetPage().GetIndex();
                        mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);

                        PopulateMenu();
                        InitializeCanvasAndGeometry();

                        mJustSwitchedFromAnotherTool = false;
                        CreateAppearance();

                        mCreated = true;

                        RemovePointer(e.Pointer);
                        mPointerDown = false;
                        mPointerID = -1;
                        mViewerCanvas?.ReleasePointerCaptures();

                        mPt1.x = m_cnPt1.x;
                        mPt2.y = m_cnPt2.y;
                    }
                }
                else
                {
                    RemovePointer(e.Pointer);
                    mPointerDown = false;
                    mPointerID = -1;
                    mViewerCanvas?.ReleasePointerCaptures();

                    mPt1.x = m_cnPt1.x;
                    mPt2.y = m_cnPt2.y;

                    BackToPan();
                }
            }

            this.Children.Remove(mPath);
            this.Children.Remove(mLine);

            // Cleanup            
            mPointerDown = false;
            mStartDrawing = false;
            mHasFirstPoint = false;

            EnableScrolling();

            base.PointerReleasedHandler(sender, e);
            return false;
        }
        #endregion

        #region Utilities
        
        private void ProcessLassoFindAndAnnotSelection()
        {
            var annot_count = mPDFView.GetDoc().GetPage(mDownPageNumber).GetNumAnnots();
            if (annot_count == 0)
                return;            

            var sharedCanvasDevice = CanvasDevice.GetSharedDevice();
            CanvasPathBuilder pathBuilder = new CanvasPathBuilder(sharedCanvasDevice);
            if (mPath.Data is PathGeometry pathGeo)
            {
                // Use CanvasGeometry to check if fill contains the annotation rectangle
                pathBuilder.BeginFigure((float)mPathFigure.StartPoint.X, (float)mPathFigure.StartPoint.Y);
                var segments = pathGeo.Figures[0].Segments;
                for (int i = 0; i < segments.Count; i++)
                {
                    var line = segments[i] as LineSegment;
                    pathBuilder.AddLine((float)line.Point.X, (float)line.Point.Y);
                }

                pathBuilder.EndFigure(CanvasFigureLoop.Closed);
                CanvasGeometry tempCanvasGeo = CanvasGeometry.CreatePath(pathBuilder);

                mAnnotIsTextMarkup = false;

                for (int i = 0; i < annot_count; i++)
                {
                    var annot = mPDFView.GetDoc().GetPage(mDownPageNumber).GetAnnot(i);

                    if (annot.IsValid() == false)
                        continue;

                    var annotType = annot.GetAnnotType();
                    
                    if (IsAnnotGroupSelectable(annotType) == false)
                        continue;

                    /* Note: Now we check the annotation bounds, convert to screen points and
                     * then check if the annot position intersects the lasso shape */
                    var markup = new Annots.Markup(annot);
                    var annot_rect = markup.GetContentRect();

                    PDFDouble x1 = new PDFDouble(annot_rect.x1);
                    PDFDouble x2 = new PDFDouble(annot_rect.x2);
                    PDFDouble y1 = new PDFDouble(annot_rect.y1);
                    PDFDouble y2 = new PDFDouble(annot_rect.y2);

                    // Get coordinates of two opposite points in screen space
                    mPDFView.ConvPagePtToScreenPt(x1, y1, mDownPageNumber);
                    mPDFView.ConvPagePtToScreenPt(x2, y2, mDownPageNumber);
                    mPDFView.ConvScreenPtToAnnotationCanvasPt(x1, y1);
                    mPDFView.ConvScreenPtToAnnotationCanvasPt(x2, y2);

                    var x1_pos = (float)x1.Value;
                    var x2_pos = (float)x2.Value;

                    var y1_pos = (float)y1.Value;
                    var y2_pos = (float)y2.Value;

                    /* Note: we need to create a win2d rectangle in order to check if
                     * there is an intersection with the lasso shape */
                    var annot_CreatedRectangle = CanvasGeometry.CreateRectangle(sharedCanvasDevice,
                        x1_pos, y1_pos,
                        Math.Abs(x1_pos - x2_pos),
                        -Math.Abs(y1_pos - y2_pos));

                    var bb = annot_CreatedRectangle.ComputeBounds();

                    var bb2 = tempCanvasGeo.ComputeBounds();

                    var combineResult = tempCanvasGeo.CombineWith(annot_CreatedRectangle,
                        Matrix3x2.CreateTranslation(Vector2.One),
                        CanvasGeometryCombine.Intersect);

                    var combineBounds = combineResult.ComputeBounds();
                    if (combineBounds.IsEmpty == false)
                    {
                        if (DoesAnnotationSupportResize(annot) == false)
                            mAnnotIsTextMarkup = true;

                        switch (annotType)
                        {
                            case AnnotType.e_Polygon:
                            case AnnotType.e_Polyline:
                                if (CheckPolyAnnotWithinSelection(annot, tempCanvasGeo))
                                {
                                    mSelectedAnnotations.Add(annot, mDownPageNumber);
                                }
                                continue;

                            case AnnotType.e_Circle:
                                var circle_radiusX = Math.Abs(x2_pos - x1_pos) / 2;
                                var circle_radiusY = Math.Abs(y2_pos - y1_pos) / 2;

                                var circle_centerX = circle_radiusX + x1_pos;
                                var circle_centerY = circle_radiusY + y2_pos;

                                var ellipse = CanvasGeometry.CreateEllipse(sharedCanvasDevice,
                                    circle_centerX,
                                    circle_centerY,
                                    circle_radiusX,
                                    circle_radiusY);

                                var circle_geo_result = tempCanvasGeo.CombineWith(ellipse, Matrix3x2.CreateTranslation(Vector2.Zero), CanvasGeometryCombine.Intersect);
                                var circle_bounds = circle_geo_result.ComputeBounds();

                                if (circle_bounds.IsEmpty == false)
                                {
                                    mSelectedAnnotations.Add(annot, mDownPageNumber);
                                }

                                continue;

                            case AnnotType.e_Ink:

                                var inkAnnot = annot as Annots.Ink;
                                if (inkAnnot == null)
                                    return;

                                InkStrokeBuilder inkStrokeBuilder = new InkStrokeBuilder();
                                
                                List<InkStroke> inkStrokes = new List<InkStroke>();
                                List<UIPoint> points = new List<UIPoint>();
                                
                                var inkPathCount = inkAnnot.GetPathCount();

                                for (int j = 0; j < inkPathCount; j++)
                                {
                                    var inkPtCount = inkAnnot.GetPointCount(j);
                                    for (int o = 0; o < inkPtCount; o++)
                                    {
                                        var inkPt = inkAnnot.GetPoint(j, o);

                                        PDFDouble ink_x = new PDFDouble(inkPt.x);
                                        PDFDouble ink_y = new PDFDouble(inkPt.y);

                                        // Get coordinates of two opposite points in screen space
                                        mPDFView.ConvPagePtToScreenPt(ink_x, ink_y, mDownPageNumber);
                                        mPDFView.ConvScreenPtToAnnotationCanvasPt(ink_x, ink_y);

                                        points.Add(new UIPoint(ink_x.Value, ink_y.Value));
                                    }

                                    inkStrokes.Add(inkStrokeBuilder.CreateStroke(points));
                                }
                                
                                var inkGeo = CanvasGeometry.CreateInk(sharedCanvasDevice, inkStrokes);

                                var ink_geo_result = tempCanvasGeo.CombineWith(inkGeo, Matrix3x2.CreateTranslation(Vector2.Zero), CanvasGeometryCombine.Intersect);
                                var ink_bounds = ink_geo_result.ComputeBounds();

                                if (ink_bounds.IsEmpty == false)
                                {
                                    System.Diagnostics.Debug.WriteLine("===> Ink Found");

                                    mSelectedAnnotations.Add(annot, mDownPageNumber);
                                }
                                else
                                {
                                    System.Diagnostics.Debug.WriteLine("===> Ink NOT Found");
                                }

                                break;

                            case AnnotType.e_Line:

                                var lineAnnot = annot as pdftron.PDF.Annots.Line;
                                if (lineAnnot == null)
                                    continue;

                                var startPt = lineAnnot.GetStartPoint();
                                var endPt = lineAnnot.GetEndPoint();

                                PDFDouble start_x = new PDFDouble(startPt.x);
                                PDFDouble start_y = new PDFDouble(startPt.y);

                                PDFDouble end_x = new PDFDouble(endPt.x);
                                PDFDouble end_y = new PDFDouble(endPt.y);

                                mPDFView.ConvPagePtToScreenPt(start_x, start_y, mDownPageNumber);
                                mPDFView.ConvScreenPtToAnnotationCanvasPt(start_x, start_y);

                                mPDFView.ConvPagePtToScreenPt(end_x, end_y, mDownPageNumber);
                                mPDFView.ConvScreenPtToAnnotationCanvasPt(end_x, end_y);                                                                                 

                                CanvasPathBuilder linePath = new CanvasPathBuilder(sharedCanvasDevice);
                                linePath.BeginFigure((float)start_x.Value, (float)start_y.Value);
                                linePath.AddLine((float)end_x.Value, (float)end_y.Value);
                                linePath.EndFigure(CanvasFigureLoop.Open);

                                var line = CanvasGeometry.CreatePath(linePath);
                                var line_geo_result = tempCanvasGeo.CompareWith(line);

                                if (line_geo_result == CanvasGeometryRelation.Overlap
                                    || line_geo_result == CanvasGeometryRelation.Contains)
                                {
                                    mSelectedAnnotations.Add(annot, mDownPageNumber);
                                }
                                break;

                            default:
                                /* Note: any annot with rectangle format can just be added right 
                                 * away with no extra check and processing */
                                mSelectedAnnotations.Add(annot, mDownPageNumber);
                                break;
                        }
                    }
                }

                FindAndAddExistingGroupAnnotations();

                PDFRect selectedRectInPage = null;

                // Go over all selected annotations and draw a big group rectangle
                foreach (KeyValuePair<IAnnot, int> annot in mSelectedAnnotations)
                {
                    PDFRect annotRect = mPDFView.GetPageRectForAnnot(annot.Key, mDownPageNumber);
                    annotRect.Normalize();

                    if (selectedRectInPage == null)
                    {
                        selectedRectInPage = new PDFRect(annotRect.x1, annotRect.y1, annotRect.x2, annotRect.y2);
                    }
                    else
                    {
                        if (annotRect.x1 < selectedRectInPage.x1)
                        {
                            selectedRectInPage.x1 = annotRect.x1;
                        }

                        if (annotRect.y1 < selectedRectInPage.y1)
                        {
                            selectedRectInPage.y1 = annotRect.y1;
                        }

                        if (annotRect.x2 > selectedRectInPage.x2)
                        {
                            selectedRectInPage.x2 = annotRect.x2;
                        }

                        if (annotRect.y2 > selectedRectInPage.y2)
                        {
                            selectedRectInPage.y2 = annotRect.y2;
                        }
                    }
                }

                mGroupAnnotations = mSelectedAnnotations.Keys.ToList();

                mSelectedRectInPage = selectedRectInPage;
                SetMAnnot();
            }
        }

        /// <summary>
        /// Check if annot is supported to be selected in a group
        /// </summary>
        private bool IsAnnotGroupSelectable(AnnotType annot)
        {
            switch (annot)
            {
                case AnnotType.e_Link:
                case AnnotType.e_Widget:
                    return false;
            }

            return true;
        }

        private bool CheckPolyAnnotWithinSelection(IAnnot annot, CanvasGeometry canvasGeometry)
        {
            var poly = annot as IPolyLine;
            if (poly == null)
                return false;

            var vertexPoints = poly.GetVertexCount();

            Vector2[] vPoints = new Vector2[vertexPoints];

            for (int j = 0; j < vertexPoints; j++)
            {
                var vPoint = poly.GetVertex(j);

                PDFDouble v_x = new PDFDouble(vPoint.x);
                PDFDouble v_y = new PDFDouble(vPoint.y);

                // Get coordinates of two opposite points in screen space
                mPDFView.ConvPagePtToScreenPt(v_x, v_y, mAnnotPageNum);
                mPDFView.ConvScreenPtToAnnotationCanvasPt(v_x, v_y);

                vPoints[j].X = (float)v_x.Value;
                vPoints[j].Y = (float)v_y.Value;
            }

            var polygon = CanvasGeometry.CreatePolygon(CanvasDevice.GetSharedDevice(), vPoints);
            var poly_geo_result = canvasGeometry.CombineWith(polygon, Matrix3x2.CreateTranslation(Vector2.Zero), CanvasGeometryCombine.Intersect);
            var poly_bounds = poly_geo_result.ComputeBounds();

            if (poly_bounds.IsEmpty == false)
            {
                // select polygon
                System.Diagnostics.Debug.WriteLine("======> Poly Found! ");                
                return true;
            }
            else
            {
                System.Diagnostics.Debug.WriteLine("======> Poly NOT Found! ");                
            }

            return false;
        }

        private bool CheckCircleAnnotWithinSelection(IAnnot annot, CanvasGeometry canvasGeometry)
        {
            var circle = annot as Annots.Circle;
            if (circle == null)
                return false;

            

            return false;
        }

        private void CreateAnnotationGroup()
        {
            // first find primary annot
            // here we assume the last one created is the primary
            List<IAnnot> annotations = new List<IAnnot>(mSelectedAnnotations.Keys.ToList());
            if ((annotations == null) || (annotations.Count == 0))
            {
                return;
            }

            IAnnot primaryAnnotation = annotations[annotations.Count - 1];

            try
            {
                UtilityFunctions.CreateAnnotationGroup(mPDFView, primaryAnnotation, annotations);
                mToolManager.RaiseAnnotationGroupAddedEvent(mSelectedAnnotations);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            BackToPan();
        }

        protected override PDFRect FindSelectedAnnotationsAndRect()
        {
            PDFRect pageRect = GetShapeBoundingBox();
            mSelectedAnnotations.Clear();
            List<IAnnot> annotsInPage = UtilityFunctions.GetAnnotationsOnPage(mPDFView, mDownPageNumber);
            PDFRect selectedRectInPage = null;

            foreach (IAnnot annot in annotsInPage)
            {
                if (mAnnot == null)
                    mAnnot = annot;

                PDFRect annotRect = mPDFView.GetPageRectForAnnot(annot, mDownPageNumber);
                annotRect.Normalize();

                PDFRect dummyPageRect = new PDFRect(pageRect.x1, pageRect.y1, pageRect.x2, pageRect.y2);

                if (dummyPageRect.IntersectRect(dummyPageRect, annotRect))
                {
                    if (DoesAnnotationSupportResize(annot) == false)
                        mAnnotIsTextMarkup = true;

                    //mSelectedAnnotations.Add(annot, mDownPageNumber);
                }
            }

            // Go over all selected annotations and draw a big group rectangle
            foreach (KeyValuePair<IAnnot, int> annot in mSelectedAnnotations)
            {
                PDFRect annotRect = mPDFView.GetPageRectForAnnot(annot.Key, mDownPageNumber);
                annotRect.Normalize();

                if (selectedRectInPage == null)
                {
                    selectedRectInPage = new PDFRect(annotRect.x1, annotRect.y1, annotRect.x2, annotRect.y2);
                }
                else
                {
                    if (annotRect.x1 < selectedRectInPage.x1)
                    {
                        selectedRectInPage.x1 = annotRect.x1;
                    }

                    if (annotRect.y1 < selectedRectInPage.y1)
                    {
                        selectedRectInPage.y1 = annotRect.y1;
                    }

                    if (annotRect.x2 > selectedRectInPage.x2)
                    {
                        selectedRectInPage.x2 = annotRect.x2;
                    }

                    if (annotRect.y2 > selectedRectInPage.y2)
                    {
                        selectedRectInPage.y2 = annotRect.y2;
                    }
                }
            }
            mGroupAnnotations = mSelectedAnnotations.Keys.ToList();
            SetMAnnot();
            return selectedRectInPage;
        }

        private PDFRect GetShapeBoundingBox()
        {
            if (mPath.Data is PathGeometry pathGeo)
            {

            }

            // Computes the bounding box of the rubber band in page space.
            PDFDouble x1 = new PDFDouble(m_cnPt1.x);
            PDFDouble y1 = new PDFDouble(m_cnPt1.y);
            PDFDouble x2 = new PDFDouble(m_cnPt2.x);
            PDFDouble y2 = new PDFDouble(m_cnPt2.y);

            mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
            mPDFView.ConvScreenPtToPagePt(x1, y1, mDownPageNumber);
            mPDFView.ConvScreenPtToPagePt(x2, y2, mDownPageNumber);

            PDFRect rect;
            try
            {
                rect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);
                rect.Normalize();
                return rect;
            }
            catch (Exception)
            {
                return null;
            }
        }

        private void FindAndAddExistingGroupAnnotations()
        {
            //find all annotations part of a group
            Dictionary<string, IAnnot> annotationsInGroup = new Dictionary<string, IAnnot>();
            foreach (KeyValuePair<IAnnot, int> annotation in mSelectedAnnotations)
            {
                string text = "";
                try
                {
                    var id = annotation.Key.GetUniqueID();
                    if (id != null)
                    {
                        text = id.GetAsPDFText();
                    }
                    else
                    {
                        mIsUngroupAllowed = false;
                        continue;
                    }
                }
                catch (NullReferenceException)
                {
                    mIsUngroupAllowed = false;
                    continue;
                }

                if (!annotationsInGroup.ContainsKey(annotation.Key.GetUniqueID().GetAsPDFText()))
                {
                    bool isRtReply = false;
                    try
                    {
                        isRtReply = UtilityFunctions.IsReplyTypeReply(annotation.Key);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }

                    if ((mAnnot != null) && (!isRtReply))
                    {
                        KeyValuePair<ToolType, List<IAnnot>> pair = CanSelectGroupAnnotation(mPDFView, annotation.Key, annotation.Value);

                        if (pair.Value != null)
                        {
                            foreach (IAnnot item in pair.Value)
                            {
                                if (!annotationsInGroup.ContainsKey(item.GetUniqueID().GetAsPDFText()))
                                {
                                    annotationsInGroup.Add(item.GetUniqueID().GetAsPDFText(), item);
                                }
                            }
                        }
                    }
                }
            }

            //find if there is any left that is not part of a group
            foreach (KeyValuePair<IAnnot, int> annotation in mSelectedAnnotations)
            {
                string text = "";
                try
                {
                    var id = annotation.Key.GetUniqueID();
                    if (id != null)
                    {
                        text = id.GetAsPDFText();
                    }
                    else
                    {
                        mIsUngroupAllowed = false;
                        continue;
                    }
                }
                catch (NullReferenceException)
                {
                    mIsUngroupAllowed = false;
                    continue;
                }

                if (!annotationsInGroup.ContainsKey(annotation.Key.GetUniqueID().GetAsPDFText()))
                {
                    mIsUngroupAllowed = false;
                    break;
                }
            }

            //add missing annotations from groups to mSelectedAnnotations
            foreach (KeyValuePair<string, IAnnot> item in annotationsInGroup)
            {
                bool found = false;

                foreach (KeyValuePair<IAnnot, int> selectedItem in mSelectedAnnotations)
                {
                    if (selectedItem.Key.GetUniqueID().GetAsPDFText() == item.Key)
                    {
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    mSelectedAnnotations.Add(item.Value, item.Value.GetPage().GetIndex());
                }
            }
        }
        #endregion

        #region Handlers For Popup Menu

        private bool mIsUngroupAllowed = true;

        /// <summary>
        /// Adds in the menu options according to the type of annotation
        /// </summary>
        protected override void PopulateMenu()
        {
            if (mAnnot != null)
            {
                mMenuTitles = null;
                try
                {
                    mMenuTitles = new Dictionary<string, string>();
                    //locks the document first as accessing annotation/doc information isn't thread safe.
                    mPDFView.DocLockRead();
                    mMenuTitles.Add("done", ResourceHandler.GetString("EditOption_Done"));
                    mMenuTitles.Add("group", ResourceHandler.GetString("EditOption_Group"));

                    if (mIsUngroupAllowed)
                    {
                        mMenuTitles.Add("ungroup", ResourceHandler.GetString("EditOption_Ungroup"));
                    }

                    mMenuTitles.Add("flatten", ResourceHandler.GetString("EditOption_Flatten"));
                    mMenuTitles.Add("opacity", ResourceHandler.GetString("EditOption_Opacity"));
                    mMenuTitles.Add("fill color", ResourceHandler.GetString("EditOption_FillColor"));
                    mMenuTitles.Add("color", ResourceHandler.GetString("EditOption_Color"));
                    mMenuTitles.Add("delete", ResourceHandler.GetString("EditOption_Delete"));

                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlockRead();
                }

                if (mMenuTitles != null && !mAnnotIsSticky) // if sticky, we don't create menu until it's needed
                {
                    mContextMenu = new Controls.ContextMenuControl(mPDFView, mMenuTitles);
                    mContextMenu.ItemClicked += OnCommandMenuClicked;
                }
            }
        }

        internal override void OnCommandMenuClicked(string title)
        {
            if (mAnnot == null)
            {
                SetMAnnot();
            }

            mIsShowingCommandMenu = false;
            HideContextMenu(mContextMenu);

            if (title.Equals("done", StringComparison.OrdinalIgnoreCase))
            {
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("group", StringComparison.OrdinalIgnoreCase))
            {
                CreateAnnotationGroup();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("ungroup", StringComparison.OrdinalIgnoreCase))
            {
                UngroupAnnotations();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("delete", StringComparison.OrdinalIgnoreCase))
            {
                DeleteGroupAnnotations();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("fill color", StringComparison.OrdinalIgnoreCase))
            {
                mIsFillColor = true;
                if (mAnnot.GetAnnotType() == AnnotType.e_FreeText)
                {
                    mEffectiveBrush = e_text_fill;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_fill;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("opacity", StringComparison.OrdinalIgnoreCase))
            {
                HandleOpacity();
            }
            else if (title.Equals("color", StringComparison.OrdinalIgnoreCase))
            {
                mIsFillColor = false;
                AnnotType aType = mAnnot.GetAnnotType();
                if (aType == AnnotType.e_Line || aType == AnnotType.e_Ink || aType == AnnotType.e_Polyline)
                {
                    mEffectiveBrush = e_stroke;
                    ShowColorPicker();
                }
                else if (aType == AnnotType.e_Highlight)
                {
                    mEffectiveBrush = e_highlight;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_text_markup;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("flatten", StringComparison.OrdinalIgnoreCase))
            {
                HandleFlatten();
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }
        }

        #endregion Handlers For Popup Menu

        #region Overrided Events
        internal override bool PointerCaptureLostHandler(object sender, PointerRoutedEventArgs e)
        {
            mPointerID = -1;
            return true;
        }

        internal override bool HoldingHandler(object sender, HoldingRoutedEventArgs e)
        {
            return true;
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)
        {
            return false;
        }

        internal override bool OnPageNumberChanged(int current_page, int num_pages)
        {
            return false;
        }

        internal override bool PointerCanceledHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            return true;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            return false;
        }

        internal override void OnCreate()
        {
            //Note: Do nothing - Ignore to avoid parent/inherited logic
        }
        #endregion
    }
}
