﻿using System;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Windows.Storage;
using Windows.Storage.FileProperties;
using Windows.UI.Xaml;

namespace pdftron.PDF.Tools.Utilities
{
    internal class AutoSaveHelper
    {
        private PDFDoc _pdfDoc;
        private TimeSpan _interval;
        private DispatcherTimer _saveTimer;
        private bool _hasSavedSinceGettingDoc;

        private DocumentWasSavedHandler _documentWasSaved;
        public delegate void DocumentWasSavedHandler(DateTimeOffset newModifiedDate, ulong newFilesize);
        public event DocumentWasSavedHandler DocumentWasSaved
        {
            add
            {
                if (_documentWasSaved?.GetInvocationList().Contains(value) != true)
                    _documentWasSaved += value;
            }
            remove => _documentWasSaved -= value;
        }

        private CommitChangesAsyncDelegate _commitChangesAsync;
        public delegate Task CommitChangesAsyncDelegate();
        public event CommitChangesAsyncDelegate CommitChangesAsync
        {
            add
            {
                if (_commitChangesAsync?.GetInvocationList().Contains(value) != true)
                    _commitChangesAsync += value;
            }
            remove => _commitChangesAsync -= value;
        }

        /// <summary>
        /// Sets the document to save to
        /// </summary>
        public PDFDoc PDFDoc
        {
            set
            {
                _hasSavedSinceGettingDoc = false;
                _pdfDoc = value;
                Stop();
                Start();
            }
            get => _pdfDoc;
        }

        /// <summary>
        /// Sets the interval in seconds between auto-saves.
        /// </summary>
        public double Interval
        {
            set
            {
                _interval = TimeSpan.FromSeconds(value);
                if (_saveTimer != null)
                    _saveTimer.Interval = _interval;
            }
        }

        public bool IsPaused { get; set; }

        /// <summary>
        /// This variable lets you know if the auto-saver has saved the document, in case you want to display a notification.
        /// </summary>
        public bool HasSavedSinceGettingDoc => _hasSavedSinceGettingDoc;

        /// <summary>
        /// The current file.
        /// </summary>
        public StorageFile CurrentFile { get; set; }

        /// <summary>
        /// The current temporary file.
        /// </summary>
        public StorageFile TemporaryFile { get; set; }

        /// <summary>
        /// Creates an AutoSaveHelper that will save every 30 seconds.
        /// </summary>
        public AutoSaveHelper() => Init(30);

        /// <summary>
        /// Creates an AutoSaveHelper that will save every interval seconds.
        /// </summary>
        public AutoSaveHelper(double intervalInSeconds) => Init(intervalInSeconds);

        /// <summary>
        /// Stops the AutoSaveHelper
        /// </summary>
        public void Stop() => _saveTimer?.Stop();

        /// <summary>
        /// Starts the AutoSaveHelper
        /// </summary>
        public void Start() => _saveTimer?.Start();

        private void Init(double interval)
        {
            _saveTimer = new DispatcherTimer();
            _saveTimer.Tick += SaveTimer_Tick;
            _interval = TimeSpan.FromSeconds(interval);
            _saveTimer.Interval = _interval;
            IsPaused = false;
        }

        private async void SaveTimer_Tick(object sender, object e)
        {
            if (!IsPaused)
                await SaveAsync();
        }

        private async Task SaveAsync()
        {
            if (_pdfDoc != null && CurrentFile != null && TemporaryFile != null)
            {
                bool locked = false;
                try
                {
                    locked = _pdfDoc.TryLock();
                    if (locked && _pdfDoc.IsModified())
                    {
                        _pdfDoc.Unlock();
                        locked = false;
                        DateTime preSave = DateTime.Now;

                        Task commitTask = _commitChangesAsync?.Invoke();
                        if (commitTask != null)
                            await commitTask;

                        await _pdfDoc.SaveAsync(SDF.SDFDocSaveOptions.e_incremental);
                        if (!DocumentManager.IsReady)
                            await DocumentManager.Instance.InitializeInstanceAsync();

                        await DocumentManager.Instance.AddChangesToOriginalAsync(CurrentFile, TemporaryFile);
                        BasicProperties props = await TemporaryFile.GetBasicPropertiesAsync();

                        TimeSpan saveTime = DateTime.Now.Subtract(preSave);
                        _hasSavedSinceGettingDoc = true;
                        _documentWasSaved?.Invoke(props.DateModified, props.Size);
                    }
                }
                catch (Exception ex)
                {
                    Common.PDFNetException pdfnetE = new Common.PDFNetException(ex.HResult);
                    if (pdfnetE.IsPDFNetException)
                    {
                        Debug.WriteLine("Error: " + pdfnetE.ToString());
                    }
                    else
                    {
                        Debug.WriteLine("Error: " + ex.ToString());
                    }
                }
                finally
                {
                    if (locked)
                        _pdfDoc.Unlock();
                }

            }
        }
    }
}
