﻿using pdftron.SDF;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools.Utilities
{
    /// <summary>
    /// Measure class with all PDF measure specs
    /// </summary>
    public class MeasureUtils
    {
        #region Const and Default values
        public const string U_PT = "pt";
        public const string U_IN = "in";
        public const string U_MM = "mm";
        public const string U_CM = "cm";
        public const string U_M = "m";
        public const string U_KM = "km";
        public const string U_FT = "ft";
        public const string U_YD = "yd";
        public const string U_MI = "mi";

        public static string DEFAULT_PRECISION = "1";
        public static string DEFAULT_BASE_UNIT = "pt";
        public static string DEFAULT_BASE_SCALE = "1.0";
        public static string DEFAULT_TRANSLATE_UNIT = "pt";
        public static string DEFAULT_TRANSLATE_SCALE = "1.0";
        #endregion

        #region PDF Specs and Utilities

        // Keys
        public const string K_Measure = "Measure";
        public const string K_IT = "IT";

        // Names
        public const string K_LineDimension = "LineDimension";
        public const string K_PolyLineDimension = "PolyLineDimension";
        public const string K_PolygonDimension = "PolygonDimension";

        // Dict key types
        public const string KEY_TYPE = "Type";
        public const string KEY_SCALE = "R";
        public const string KEY_AXIS = "X";
        public const string KEY_DISTANCE = "D";
        public const string KEY_AREA = "A";
        public const string KEY_UNIT = "U";
        public const string KEY_FACTOR = "C";
        public const string KEY_DECIMAL_SYMBOL = "RD";
        public const string KEY_THOUSAND_SYMBOL = "RT";
        public const string KEY_PRECISION = "D";
        public const string KEY_DISPLAY = "F";
        public const string KEY_UNIT_PREFIX = "PS";
        public const string KEY_UNIT_SUFFIX = "SS";
        public const string KEY_UNIT_POSITION = "O";

        /// <summary>
        /// Add Measure information accordingly to the PDF specs
        /// </summary>
        /// <param name="annot">The annotation to get measure data</param>
        /// <param name="measureInfo">The measure data</param>
        public static void SetMeasurementInfo(IAnnot annot, IMeasure measureInfo)
        {
            if (annot == null || !annot.IsValid())
                return;

            var annotType = annot.GetAnnotType();
            if (annotType != AnnotType.e_Line  && annotType != AnnotType.e_Polyline && annotType != AnnotType.e_Polygon)
                return;

            // Get SDF obj so it can get measure info added
            var obj = annot.GetSDFObj();

            // IT
            string it = K_LineDimension;
            if (annotType == AnnotType.e_Polyline)
                it = K_PolyLineDimension;
            else if (annotType == AnnotType.e_Polygon)
                it = K_PolygonDimension;

            // Add General Measure Data
            obj.PutName(K_IT, it);
            Obj measureObj = obj.PutDict(K_Measure);
            measureObj.PutName(KEY_TYPE, K_Measure);
            measureObj.PutString(KEY_SCALE, measureInfo.Scale);

            // Create Measure Types Obj
            Obj axisArray = measureObj.PutArray(KEY_AXIS);
            Obj distanceArray = measureObj.PutArray(KEY_DISTANCE);
            Obj areaArray = measureObj.PutArray(KEY_AREA);

            Obj axis = axisArray.PushBackDict();
            Obj distance = distanceArray.PushBackDict();
            Obj area = areaArray.PushBackDict();

            if (annotType == AnnotType.e_Line)
            {
                LoadMeasureInfoToObj(axis, new MeasureInfo 
                { 
                    BaseFactor = measureInfo.BaseFactor,
                    Unit = measureInfo.BaseUnit
                });
                LoadMeasureInfoToObj(distance, new MeasureInfo 
                {
                    BaseFactor = measureInfo.TranslateFactor,
                    Precision = measureInfo.Precision,
                    Unit = measureInfo.Unit

                });
                LoadMeasureInfoToObj(area, new MeasureInfo { Unit = "sq in" });
            }
            else if (annotType == AnnotType.e_Polyline)
            {
                LoadMeasureInfoToObj(distance, new MeasureInfo
                {
                    BaseFactor = measureInfo.TranslateFactor,
                    Precision = measureInfo.Precision,
                    Unit = measureInfo.Unit
                });
                LoadMeasureInfoToObj(axis, new MeasureInfo
                {
                    BaseFactor = measureInfo.BaseFactor,
                    Unit = measureInfo.BaseUnit
                });
                LoadMeasureInfoToObj(area, new MeasureInfo { Unit = "sq in" });
            }
            else if (annotType == AnnotType.e_Polygon)
            {                
                LoadMeasureInfoToObj(axis, new MeasureInfo
                {
                    BaseFactor = measureInfo.TranslateFactor,
                    Precision = measureInfo.Precision,
                    Unit = measureInfo.Unit
                });
                LoadMeasureInfoToObj(distance, new MeasureInfo 
                {
                    Precision = measureInfo.Precision,
                    Unit = measureInfo.Unit
                });

                measureInfo.Unit = "sq " + measureInfo.Unit;
                LoadMeasureInfoToObj(area, measureInfo);
            }
        }

        private static void LoadMeasureInfoToObj(Obj obj, IMeasure measureInfo)
        {
            obj.PutNumber(KEY_FACTOR, measureInfo.BaseFactor);
            obj.PutNumber(KEY_PRECISION, measureInfo.Precision);
            obj.PutName(KEY_DISPLAY, measureInfo.Display);
            obj.PutString(KEY_DECIMAL_SYMBOL, measureInfo.DecimalSymbol);
            obj.PutString(KEY_THOUSAND_SYMBOL, measureInfo.ThousandSymbol);
            obj.PutString(KEY_UNIT_SUFFIX, measureInfo.UnitSuffix);
            obj.PutString(KEY_UNIT, measureInfo.Unit);
            obj.PutString(KEY_UNIT_PREFIX, measureInfo.UnitPrefix);
            obj.PutName(KEY_UNIT_POSITION, measureInfo.UnitPosition);
        }

        public static MeasureInfoGroup GetAnnotMeasureInfo(IAnnot annot)
        {
            if (annot == null || !annot.IsValid())
                return null;

            var annotType = annot.GetAnnotType();
            if ((annotType != AnnotType.e_Line)
                && (annotType != AnnotType.e_Polygon)
                && (annotType != AnnotType.e_Polyline))
            {
                return null;
            }

            MeasureInfoGroup measureInfoGroup = null;

            var obj = annot.GetSDFObj();

            Obj measureObj = obj.FindObj(K_Measure);
            if (measureObj != null)
            {
                measureInfoGroup = new MeasureInfoGroup();

                Obj scaleObj = measureObj.FindObj(KEY_SCALE);
                if (scaleObj != null && scaleObj.IsString())
                {
                    measureInfoGroup.Scale = scaleObj.GetAsPDFText();
                }                

                // Axis
                Obj axisArray = measureObj.FindObj(KEY_AXIS);
                measureInfoGroup.AxisInfo = GetMeasureIngo(axisArray);

                // Distance
                Obj distanceArray = measureObj.FindObj(KEY_DISTANCE);
                measureInfoGroup.DistanceInfo = GetMeasureIngo(distanceArray);

                // Area
                Obj areaArray = measureObj.FindObj(KEY_AREA);
                measureInfoGroup.AreaInfo = GetMeasureIngo(areaArray);
            }

            return measureInfoGroup;
        }

        private static MeasureInfo GetMeasureIngo(Obj obj)
        {
            MeasureInfo measureInfo = new MeasureInfo();

            if (obj != null && obj.IsArray() && obj.Size() > 0)
            {
                Obj axis = obj.GetAt(0);
                if (axis != null && axis.IsDict())
                {
                    Obj factor = axis.FindObj(KEY_FACTOR);
                    if (factor != null && factor.IsNumber())
                    {
                        measureInfo.BaseFactor = factor.GetNumber();
                    }
                    Obj precision = axis.FindObj(KEY_PRECISION);
                    if (precision != null && precision.IsNumber())
                    {
                        measureInfo.Precision = (int)precision.GetNumber();
                    }
                    Obj display = axis.FindObj(KEY_DISPLAY);
                    if (display != null && display.IsName())
                    {
                        measureInfo.Display = display.GetName();
                    }
                    Obj decimalSymbol = axis.FindObj(KEY_DECIMAL_SYMBOL);
                    if (decimalSymbol != null && decimalSymbol.IsString())
                    {
                        measureInfo.DecimalSymbol = decimalSymbol.GetAsPDFText();
                    }
                    Obj thousandSymbol = axis.FindObj(KEY_THOUSAND_SYMBOL);
                    if (thousandSymbol != null && thousandSymbol.IsString())
                    {
                        measureInfo.ThousandSymbol = thousandSymbol.GetAsPDFText();
                    }
                    Obj unitSuffix = axis.FindObj(KEY_UNIT_SUFFIX);
                    if (unitSuffix != null && unitSuffix.IsString())
                    {
                        measureInfo.UnitSuffix = unitSuffix.GetAsPDFText();
                    }
                    Obj unit = axis.FindObj(KEY_UNIT);
                    if (unit != null && unit.IsString())
                    {
                        measureInfo.Unit = unit.GetAsPDFText();
                    }
                    Obj unitPrefix = axis.FindObj(KEY_PRECISION);
                    if (unitPrefix != null && unitPrefix.IsString())
                    {
                        measureInfo.UnitPrefix = unitPrefix.GetAsPDFText();
                    }
                    Obj unitPosition = axis.FindObj(KEY_UNIT_POSITION);
                    if (unitPosition != null && unitPosition.IsName())
                    {
                        measureInfo.UnitPosition = unitPosition.GetName();
                    }
                }
            }

            return measureInfo;
        }

        /// <summary>
        /// Check if annotation has Measure key set 
        /// Only valid for Line, Polygon and Polyline annotations
        /// </summary>
        /// <param name="annot">Annotation</param>
        /// <returns>True if annot has Measure key set</returns>
        public static bool IsMeasureAnnot(IAnnot annot)
        {
            if (annot == null || !annot.IsValid())
                return false;

            var annotType = annot.GetAnnotType();
            if ((annotType != AnnotType.e_Line)
                && (annotType != AnnotType.e_Polygon)
                && (annotType != AnnotType.e_Polyline))
            {
                return false;
            }

            var obj = annot.GetSDFObj();
            Obj measureObj = obj.FindObj(K_Measure);
            if (measureObj == null)
                return false;
            else
                return true;
        }

        /// <summary>
        /// Check if annotations is Ruler Measure type
        /// </summary>
        /// <param name="annot">Annotation</param>
        /// <returns>True or False</returns>
        public static bool IsRulerMeasure(IAnnot annot)
        {
            var itField = GetIT(annot);
            return itField != null && itField.Equals(K_LineDimension); 
        }

        /// <summary>
        /// Check if annotations is Area Measure type
        /// </summary>
        /// <param name="annot">Annotation</param>
        /// <returns>True or False</returns>
        public static bool IsAreaMeasure(IAnnot annot)
        {
            var itField = GetIT(annot);
            return itField != null && itField.Equals(K_PolygonDimension);
        }

        /// <summary>
        /// Check if annotations is Perimeter Measure type
        /// </summary>
        /// <param name="annot">Annotation</param>
        /// <returns>True or False</returns>
        public static bool IsPerimeterMeasure(IAnnot annot)
        {
            var itField = GetIT(annot);
            return itField != null && itField.Equals(K_PolyLineDimension);
        }

        #endregion

        #region Utilities

        /// <summary>
        /// Get the IT -> Measure Annotation Name
        /// </summary>
        /// <param name="annot">Annotation</param>
        /// <returns>Annotation name</returns>
        public static string GetIT(IAnnot annot)
        {
            if (annot == null || !annot.IsValid())
                return null;

            Obj obj = annot.GetSDFObj();
            Obj itObj = obj.FindObj(K_IT);
            if (itObj != null && itObj.IsName())
                return itObj.GetName();
            
            return null;
        }

        /// <summary>
        /// Convert the page points to a final value and adds the unit
        /// </summary>
        /// <param name="actual_pg_pts">Page pts value to be converted</param>
        /// <param name="baseScale">The base scale</param>
        /// <param name="baseUnit">The base unit (in, cm, mm, etc)</param>
        /// <param name="translateScale">The translated scale</param>
        /// <param name="translateUnit">The translated unit (in, cm, mm, etc)</param>
        /// <param name="precision">Decimals to display</param>
        /// <returns></returns>
        public static string ConvertToFinalVal(double actual_pg_pts, string baseScale, string baseUnit, string translateScale, string translateUnit, string precision)
        {
            string result = string.Empty;
            double base_scale_factor;
            double trans_scale_factor;

            if (double.TryParse(baseScale, out base_scale_factor) == false)
                return result;

            if (double.TryParse(translateScale, out trans_scale_factor) == false)
                return result;

            // Proper convert base value considering scale and unit type
            var base_conv_val = ConvertPagePtsTo(GetPageUnitFromString(baseUnit), actual_pg_pts, base_scale_factor);
            // Translate base converted val to translated one considering the factor
            var trans_conv_val = base_conv_val * trans_scale_factor;
            // Proper format value using precision
            result = FormatDecimals(trans_conv_val, precision);
            // Add translated unit
            result += " " + translateUnit;
            return result;
        }

        /// <summary>
        /// 
        /// NOTE: Page Points by default is 1/72 inch
        /// https://www.pdftron.com/documentation/uwp/guides/features/viewer/coordinates
        /// </summary>
        /// <param name="pageUnit"></param>
        /// <param name="actual_pg_pts"></param>
        /// <returns></returns>
        private static double ConvertPagePtsTo(PageUnit pageUnit, double actual_pg_pts, double factor)
        {
            double val_result = actual_pg_pts;
            var val_inches = actual_pg_pts / 72;

            switch (pageUnit)
            {
                case PageUnit.cm:
                    val_result = (val_inches * 2.54) / factor; // 1 inch -> 2.54cm
                    break;

                case PageUnit.mm:
                    val_result = (val_inches * 25.4) / factor; // 1 inch -> 25.4 mm
                    break;

                case PageUnit.pt:
                    return (actual_pg_pts / factor);

                case PageUnit.inch:
                    val_result = val_inches / factor;
                    break;
            }

            return val_result;
        }

        /// <summary>
        /// Convert scales to PDF spec Factor
        /// </summary>
        /// <param name="pgUnit">The measurement unit (in, cm, mm, etc)</param>
        /// <param name="toolType">Measure tool type</param>
        /// <param name="bScale">The base scale</param>
        /// <param name="tScale">The translated scale</param>
        /// <returns></returns>
        public static double ConvScaleToPDFFactor(PageUnit pgUnit, ToolType toolType, string bScale, string tScale)
        {
            double result = 0.01388888; // pt: 1/72 in
            
            double dBScale;
            double.TryParse(bScale, out dBScale);

            double dTScale;
            double.TryParse(tScale, out dTScale);

            switch (pgUnit)
            {
                case PageUnit.pt:
                    if (toolType == ToolType.e_ruler_create)
                        result = 1 / dBScale;
                    else
                        result = 1 / dBScale * dTScale;
                    break;

                case PageUnit.inch:
                    if (toolType == ToolType.e_ruler_create)
                        result = (result / dBScale);
                    else
                        result = (result / dBScale) * dTScale;
                    break;

                case PageUnit.cm:
                    if (toolType == ToolType.e_ruler_create)
                        result = (result * 2.54) / dBScale;
                    else
                        result = ((result * 2.54) / dBScale) * dTScale;
                    break;

                case PageUnit.mm:
                    if (toolType == ToolType.e_ruler_create)
                        result = (result * 25.4) / dBScale;
                    else
                        result = ((result * 25.4) / dBScale) * dTScale;
                    break;
            }

            return result;
        }

        /// <summary>
        /// Convert string precision from the UI to PDF specs precision
        /// IE.: 0.01 => 100, 0.001 => 1000
        /// </summary>
        /// <param name="precision"></param>
        /// <returns></returns>
        public static double ConvStrToPDFPrecision(string precision)
        {
            int decimals = precision.Substring(precision.IndexOf(".") + 1).Length;
            switch (decimals)
            {
                default:
                case 0:
                    return 1;
                case 1:
                    return 10;
                case 2:
                    return 100;
                case 3:
                    return 1000;
                case 4:
                    return 10000;
            }
        }

        /// <summary>
        /// Convert PDF spec precision to UI format
        /// I.e.: 100 => 0.01
        /// </summary>
        /// <param name="precision"></param>
        /// <returns></returns>
        public static string ConvPDFPrecisionToStr(double precision)
        {
            int div = precision.ToString().Length;
            switch (div)
            {
                case 1:
                    return "1";
                case 2:
                    return "0.1";
                case 3:
                    return "0.01";
                case 4:
                    return "0.001";
                case 5:
                    return "0.0001";
            }
            return "1";
        }

        /// <summary>
        /// Return PageUnit enum based on string unit value
        /// </summary>
        /// <param name="val">Unit: pt, in, mm, cm, m, km, ft, yd, mi</param>
        /// <returns>PageUnit value</returns>
        public static PageUnit GetPageUnitFromString(string val)
        {
            switch(val)
            {
                case U_PT:
                    return PageUnit.pt;

                case U_IN:
                    return PageUnit.inch;

                case U_MM:
                    return PageUnit.mm;

                case U_CM:
                    return PageUnit.cm;

                case U_M:
                    return PageUnit.m;

                case U_KM:
                    return PageUnit.km;

                case U_FT:
                    return PageUnit.ft;

                case U_YD:
                    return PageUnit.yd;

                case U_MI:
                    return PageUnit.mi;
            }

            return PageUnit.pt;
        }

        /// <summary>
        /// Utility method to format string decimal
        /// </summary>
        /// <param name="val">Value</param>
        /// <param name="format">String format</param>
        /// <returns></returns>
        public static string FormatDecimals(double val, string format)
        {
            string result_val = string.Empty;
            string str_format = string.Empty;

            for (int i = 0; i < format.Length; i++)
            {
                if (format[i] == '.')
                    str_format += format[i];
                else
                    str_format += '#';
            }

            result_val = val.ToString(str_format);
            return result_val;
        }

        /// <summary>
        /// Calculate and return the total perimeter of all points
        /// </summary>
        /// <param name="points">Points to get the perimeter</param>
        /// <returns>The total perimeter</returns>
        public static double CalcPerimeter(List<Point> points)
        {
            double perimeter = 0;
            Point prevPoint = null;
            foreach (Point point in points)
            {
                if (prevPoint != null)
                {
                    perimeter += Math.Sqrt(Math.Pow((point.x - prevPoint.x), 2) + Math.Pow((point.y - prevPoint.y), 2));
                }
                prevPoint = point;
            }
            return perimeter;
        }

        /// <summary>
        /// Calculate and return the area
        /// </summary>
        /// <param name="points">Points to get the perimeter</param>
        /// <returns>The total perimeter</returns>
        public static double CalcArea(List<Point> points)
        {
            int numPoints = points.Count;
            double area = 0;

            for (int i = 0; i < numPoints; i++)
            {
                Point point = points[i];
                double addX = point.x;
                double addY = points[(i == numPoints - 1 ? 0 : i + 1)].y;
                double subX = points[(i == numPoints - 1 ? 0 : i + 1)].x;
                double subY = point.y;
                area += addX * addY - subX * subY;
            }
            return Math.Abs(area) / 2;
        }

        /// <summary>
        /// It updates the area of the Polygon annotation if it has measurement data
        /// </summary>
        /// <param name="polygon">Polygon annot</param>
        public static void UpdateAreaAnnot(Annots.Polygon polygon)
        {
            if ((polygon == null) || (!polygon.IsValid()))
                return;

            if (IsAreaMeasure(polygon) == false)
                return;

            List<Point> points = new List<Point>();
            for (int i = 0; i < polygon.GetVertexCount(); i++)
                points.Add(polygon.GetVertex(i));

            double area = CalcArea(points);

            // Calculate final conversion and add translated unit        
            var groupInfo = GetAnnotMeasureInfo(polygon);
            var val = area * groupInfo.AreaInfo.BaseFactor;
            var strVal = FormatDecimals(val, MeasureUtils.ConvPDFPrecisionToStr(groupInfo.AreaInfo.Precision)) + " " + groupInfo.AreaInfo.Unit;

            polygon.SetContents(strVal);
        }

        /// <summary>
        /// Ut Updates the perimeter of the Polyline annotation if it has measurement data
        /// </summary>
        /// <param name="polyline">Annotation polyline</param>
        public static void UpdatePerimeterAnnot(Annots.PolyLine polyline)
        {
            if ((polyline == null) || (!polyline.IsValid()))
                return;

            if (IsPerimeterMeasure(polyline) == false)
                return;

            List<Point> points = new List<Point>();
            for (int i = 0; i < polyline.GetVertexCount(); i++)
                points.Add(polyline.GetVertex(i));

            double perimeter = MeasureUtils.CalcPerimeter(points);

            // Calculate final conversion and add translated unit  
            var groupInfo = MeasureUtils.GetAnnotMeasureInfo(polyline);
            var val = perimeter * groupInfo.AxisInfo.BaseFactor;
            var strVal = MeasureUtils.FormatDecimals(val, MeasureUtils.ConvPDFPrecisionToStr(groupInfo.DistanceInfo.Precision)) + " " + groupInfo.DistanceInfo.Unit;

            polyline.SetContents(strVal);
        }

        #endregion
    }

    public enum PageUnit
    {
        pt,
        inch,
        mm,
        cm,
        m,
        km,
        ft,
        yd,
        mi
    }

    public enum MeasureInfoType
    {
        Precision,
        BaseUnit,
        BaseScale,
        TranslateUnit,
        TranslateScale
    }

    public class MeasureInfo : IMeasure
    {
        public double Precision { get; set; } = 100;
        public double BaseFactor { get; set; } = 1;
        public double TranslateFactor { get; set; } = 1;
        public string Display { get; set; } = "D";
        public string DecimalSymbol { get; set; } = ".";
        public string ThousandSymbol { get; set; } = ",";
        public string Unit { get; set; } = "in";
        public string UnitSuffix { get; set; } = string.Empty;
        public string UnitPrefix { get; set; } = string.Empty;
        public string BaseUnit { get; set; } = string.Empty;
        public string Scale { get; set; } = string.Empty;
        public string UnitPosition { get; set; } = "S";
    }

    public class MeasureInfoGroup
    {
        public string Scale { get; set; }
        public MeasureInfo AxisInfo { get; set; }
        public MeasureInfo DistanceInfo { get; set; }
        public MeasureInfo AreaInfo { get; set; }
    }
}
