﻿using pdftron.PDF.Tools.Controls.Compare;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.ApplicationModel.Core;
using Windows.Foundation;
using Windows.Storage;
using Windows.UI.Core;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;


/* Some design infomation
 * 
 * For this implementation we use  CoreApplication.CreateNewView() which creates a new window and thread,
 * so now we need to ensure to know which thread we are in roder to execute UI updates.
 * 
 * Mainly what we need to ensure no thread sync exception happens is to call "Window.Current.Dispatcher.RunAsync"
 */

namespace pdftron.PDF.Tools.Utilities
{
    /// <summary>
    /// Interface used to allow a UserControl to be created on a new Window using
    /// MultiWindowManager
    /// </summary>
    public interface IMultiWindowUserControl
    { }

    /// <summary>
    /// MultiWindowManager manages the creation and disposing of secondary windows of Xodo, and 
    /// </summary>
    public class MultiWindowManager
    {
        private static int _mainAppViewId;
        private static int _maxAllowedWindows = 5;
        private static Dictionary<int, string> _activeAppViews;
        private static MultiWindowManager _current;

        private MultiWindowManager()
        {
            if (_activeAppViews == null)
            {
                _mainAppViewId = ApplicationView.GetForCurrentView().Id;
                _activeAppViews = new Dictionary<int, string>() { { _mainAppViewId, " " } };
            }
        }

        /// <summary>
        /// Retrieve the MultiWindowManager single instance
        /// </summary>
        public static MultiWindowManager Instance 
        {
            get 
            {
                if (_current == null)
                    _current = new MultiWindowManager();

                return _current;
            }
        }

        #region Private Methods

        /// <summary>
        /// Tracks an application view id
        /// </summary>
        /// <param name="windowId"></param>
        private void TrackWindow(int appViewId)
        {
            if (_activeAppViews == null)
                return;

            System.Diagnostics.Debug.WriteLine("===> Debug: Adding Window: App View Id: " + appViewId.ToString());

            _activeAppViews.Add(appViewId, "");
        }

        #endregion

        #region Public Methods

        public class MultiWindowCreationResult
        {
            public bool IsWindowCreated { get; set; } = false;
            public int AppViewId { get; set; }
        }

        /// <summary>
        /// Initialize a new Window and create a new UserControl instance based on the Type
        /// Example: typeof(CustomUserControl)
        /// NOTE: it must implement IMultiWindowUserControl
        /// </summary>
        /// <param name="classType">It must derive from Windows.UI.Xaml.Controls.UserControl</param>
        /// <param name="controlType">Type of UserControl creation</param>
        /// <returns></returns>
        public async Task<MultiWindowCreationResult> TryCreateNewWindowAsync(System.Type classType)
        {
            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Multi Window", new Dictionary<string, string>() { ["Create New Window"] = "Type: " + classType.Name });

            if (classType.GetInterfaces().Contains(typeof(IMultiWindowUserControl)) == false)
                return new MultiWindowCreationResult();

            if (classType.BaseType != typeof(Windows.UI.Xaml.Controls.UserControl) && classType != typeof(Windows.UI.Xaml.Controls.UserControl))
                return new MultiWindowCreationResult();

            // Block creation requests after reaching max allowed windows
            if (_activeAppViews.Count >= _maxAllowedWindows)
                return new MultiWindowCreationResult();

            int currentAppViewId = -1;
            bool creationRslt = false;

            // Create new application view in a different thread
            await CoreApplication.CreateNewView().Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
            {
                Frame frame = new Frame();

                // Create a mew Control based on the type passed which must implement IMultiWindowsUserControl
                var ctrl = (IMultiWindowUserControl)Activator.CreateInstance(classType);                
                if (ctrl == null)
                {
                    creationRslt = false;
                    return;
                }

                if (frame.Navigate(typeof(NewWindowPage), ctrl) == false)
                {
                    creationRslt = false;
                    return;
                }

                Window.Current.Content = frame;
                Window.Current.Activate();                

                currentAppViewId = ApplicationView.GetForCurrentView().Id;

                creationRslt = true;
            });

            if (currentAppViewId == -1 || creationRslt == false)
                return new MultiWindowCreationResult();

            //Attempt to show the newly created application view
            if (await ApplicationViewSwitcher.TryShowAsStandaloneAsync(currentAppViewId))
            {
                TrackWindow(currentAppViewId);
                return new MultiWindowCreationResult
                {
                    IsWindowCreated = true,
                    AppViewId = currentAppViewId
                };
            }

            return new MultiWindowCreationResult();
        }

        /// <summary>
        /// Remove the Windows / App View reference from the manager's list
        /// </summary>
        /// <param name="id">App View Id</param>
        /// <returns>True if found and removed, false otherwise</returns>
        public bool UntrackWindowById(int id)
        {
            if (_activeAppViews == null)
                return false;

            if (id == _mainAppViewId)
                return false;

                System.Diagnostics.Debug.WriteLine("===> Debug: Removing Windows: App View Id: " + id.ToString());

            return _activeAppViews.Remove(id);
        }

        /// <summary>
        /// Close a Windwos / App View by Id
        /// </summary>
        /// <param name="id">App View Id</param>
        /// <returns>True if closed with success, False otherwise</returns>
        public async Task<bool> CloseWindowByIdAsync(int id)
        {
            if (_activeAppViews == null)
                return false;

            if (id == _mainAppViewId)
                return false;

            await ApplicationViewSwitcher.SwitchAsync(_mainAppViewId, id, ApplicationViewSwitchingOptions.ConsolidateViews);

            return true;
        }

        /// <summary>
        /// Attempts to close all secondary application views launched from Xodo
        /// </summary>
        /// <returns></returns>
        public async Task CloseAllChildWindowsAsync()
        {
            if (_activeAppViews == null)
                return;

            foreach (int windowId in _activeAppViews.Keys.ToList())
            {
                if (windowId == _mainAppViewId)
                    continue;

                await ApplicationViewSwitcher.SwitchAsync(_mainAppViewId, windowId, ApplicationViewSwitchingOptions.ConsolidateViews);
                UntrackWindowById(windowId);
            }
        }

        /// <summary>
        /// Switch from main Window to a "child" Window by Id
        /// </summary>
        /// <param name="toViewId">App View ID</param>
        public async Task SwitchWindowByIdAsync(int toViewId)
        {
            if (_activeAppViews == null)
                return;

            if (_activeAppViews.Keys.Contains(toViewId) == false)
                return;

            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Multi Window", new Dictionary<string, string>() { ["Command"] = "Switch Window By ID"});

            await ApplicationViewSwitcher.SwitchAsync(toViewId, _mainAppViewId, ApplicationViewSwitchingOptions.Default);
        }

        #endregion

        #region Public Properties
        /// <summary>
        /// Number of opened Windows including the Main Application
        /// </summary>
        public int OpenedWindowsCount
        {
            get { return _activeAppViews.Count; }
        }
        #endregion
    }
}
