﻿using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools
{
    public class NavigationState
    {
        public NavigationState(int page, double zoom, double horizontalPos, double verticalPos)
        {
            Page = page;
            Zoom = zoom;
            HorizontalScrollPosition = horizontalPos;
            VerticalScrollPosition = verticalPos;
            IsWeak = false;
        }

        public int Page { get; private set; }
        public double Zoom { get; private set; }
        public double HorizontalScrollPosition { get; private set; }
        public double VerticalScrollPosition { get; private set; }

        internal bool HasPositionState {  get { return !double.IsNaN(Zoom) && !double.IsNaN(HorizontalScrollPosition) && !double.IsNaN(VerticalScrollPosition); } }
        /// <summary>
        /// This flag governs if the state should be saved if the user navigates using a regular page swipe (not weak)
        /// or if it should be discarded (weak).
        /// </summary>
        internal bool IsWeak { get; set; }

        public static NavigationState TakeSnapshot(PDFViewCtrl pdfView)
        {
            return new NavigationState(pdfView.GetCurrentPage(), pdfView.GetZoom(), pdfView.GetHScrollPos(), pdfView.GetVScrollPos());
        }

        public static NavigationState TakeSnapshot(ReflowViewModel reflow)
        {
            return new NavigationState(reflow.CurrentPageNumber, double.NaN, double.NaN, double.NaN);
        }

        /// <summary>
        /// This will return true if the two views are close enough to the same that we don't need both in the navigation stack.
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public override bool Equals(object obj)
        {
            NavigationState other = obj as NavigationState;
            if (other != null)
            {
                if (Page == other.Page // same page, no other values
                    && double.IsNaN(Zoom) && double.IsNaN(other.Zoom)
                    && double.IsNaN(HorizontalScrollPosition) && double.IsNaN(other.HorizontalScrollPosition)
                    && double.IsNaN(VerticalScrollPosition) && double.IsNaN(other.VerticalScrollPosition))
                {
                    return true;
                }
                if (double.IsNaN(Zoom) || double.IsNaN(other.Zoom) // at least one invalid value, but not all and the same page
                    || double.IsNaN(HorizontalScrollPosition) || double.IsNaN(other.HorizontalScrollPosition)
                    || double.IsNaN(VerticalScrollPosition) || double.IsNaN(other.VerticalScrollPosition))
                {
                    return false;
                }
                return Page == other.Page &&
                    (Math.Max(Zoom, other.Zoom) / Math.Min(Zoom, other.Zoom)) < 1.01 // dividing zooms is close to 1
                    && Math.Abs(HorizontalScrollPosition - other.HorizontalScrollPosition) < 2
                    && Math.Abs(VerticalScrollPosition - other.VerticalScrollPosition) < 2;
            }
            return base.Equals(obj);
        }

        /// <summary>
        /// It makes no sense to hash this class. So we throw an exception.
        /// Equality is meant to determine whether two visual starts are similar enough that it's not worth
        /// generating two separate states.
        /// </summary>
        /// <returns></returns>
        /// <exception cref="NotSupportedException"></exception>
        public override int GetHashCode() => throw new NotSupportedException();

        public void CleanPositionData()
        {
            HorizontalScrollPosition = double.NaN;
            VerticalScrollPosition = double.NaN;
            Zoom = double.NaN;
        }

        public override string ToString()
        {
            if (double.IsNaN(Zoom) ||
                double.IsNaN(HorizontalScrollPosition) ||
                double.IsNaN(VerticalScrollPosition))
            {
                return string.Format("Page {0}, no other state", Page);
            }
            return string.Format("Page {0}, zoom {1}, at ({2}, {3})", Page, (int)(Zoom * 100), (int)HorizontalScrollPosition, (int)VerticalScrollPosition);
        }
    }

    public class NavigationStack
    {
        public enum Navigators
        {
            None = 0,
            PDFViewCtrl,
            ReflowView,
            Both,
        }

        public Navigators _CurrentNavigator = Navigators.None;
        public Navigators CurrentNavigator
        {
            get { return _CurrentNavigator; }
            set
            {
                if (_CurrentNavigator != value)
                {
                    if ((value == Navigators.PDFViewCtrl || value == Navigators.Both) && PDFViewCtrl == null)
                    {
                        throw new ArgumentException("Can not set the PDFViewCtrl as navigator when there is no PDFViewCtrl");
                    }
                    else if ((value == Navigators.ReflowView || value == Navigators.Both) && ReflowViewModel == null)
                    {
                        throw new ArgumentException("Can not set the ReflowView as navigator when there is no ReflowViewModel");
                    }
                    _CurrentNavigator = value;
                    if (_CurrentNavigator == Navigators.ReflowView)
                    {
                        CleanPositionData();
                    }
                }
            }
        }

        private PDFViewCtrl _PDFViewCtrl;
        private PDFViewCtrl PDFViewCtrl
        {
            get { return _PDFViewCtrl; }
            set
            {
                if(_PDFViewCtrl != value)
                {
                    if (_PDFViewCtrl != null)
                    {
                        _PDFViewCtrl.OnPageNumberChanged -= PDFViewCtrl_OnPageNumberChanged;
                        _PDFViewCtrl.OnLayoutChanged -= PDFViewCtrl_OnLayoutChanged;
                    }
                    _PDFViewCtrl = value;
                    if (_PDFViewCtrl != null)
                    {
                        _PDFViewCtrl.OnPageNumberChanged += PDFViewCtrl_OnPageNumberChanged;
                        _PDFViewCtrl.OnLayoutChanged += PDFViewCtrl_OnLayoutChanged;
                        _PagePresentationMode = _PDFViewCtrl.GetPagePresentationMode();
                        _Rotation = _PDFViewCtrl.GetRotation();
                        if (ReflowViewModel == null)
                        {
                            CurrentNavigator = Navigators.PDFViewCtrl;
                        }
                        else
                        {
                            CurrentNavigator = Navigators.Both;
                        }
                    }
                    if (_PDFViewCtrl == null)
                    {
                        if (CurrentNavigator == Navigators.Both)
                        {
                            CurrentNavigator = Navigators.ReflowView;
                        }
                        if (CurrentNavigator == Navigators.PDFViewCtrl)
                        {
                            CurrentNavigator = Navigators.None;
                        }
                    }
                    if (PDFViewCtrl == null && ReflowViewModel == null)
                    {
                        Clear();
                    }
                }
            }
        }

        private ReflowViewModel _ReflowViewModel = null;
        public ReflowViewModel ReflowViewModel
        {
            get { return _ReflowViewModel; }
            set
            {
                if (_ReflowViewModel != value)
                {
                    if (_ReflowViewModel != null)
                    {
                        _ReflowViewModel.PageChanged -= ReflowViewModel_PageChanged;
                        _ReflowViewModel.PageChangeIntended -= ReflowViewModel_PageChangeIntended;
                    }
                    _ReflowViewModel = value;
                    if (_ReflowViewModel != null)
                    {
                        _ReflowViewModel.PageChanged += ReflowViewModel_PageChanged;
                        _ReflowViewModel.PageChangeIntended += ReflowViewModel_PageChangeIntended;
                        if (PDFViewCtrl == null)
                        {
                            CurrentNavigator = Navigators.ReflowView;
                        }
                        else
                        {
                            CurrentNavigator = Navigators.Both;
                        }
                    }
                    if (_ReflowViewModel == null)
                    {
                        if (CurrentNavigator == Navigators.Both)
                        {
                            CurrentNavigator = Navigators.PDFViewCtrl;
                        }
                        if (CurrentNavigator == Navigators.ReflowView)
                        {
                            CurrentNavigator = Navigators.None;
                        }
                    }
                    if (PDFViewCtrl == null && ReflowViewModel == null)
                    {
                        Clear();
                    }
                }
            }
        }

        private Stack<NavigationState> _UndoStack = new Stack<NavigationState>();
        private Stack<NavigationState> _RedoStack = new Stack<NavigationState>();
        private NavigationState _CurrentState = null;

        private PDFViewCtrlPagePresentationMode _PagePresentationMode = PDFViewCtrlPagePresentationMode.e_single_continuous;
        private PDFViewCtrlPagePresentationMode PagePresentationMode
        {
            get { return _PagePresentationMode; }
            set
            {
                if (_PagePresentationMode != value)
                {
                    _PagePresentationMode = value;
                    CleanPositionData();
                }
            }
        }

        private PageRotate _Rotation = PageRotate.e_0;
        private PageRotate Rotation
        {
            get { return _Rotation; }
            set
            {
                if (_Rotation != value)
                {
                    _Rotation = value;
                    CleanPositionData();
                }
            }
        }

        public delegate void NavigationStackChangedDelegate();
        public event NavigationStackChangedDelegate NavigationStackChanged;

        public NavigationStack()
        {
            CurrentNavigator = Navigators.None;
        }

        public NavigationStack(PDFViewCtrl ctrl)
        {
            PDFViewCtrl = ctrl;
            _PagePresentationMode = ctrl.GetPagePresentationMode();
            _Rotation = ctrl.GetRotation();
        }



        public NavigationStack(ReflowViewModel reflow)
        {
            ReflowViewModel = reflow;
        }

        public void Unsubscribe()
        {
            PDFViewCtrl = null;
        }

        #region Actions on PDFViewCtrl

        private bool _ExecutingAction = false;

        /// <summary>
        /// Wrapps all the actions that affect the view of the PDFViewCtrl.
        /// This simply lets us avoid rewriting a whole bunch of code.
        /// </summary>
        /// <typeparam name="R"></typeparam>
        /// <typeparam name="P"></typeparam>
        /// <param name="param"></param>
        /// <param name="defaultRet"></param>
        /// <param name="act"></param>
        /// <returns></returns>
        private R GenericActionWrapper<R, P>(P param, R defaultRet, Func<P, R> act)
        {
            _ExecutingAction = true;

            try
            {
                NavigationState beforeState = null;
                NavigationState afterState = null;
                R actionSuccess = defaultRet;
                if (CurrentNavigator == Navigators.PDFViewCtrl || CurrentNavigator == Navigators.Both)
                {
                    beforeState = NavigationState.TakeSnapshot(PDFViewCtrl);
                    actionSuccess = act(param);
                    afterState = NavigationState.TakeSnapshot(PDFViewCtrl);
                }
                else if (CurrentNavigator == Navigators.ReflowView)
                {
                    beforeState = NavigationState.TakeSnapshot(ReflowViewModel);
                    actionSuccess = act(param);
                    afterState = NavigationState.TakeSnapshot(ReflowViewModel);
                }
                RegisterJump(beforeState, afterState);
                if (CurrentNavigator == Navigators.Both)
                {
                    SyncReflowViewModel(afterState);
                }

                return actionSuccess;
            }
            finally
            {
                DoneExecutingAction();
            }
        }

        /// <summary>
        /// When the PDFViewCtrl changes page, the OnPageNumberChanged is not raised immediately. It is put on the
        /// dispatcher at normal priority and raised after the current block of code has finished. This means that
        /// we need to ensure that _ExecutingAction is reset after this happens, so we add the same delay here.
        /// </summary>
        /// <remarks> This should not cause problems, because two actions on the navigation stack should not happen
        /// in one sequence of execution.
        /// </remarks>
        private async void DoneExecutingAction()
        {
            await _PDFViewCtrl.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                _ExecutingAction = false;
            });
        }

        /// <summary>
        /// Same as PDFViewCtrl.ExecuteAction but records state information for the navigation stack.
        /// A PDFViewCtrl is required for this to work.
        /// </summary>
        /// <param name="actionParam">The action to execute</param>
        public void ExecuteAction(ActionParameter actionParam)
        {
            bool result = GenericActionWrapper<bool, ActionParameter>(actionParam, true, (p) =>
            {
                if (PDFViewCtrl == null)
                {
                    return false;
                }
                PDFViewCtrl.ExecuteAction(p);
                if (CurrentNavigator == Navigators.Both || CurrentNavigator == Navigators.ReflowView)
                {
                    ReflowViewModel.SetPageNumber(PDFViewCtrl.GetCurrentPage());
                }
                if (CurrentNavigator == Navigators.ReflowView)
                {
                    PDFViewCtrl.CancelRendering();
                }
                return true;
            });
        }

        /// <summary>
        /// Same as PDFViewCtrl.SetCurrentPage or ReflowViewModel.SetPageNumber
        /// but records state information for the navigation stack.
        /// </summary>
        /// <param name="pageNumber">The page number to go to</param>
        /// <returns></returns>
        public bool SetCurrentPage(int pageNumber)
        {
            bool result = GenericActionWrapper<bool, int>(pageNumber, false, (p) =>
            {
                if (CurrentNavigator == Navigators.ReflowView)
                {
                    return ReflowViewModel.SetPageNumber(pageNumber);
                }
                else if (CurrentNavigator != Navigators.None)
                {
                    return PDFViewCtrl.SetCurrentPage(p);
                }
                return false;
            });
            return result;
        }

        #endregion Actions on PDFViewCtrl


        #region State Actions

        /// <summary>
        ///  Will push the NavigationState, unless it's Equal to the previous one.
        /// </summary>
        private void PushState(NavigationState state)
        {
            if (!(_RedoStack.Count > 0 && _RedoStack.Peek().Equals(state)) &&
                !(_UndoStack.Count > 0 && _UndoStack.Peek().Equals(state)))
            {
                if (_CurrentState != null && !_CurrentState.Equals(state)
                    && !(!_CurrentState.HasPositionState && _CurrentState.Page == state.Page))
                {

                    _CurrentState.IsWeak = false;
                    _UndoStack.Push(_CurrentState);
                }
                _RedoStack.Clear();
                _CurrentState = state;
                if (_UndoStack.Count > 0 &&
                    !_UndoStack.Peek().HasPositionState &&
                    _UndoStack.Peek().Page == _CurrentState.Page)
                {
                    _UndoStack.Pop();
                }
            }
        }

        #endregion State Actions


        #region Public API

        /// <summary>
        /// True if it is possible to undo an action
        /// </summary>
        public bool IsUndoEnabled { get { return _UndoStack.Count > 0; } }

        /// <summary>
        /// True if it is possible to redo an action
        /// </summary>
        public bool IsRedoEnabled { get { return _RedoStack.Count > 0; } }

        /// <summary>
        /// Undoes a navigation action
        /// </summary>
        public void Undo()
        {
            if (_UndoStack.Count == 0)
            {
                throw new Exception("There is no undo state to move to");
            }
            if (_CurrentState != null)
            {
                if (_RedoStack.Count == 0 || !_RedoStack.Peek().Equals(_CurrentState))
                {
                    _RedoStack.Push(_CurrentState);
                }
            }
            NavigationState currentState = TakeSnapshotForNavigator();
            if (_RedoStack.Count == 0 || _RedoStack.Peek().Page != currentState.Page)
            {
                _RedoStack.Push(currentState);
            }
            NavigationState backState = _UndoStack.Pop();
            backState.IsWeak = true;
            _CurrentState = backState;
            if (!currentState.Equals(_CurrentState))
            {
                try
                {
                    _ExecutingAction = true;
                    SyncNavigators(_CurrentState);
                }
                catch (Exception)
                {

                }
                finally
                {
                    DoneExecutingAction();
                }
            }
            NavigationStackChanged?.Invoke();
        }

        /// <summary>
        /// Redoes a navigation action
        /// </summary>
        public void Redo()
        {
            if (_RedoStack.Count == 0)
            {
                throw new Exception("There is no undo state to move to");
            }
            if (_CurrentState != null)
            {
                if (_UndoStack.Count == 0 || !_UndoStack.Peek().Equals(_CurrentState))
                {
                    _UndoStack.Push(_CurrentState);
                }
            }
            _CurrentState = _RedoStack.Pop();
            _CurrentState.IsWeak = true;
            NavigationState currentState = TakeSnapshotForNavigator();
            if (!currentState.Equals(_CurrentState))
            {
                try
                {
                    _ExecutingAction = true;
                    SyncNavigators(_CurrentState);
                }
                catch (Exception)
                {

                }
                finally
                {
                    DoneExecutingAction();
                }
            }
            NavigationStackChanged?.Invoke();
        }

        /// <summary>
        /// Used to register a switch from one NavigationState to Another
        /// For example, when creating composite interactions, like setting page number and then setting scroll position
        /// this function can be used with a NavigationState.TakeSnapshot before and after the composite interaction
        /// </summary>
        /// <param name="beforeState"></param>
        /// <param name="afterState"></param>
        public void RegisterJump(NavigationState beforeState, NavigationState afterState)
        {
            if (!beforeState.Equals(afterState))
            {
                PushState(beforeState);
                PushState(afterState);
                if (ReflowViewModel != null && CurrentNavigator == Navigators.ReflowView || CurrentNavigator == Navigators.Both)
                {
                    ReflowViewModel.SetPageNumber(afterState.Page);
                }
                NavigationStackChanged?.Invoke();
            }
        }

        /// <summary>
        /// Removes specific data related to zoom and scroll positions. Useful after, for example, changing page presentation mode
        /// </summary>
        public void CleanPositionData()
        {
            foreach (NavigationState state in _UndoStack)
            {
                state.CleanPositionData();
            }
            _RedoStack.Clear();
            if (_UndoStack.Count > 1)
            {
                Stack<NavigationState> states = new Stack<NavigationState>();
                states.Push(_UndoStack.Pop());
                while (_UndoStack.Count > 0)
                {
                    NavigationState state = _UndoStack.Pop();
                    if (!state.Equals(states.Peek()))
                    {
                        states.Push(state);
                    }
                }
                while (states.Count > 0)
                {
                    _UndoStack.Push(states.Pop());
                }
                if (_CurrentState != null && _CurrentState.Page == _UndoStack.Peek().Page)
                {
                    _UndoStack.Pop();
                }
            }
            NavigationStackChanged?.Invoke();
        }

        /// <summary>
        /// Empties the Navigation Stack
        /// </summary>
        public void Clear()
        {
            _UndoStack.Clear();
            _RedoStack.Clear();
            NavigationStackChanged?.Invoke();
        }

        #endregion Public API


        #region PDFViewCtrl Handling

        private void PDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            if (!_ExecutingAction)
            {
                if (_CurrentState == null || _CurrentState.Page != current_page)
                {
                    if (_CurrentState != null && !_CurrentState.IsWeak)
                    {
                        _UndoStack.Push(_CurrentState);
                    }
                    _CurrentState = null;
                    _RedoStack.Clear();
                }
                NavigationStackChanged?.Invoke();
            }
        }

        private void PDFViewCtrl_OnLayoutChanged()
        {
            PagePresentationMode = PDFViewCtrl.GetPagePresentationMode();
            Rotation = PDFViewCtrl.GetRotation();
        }

        private void ReflowViewModel_PageChanged(int pageNumber, int totalPages)
        {
            PDFViewCtrl_OnPageNumberChanged(pageNumber, totalPages);
        }

        private bool ReflowViewModel_PageChangeIntended(int pageNumber)
        {
            return SetCurrentPage(pageNumber);
        }

        private void SyncNavigators(NavigationState state)
        {
            if ((CurrentNavigator == Navigators.PDFViewCtrl || CurrentNavigator == Navigators.Both) && PDFViewCtrl != null)
            {
                SyncPDFViewCtrl(state);
            }
            if ((CurrentNavigator == Navigators.ReflowView || CurrentNavigator == Navigators.Both) && ReflowViewModel != null)
            {
                SyncReflowViewModel(state);
            }
        }

        private void SyncPDFViewCtrl(NavigationState state)
        {
            PDFViewCtrl.SetCurrentPage(state.Page);
            if (!double.IsNaN(state.Zoom) &&
                !double.IsNaN(state.HorizontalScrollPosition) &&
                !double.IsNaN(state.VerticalScrollPosition))
            {
                PDFViewCtrl.SetZoom(state.Zoom);
                PDFViewCtrl.SetHScrollPos(state.HorizontalScrollPosition);
                PDFViewCtrl.SetVScrollPos(state.VerticalScrollPosition);
            }
        }

        private void SyncReflowViewModel(NavigationState state)
        {
            ReflowViewModel.SetPageNumber(state.Page);
        }

        private NavigationState TakeSnapshotForNavigator()
        {
            if (CurrentNavigator == Navigators.ReflowView && ReflowViewModel != null)
            {
                return NavigationState.TakeSnapshot(ReflowViewModel);
            }
            else if ((CurrentNavigator == Navigators.PDFViewCtrl || CurrentNavigator == Navigators.Both)
                && PDFViewCtrl != null)
            {
                return NavigationState.TakeSnapshot(PDFViewCtrl);
            }
            return null;
        }

        #endregion PDFViewCtrl Handling

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder("NavigationStack State:" + Environment.NewLine);
            foreach (NavigationState state in _UndoStack.Reverse())
            {
                sb.AppendLine(string.Format("UndoStack: {0}", state.ToString()));
            }
            if (_CurrentState != null)
            {
                sb.AppendLine(string.Format("Current:   {0}{1}", _CurrentState.ToString(), (_CurrentState.IsWeak ? " is weak" : "")));
            }
            foreach (NavigationState state in _RedoStack)
            {
                sb.AppendLine(string.Format("RedoStack: {0}", state.ToString()));
            }
            return sb.ToString();
        }
    }
}
