﻿using pdftron.PDF.Annots;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

using UIPopup = Windows.UI.Xaml.Controls.Primitives.Popup;
using UIRect = Windows.Foundation.Rect;
using UIPoint = Windows.Foundation.Point;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Utilities
{
    public sealed partial class NoteDialogPhone : NoteDialogBase
    {
        public enum InputMode
        {
            SingleLineText,
            MultiLineText,
            Password,
        }

        private InputMode _TextInputMode = InputMode.MultiLineText;

        public InputMode TextInputMode
        {
            get { return _TextInputMode; }
            set
            {
                _TextInputMode = value;
                switch (value)
                {
                    case InputMode.MultiLineText:
                        MainTextBox.Visibility = Windows.UI.Xaml.Visibility.Visible;
                        SingleLineTextBlock.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                        PassWordBox.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                        break;
                    case InputMode.SingleLineText:
                        MainTextBox.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                        SingleLineTextBlock.Visibility = Windows.UI.Xaml.Visibility.Visible;
                        PassWordBox.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                        break;
                    case InputMode.Password:
                        MainTextBox.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                        SingleLineTextBlock.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                        PassWordBox.Visibility = Windows.UI.Xaml.Visibility.Visible;
                        break;
                }
            }
        }

        /// <summary>
        /// The text in the user has filled in.
        /// </summary>
        public string AnnotText 
        { 
            get 
            { 
                switch (TextInputMode)
                {
                    case InputMode.MultiLineText:
                        return MainTextBox.Text;
                    case InputMode.SingleLineText:
                        return SingleLineTextBlock.Text;
                    case InputMode.Password:
                        return PassWordBox.Password;
                }
                return string.Empty;
            } 
            set
            {
                if (value != null)
                {
                    MainTextBox.Text = value;
                    SingleLineTextBlock.Text = value;
                    PassWordBox.Password = value;

                    switch (TextInputMode)
                    {
                        case InputMode.MultiLineText:
                            MainTextBox.Focus(FocusState.Programmatic);
                            MainTextBox.SelectionStart = value.Length;
                            break;
                        case InputMode.SingleLineText:
                            SingleLineTextBlock.Focus(FocusState.Programmatic);
                            SingleLineTextBlock.SelectionStart = value.Length;
                            break;
                        case InputMode.Password:
                            PassWordBox.Focus(FocusState.Programmatic);
                            break;
                    }
                }
            }
        }

        UIPopup _Popup;
        SizeChangedEventHandler _SizeHandler;

        public NoteDialogPhone()
        {
            this.InitializeComponent();

            TextInputMode = _TextInputMode;
        }

        public NoteDialogPhone(string oldText, bool canDelete, bool canSaveEmpty, PDFViewCtrl ctrl, ToolManager toolManager)
            : base(oldText, canDelete, canSaveEmpty, ctrl, toolManager)
        {
            this.InitializeComponent();
            TextInputMode = _TextInputMode;

            this.Loaded += NoteDialog_Loaded;
            _SizeHandler = new SizeChangedEventHandler(_PDFViewCtrl_SizeChanged);
            _PDFViewCtrl.SizeChanged += _SizeHandler;

            MainTextBox.Text = _OriginalText;
            MainTextBox.SelectionStart = MainTextBox.Text.Length;
            MainTextBox.TextChanged += MainTextBox_TextChanged;
            SingleLineTextBlock.Text = _OriginalText;
            SingleLineTextBlock.SelectionStart = MainTextBox.Text.Length;
            SingleLineTextBlock.TextChanged += MainTextBox_TextChanged;
            PassWordBox.Password = _OriginalText;
            PassWordBox.PasswordChanged += PassWordBox_PasswordChanged;

            if (canDelete)
            {
                DeleteOrCloseHost.Visibility = Visibility.Visible;
                SaveOrCancelHost.Visibility = Visibility.Collapsed;
            }
            if (canSaveEmpty)
            {
                SaveButton.IsEnabled = true;
            }

            CreatePopup();
            _ToolManager.mCurrentNoteDialog = this;
            FocusWhenOpened = true;
        }

 
        private void MainTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            _TextHasChanged = true;
            DeleteOrCloseHost.Visibility = Visibility.Collapsed;
            SaveOrCancelHost.Visibility = Visibility.Visible;
            SaveButton.IsEnabled = true;
        }

        private void PassWordBox_PasswordChanged(object sender, RoutedEventArgs e)
        {
            _TextHasChanged = true;
            DeleteOrCloseHost.Visibility = Visibility.Collapsed;
            SaveOrCancelHost.Visibility = Visibility.Visible;
            SaveButton.IsEnabled = true;
        }

        private void CloseInternal(NotedialogResults result)
        {
            if (_SizeHandler != null)
            {
                _PDFViewCtrl.SizeChanged -= _SizeHandler;
            }
            _PDFViewCtrl.IsEnabled = true;
            _Popup.IsOpen = false;
            _ToolManager.mCurrentNoteDialog = null;

            string savedResult = MainTextBox.Text;
            if (TextInputMode == InputMode.SingleLineText)
            {
                savedResult = SingleLineTextBlock.Text;
            }
            else if (TextInputMode == InputMode.Password)
            {
                savedResult = PassWordBox.Password;
            }

            RaiseNoteClosedEvent(result, savedResult);
        }

        public override void Close()
        {
            CloseInternal(NotedialogResults.Canceled);
            base.Close();
        }


        void NoteDialog_Loaded(object sender, RoutedEventArgs e)
        {
            if (FocusWhenOpened)
            {
                switch (TextInputMode)
                {
                    case InputMode.MultiLineText:
                        MainTextBox.Focus(FocusState.Programmatic);
                        break;
                    case InputMode.SingleLineText:
                        SingleLineTextBlock.Focus(FocusState.Programmatic);
                        break;
                    case InputMode.Password:
                        PassWordBox.Focus(FocusState.Programmatic);
                        break;
                }   
            }
        }
        
        void _PDFViewCtrl_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            PositionPopup();
        }

        private void CreatePopup()
        {
            _Popup = new UIPopup();
            _Popup.Child = this;
            _Popup.IsOpen = true;
            PositionPopup();

            MainTextBox.SelectionStart = MainTextBox.Text.Length;
        }

        private void PositionPopup()
        {
            UIRect rect = UtilityFunctions.GetElementRect(_PDFViewCtrl);

            _Popup.HorizontalOffset = rect.X;
            //_Popup.VerticalOffset = rect.Y;

            MainGrid.Width = rect.Width;
            if (rect.Width > rect.Height) // landscape mode
            {
                VisualStateManager.GoToState(this, "WideLayout", true);
            }
            else
            {
                VisualStateManager.GoToState(this, "TallLayout", true);
            }
        }

        private void SaveButton_Click(object sender, RoutedEventArgs e)
        {
            CloseInternal(NotedialogResults.Changed);
            Close();
        }

        private void CancelButton_Click(object sender, RoutedEventArgs e)
        {
            Close();
        }


        private void DeleteButton_Click(object sender, RoutedEventArgs e)
        {
            CloseInternal(NotedialogResults.Deleted);
        }
    }
}
