﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools.Utilities
{
    public class PDFDocCreator
    {
        static double LINE_SHADE = 0.35;
        static double MARGIN_RED = 1.0;
        static double MARGIN_GREEN = 0.5;
        static double MARGIN_BLUE = 0.5;

        static uint WHITE_COLOR = 0xFFFFFFFF;

        /// <summary>
        /// Create a PDF document of a specific type e number of pages
        /// </summary>
        /// <param name="paperType">Type of paper</param>
        /// <param name="width">The width in page points</param>
        /// <param name="height">The height in page points</param>
        /// <param name="pages">Number of pages</param>
        /// <returns>Generated PDFDoc</returns>
        public static PDFDoc Create(PaperType paperType, double width, double height, int pages)
        {
            double width_in = PagePointToInches(width);
            double height_in = PagePointToInches(height);

            double lineShade = LINE_SHADE;

            double bgR = ((WHITE_COLOR & 0x00FF0000) >> 16) / 255.0;
            double bgG = ((WHITE_COLOR & 0x0000FF00) >> 8) / 255.0;
            double bgB = (WHITE_COLOR & 0x000000FF) / 255.0;

            PDFDoc doc;

            switch (paperType)
            {
                case PaperType.Dotted:
                    doc = PDFDocGenerator.GenerateDottedPaperDoc(width_in, height_in, 0.25, 2.0, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Graph:
                    doc = PDFDocGenerator.GenerateGraphPaperDoc(width_in, height_in, 0.25, 0.45, 1.7, 5, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Grid:
                    doc = PDFDocGenerator.GenerateGridPaperDoc(width_in, height_in, 0.25, 0.45, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Iso:
                    doc = PDFDocGenerator.GenerateIsometricDottedPaperDoc(width_in, height_in, 0.25, 2.0, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Lined:
                    double leftMarginR = MARGIN_RED * 0.7;
                    double leftMarginG = MARGIN_GREEN * 0.7;
                    double leftMarginB = MARGIN_BLUE * 0.7;

                    double rightMarginR = MARGIN_RED * 0.45;
                    double rightMarginG = MARGIN_GREEN * 0.45;
                    double rightMarginB = MARGIN_BLUE * 0.45;

                    doc = PDFDocGenerator.GenerateLinedPaperDoc(width_in, height_in, 0.25, 0.45, lineShade,
                        lineShade, lineShade, 1.2, leftMarginR, leftMarginG, leftMarginB,
                        rightMarginR, rightMarginG, rightMarginB, bgR, bgG, bgB, 0.85, 0.35);
                    break;

                case PaperType.Music:
                    doc = PDFDocGenerator.GenerateMusicPaperDoc(width_in, height_in, 0.5, 10, 6.5, 0.25, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                default:
                case PaperType.Blank:
                    doc = PDFDocGenerator.GenerateBlankPaperDoc(width_in, height_in, bgR, bgG, bgB);
                    break;
            }

            for (int i = 1; i < pages; i++)
                doc.PagePushBack(doc.GetPage(1));

            return doc;
        }


        public static double PagePointToInches(double points)
        {
            return points / 72;
        }
    }

    public enum PaperType
    {
        Blank,
        Lined,
        Grid,
        Graph,
        Dotted,
        Iso,
        Music
    }

    public class PaperTypeHolder
    {
        public PaperType PaperType { get; set; }

        public override string ToString() => PaperType.ToString("g");

        public PaperTypeHolder(PaperType type)
        {
            PaperType = type;
        }
    }
}
