﻿using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.UI.Popups;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Utilities
{
    public class SaveOnExitHelper
    {
        /// <summary>
        /// It checks the level of saving needed for the selected tab
        /// </summary>
        /// <param name="tab">The tab to be checked</param>
        /// <returns></returns>
        public static SaveNeededStates CheckTabSavingRequirements(PdfViewCtrlTabInfo tab)
        {
            if (tab == null)
                return SaveNeededStates.none;

            if (ViewerSettings.AutoSaveOn && tab.IsDocumentModifiedSinceLastSave &&
                (tab.DocumentState == OpenedDocumentStates.Normal
                || tab.DocumentState == OpenedDocumentStates.NeedsFullSave))
            {
                return SaveNeededStates.saveNeeded;
            }
            else if (tab.DocumentState == OpenedDocumentStates.CorruptedAndModified
                || tab.DocumentState == OpenedDocumentStates.Created
                || (tab.DocumentState == OpenedDocumentStates.ReadOnly && tab.HasDocumentBeenModifiedSinceOpening)
                || (tab.DocumentState == OpenedDocumentStates.Normal && tab.IsDocumentModifiedSinceLastSave)
                || (tab.DocumentState == OpenedDocumentStates.NeedsFullSave && tab.IsDocumentModifiedSinceLastSave)
                || (tab.DocumentState == OpenedDocumentStates.NonePDF && tab.IsDocumentModifiedSinceLastSave)
                )
            {
                return SaveNeededStates.saveAsNeeded;
            }

            return SaveNeededStates.none;
        }

        /// <summary>
        /// It starts the saving process on the selected tab
        /// </summary>
        /// <param name="tab"></param>
        /// <param name="currentLoader"></param>
        /// <param name="includeFullSave"></param>
        /// <param name="alwaysSave">Means that we want to save to the original file, not just to the temporary one</param>
        /// <returns></returns>
        public static async Task SaveTabHelperAsync(PdfViewCtrlTabInfo tab, ResourceLoader currentLoader, bool includeFullSave = false, bool alwaysSave = false, bool showErrors = true)
        {
            if (tab == null)
                tab = TabControlViewModel.Instance.SelectedTab;

            if (tab == null)
                return;

            SaveResult result = await SaveTabToOriginalAsync(tab, includeFullSave, alwaysSave).ConfigureAwait(true);
            if (result != SaveResult.e_normal && showErrors)
            {
                string messageString = string.Format(currentLoader.GetString("ViewerPage_TabDialog_SaveError_AccessDenied"), tab.Title);
                if (result == SaveResult.e_unknown_error)
                {
                    messageString = string.Format(currentLoader.GetString("ViewerPage_TabDialog_SaveError"), tab.Title);
                }

                ContentDialog dialog = new ContentDialog
                {
                    Title = currentLoader.GetString("ViewerPage_TabDialog_SaveError_Title"),
                    Content = messageString,
                    CloseButtonText = currentLoader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
            }
        }

        public static async Task<SaveResult> SaveTabToOriginalAsync(PdfViewCtrlTabInfo tab, bool includeFullSave = false, bool alwaysSave = false)
        {
            if (tab == null)
                tab = TabControlViewModel.Instance.SelectedTab;

            if (tab == null)
                return SaveResult.e_normal;

            SaveResult result = SaveResult.e_normal;
            try
            {
                if (tab.DocumentState != OpenedDocumentStates.Corrupted && tab.DocumentState != OpenedDocumentStates.CorruptedAndModified)
                {
                    result = SaveResult.e_unknown_error;

                    if (tab.Doc != null)
                    {
                        await tab.SaveDocumentImplAsync();
                    }

                    if ((tab.DocumentState == OpenedDocumentStates.Normal && (ViewerSettings.AutoSaveOn || alwaysSave))
                        || (tab.DocumentState == OpenedDocumentStates.NeedsFullSave && includeFullSave))
                    {
                        await DocumentManager.Instance.AddChangesToOriginalAsync(tab.OriginalFile, tab.PdfFile);
                        Tuple<DateTimeOffset, ulong> lastModifiedAndSize = await pdftron.PDF.Tools.UtilityFunctions.GetDateModifiedAndSizeAsync(tab.OriginalFile);
                        tab.IsDocumentModifiedSinceLastSave = false;
                        tab.MetaData.LastModifiedDate = lastModifiedAndSize.Item1;
                        tab.MetaData.FileSize = lastModifiedAndSize.Item2;
                        tab.SystemLastModifiedDate = lastModifiedAndSize.Item1;
                        tab.SystemFileSize = lastModifiedAndSize.Item2;
                        TabControlViewModel.Instance.SaveTabState();
                    }

                    result = SaveResult.e_normal;
                }
            }
            catch (UnauthorizedAccessException)
            {
                result = SaveResult.e_access_denied;
            }
            catch (Exception)
            {
                result = SaveResult.e_unknown_error;
            }

            return result;
        }
    }
}
